Type: Package
Package: tidyfinance
Title: Tidy Finance Helper Functions
Version: 0.5.0
Authors@R: c(
    person("Christoph", "Scheuch", , "christoph@tidy-intelligence.com", role = c("aut", "cre", "cph"),
           comment = c(ORCID = "0009-0004-0423-6819")),
    person("Stefan", "Voigt", , "stefan.voigt@econ.ku.dk", role = c("aut", "cph"),
           comment = c(ORCID = "0000-0001-5619-3161")),
    person("Patrick", "Weiss", , "patrickw@ru.is", role = c("aut", "cph"),
           comment = c(ORCID = "0000-0002-9282-5872")),
    person("Maximilian", "Mücke", role = "ctb",
           comment = c(ORCID = "0009-0000-9432-9795"))
  )
Description: Helper functions for empirical research in financial
    economics, addressing a variety of topics covered in Scheuch, Voigt,
    and Weiss (2023) <doi:10.1201/b23237>.  The package is designed to
    provide shortcuts for issues extensively discussed in the book,
    facilitating easier application of its concepts. For more information
    and resources related to the book, visit
    <https://www.tidy-finance.org/r/index.html>.
License: MIT + file LICENSE
URL: https://www.tidy-finance.org/r/,
        https://github.com/tidy-finance/r-tidyfinance
BugReports: https://github.com/tidy-finance/r-tidyfinance/issues
Depends: R (>= 4.1)
Imports: arrow, cli, dplyr (>= 1.1.4), dbplyr (>= 2.5.0), glue,
        jsonlite, lifecycle, lubridate (>= 1.9.3), purrr (>= 1.0.2),
        rlang (>= 1.1.3), slider (>= 0.3.1), stats, stringr, tibble,
        tidyr (>= 1.3.1), DBI (>= 1.2.2), frenchdata (>= 0.2.0), httr2
        (>= 1.0.0), RPostgres (>= 1.4.5), sandwich (>= 3.1-0)
Suggests: knitr, rmarkdown, furrr (>= 0.3.1), testthat (>= 3.2.0)
VignetteBuilder: knitr
Config/testthat/edition: 3
Encoding: UTF-8
RoxygenNote: 7.3.3
NeedsCompilation: no
Packaged: 2026-05-12 14:31:15 UTC; krise
Author: Christoph Scheuch [aut, cre, cph] (ORCID:
    <https://orcid.org/0009-0004-0423-6819>),
  Stefan Voigt [aut, cph] (ORCID:
    <https://orcid.org/0000-0001-5619-3161>),
  Patrick Weiss [aut, cph] (ORCID:
    <https://orcid.org/0000-0002-9282-5872>),
  Maximilian Mücke [ctb] (ORCID: <https://orcid.org/0009-0000-9432-9795>)
Maintainer: Christoph Scheuch <christoph@tidy-intelligence.com>
Repository: CRAN
Date/Publication: 2026-05-12 15:00:02 UTC
Built: R 4.6.0; ; 2026-05-12 19:55:02 UTC; unix

# Generated by roxygen2: do not edit by hand

export(add_lagged_columns)
export(assign_portfolio)
export(breakpoint_options)
export(compute_breakpoints)
export(compute_long_short_returns)
export(compute_portfolio_returns)
export(compute_rolling_value)
export(create_summary_statistics)
export(create_wrds_dummy_database)
export(data_options)
export(disconnect_connection)
export(download_data)
export(download_data_constituents)
export(download_data_factors_ff)
export(download_data_factors_q)
export(download_data_fred)
export(download_data_huggingface)
export(download_data_macro_predictors)
export(download_data_osap)
export(download_data_risk_free)
export(download_data_stock_prices)
export(download_data_wrds)
export(download_data_wrds_ccm_links)
export(download_data_wrds_compustat)
export(download_data_wrds_crsp)
export(download_data_wrds_fisd)
export(download_data_wrds_trace_enhanced)
export(estimate_betas)
export(estimate_fama_macbeth)
export(estimate_model)
export(filter_options)
export(filter_sorting_data)
export(get_available_huggingface_files)
export(get_wrds_connection)
export(implement_portfolio_sort)
export(join_lagged_values)
export(list_supported_indexes)
export(list_supported_types)
export(list_tidy_finance_chapters)
export(open_tidy_finance_website)
export(portfolio_sort_options)
export(set_wrds_credentials)
export(trim)
export(validate_dates)
export(winsorize)
import(DBI)
import(dplyr)
import(httr2)
import(lubridate)
import(tibble)
importFrom(RPostgres,Postgres)
importFrom(arrow,read_parquet)
importFrom(arrow,write_parquet)
importFrom(dbplyr,translate_sql)
importFrom(frenchdata,download_french_data)
importFrom(frenchdata,get_french_data_list)
importFrom(jsonlite,fromJSON)
importFrom(lifecycle,deprecated)
importFrom(purrr,map)
importFrom(purrr,map_dbl)
importFrom(purrr,partial)
importFrom(rlang,":=")
importFrom(rlang,enquo)
importFrom(rlang,quo_is_null)
importFrom(rlang,syms)
importFrom(sandwich,NeweyWest)
importFrom(stats,as.formula)
importFrom(stats,ave)
importFrom(stats,na.omit)
importFrom(stats,quantile)
importFrom(stats,setNames)
importFrom(stats,weighted.mean)
importFrom(stringr,str_match)
importFrom(tidyr,drop_na)
importFrom(tidyr,fill)
importFrom(tidyr,nest)
importFrom(tidyr,pivot_longer)
importFrom(tidyr,unnest)
importFrom(utils,read.csv)

# tidyfinance

![PyPI](https://img.shields.io/pypi/v/tidyfinance?label=pypi%20package)
![PyPI Downloads](https://img.shields.io/pypi/dm/tidyfinance)
[![python-package.yml](https://github.com/tidy-finance/py-tidyfinance/actions/workflows/python-package.yml/badge.svg)](https://github.com/tidy-finance/py-tidyfinance/actions/workflows/python-package.yml)
<!-- [![codecov.yml](https://codecov.io/gh/tidy-finance/py-tidyfinance/graph/badge.svg)](https://app.codecov.io/gh/tidy-finance/py-tidyfinance) -->
[![License:
MIT](https://img.shields.io/badge/License-MIT-yellow.svg)](https://opensource.org/licenses/MIT)

Helper functions for empirical research in financial economics, addressing a variety of topics covered in [Scheuch, Voigt, Weiss, and Frey (2024)](https://doi.org/10.1201/9781032684307). The package is designed to provide shortcuts for issues extensively discussed in the book, facilitating easier application of its concepts. For more information and resources related to the book, visit [tidy-finance.org/python](https://tidy-finance.org/python).

## Installation

You can install the release version from [PyPI](https://pypi.org/project/tidyfinance/):

```
pip install tidyfinance
```

You can install the development version from GitHub:

```
pip install "git+https://github.com/tidy-finance/py-tidyfinance"
```

## Download Open Source Data

The main functionality of the `tidyfinance` package centers around data download. You can download most of the data that we used in Tidy Finance with R using the `download_data()` function or its children. 

```python
import tidyfinance as tf
```

The function always requires a `domain` argument and depending on the domain typically also a `dataset`. For instance, to download monthly Fama-French factors, you have to provide the dataset name according to `pdr.famafrench.get_available_datasets()`:

```python
tf.download_data(
  domain="factors_ff",
  dataset="F-F_Research_Data_5_Factors_2x3_daily",
  start_date="2000-01-01", 
  end_date="2020-12-31"
)
```

For q factors, you provide the relevant file name:

```python
tf.download_data(
  domain="factors_q",
  dataset="q5_factors_monthly",
  start_date="2000-01-01", 
  end_date="2020-12-31"
)
```

To download the Welch and Goyal (2008) macroeconomic predictors for monthly, quarterly, or annual frequency:

```python
tf.download_data(
  domain="macro_predictors",
  dataset="monthly",
  start_date="2000-01-01", 
  end_date="2020-12-31"
)
```

To download data from Open Source Asset Pricing (OSAP):

```python
tf.download_data(
  domain="osap",
  start_date="2020-01-01", 
  end_date="2020-12-31"
)
```

To download multiple series from the Federal Reserve Economic Data (FRED):

```python
tf.download_data(
  domain="fred",
  series=["GDP", "CPIAUCNS"], 
  start_date="2020-01-01", 
  end_date="2020-12-31"
)
```

To download stock prices from Yahoo Finance:

```python
tf.download_data(
  domain="stock_prices",
  symbols=["AAPL", "MSFT"], 
  start_date="2020-01-01", 
  end_date="2020-12-31"
)
```

To download index constituents from selected ETF holdings: 

```python
tf.download_data(
  domain="constituents",
  index="S&P 500"
)
```

## Download WRDS Data

To access data from the [Wharton Research Data Services (WRDS)](https://wrds-www.wharton.upenn.edu/), you need to set your credentials first:

```python
tf.set_wrds_credentials()
```

To download monthly CRSP data: 

```python
tf.download_data(
  domain="wrds",
  dataset="crsp_monthly", 
  start_date="2020-01-01", 
  end_date="2020-12-31"
)
```

To download annual (or quaterly) Compustat data:

```python
tf.download_data(
  domain="wrds",
  dataset="compustat_annual", 
  start_date="2020-01-01", 
  end_date="2020-12-31"
)
```

To download the CRSP-Compustat linking table: 

```python
tf.download_data(
  domain="wrds",
  dataset="ccm_links"
)
```

To download bond characteristics from Mergent FISD:

```python
tf.download_data(
  domain="wrds",
  dataset="fisd"
)
```

To download Enhanced TRACE data for selected bonds: 

```python
tf.download_data(
  domain="wrds",
  dataset="trace_enhanced",
  cusips=["00101JAH9"],
  start_date="2019-01-01", 
  end_date="2021-12-31"
)
```

## Other Helpers

We include functions to check out content from tidy-finance.org:

```python
tf.list_tidy_finance_chapters()
tf.open_tidy_finance_website("capital-asset-pricing-model")
```

We also include (experimental) functions that can be used for different applications, but note that they might heavily change in future package versions as we try to make them more general:

```python
# Create summary statistics
help(tf.create_summary_statistics)

# Assign portfolios
help(tf.assign_portfolio)

# Estimate betas
help(tf.estimate_betas)

# Estimate Fama-MacBeth 
help(tf.estimate_fama_macbeth)

# Add lag columns
help(tf.add_lag_columns)

# Winsorize or trim 
help(tf.winsorize)
help(tf.trim)
```

#' Filter Sorting Data
#'
#' Applies sample construction filters to a data frame before portfolio
#' sorting. Filters are applied in a fixed order: financials exclusion,
#' utilities exclusion, minimum stock price, minimum size quantile, minimum
#' listing age, positive book equity, and positive earnings. An informational
#' message is emitted for each filter that actually removes at least one
#' observation.
#'
#' @param data A data frame containing the stock-level panel data to be
#'   filtered.
#' @param filter_options A list of class `tidyfinance_filter_options` created
#'   by [filter_options()]. If `NULL` (the default), the defaults from
#'   [filter_options()] are used (i.e., no filters are applied). The arguments
#'   accepted by [filter_options()] include
#'   \itemize{
#'     \item `exclude_financials` A logical indicating whether to exclude
#'       financial firms (SIC codes 6000–6799). Defaults to `FALSE`.
#'     \item `exclude_utilities` A logical indicating whether to exclude
#'       utility firms (SIC codes 4900–4999). Defaults to `FALSE`.
#'     \item `min_stock_price` A single positive numeric specifying the
#'       minimum stock price required to include an observation. `NULL` (the
#'       default) applies no price filter.
#'     \item `min_size_quantile` A single numeric strictly between 0 and 1
#'       specifying the minimum cross-sectional size quantile (based on lagged
#'       market cap) required to include an observation. `NULL` (the default)
#'       applies no size quantile filter. The cutoff is computed from NYSE
#'       stocks only; the `exchange` column (mapped via [data_options()])
#'       must be present in the data or an error is raised.
#'     \item `min_listing_age` A single non-negative integer or numeric
#'       specifying the minimum number of months a stock must have been listed
#'       in CRSP. `NULL` (the default) applies no listing age filter.
#'     \item `exclude_negative_book_equity` A logical indicating whether to
#'       exclude observations with non-positive book equity. Defaults to
#'       `FALSE`.
#'     \item `exclude_negative_earnings` A logical indicating whether to
#'       exclude observations with non-positive earnings. Defaults to `FALSE`.
#'   }
#' @param data_options A list of class `tidyfinance_data_options` (created via
#'   [data_options()]) specifying column name mappings. The `siccd` element is
#'   used to specify the SIC code column, `price` is used to specify the
#'   (adjusted) price column, `mktcap_lag` is used to specify the market
#'   capitalization column, `date` is used to specify the date column,
#'   `listing_age` is used to specify the listing age column, `be` is used to
#'   specify the book equity column, and `earnings` is used to specify the
#'   earnings column. Uses [data_options()] default if `NULL`:
#'   `"siccd" = "siccd"`, `"price" = "prc_adj"`, `"exchange" = "exchange"`
#'   `"mktcap_lag" = "mktcap_lag"`, `"date" = "date"`,
#'   `"listing_age" = "listing_age"`, `"be" = "be"`, and `"earnings" = "ib"`.
#' @param quiet A logical indicating whether informational messages should be
#'   suppressed. Defaults to `FALSE`.
#'
#' @returns The filtered data frame, preserving the class and structure of the
#'   input.
#'
#' @family portfolio functions
#' @export
#'
#' @examples
#' data <- data.frame(
#'   permno = 1:5,
#'   date = as.Date("2020-01-01"),
#'   siccd = c(6100, 2000, 4950, 3000, 6500),
#'   prc_adj = c(5, 0.5, 15, 20, 10)
#' )
#'
#' data |>
#'   filter_sorting_data(
#'     filter_options = filter_options(
#'       exclude_financials = TRUE,
#'       min_stock_price = 1
#'     )
#'   )
#'
filter_sorting_data <- function(
  data,
  filter_options = NULL,
  data_options = NULL,
  quiet = FALSE
) {
  if (!is.logical(quiet) || length(quiet) != 1 || is.na(quiet)) {
    cli::cli_abort("{.arg quiet} must be a single logical.")
  }

  if (is.null(filter_options)) {
    filter_options <- filter_options()
  }

  if (is.null(data_options)) {
    data_options <- data_options()
  }

  # exclude_financials / exclude_utilities (share the same SIC column)
  if (
    isTRUE(filter_options$exclude_financials) ||
      isTRUE(filter_options$exclude_utilities)
  ) {
    col_siccd <- data_options$siccd
    if (!col_siccd %in% colnames(data)) {
      cli::cli_abort(c(
        "Column {.val {col_siccd}} not found in {.arg data}.",
        "i" = "Set {.arg data_options$siccd} to the correct column name."
      ))
    }

    if (isTRUE(filter_options$exclude_financials)) {
      n_before <- nrow(data)
      data <- data |>
        dplyr::filter(
          is.na(.data[[col_siccd]]) |
            !(.data[[col_siccd]] >= 6000 & .data[[col_siccd]] <= 6799)
        )
      n_dropped <- n_before - nrow(data)
      if (!quiet && n_dropped > 0) {
        cli::cli_inform(
          "Filter 'exclude_financials': removed {n_dropped} observation{?s}."
        )
      }
    }

    if (isTRUE(filter_options$exclude_utilities)) {
      n_before <- nrow(data)
      data <- data |>
        dplyr::filter(
          is.na(.data[[col_siccd]]) |
            !(.data[[col_siccd]] >= 4900 & .data[[col_siccd]] <= 4999)
        )
      n_dropped <- n_before - nrow(data)
      if (!quiet && n_dropped > 0) {
        cli::cli_inform(
          "Filter 'exclude_utilities': removed {n_dropped} observation{?s}."
        )
      }
    }
  }

  # min_stock_price
  if (!is.null(filter_options$min_stock_price)) {
    col_price <- data_options$price
    if (!col_price %in% colnames(data)) {
      cli::cli_abort(c(
        "Column {.val {col_price}} not found in {.arg data}.",
        "i" = "Set {.arg data_options$price} to the correct column name."
      ))
    }
    n_before <- nrow(data)
    data <- data |>
      dplyr::filter(
        !is.na(.data[[col_price]]) &
          .data[[col_price]] >= filter_options$min_stock_price
      )
    n_dropped <- n_before - nrow(data)
    if (!quiet && n_dropped > 0) {
      cli::cli_inform(
        "Filter 'min_stock_price': removed {n_dropped} observation{?s}."
      )
    }
  }

  # min_size_quantile
  if (!is.null(filter_options$min_size_quantile)) {
    col_mktcap_lag <- data_options$mktcap_lag
    col_date <- data_options$date
    col_exchange <- data_options$exchange
    if (!col_mktcap_lag %in% colnames(data)) {
      cli::cli_abort(c(
        "Column {.val {col_mktcap_lag}} not found in {.arg data}.",
        "i" = "Set {.arg data_options$mktcap_lag} to the correct column name."
      ))
    }
    if (!col_date %in% colnames(data)) {
      cli::cli_abort(c(
        "Column {.val {col_date}} not found in {.arg data}.",
        "i" = "Set {.arg data_options$date} to the correct column name."
      ))
    }
    if (!col_exchange %in% colnames(data)) {
      cli::cli_abort(c(
        "Column {.val {col_exchange}} not found in {.arg data}.",
        "i" = paste0(
          "The size quantile cutoff is computed from NYSE stocks. ",
          "Set {.arg data_options$exchange} to the correct column name."
        )
      ))
    }
    n_before <- nrow(data)
    size_threshold <- filter_options$min_size_quantile
    size_cutoffs <- data |>
      dplyr::filter(.data[[col_exchange]] == "NYSE") |>
      dplyr::group_by(.data[[col_date]]) |>
      dplyr::summarise(
        .size_cutoff = quantile(
          .data[[col_mktcap_lag]],
          probs = size_threshold,
          na.rm = TRUE
        ),
        .groups = "drop"
      )

    dates_in_data <- data |>
      dplyr::distinct(.data[[col_date]]) |>
      dplyr::pull(col_date)

    dates_missing_cutoff <- setdiff(dates_in_data, size_cutoffs[[col_date]])
    if (length(dates_missing_cutoff) > 0) {
      cli::cli_warn(c(
        paste0(
          "Filter 'min_size_quantile': {length(dates_missing_cutoff)} ",
          "date{?s} dropped because no NYSE stocks are available to compute ",
          "the size quantile cutoff."
        ),
        "i" = "Affected date{?s}: {.val {dates_missing_cutoff}}."
      ))
    }

    data <- data |>
      check_new_col(".size_cutoff") |>
      dplyr::select(-dplyr::any_of(".size_cutoff")) |>
      dplyr::inner_join(size_cutoffs, by = col_date) |>
      dplyr::filter(
        !is.na(.data[[col_mktcap_lag]]) &
          .data[[col_mktcap_lag]] >= .size_cutoff
      ) |>
      dplyr::select(-.size_cutoff)
    n_dropped <- n_before - nrow(data)
    if (!quiet && n_dropped > 0) {
      cli::cli_inform(
        "Filter 'min_size_quantile': removed {n_dropped} observation{?s}."
      )
    }
  }

  # min_listing_age
  if (!is.null(filter_options$min_listing_age)) {
    col_listing_age <- data_options$listing_age
    if (!col_listing_age %in% colnames(data)) {
      cli::cli_abort(c(
        "Column {.val {col_listing_age}} not found in {.arg data}.",
        "i" = "Set {.arg data_options$listing_age} to the correct column name."
      ))
    }
    n_before <- nrow(data)
    data <- data |>
      dplyr::filter(
        !is.na(.data[[col_listing_age]]) &
          .data[[col_listing_age]] >= filter_options$min_listing_age
      )
    n_dropped <- n_before - nrow(data)
    if (!quiet && n_dropped > 0) {
      cli::cli_inform(
        "Filter 'min_listing_age': removed {n_dropped} observation{?s}."
      )
    }
  }

  # exclude_negative_book_equity
  if (isTRUE(filter_options$exclude_negative_book_equity)) {
    col_be <- data_options$be
    if (!col_be %in% colnames(data)) {
      cli::cli_abort(c(
        "Column {.val {col_be}} not found in {.arg data}.",
        "i" = "Set {.arg data_options$be} to the correct column name."
      ))
    }
    n_before <- nrow(data)
    data <- data |>
      dplyr::filter(
        !is.na(.data[[col_be]]) & .data[[col_be]] > 0
      )
    n_dropped <- n_before - nrow(data)
    if (!quiet && n_dropped > 0) {
      cli::cli_inform(
        paste0(
          "Filter 'exclude_negative_book_equity': ",
          "removed {n_dropped} observation{?s}."
        )
      )
    }
  }

  # exclude_negative_earnings
  if (isTRUE(filter_options$exclude_negative_earnings)) {
    col_earn <- data_options$earnings
    if (!col_earn %in% colnames(data)) {
      cli::cli_abort(c(
        "Column {.val {col_earn}} not found in {.arg data}.",
        "i" = "Set {.arg data_options$earnings} to the correct column name."
      ))
    }
    n_before <- nrow(data)
    data <- data |>
      dplyr::filter(
        !is.na(.data[[col_earn]]) & .data[[col_earn]] > 0
      )
    n_dropped <- n_before - nrow(data)
    if (!quiet && n_dropped > 0) {
      cli::cli_inform(
        paste0(
          "Filter 'exclude_negative_earnings': ",
          "removed {n_dropped} observation{?s}."
        )
      )
    }
  }

  data
}

make_sorting_data <- function(seed = 42) {
  set.seed(seed)
  data.frame(
    permno = 1:30,
    date = rep(as.Date("2020-01-01"), 30),
    exchange = c(rep("NYSE", 15), rep("NASDAQ", 15)),
    siccd = c(
      rep(6200L, 5), # financial (SIC 6000–6799)
      rep(4950L, 5), # utility  (SIC 4900–4999)
      rep(5000L, 5), # retail   (neutral)
      rep(7000L, 5), # services (neutral)
      rep(1000L, 10) # other    (neutral)
    ),
    prc_adj = c(rep(0.5, 10), rep(5, 10), rep(20, 10)),
    mktcap_lag = c(
      seq(100, 500, length.out = 15),
      seq(5000, 10000, length.out = 15)
    ),
    listing_age = c(rep(6L, 10), rep(24L, 20)),
    be = c(rep(-1, 5), rep(0.5, 5), rep(1, 20)),
    ib = c(rep(-2, 5), rep(0.3, 5), rep(1, 20))
  )
}

test_that("filter_sorting_data returns all rows when no filters are active", {
  data <- make_sorting_data()
  result <- filter_sorting_data(data, quiet = TRUE)
  expect_equal(nrow(result), nrow(data))
})

test_that("filter_sorting_data preserves all columns", {
  data <- make_sorting_data()
  result <- filter_sorting_data(data, quiet = TRUE)
  expect_equal(colnames(result), colnames(data))
})

test_that(
  paste0(
    "filter_sorting_data removes ",
    "financials when exclude_financials = TRUE"
  ),
  {
    data <- make_sorting_data()
    n_financials <- sum(data$siccd >= 6000 & data$siccd <= 6799, na.rm = TRUE)
    result <- filter_sorting_data(
      data,
      filter_options = filter_options(exclude_financials = TRUE),
      quiet = TRUE
    )
    expect_equal(nrow(result), nrow(data) - n_financials)
    expect_true(all(
      is.na(result$siccd) | result$siccd < 6000 | result$siccd > 6799
    ))
  }
)

test_that(
  paste0(
    "filter_sorting_data removes utility ",
    "firms when exclude_utilities = TRUE"
  ),
  {
    data <- make_sorting_data()
    n_utilities <- sum(data$siccd >= 4900 & data$siccd <= 4999, na.rm = TRUE)
    result <- filter_sorting_data(
      data,
      filter_options = filter_options(exclude_utilities = TRUE),
      quiet = TRUE
    )
    expect_equal(nrow(result), nrow(data) - n_utilities)
    expect_true(all(
      is.na(result$siccd) | result$siccd < 4900 | result$siccd > 4999
    ))
  }
)

test_that(
  paste0(
    "filter_sorting_data removes financials ",
    "and utilities when both excluded"
  ),
  {
    data <- make_sorting_data()
    n_fin_or_util <- sum(
      (data$siccd >= 6000 & data$siccd <= 6799) |
        (data$siccd >= 4900 & data$siccd <= 4999),
      na.rm = TRUE
    )
    result <- filter_sorting_data(
      data,
      filter_options = filter_options(
        exclude_financials = TRUE,
        exclude_utilities = TRUE
      ),
      quiet = TRUE
    )
    expect_equal(nrow(result), nrow(data) - n_fin_or_util)
  }
)

test_that(
  paste0(
    "filter_sorting_data errors when ",
    "siccd missing for financial exclusion"
  ),
  {
    data <- make_sorting_data()
    data$siccd <- NULL
    expect_error(
      filter_sorting_data(
        data,
        filter_options = filter_options(exclude_financials = TRUE)
      ),
      "siccd"
    )
  }
)

test_that("filter_sorting_data applies min_stock_price filter correctly", {
  data <- make_sorting_data()
  threshold <- 3
  n_below <- sum(!is.na(data$prc_adj) & data$prc_adj < threshold)
  result <- filter_sorting_data(
    data,
    filter_options = filter_options(min_stock_price = threshold),
    quiet = TRUE
  )
  expect_equal(nrow(result), nrow(data) - n_below)
  expect_true(all(result$prc_adj >= threshold, na.rm = TRUE))
})

test_that("filter_sorting_data errors when price missing for min_stock_price", {
  data <- make_sorting_data()
  data$prc_adj <- NULL
  expect_error(
    filter_sorting_data(
      data,
      filter_options = filter_options(min_stock_price = 5)
    ),
    "prc_adj"
  )
})

test_that(
  paste0(
    "filter_sorting_data applies ",
    "min_size_quantile filter using NYSE cutoff"
  ),
  {
    data <- make_sorting_data()
    # NYSE mktcap_lag: seq(100, 500, length.out = 15), 50th pctile = 300
    # 7 NYSE stocks fall below the cutoff;
    # all 15 NASDAQ stocks (5000–10000) pass
    nyse_cutoff <- quantile(seq(100, 500, length.out = 15), probs = 0.5)
    result <- filter_sorting_data(
      data,
      filter_options = filter_options(min_size_quantile = 0.5),
      quiet = TRUE
    )
    expect_equal(nrow(result), sum(data$mktcap_lag >= nyse_cutoff))
    expect_true(all(result$mktcap_lag >= nyse_cutoff))
  }
)

test_that(
  paste0(
    "filter_sorting_data errors when ",
    "mktcap_lag missing for min_size_quantile"
  ),
  {
    data <- make_sorting_data()
    data$mktcap_lag <- NULL
    expect_error(
      filter_sorting_data(
        data,
        filter_options = filter_options(min_size_quantile = 0.2)
      ),
      "mktcap_lag"
    )
  }
)

test_that(
  paste0(
    "filter_sorting_data errors when ",
    "exchange missing for min_size_quantile"
  ),
  {
    data <- make_sorting_data()
    data$exchange <- NULL
    expect_error(
      filter_sorting_data(
        data,
        filter_options = filter_options(min_size_quantile = 0.2)
      ),
      "exchange"
    )
  }
)

test_that("filter_sorting_data applies min_listing_age filter correctly", {
  data <- make_sorting_data()
  min_age <- 12
  n_young <- sum(!is.na(data$listing_age) & data$listing_age < min_age)
  result <- filter_sorting_data(
    data,
    filter_options = filter_options(min_listing_age = min_age),
    quiet = TRUE
  )
  expect_equal(nrow(result), nrow(data) - n_young)
  expect_true(all(result$listing_age >= min_age, na.rm = TRUE))
})

test_that("filter_sorting_data errors when listing_age column is missing", {
  data <- make_sorting_data()
  data$listing_age <- NULL
  expect_error(
    filter_sorting_data(
      data,
      filter_options = filter_options(min_listing_age = 12)
    ),
    "listing_age"
  )
})

test_that(
  paste0(
    "filter_sorting_data applies ",
    "exclude_negative_book_equity filter correctly"
  ),
  {
    data <- make_sorting_data()
    n_nonpositive <- sum(!is.na(data$be) & data$be <= 0)
    result <- filter_sorting_data(
      data,
      filter_options = filter_options(exclude_negative_book_equity = TRUE),
      quiet = TRUE
    )
    expect_equal(nrow(result), nrow(data) - n_nonpositive)
    expect_true(all(result$be > 0))
  }
)

test_that(
  paste0(
    "filter_sorting_data errors when ",
    "be missing for exclude_negative_book_equity"
  ),
  {
    data <- make_sorting_data()
    data$be <- NULL
    expect_error(
      filter_sorting_data(
        data,
        filter_options = filter_options(exclude_negative_book_equity = TRUE)
      ),
      "be"
    )
  }
)

test_that(
  paste0(
    "filter_sorting_data applies ",
    "exclude_negative_earnings filter correctly"
  ),
  {
    data <- make_sorting_data()
    n_nonpositive <- sum(!is.na(data$ib) & data$ib <= 0)
    result <- filter_sorting_data(
      data,
      filter_options = filter_options(exclude_negative_earnings = TRUE),
      quiet = TRUE
    )
    expect_equal(nrow(result), nrow(data) - n_nonpositive)
    expect_true(all(result$ib > 0))
  }
)

test_that(
  paste0(
    "filter_sorting_data emits a message mentioning ",
    "exclude_negative_earnings when quiet = FALSE and rows are removed"
  ),
  {
    data <- make_sorting_data()
    expect_message(
      filter_sorting_data(
        data,
        filter_options = filter_options(exclude_negative_earnings = TRUE),
        quiet = FALSE
      ),
      "exclude_negative_earnings"
    )
  }
)

test_that(
  paste0(
    "filter_sorting_data errors when earnings ",
    "column is missing for exclude_negative_earnings"
  ),
  {
    data <- make_sorting_data()
    data$ib <- NULL
    expect_error(
      filter_sorting_data(
        data,
        filter_options = filter_options(exclude_negative_earnings = TRUE)
      ),
      "ib"
    )
  }
)

test_that(
  paste0(
    "filter_sorting_data emits a message ",
    "when quiet=FALSE and rows removed"
  ),
  {
    data <- make_sorting_data()
    expect_message(
      filter_sorting_data(
        data,
        filter_options = filter_options(exclude_financials = TRUE),
        quiet = FALSE
      )
    )
  }
)

test_that("filter_sorting_data suppresses messages when quiet = TRUE", {
  data <- make_sorting_data()
  expect_silent(
    filter_sorting_data(
      data,
      filter_options = filter_options(exclude_financials = TRUE),
      quiet = TRUE
    )
  )
})

test_that("filter_sorting_data errors for invalid quiet argument", {
  data <- make_sorting_data()
  expect_error(filter_sorting_data(data, quiet = "yes"), "quiet")
  expect_error(filter_sorting_data(data, quiet = NA), "quiet")
  expect_error(filter_sorting_data(data, quiet = 1), "quiet")
})

test_that(
  paste0(
    "filter_sorting_data respects custom ",
    "column names via data_options"
  ),
  {
    data <- make_sorting_data()
    names(data)[names(data) == "prc_adj"] <- "stock_price"
    result <- filter_sorting_data(
      data,
      filter_options = filter_options(min_stock_price = 3),
      data_options = data_options(price = "stock_price"),
      quiet = TRUE
    )
    expect_true(all(result$stock_price >= 3, na.rm = TRUE))
  }
)

test_that(
  paste0(
    "filter_sorting_data warns when dates are dropped ",
    "due to missing NYSE quantile cutoff"
  ),
  {
    # date1 has NYSE stocks -> quantile cutoff can be computed
    # date2 has no NYSE stocks -> no cutoff, all rows for date2 are dropped
    date1 <- as.Date("2020-01-01")
    date2 <- as.Date("2020-02-01")
    data <- data.frame(
      permno = 1:6,
      date = c(rep(date1, 4), rep(date2, 2)),
      exchange = c("NYSE", "NYSE", "NASDAQ", "NASDAQ", "NASDAQ", "NASDAQ"),
      mktcap_lag = c(100, 200, 150, 250, 300, 400)
    )
    expect_warning(
      result <- filter_sorting_data(
        data,
        filter_options = filter_options(min_size_quantile = 0.5)
      ),
      regexp = "min_size_quantile"
    )
    # all rows for date2 must be gone
    expect_true(all(result$date == date1))
  }
)

test_that("filter_sorting_data with NA siccd values preserves those rows", {
  data <- make_sorting_data()
  data$siccd[1:3] <- NA_integer_
  result <- filter_sorting_data(
    data,
    filter_options = filter_options(exclude_financials = TRUE),
    quiet = TRUE
  )
  # NA rows should be kept (not removed)
  expect_true(any(is.na(result$siccd)))
})

set.seed(42)

# Minimal stock-return panel used throughout these tests
panel <- tibble::tibble(
  permno = rep(1:20, each = 6),
  date = rep(
    seq.Date(as.Date("2020-01-01"), by = "month", length.out = 6),
    20
  ),
  ret_excess = rnorm(120),
  mktcap_lag = runif(120, 100, 1000),
  bm = rnorm(120)
)

test_that("check_new_col does not warn for genuinely new column names", {
  expect_no_warning(check_new_col(panel, "bm_lag"))
})

test_that("check_new_col does not warn when cols is empty", {
  expect_no_warning(check_new_col(panel, character(0)))
})

test_that("check_new_col warns when ret_excess would be overwritten", {
  expect_warning(
    check_new_col(panel, "ret_excess"),
    "[Oo]verwrit"
  )
})

test_that(
  paste0(
    "check_new_col warns when multiple ",
    "portfolio columns would be overwritten"
  ),
  {
    data_with_portfolios <- tibble::add_column(
      panel,
      portfolio_main = NA_integer_,
      portfolio_secondary = NA_integer_
    )
    expect_warning(
      check_new_col(
        data_with_portfolios,
        c("portfolio_main", "portfolio_secondary")
      ),
      "[Oo]verwrit"
    )
  }
)

test_that("check_new_col warning message contains the offending column name", {
  expect_warning(
    check_new_col(panel, "mktcap_lag"),
    "mktcap_lag"
  )
})

test_that(
  paste0(
    "check_new_col uses the 'Overwriting' ",
    "branch when caller has no name"
  ),
  {
    expect_warning(
      (function() check_new_col(panel, "ret_excess"))(),
      "Overwriting"
    )
  }
)

test_that(
  paste0(
    "check_new_col emits exactly one warning ",
    "even for multiple overlapping cols"
  ),
  {
    data_with_both <- tibble::add_column(
      panel,
      bm_lag = NA_real_,
      ret_excess_lag = NA_real_
    )
    w <- capture_warnings(
      check_new_col(data_with_both, c("bm_lag", "ret_excess_lag"))
    )
    expect_length(w, 1)
  }
)

test_that(
  paste0(
    "check_new_col warns only for existing ",
    "columns, not new ones in the same call"
  ),
  {
    w <- capture_warnings(check_new_col(panel, c("bm", "bm_lag")))
    expect_length(w, 1)
  }
)

test_that("check_new_col returns the input data unchanged when no overlap", {
  result <- check_new_col(panel, "bm_lag")
  expect_identical(result, panel)
})

test_that(
  paste0(
    "check_new_col returns the input data ",
    "unchanged when overlap triggers a warning"
  ),
  {
    result <- suppressWarnings(check_new_col(panel, "ret_excess"))
    expect_identical(result, panel)
  }
)

test_that("add_lagged_columns warns when bm_lag already exists in data", {
  data_with_lag <- tibble::add_column(panel, bm_lag = NA_real_)
  expect_warning(
    add_lagged_columns(
      data_with_lag,
      cols = "bm",
      lag = months(3),
      by = "permno"
    ),
    "overwrites"
  )
})

test_that(
  paste0(
    "add_lagged_columns warns when .upper ",
    "already exists and max_lag is used"
  ),
  {
    data_with_upper <- tibble::add_column(panel, .upper = NA)
    expect_warning(
      add_lagged_columns(
        data_with_upper,
        cols = "bm",
        lag = months(1),
        max_lag = months(3),
        by = "permno"
      ),
      "overwrites"
    )
  }
)

test_that(
  paste0(
    "compute_long_short_returns warns when ",
    "..date already exists in data"
  ),
  {
    portfolio_returns <- tibble::tibble(
      date = rep(
        seq.Date(as.Date("2020-01-01"), by = "month", length.out = 6),
        each = 5
      ),
      portfolio = rep(1L:5L, 6),
      ret_excess = rnorm(30)
    )
    portfolio_returns[["..date"]] <- NA
    expect_warning(
      compute_long_short_returns(portfolio_returns),
      "overwrites"
    )
  }
)

test_that(
  paste0(
    "filter_sorting_data warns when ",
    ".size_cutoff already exists in data"
  ),
  {
    data_with_cutoff <- tibble::add_column(
      panel,
      .size_cutoff = NA_real_,
      exchange = "NYSE"
    )
    expect_warning(
      filter_sorting_data(
        data_with_cutoff,
        filter_options = filter_options(min_size_quantile = 0.2),
        quiet = TRUE
      ),
      "overwrites"
    )
  }
)

test_that(
  paste0(
    "join_lagged_values warns when new_data ",
    "already has .year and ff_adjustment is TRUE"
  ),
  {
    original <- dplyr::select(panel, permno, date, ret_excess, mktcap_lag)
    new <- dplyr::mutate(
      dplyr::select(panel, permno, date, bm),
      .year = NA_real_
    )

    expect_warning(
      join_lagged_values(
        original_data = original,
        new_data = new,
        id_keys = "permno",
        min_lag = months(1),
        max_lag = months(3),
        ff_adjustment = TRUE
      ),
      "overwrites"
    )
  }
)

test_that(
  paste0(
    "compute_portfolio_returns warns ",
    "when portfolio already exists in data"
  ),
  {
    data_with_portfolio <- tibble::add_column(panel, portfolio = NA_integer_)
    expect_warning(
      compute_portfolio_returns(
        data_with_portfolio,
        sorting_variable = "bm",
        sorting_method = "univariate",
        breakpoint_options_main = breakpoint_options(n_portfolios = 3),
        data_options = data_options(id = "permno", date = "date")
      ),
      "overwrites"
    )
  }
)

#' Open Tidy Finance Website or Specific Chapter in Browser
#'
#' Opens the main Tidy Finance website or a specific chapter within the site in
#' the user's default web browser. If a chapter is specified, the function
#' constructs the URL to access the chapter directly.
#'
#' @param chapter An optional character string specifying the chapter to open.
#'   If `NULL` (the default), the function opens the main page of Tidy Finance
#'   with R. If a chapter name is provided (e.g., "beta-estimation"), the
#'   function opens the corresponding chapter's page (e.g.,
#'   "beta-estimation.html"). If the chapter name does not exist, then the
#'   function opens the main page.
#'
#' @returns Invisible `NULL`. The function is called for its side effect of
#'   opening a web page.
#'
#' @family utility functions
#' @export
#'
#' @examples
#' open_tidy_finance_website()
#' open_tidy_finance_website("beta-estimation")
#'
open_tidy_finance_website <- function(chapter = NULL) {
  base_url <- "https://www.tidy-finance.org/r/"
  if (!is.null(chapter)) {
    tidy_finance_chapters <- list_tidy_finance_chapters()
    if (chapter %in% tidy_finance_chapters) {
      final_url <- paste0(base_url, chapter, ".html")
    } else {
      final_url <- base_url
    }
  } else {
    final_url <- base_url
  }
  utils::browseURL(final_url)
  invisible(NULL)
}

#' Disconnect Database Connection
#'
#' Safely disconnects an established database connection using the DBI package.
#'
#' @param con A database connection object created by DBI::dbConnect or any
#'   similar function that establishes a connection to a database.
#'
#' @returns `TRUE`, invisibly. Throws an error if the disconnection fails.
#'
#' @family WRDS functions
#' @export
#'
#' @examples
#' \dontrun{
#' con <- get_wrds_connection()
#' disconnect_connection(con)
#' }
disconnect_connection <- function(con) {
  DBI::dbDisconnect(con)
}

#' Check That New Columns Do Not Overwrite Existing Data
#'
#' Internal helper used before creating columns in a data frame to warn about
#' silently overwriting user-supplied data. Warns with an informative message
#' if any of the supplied column names already exist in `data`.
#'
#' @param data A data frame (or tibble) that is about to receive new columns.
#' @param cols A character vector of one or more column names that are about
#'   to be created.
#'
#' @returns Invisibly returns `data` to allow the helper to be used inside a
#'   pipe. When a column would be overwritten, the function issues a warning.
#'
#' @keywords internal
#' @noRd
check_new_col <- function(data, cols, call = rlang::caller_env()) {
  existing <- cols[cols %in% colnames(data)]
  if (length(existing) > 0L) {
    caller_name <- rlang::call_name(rlang::caller_call())
    cli::cli_warn(
      c(
        if (!is.null(caller_name)) {
          "{.fn {caller_name}} overwrites {length(existing)} existing column{?s}: {.val {existing}}." # nolint
        } else {
          "Overwriting {length(existing)} existing column{?s}: {.val {existing}}." # nolint
        },
        "i" = "Consider renaming or removing existing columns."
      ),
      call = call,
    )
  }
  invisible(data)
}

#' Download and Process Fama-French Factor Data
#'
#' Downloads and processes Fama-French factor data based on the specified
#' dataset name and date range. The function requires the `frenchdata`
#' package to download the data. It processes the raw data into a structured
#' format, including date conversion, scaling factor values, and filtering by
#' the specified date range.
#'
#' If there are multiple tables in the raw Fama-French data (e.g.,
#' value-weighted and equal-weighted returns), then the function only returns
#' the first table because these are the most popular. Please use the
#' `frenchdata` package directly if you need less commonly used tables.
#'
#' @param dataset The name of the Fama-French dataset to download (e.g.,
#'   "Fama/French 3 Factors").
#' @param start_date Optional. A character string or Date object in
#'   "YYYY-MM-DD" format specifying the start date for the data. If not
#'   provided, the full dataset is returned.
#' @param end_date Optional. A character string or Date object in "YYYY-MM-DD"
#'   format specifying the end date for the data. If not provided, the full
#'   dataset is returned.
#' @param type `r lifecycle::badge("deprecated")` Use `dataset` instead.
#'
#' @returns A tibble with processed factor data, including the date, risk-free
#'   rate, market excess return, and other factors, filtered by the specified
#'   date range.
#'
#' @references
#'   Fama, E. F., & French, K. R. (1993). Common risk factors in the returns on
#'   stocks and bonds. *Journal of Financial Economics*, 33(1), 3-56.
#'   \doi{10.1016/0304-405X(93)90023-5}
#'
#'   Fama, E. F., & French, K. R. (2015). A five-factor asset pricing model.
#'   *Journal of Financial Economics*, 116(1), 1-22.
#'   \doi{10.1016/j.jfineco.2014.10.010}
#'
#'   Carhart, M. M. (1997). On persistence in mutual fund performance.
#'   *Journal of Finance*, 52(1), 57-82.
#'   \doi{10.1111/j.1540-6261.1997.tb03808.x}
#'
#' @family download functions
#' @export
#'
#' @examples
#' \donttest{
#'   download_data_factors_ff(
#'     "Fama/French 3 Factors", "2000-01-01", "2020-12-31"
#'   )
#'   download_data_factors_ff(
#'     "10 Industry Portfolios", "2000-01-01", "2020-12-31"
#'   )
#' }
download_data_factors_ff <- function(
  dataset = NULL,
  start_date = NULL,
  end_date = NULL,
  type = deprecated()
) {
  # Handle explicit type argument
  if (lifecycle::is_present(type)) {
    lifecycle::deprecate_warn(
      when = "0.5.0",
      what = "download_data_factors_ff(type)",
      with = "download_data_factors_ff(dataset)"
    )
    parsed <- parse_type_to_domain_dataset(type)
    dataset <- parsed$dataset
  }

  # Handle legacy type passed as dataset argument
  if (!is.null(dataset) && is_legacy_type_ff(dataset)) {
    lifecycle::deprecate_warn(
      when = "0.5.0",
      what = "download_data_factors_ff(type)",
      with = "download_data_factors_ff(dataset)",
      details = paste0(
        "Column type should be replaced with dataset name. ",
        "Use `list_supported_types(domain = 'Fama-French')` to see the mapping."
      )
    )
    parsed <- parse_type_to_domain_dataset(dataset)
    dataset <- parsed$dataset
  }

  if (is.null(dataset)) {
    cli::cli_abort("Argument {.arg dataset} is required.")
  }

  check_supported_dataset_ff(dataset)

  dates <- validate_dates(start_date, end_date)
  start_date <- dates$start_date
  end_date <- dates$end_date

  raw_data <- handle_download_error(
    function() suppressMessages(frenchdata::download_french_data(dataset)),
    fallback = tibble(
      date = Date()
    )
  )

  if (!inherits(raw_data, "french_dataset")) {
    cli::cli_inform("Returning an empty data set due to download failure.")
    return(raw_data)
  }

  raw_data <- raw_data$subsets$data[[1]]

  frequency <- determine_frequency_ff(dataset)

  if (frequency == "monthly") {
    processed_data <- raw_data |>
      mutate(date = floor_date(ymd(paste0(date, "01")), "month"))
  } else if (frequency %in% c("daily", "weekly")) {
    processed_data <- raw_data |>
      mutate(date = ymd(date))
  } else {
    cli::cli_abort(
      "This dataset has neither daily, weekly, nor monthly frequency."
    )
  }

  processed_data <- processed_data |>
    mutate(
      across(-date, ~ na_if(., -99.99)),
      across(-date, ~ na_if(., -999)),
      across(-date, ~ . / 100)
    )

  colnames_lower <- tolower(colnames(processed_data))
  colnames_clean <- gsub("-rf", "_excess", colnames_lower, fixed = TRUE)
  colnames_clean <- gsub("rf", "risk_free", colnames_clean, fixed = TRUE)
  colnames(processed_data) <- colnames_clean

  if (!is.null(start_date) && !is.null(end_date)) {
    processed_data <- processed_data |>
      filter(between(date, start_date, end_date))
  }

  processed_data
}

#' Download and Process Global Q Factor Data
#'
#' Downloads and processes Global Q factor data based on the
#' specified dataset, date range, and source URL. The processing
#' includes date conversion, renaming variables to a standardized
#' format, scaling factor values, and filtering by the specified
#' date range.
#'
#' @param dataset The name of the dataset to download (e.g.,
#'   "q5_factors_daily_2023.csv", "q5_factors_monthly_2023.csv").
#' @param start_date Optional. A character string or Date object in
#'   "YYYY-MM-DD" format specifying the start date for the data. If
#'   not provided, the full dataset is returned.
#' @param end_date Optional. A character string or Date object in
#'   "YYYY-MM-DD" format specifying the end date for the data. If not
#'   provided, the full dataset is returned.
#' @param type `r lifecycle::badge("deprecated")` Use `dataset`
#'   instead.
#' @param url The base URL from which to download the dataset files.
#'
#' @returns A tibble with processed factor data, including the date,
#'   risk-free rate, market excess return, and other factors,
#'   filtered by the specified date range.
#'
#' @references
#'   Hou, K., Xue, C., & Zhang, L. (2015). Digesting anomalies: An investment
#'   approach. *Review of Financial Studies*, 28(3), 650-705.
#'   \doi{10.1093/rfs/hhu068}
#'
#'   Hou, K., Mo, H., Xue, C., & Zhang, L. (2019). Which factors?
#'   *Review of Finance*, 23(1), 1-35.
#'   \doi{10.1093/rof/rfy032}
#'
#' @family download functions
#' @export
#'
#' @examples
#' \donttest{
#' download_data_factors_q("q5_factors_daily_2024", "2020-01-01", "2020-12-31")
#' download_data_factors_q("q5_factors_annual_2024")
#' }
download_data_factors_q <- function(
  dataset = NULL,
  start_date = NULL,
  end_date = NULL,
  type = deprecated(),
  url = "https://global-q.org/uploads/1/2/2/6/122679606/"
) {
  # Handle explicit type argument
  if (lifecycle::is_present(type)) {
    lifecycle::deprecate_warn(
      when = "0.5.0",
      what = "download_data_factors_q(type)",
      with = "download_data_factors_q(dataset)"
    )
    parsed <- parse_type_to_domain_dataset(type)
    dataset <- parsed$dataset
  }

  # Handle legacy type passed as dataset argument
  if (!is.null(dataset) && is_legacy_type_q(dataset)) {
    lifecycle::deprecate_warn(
      when = "0.5.0",
      what = "download_data_factors_q(type)",
      with = "download_data_factors_q(dataset)",
      details = paste0(
        "Column type should be replaced with dataset name. ",
        "Use `list_supported_types(domain = 'Global Q')` to see the mapping."
      )
    )
    parsed <- parse_type_to_domain_dataset(dataset)
    dataset <- parsed$dataset
  }

  if (is.null(dataset)) {
    cli::cli_abort("Argument {.arg dataset} is required.")
  }

  check_supported_dataset_q(dataset)

  dates <- validate_dates(start_date, end_date)
  start_date <- dates$start_date
  end_date <- dates$end_date

  raw_data <- handle_download_error(
    function(url) {
      suppressWarnings(
        suppressMessages(utils::read.csv(url)) |> as_tibble()
      )
    },
    paste0(url, dataset, ".csv"),
    fallback = tibble(
      date = Date()
    )
  )

  if (nrow(raw_data) == 0) {
    cli::cli_inform("Returning an empty data set due to download failure.")
    return(raw_data)
  }

  frequency <- determine_frequency_q(dataset)

  if (frequency == "monthly") {
    processed_data <- raw_data |>
      mutate(date = ymd(paste(year, month, "01", sep = "-"))) |>
      select(-c(year, month))
  } else if (frequency == "daily") {
    processed_data <- raw_data |>
      mutate(DATE = ymd(DATE))
  } else if (frequency == "annual") {
    processed_data <- raw_data |>
      mutate(date = year)
  } else if (frequency %in% c("weekly", "quarterly")) {
    processed_data <- raw_data |>
      mutate(date = ymd(paste(year, month, day, sep = "-"))) |>
      select(-c(year, month, day))
  }

  processed_data <- processed_data |>
    rename_with(~ sub("R_", "", ., fixed = TRUE)) |>
    rename_with(tolower) |>
    mutate(across(-date, ~ . / 100)) |>
    select(date, risk_free = f, mkt_excess = mkt, everything())

  if (!is.null(start_date) && !is.null(end_date)) {
    processed_data <- processed_data |>
      filter(between(date, start_date, end_date))
  }

  processed_data
}

# Helper functions --------------------------------------------------------

#' Check if a string is a legacy Fama-French type
#' @noRd
is_legacy_type_ff <- function(x) {
  ff_types <- dplyr::bind_rows(
    list_supported_types_ff(),
    list_supported_types_ff_legacy()
  )
  x %in% ff_types$type
}

#' Check if a string is a legacy Global Q type
#' @noRd
is_legacy_type_q <- function(x) {
  q_types <- list_supported_types_q()
  x %in% q_types$type
}

#' Determine frequency from Fama-French dataset name
#' @noRd
determine_frequency_ff <- function(dataset) {
  if (grepl("\\[Daily\\]", dataset)) {
    "daily"
  } else if (grepl("\\[Weekly\\]", dataset)) {
    "weekly"
  } else {
    "monthly"
  }
}

#' Determine frequency from Global Q dataset name
#' @noRd
determine_frequency_q <- function(dataset) {
  if (grepl("daily", dataset, ignore.case = TRUE)) {
    "daily"
  } else if (grepl("weekly", dataset, ignore.case = TRUE)) {
    "weekly"
  } else if (grepl("monthly", dataset, ignore.case = TRUE)) {
    "monthly"
  } else if (grepl("quarterly", dataset, ignore.case = TRUE)) {
    "quarterly"
  } else if (grepl("annual", dataset, ignore.case = TRUE)) {
    "annual"
  } else {
    cli::cli_abort(
      "Cannot determine frequency from dataset name: {.val {dataset}}"
    )
  }
}

#' Check if Fama-French dataset is supported
#' @noRd
check_supported_dataset_ff <- function(dataset) {
  ff_types <- dplyr::bind_rows(
    list_supported_types_ff(),
    list_supported_types_ff_legacy()
  )

  if (!dataset %in% ff_types$dataset_name) {
    cli::cli_abort(c(
      "Unsupported Fama-French dataset: {.val {dataset}}",
      "i" = "Use {.fn list_supported_types} with {.arg domain = 'Fama-French'}",
      "to see available datasets."
    ))
  }
}

#' Check if Global Q dataset is supported
#' @noRd
check_supported_dataset_q <- function(dataset) {
  q_types <- list_supported_types_q()

  if (!dataset %in% q_types$dataset_name) {
    cli::cli_abort(c(
      "Unsupported Global Q dataset: {.val {dataset}}",
      "i" = "Use {.fn list_supported_types} with {.arg domain = 'Global Q'}",
      "to see available datasets."
    ))
  }
}

test_that("download_data_factors_q gracefully handles broken url", {
  skip_if_offline()
  skip_on_cran()
  expect_message(
    download_data_factors_q(
      dataset = "q5_factors_daily_2024",
      start_date = "2020-01-01",
      end_date = "2022-12-01",
      url = "test"
    ),
    "Returning an empty data set due to download failure."
  )
})

test_that("download_data_factors_q handles start_date after end_date", {
  expect_error(
    download_data_factors_q(
      dataset = "q5_factors_daily_2024",
      start_date = "2021-12-31",
      end_date = "2020-01-01"
    ),
    regexp = "`start_date` cannot be after `end_date`"
  )
})

#' Download and Process Macro Predictor Data
#'
#' Downloads and processes macroeconomic predictor data based on the
#' specified dataset (monthly, quarterly, or annual), date range, and
#' source URL. The function downloads the data from a Google Sheets
#' export link. It processes the raw data into a structured format,
#' calculating additional financial metrics and filtering by the
#' specified date range.
#'
#' @param dataset The dataset to download ("monthly", "quarterly",
#'   "annual").
#' @param start_date Optional. A character string or Date object in
#'   "YYYY-MM-DD" format specifying the start date for the data. If
#'   not provided, the full dataset is returned.
#' @param end_date Optional. A character string or Date object in
#'   "YYYY-MM-DD" format specifying the end date for the data. If not
#'   provided, the full dataset is returned.
#' @param type `r lifecycle::badge("deprecated")` Use `dataset`
#'   instead.
#' @param sheet_id The Google Sheets ID from which to download the
#'   dataset, with the default "1bM7vCWd3WOt95Sf9qjLPZjoiafgF_8EG".
#'
#' @returns A tibble with processed data, filtered by the specified
#'   date range and including financial metrics.
#'
#' @references
#'   Welch, I., & Goyal, A. (2008). A comprehensive look at the empirical
#'   performance of equity premium prediction. *Review of Financial Studies*,
#'   21(4), 1455-1508. \doi{10.1093/rfs/hhm014}
#'
#' @family download functions
#' @export
#'
#' @examples
#' \donttest{
#'   download_data_macro_predictors("monthly")
#'   download_data_macro_predictors("quarterly", "2000-01-01", "2020-12-31")
#' }
#'
download_data_macro_predictors <- function(
  dataset = NULL,
  start_date = NULL,
  end_date = NULL,
  type = deprecated(),
  sheet_id = "1bM7vCWd3WOt95Sf9qjLPZjoiafgF_8EG"
) {
  # Handle explicit type argument
  if (lifecycle::is_present(type)) {
    lifecycle::deprecate_warn(
      when = "0.5.0",
      what = "download_data_macro_predictors(type)",
      details = "Use the `dataset` argument instead."
    )
    dataset <- sub("^macro_predictors_", "", type)
  }

  # Handle legacy type passed as dataset argument
  if (!is.null(dataset) && is_legacy_type_macro_predictors(dataset)) {
    lifecycle::deprecate_warn(
      when = "0.5.0",
      what = "download_data_macro_predictors(type)",
      details = paste0(
        "The `type` argument is deprecated. ",
        "Use `dataset` instead (e.g., 'monthly'",
        "instead of 'macro_predictors_monthly')."
      )
    )
    dataset <- sub("^macro_predictors_", "", dataset)
  }

  if (is.null(dataset)) {
    cli::cli_abort("Argument {.arg dataset} is required.")
  }

  check_supported_dataset_macro_predictors(dataset)

  dates <- validate_dates(start_date, end_date)
  start_date <- dates$start_date
  end_date <- dates$end_date

  build_macro_predictors_url <- function(sheet_name) {
    paste0(
      "https://docs.google.com/spreadsheets/d/",
      sheet_id,
      "/gviz/tq?tqx=out:csv&sheet=",
      sheet_name
    )
  }

  if (dataset == "monthly") {
    raw_data <- handle_download_error(
      function() {
        suppressWarnings(
          as_tibble(read.csv(build_macro_predictors_url("Monthly")))
        )
      },
      fallback = tibble()
    )

    if (nrow(raw_data) == 0) {
      cli::cli_inform("Returning an empty data set due to download failure.")
      return(raw_data)
    }

    processed_data <- raw_data |>
      mutate(date = ym(yyyymm))
  } else if (dataset == "quarterly") {
    raw_data <- handle_download_error(
      function() {
        suppressWarnings(
          as_tibble(read.csv(build_macro_predictors_url("Quarterly")))
        )
      },
      fallback = tibble()
    )

    if (nrow(raw_data) == 0) {
      cli::cli_inform("Returning an empty data set due to download failure.")
      return(raw_data)
    }

    processed_data <- raw_data |>
      mutate(
        year = substr(yyyyq, 1, 4),
        quarter = substr(yyyyq, 5, 5),
        month = as.integer(quarter) * 3 - 2,
        date = as.Date(paste0(year, "-", month, "-01"))
      )
  } else if (dataset == "annual") {
    raw_data <- handle_download_error(
      function() {
        suppressWarnings(
          as_tibble(read.csv(build_macro_predictors_url("Annual")))
        )
      },
      fallback = tibble()
    )

    if (nrow(raw_data) == 0) {
      cli::cli_inform("Returning an empty data set due to download failure.")
      return(raw_data)
    }

    processed_data <- raw_data |>
      mutate(date = as.Date(paste0(yyyy, "-01-01")))
  }

  processed_data <- processed_data |>
    mutate(
      across(where(is.character), ~ as.numeric(gsub("\\,", "", .))),
      IndexDiv = Index + D12,
      logret = log(IndexDiv) - log(lag(IndexDiv)),
      rp_div = lead(logret - Rfree, 1),
      log_d12 = log(D12),
      log_e12 = log(E12),
      dp = log_d12 - log(Index),
      dy = log_d12 - log(lag(Index)),
      ep = log_e12 - log(Index),
      de = log_d12 - log_e12,
      tms = lty - tbl,
      dfy = BAA - AAA
    ) |>
    select(
      date,
      rp_div,
      dp,
      dy,
      ep,
      de,
      svar,
      bm = `b.m`,
      ntis,
      tbl,
      lty,
      ltr,
      tms,
      dfy,
      infl
    ) |>
    tidyr::drop_na()

  if (!is.null(start_date) && !is.null(end_date)) {
    processed_data <- processed_data |>
      filter(between(date, start_date, end_date))
  }

  processed_data
}

#' Check if a string is a legacy macro predictors type
#' @noRd
is_legacy_type_macro_predictors <- function(x) {
  grepl("^macro_predictors_", x)
}

#' Check if macro predictors dataset is supported
#' @noRd
check_supported_dataset_macro_predictors <- function(dataset) {
  supported_datasets <- c("monthly", "quarterly", "annual")

  if (!dataset %in% supported_datasets) {
    cli::cli_abort(c(
      "Unsupported macro predictors dataset: {.val {dataset}}",
      "i" = "Supported datasets: {.val {supported_datasets}}"
    ))
  }
}

#' Download Constituent Data
#'
#' Downloads and processes the constituent data for a specified financial
#' index. The data is fetched from a remote CSV file, filtered, and cleaned to
#' provide relevant information about constituents.
#'
#' @details
#' The function retrieves the URL of the CSV file for the specified index from
#' ETF sites, then sends an HTTP GET request to download the CSV file, and
#' processes the CSV file to extract equity constituents.
#'
#' The approach is inspired by `tidyquant::tq_index()`, which uses a different
#' wrapper around other ETFs.
#'
#' @param index A character string specifying the name of the financial index
#'   for which to download constituent data. The index must be one of the
#'   supported indexes listed by [list_supported_indexes()].
#'
#' @returns A tibble with five columns:
#' \describe{
#'   \item{symbol}{The ticker symbol of the equity constituent.}
#'   \item{name}{The name of the equity constituent.}
#'   \item{location}{The location where the company is based.}
#'   \item{exchange}{The exchange where the equity is traded.}
#'   \item{currency}{The currency in which the equity is traded, derived
#'     from the exchange.}
#' }
#' The tibble is filtered to exclude non-equity entries, blacklisted symbols,
#' empty names, and any entries containing the index name or "CASH".
#'
#' @family download functions
#' @export
#'
#' @examples
#' \donttest{
#'   download_data_constituents("DAX")
#' }
#'
download_data_constituents <- function(index) {
  symbol_blacklist <- c("", "USD", "GXU4", "EUR", "MARGIN_EUR", "MLIFT")

  supported_indexes <- list_supported_indexes()

  if (!(index %in% supported_indexes$index)) {
    cli::cli_abort(
      paste(
        "The index '{index}' is not supported.",
        "Please use one of the supported indexes:",
        "{paste(supported_indexes$index, collapse = ', ')}"
      )
    )
  }

  url <- supported_indexes$url[supported_indexes$index == index]

  user_agent <- get_random_user_agent()

  response <- handle_download_error(
    function() {
      httr2::request(url) |>
        httr2::req_error(is_error = \(resp) FALSE) |>
        httr2::req_user_agent(user_agent) |>
        httr2::req_perform()
    },
    fallback = NULL
  )

  if (!is.null(response)) {
    if (response$status_code != 200) {
      cli::cli_abort(
        paste(
          "Failed to download data for index {.arg index}.",
          "Please check the index name or try again later."
        )
      )
    } else {
      constituents_raw <- suppressWarnings(httr2::resp_body_string(response)) |>
        textConnection() |>
        read.csv(
          skip = supported_indexes$skip[supported_indexes$index == index]
        )

      column_names <- paste0(colnames(constituents_raw), collapse = "|")

      if (grepl(column_names, "Anlageklasse")) {
        constituents_processed <- constituents_raw |>
          filter(Anlageklasse == "Aktien") |>
          select(
            symbol = "Emittententicker",
            name = "Name",
            location = "Standort",
            exchange = "B\u00f6rse"
          )
      }
      if (grepl(column_names, "Asset.Class")) {
        constituents_processed <- constituents_raw |>
          filter(Asset.Class == "Equity") |>
          select(
            symbol = "Ticker",
            name = "Name",
            location = "Location",
            exchange = "Exchange"
          )
      }

      constituents_processed <- constituents_processed |>
        filter(!symbol %in% symbol_blacklist) |>
        filter(name != "") |>
        filter(!grepl(tolower(index), tolower(name))) |>
        filter(!grepl("CASH", name)) |>
        filter(!grepl(tolower(gsub("\\s+", "", index)), tolower(name))) |>
        as_tibble() |>
        mutate(
          symbol = case_when(
            name == "NATIONAL BANK OF CANADA" ~ "NA",
            TRUE ~ symbol
          ),
          symbol = gsub(" ", "-", symbol),
          symbol = gsub("/", "-", symbol)
        ) |>
        mutate(
          symbol = case_when(
            exchange %in% c("Xetra", "Deutsche B\u00f6rse AG") ~
              paste0(symbol, ".DE"),
            exchange == "Boerse Berlin" ~ paste0(symbol, ".BE"),
            exchange == "Borsa Italiana" ~ paste0(symbol, ".MI"),
            exchange == "Nyse Euronext - Euronext Paris" ~
              paste0(symbol, ".PA"),
            exchange == "Euronext Amsterdam" ~ paste0(symbol, ".AS"),
            exchange == "Nasdaq Omx Helsinki Ltd." ~ paste0(symbol, ".HE"),
            exchange == "Singapore Exchange" ~ paste0(symbol, ".SI"),
            exchange == "Asx - All Markets" ~ paste0(symbol, ".AX"),
            exchange == "London Stock Exchange" ~ paste0(symbol, ".L"),
            exchange == "SIX Swiss Exchange" ~ paste0(symbol, ".SW"),
            exchange == "Tel Aviv Stock Exchange" ~ paste0(symbol, ".TA"),
            exchange == "Tokyo Stock Exchange" ~ paste0(symbol, ".T"),
            exchange == "Hong Kong Stock Exchange" ~ paste0(symbol, ".HK"),
            exchange == "Toronto Stock Exchange" ~ paste0(symbol, ".TO"),
            exchange == "Euronext Brussels" ~ paste0(symbol, ".BR"),
            exchange == "Euronext Lisbon" ~ paste0(symbol, ".LS"),
            exchange == "Bovespa" ~ paste0(symbol, ".SA"),
            exchange == "Mexican Stock Exchange" ~ paste0(symbol, ".MX"),
            exchange == "Stockholm Stock Exchange" ~ paste0(symbol, ".ST"),
            exchange == "Oslo Stock Exchange" ~ paste0(symbol, ".OL"),
            exchange == "Johannesburg Stock Exchange" ~ paste0(symbol, ".J"),
            exchange == "Korea Exchange" ~ paste0(symbol, ".KS"),
            exchange == "Shanghai Stock Exchange" ~ paste0(symbol, ".SS"),
            exchange == "Shenzhen Stock Exchange" ~ paste0(symbol, ".SZ"),
            TRUE ~ symbol
          )
        ) |>
        mutate(symbol = gsub("\\.\\.", "\\.", symbol)) |>
        mutate(
          currency = case_when(
            exchange %in%
              c("Xetra", "Boerse Berlin", "Deutsche B\u00f6rse AG") ~
              "EUR",
            exchange == "Borsa Italiana" ~ "EUR",
            exchange == "Nyse Euronext - Euronext Paris" ~ "EUR",
            exchange == "Euronext Amsterdam" ~ "EUR",
            exchange == "Nasdaq Omx Helsinki Ltd." ~ "EUR",
            exchange == "Singapore Exchange" ~ "SGD",
            exchange == "Asx - All Markets" ~ "AUD",
            exchange == "London Stock Exchange" ~ "GBP",
            exchange == "SIX Swiss Exchange" ~ "CHF",
            exchange == "Tel Aviv Stock Exchange" ~ "ILS",
            exchange == "Tokyo Stock Exchange" ~ "JPY",
            exchange == "Hong Kong Stock Exchange" ~ "HKD",
            exchange == "Toronto Stock Exchange" ~ "CAD",
            exchange == "Euronext Brussels" ~ "EUR",
            exchange == "Euronext Lisbon" ~ "EUR",
            exchange == "Bovespa" ~ "BRL",
            exchange == "Mexican Stock Exchange" ~ "MXN",
            exchange == "Stockholm Stock Exchange" ~ "SEK",
            exchange == "Oslo Stock Exchange" ~ "NOK",
            exchange == "Johannesburg Stock Exchange" ~ "ZAR",
            exchange == "Korea Exchange" ~ "KRW",
            exchange == "Shanghai Stock Exchange" ~ "CNY",
            exchange == "Shenzhen Stock Exchange" ~ "CNY",
            TRUE ~ "USD"
          )
        )

      constituents_processed
    }
  }
}

test_that("download_data_constituents works for a supported index", {
  skip_if_offline()
  skip_on_cran()
  index <- "Dow Jones Industrial Average"

  expect_silent({
    constituents_data <- download_data_constituents(index)
  })

  expect_true(all(c("symbol", "name") %in% colnames(constituents_data)))

  expect_true(all(!is.na(constituents_data$symbol)))
  expect_true(all(!is.na(constituents_data$name)))
})

test_that(
  paste("download_data_constituents handles unavailable resource gracefully"),
  {
    mock_list_supported_indexes <- function() {
      tibble::tibble(
        index = c("SAMPLE_INDEX"),
        url = c("http://nonexistent-url.example.com"),
        skip = c(0)
      )
    }

    with_mocked_bindings(
      list_supported_indexes = mock_list_supported_indexes,
      {
        expect_message(
          download_data_constituents(index = "SAMPLE_INDEX"),
          regex = paste(
            "The resource may not be available, or the URL may have changed"
          )
        )
      }
    )
  }
)

#' @keywords internal
#' @import dplyr
#' @import lubridate
#' @import tibble
#' @import DBI
#' @import httr2
"_PACKAGE"

## usethis namespace: start
#' @importFrom lifecycle deprecated
#' @importFrom purrr partial map map_dbl
#' @importFrom stats quantile setNames weighted.mean as.formula na.omit ave
#' @importFrom utils read.csv
#' @importFrom tidyr pivot_longer unnest nest fill drop_na
#' @importFrom rlang := enquo syms quo_is_null
#' @importFrom jsonlite fromJSON
#' @importFrom stringr str_match
#' @importFrom frenchdata download_french_data get_french_data_list
#' @importFrom arrow read_parquet write_parquet
#' @importFrom RPostgres Postgres
#' @importFrom sandwich NeweyWest
#' @importFrom dbplyr translate_sql
## usethis namespace: end
NULL

#' Download Data from WRDS CRSP
#'
#' Downloads and processes stock return data from the CRSP database for a
#' specified period. Users can choose between monthly and daily datasets.
#' The function also adjusts returns for delisting and calculates market
#' capitalization and excess returns over the risk-free rate.
#'
#' @param dataset A string specifying the CRSP dataset to download:
#'   "crsp_monthly" or "crsp_daily".
#' @param start_date Optional. A character string or Date object in
#'   "YYYY-MM-DD" format specifying the start date for the data. If not
#'   provided, a subset of the dataset is returned.
#' @param end_date Optional. A character string or Date object in
#'   "YYYY-MM-DD" format specifying the end date for the data. If not
#'   provided, a subset of the dataset is returned.
#' @param type `r lifecycle::badge("deprecated")` Use `dataset` instead.
#' @param batch_size An optional integer specifying the batch size for
#'   processing daily data, with a default of 500.
#' @param version An optional character specifying which CRSP version to
#'   use. "v2" (the default) uses the updated second version of CRSP, and
#'   "v1" downloads the legacy version of CRSP.
#' @param additional_columns Additional columns from the CRSP monthly or
#'   daily data as a character vector.
#' @param add_ccm_links A logical indicating whether CRSP-Compustat links
#'   should be added automatically using [download_data_wrds_ccm_links()].
#' @param adjust_volume A logical indicating whether daily CRSP trading
#'   volume data should be adjusted according to Gao & Ritter (2010).
#'
#' @returns A data frame containing CRSP stock returns, adjusted for
#'   delistings, along with calculated market capitalization and excess returns
#'   over the risk-free rate. The structure of the returned data frame depends
#'   on the selected dataset.
#'
#' @references
#'   Gao, X., & Ritter, J. R. (2010). The marketing of seasoned equity
#'   offerings. *Journal of Financial Economics*, 97(1), 33-52.
#'   \doi{10.1016/j.jfineco.2010.03.007}
#'
#' @family WRDS functions
#' @export
#'
#' @examples
#' \dontrun{
#' crsp_monthly <- download_data_wrds_crsp(
#'   "crsp_monthly",
#'   "2020-11-01",
#'   "2020-12-31"
#' )
#' crsp_daily <- download_data_wrds_crsp(
#'   "crsp_daily",
#'   "2020-12-01",
#'   "2020-12-31"
#' )
#'
#' # Add additional columns
#' download_data_wrds_crsp(
#'   "crsp_monthly",
#'   "2020-11-01",
#'   "2020-12-31",
#'   additional_columns = c("mthvol", "mthvolflg")
#' )
#' }
download_data_wrds_crsp <- function(
  dataset = NULL,
  start_date = NULL,
  end_date = NULL,
  type = deprecated(),
  batch_size = 500,
  version = "v2",
  additional_columns = NULL,
  add_ccm_links = FALSE,
  adjust_volume = FALSE
) {
  # Handle explicit type argument
  if (lifecycle::is_present(type)) {
    lifecycle::deprecate_warn(
      when = "0.5.0",
      what = "download_data_wrds_crsp(type)",
      details = "Use the `dataset` argument instead."
    )
    dataset <- sub("^wrds_", "", type)
  }

  # Handle legacy type passed as dataset argument
  if (!is.null(dataset) && is_legacy_type_wrds(dataset)) {
    lifecycle::deprecate_warn(
      when = "0.5.0",
      what = "download_data_wrds_crsp(type)",
      details = paste0(
        "The `type` argument is deprecated. ",
        "Use `dataset` instead (e.g., 'crsp_monthly' instead",
        "of 'wrds_crsp_monthly')."
      )
    )
    dataset <- sub("^wrds_", "", dataset)
  }

  if (is.null(dataset)) {
    cli::cli_abort("Argument {.arg dataset} is required.")
  }

  check_supported_dataset_wrds_crsp(dataset)

  batch_size <- as.integer(batch_size)
  if (batch_size <= 0) {
    cli::cli_abort("{.arg batch_size} must be an integer larger than 0.")
  }

  if (!(version %in% c("v1", "v2"))) {
    cli::cli_abort(
      "{.arg version} must be a character equal to {.str v1} or {.str v2}."
    )
  }

  dates <- validate_dates(start_date, end_date, use_default_range = TRUE)
  start_date <- dates$start_date
  end_date <- dates$end_date

  con <- get_wrds_connection()

  if (dataset == "crsp_monthly") {
    if (version == "v1") {
      msf_db <- tbl(con, I("crsp.msf"))
      msenames_db <- tbl(con, I("crsp.msenames"))
      msedelist_db <- tbl(con, I("crsp.msedelist"))

      msf_db_columns <- c(
        "permno",
        colnames(msf_db)[-which(colnames(msf_db) %in% colnames(msenames_db))]
      )

      first_crsp_date <- msenames_db |>
        group_by(permno) |>
        summarise(first_crsp_date = min(namedt, na.rm = TRUE)) |>
        collect()

      crsp_monthly <- msf_db |>
        filter(between(date, start_date, end_date)) |>
        select(all_of(msf_db_columns)) |>
        inner_join(
          msenames_db |>
            filter(shrcd %in% c(10, 11)),
          join_by(permno)
        ) |>
        filter(between(date, namedt, nameendt)) |>
        mutate(calculation_date = date, date = floor_date(date, "month")) |>
        left_join(
          msedelist_db |>
            select(permno, dlstdt, dlret, dlstcd) |>
            mutate(date = floor_date(dlstdt, "month")),
          join_by(permno, date)
        ) |>
        select(
          permno,
          date,
          calculation_date,
          ret,
          shrout,
          altprc,
          cfacpr,
          exchcd,
          siccd,
          dlret,
          dlstcd,
          all_of(additional_columns)
        ) |>
        collect() |>
        mutate(
          date = ymd(date),
          shrout = shrout * 1000
        )

      disconnect_connection(con)

      crsp_monthly <- crsp_monthly |>
        left_join(first_crsp_date, by = "permno") |>
        mutate(
          listing_age = pmax(
            as.integer(
              lubridate::interval(first_crsp_date, date) %/% months(1)
            ),
            0L
          )
        ) |>
        select(-first_crsp_date)

      crsp_monthly <- crsp_monthly |>
        mutate(
          mktcap = abs(shrout * altprc) / 10^6,
          mktcap = na_if(mktcap, 0)
        )

      mktcap_lag <- crsp_monthly |>
        mutate(date = date %m+% months(1)) |>
        select(permno, date, mktcap_lag = mktcap)

      crsp_monthly <- crsp_monthly |>
        left_join(mktcap_lag, join_by(permno, date))

      crsp_monthly <- crsp_monthly |>
        mutate(
          exchange = case_when(
            exchcd %in% c(1, 31) ~ "NYSE",
            exchcd %in% c(2, 32) ~ "AMEX",
            exchcd %in% c(3, 33) ~ "NASDAQ",
            .default = "Other"
          )
        )

      crsp_monthly <- crsp_monthly |>
        mutate(
          industry = case_when(
            siccd >= 1 & siccd <= 999 ~ "Agriculture",
            siccd >= 1000 & siccd <= 1499 ~ "Mining",
            siccd >= 1500 & siccd <= 1799 ~ "Construction",
            siccd >= 2000 & siccd <= 3999 ~ "Manufacturing",
            siccd >= 4000 & siccd <= 4899 ~ "Transportation",
            siccd >= 4900 & siccd <= 4999 ~ "Utilities",
            siccd >= 5000 & siccd <= 5199 ~ "Wholesale",
            siccd >= 5200 & siccd <= 5999 ~ "Retail",
            siccd >= 6000 & siccd <= 6799 ~ "Finance",
            siccd >= 7000 & siccd <= 8999 ~ "Services",
            siccd >= 9000 & siccd <= 9999 ~ "Public",
            .default = "Missing"
          )
        )

      crsp_monthly <- crsp_monthly |>
        mutate(
          ret_adj = case_when(
            is.na(dlstcd) ~ ret,
            !is.na(dlstcd) & !is.na(dlret) ~ dlret,
            dlstcd %in%
              c(500, 520, 580, 584) |
              (dlstcd >= 551 & dlstcd <= 574) ~
              -0.30,
            dlstcd == 100 ~ ret,
            .default = -1
          )
        ) |>
        select(-c(dlret, dlstcd))

      crsp_monthly <- crsp_monthly |>
        mutate(
          prc_adj = abs(na_if(altprc, 0)) / cfacpr,
          prc_adj = if_else(is.infinite(prc_adj), NA_real_, prc_adj)
        )

      risk_free_monthly <- download_data_risk_free(
        start_date = start_date,
        end_date = end_date
      )

      crsp_monthly <- crsp_monthly |>
        left_join(risk_free_monthly, join_by(date)) |>
        mutate(
          ret_excess = ret_adj - risk_free
        ) |>
        select(-risk_free)

      processed_data <- crsp_monthly |>
        tidyr::drop_na(ret_excess, mktcap)
    } else {
      msf_db <- tbl(con, I("crsp.msf_v2"))
      stksecurityinfohist_db <- tbl(con, I("crsp.stksecurityinfohist"))

      msf_db_columns <- c(
        "permno",
        colnames(msf_db)[
          -which(colnames(msf_db) %in% colnames(stksecurityinfohist_db))
        ]
      )

      first_crsp_date <- stksecurityinfohist_db |>
        group_by(permno) |>
        summarise(first_crsp_date = min(secinfostartdt, na.rm = TRUE)) |>
        collect()

      crsp_monthly <- msf_db |>
        filter(between(mthcaldt, start_date, end_date)) |>
        select(all_of(msf_db_columns)) |>
        inner_join(
          stksecurityinfohist_db |>
            filter(
              sharetype == "NS" &
                securitytype == "EQTY" &
                securitysubtype == "COM" &
                usincflg == "Y" &
                issuertype %in% c("ACOR", "CORP") &
                primaryexch %in% c("N", "A", "Q") &
                conditionaltype %in% c("RW", "NW") &
                tradingstatusflg == "A"
            ),
          join_by(permno)
        ) |>
        filter(between(mthcaldt, secinfostartdt, secinfoenddt)) |>
        mutate(date = floor_date(mthcaldt, "month")) |>
        select(
          permno,
          date,
          calculation_date = mthcaldt,
          ret = mthret,
          shrout,
          prc = mthprc,
          primaryexch,
          siccd,
          all_of(additional_columns)
        ) |>
        collect() |>
        mutate(
          date = ymd(date),
          shrout = shrout * 1000
        )

      disconnect_connection(con)

      crsp_monthly <- crsp_monthly |>
        left_join(first_crsp_date, by = "permno") |>
        mutate(
          listing_age = pmax(
            as.integer(
              lubridate::interval(first_crsp_date, date) %/% months(1)
            ),
            0L
          )
        ) |>
        select(-first_crsp_date)

      crsp_monthly <- crsp_monthly |>
        mutate(
          mktcap = shrout * prc / 10^6,
          mktcap = na_if(mktcap, 0)
        )

      mktcap_lag <- crsp_monthly |>
        mutate(date = date %m+% months(1)) |>
        select(permno, date, mktcap_lag = mktcap)

      crsp_monthly <- crsp_monthly |>
        left_join(mktcap_lag, join_by(permno, date))

      crsp_monthly <- crsp_monthly |>
        mutate(
          exchange = case_when(
            primaryexch == "N" ~ "NYSE",
            primaryexch == "A" ~ "AMEX",
            primaryexch == "Q" ~ "NASDAQ",
            .default = "Other"
          )
        )

      crsp_monthly <- crsp_monthly |>
        mutate(
          industry = case_when(
            siccd >= 1 & siccd <= 999 ~ "Agriculture",
            siccd >= 1000 & siccd <= 1499 ~ "Mining",
            siccd >= 1500 & siccd <= 1799 ~ "Construction",
            siccd >= 2000 & siccd <= 3999 ~ "Manufacturing",
            siccd >= 4000 & siccd <= 4899 ~ "Transportation",
            siccd >= 4900 & siccd <= 4999 ~ "Utilities",
            siccd >= 5000 & siccd <= 5199 ~ "Wholesale",
            siccd >= 5200 & siccd <= 5999 ~ "Retail",
            siccd >= 6000 & siccd <= 6799 ~ "Finance",
            siccd >= 7000 & siccd <= 8999 ~ "Services",
            siccd >= 9000 & siccd <= 9999 ~ "Public",
            .default = "Missing"
          )
        )

      risk_free_monthly <- download_data_risk_free(
        start_date = start_date,
        end_date = end_date
      )

      crsp_monthly <- crsp_monthly |>
        left_join(risk_free_monthly, join_by(date)) |>
        mutate(
          ret_excess = ret - risk_free
        ) |>
        select(-risk_free)

      processed_data <- crsp_monthly |>
        tidyr::drop_na(ret_excess, mktcap)
    }
  } else if (dataset == "crsp_daily") {
    if (version == "v1") {
      if (isTRUE(adjust_volume)) {
        if (!all(c("prc", "vol", "cfacpr", "exchcd") %in% additional_columns)) {
          cli::cli_abort(
            paste(
              "{.val prc}, {.val vol}, {.val exchcd},",
              "and {.val cfacpr} must be contained in",
              "{.arg additional_columns} for {.arg adjust_volume = TRUE}."
            )
          )
        }
      }

      dsf_db <- tbl(con, I("crsp.dsf")) |>
        filter(between(date, start_date, end_date))
      msenames_db <- tbl(con, I("crsp.msenames"))
      msedelist_db <- tbl(con, I("crsp.msedelist"))

      dsf_db_columns <- c(
        "permno",
        colnames(dsf_db)[-which(colnames(dsf_db) %in% colnames(msenames_db))]
      )

      permnos <- dsf_db |>
        distinct(permno) |>
        pull()

      risk_free_daily <- download_data_risk_free(
        start_date = start_date,
        end_date = end_date,
        frequency = "daily"
      )

      batches <- ceiling(length(permnos) / batch_size)

      crsp_daily_list <- list()

      cli::cli_progress_bar(
        "Downloading batches",
        total = batches,
        clear = TRUE
      )

      for (j in 1:batches) {
        permno_batch <- permnos[
          ((j - 1) * batch_size + 1):min(j * batch_size, length(permnos))
        ]

        crsp_daily_sub <- dsf_db |>
          filter(permno %in% permno_batch) |>
          select(all_of(dsf_db_columns)) |>
          inner_join(
            msenames_db |>
              filter(shrcd %in% c(10, 11)),
            join_by(permno)
          ) |>
          filter(between(date, namedt, nameendt)) |>
          select(permno, date, ret, all_of(additional_columns)) |>
          collect() |>
          tidyr::drop_na(permno, date, ret)

        if (nrow(crsp_daily_sub) > 0) {
          msedelist_sub <- msedelist_db |>
            filter(permno %in% permno_batch) |>
            select(permno, dlstdt, dlret) |>
            collect() |>
            tidyr::drop_na()

          crsp_daily_sub <- crsp_daily_sub |>
            left_join(msedelist_sub, join_by(permno, date == dlstdt)) |>
            bind_rows(
              msedelist_sub |>
                anti_join(crsp_daily_sub, join_by(permno, dlstdt == date))
            ) |>
            mutate(
              ret = if_else(!is.na(dlret), dlret, ret),
              date = if_else(!is.na(dlstdt), dlstdt, date)
            ) |>
            select(-c(dlret, dlstdt)) |>
            left_join(
              msedelist_sub |>
                select(permno, dlstdt),
              join_by(permno)
            ) |>
            mutate(dlstdt = tidyr::replace_na(dlstdt, as.Date(end_date))) |>
            filter(date <= dlstdt) |>
            select(-dlstdt)

          crsp_daily_sub <- crsp_daily_sub |>
            left_join(risk_free_daily, join_by(date)) |>
            mutate(
              ret_excess = ret - risk_free
            ) |>
            select(-risk_free)

          if (isTRUE(adjust_volume)) {
            # Gao and Ritter (2010) volume adjustment for NASDAQ trading volume
            gr_date_1 <- as.Date("2001-02-01") # nolint: object_usage_linter
            gr_date_2 <- as.Date("2002-01-01") # nolint: object_usage_linter
            gr_date_3 <- as.Date("2004-01-01") # nolint: object_usage_linter

            crsp_daily_sub <- crsp_daily_sub |>
              mutate(
                vol = na_if(vol, -99),
                prc = na_if(prc, 0),
                prc_adj = abs(prc) / cfacpr,
                prc_adj = if_else(is.infinite(prc_adj), NA_real_, prc_adj)
              ) |>
              mutate(
                vol_adj = case_when(
                  exchcd == 3 & date < gr_date_1 ~ vol / 2.0,
                  exchcd == 3 & date >= gr_date_1 & date < gr_date_2 ~ vol /
                    1.8,
                  exchcd == 3 & date >= gr_date_2 & date < gr_date_3 ~ vol /
                    1.6,
                  exchcd == 3 & date >= gr_date_3 ~ vol / 1.0,
                  .default = vol
                )
              )
          }

          crsp_daily_list[[j]] <- crsp_daily_sub
        }
        cli::cli_progress_update()
      }

      disconnect_connection(con)

      processed_data <- bind_rows(crsp_daily_list)
    } else {
      if (isTRUE(adjust_volume)) {
        if (
          !all(
            c("dlyprc", "dlyvol", "dlyfacprc", "primaryexch") %in%
              additional_columns
          )
        ) {
          cli::cli_abort(
            paste(
              "{.val dlyprc}, {.val dlyvol}, {.val primaryexch},",
              "and {.val dlyfacprc} must be contained in",
              "{.arg additional_columns} for {.arg adjust_volume = TRUE}."
            )
          )
        }
      }

      dsf_db <- tbl(con, I("crsp.dsf_v2")) |>
        filter(between(dlycaldt, start_date, end_date))
      stksecurityinfohist_db <- tbl(con, I("crsp.stksecurityinfohist"))

      dsf_db_columns <- c(
        "permno",
        colnames(dsf_db)[
          -which(colnames(dsf_db) %in% colnames(stksecurityinfohist_db))
        ]
      )

      permnos <- dsf_db |>
        distinct(permno) |>
        pull()

      risk_free_daily <- download_data_risk_free(
        start_date = start_date,
        end_date = end_date,
        frequency = "daily"
      )

      batches <- ceiling(length(permnos) / batch_size)

      crsp_daily_list <- list()

      cli::cli_progress_bar(
        "Downloading batches",
        total = batches,
        clear = TRUE
      )

      for (j in 1:batches) {
        permno_batch <- permnos[
          ((j - 1) * batch_size + 1):min(j * batch_size, length(permnos))
        ]

        crsp_daily_sub <- dsf_db |>
          filter(permno %in% permno_batch) |>
          select(all_of(dsf_db_columns)) |>
          inner_join(
            stksecurityinfohist_db |>
              filter(
                sharetype == "NS" &
                  securitytype == "EQTY" &
                  securitysubtype == "COM" &
                  usincflg == "Y" &
                  issuertype %in% c("ACOR", "CORP") &
                  primaryexch %in% c("N", "A", "Q") &
                  conditionaltype %in% c("RW", "NW") &
                  tradingstatusflg == "A"
              ),
            join_by(permno)
          ) |>
          filter(between(dlycaldt, secinfostartdt, secinfoenddt)) |>
          select(
            permno,
            date = dlycaldt,
            ret = dlyret,
            all_of(additional_columns)
          ) |>
          collect() |>
          tidyr::drop_na(permno, date, ret)

        if (nrow(crsp_daily_sub) > 0) {
          crsp_daily_sub <- crsp_daily_sub |>
            left_join(risk_free_daily, join_by(date)) |>
            mutate(
              ret_excess = ret - risk_free
            ) |>
            select(-risk_free)

          if (isTRUE(adjust_volume)) {
            crsp_daily_sub <- crsp_daily_sub |>
              group_by(permno) |>
              arrange(date) |>
              mutate(
                cfacpr = cumprod(dlyfacprc)
              ) |>
              ungroup()

            gr_date_1 <- as.Date("2001-02-01")
            gr_date_2 <- as.Date("2002-01-01")
            gr_date_3 <- as.Date("2004-01-01")

            crsp_daily_sub <- crsp_daily_sub |>
              mutate(
                vol = na_if(dlyvol, -99),
                prc = na_if(dlyprc, 0),
                prc_adj = abs(prc) / cfacpr,
                prc_adj = if_else(is.infinite(prc_adj), NA_real_, prc_adj)
              ) |>
              mutate(
                vol_adj = case_when(
                  primaryexch == "Q" & date < gr_date_1 ~ vol / 2.0,
                  primaryexch == "Q" &
                    date >= gr_date_1 &
                    date < gr_date_2 ~ vol /
                    1.8,
                  primaryexch == "Q" &
                    date >= gr_date_2 &
                    date < gr_date_3 ~ vol /
                    1.6,
                  primaryexch == "Q" & date >= gr_date_3 ~ vol / 1.0,
                  .default = vol
                )
              ) |>
              select(-c(dlyvol, dlyprc, dlyfacprc))
          }

          crsp_daily_list[[j]] <- crsp_daily_sub
        }
        cli::cli_progress_update()
      }

      disconnect_connection(con)

      processed_data <- bind_rows(crsp_daily_list)
    }
  } else {
    cli::cli_abort("Unsupported CRSP dataset: {.val {dataset}}")
  }

  if (isTRUE(add_ccm_links)) {
    ccm_links <- download_data_wrds_ccm_links()

    valid_links <- processed_data |>
      inner_join(
        ccm_links,
        join_by(permno),
        relationship = "many-to-many",
        multiple = "all"
      ) |>
      filter(!is.na(gvkey) & (date >= linkdt & date <= linkenddt)) |>
      select(permno, gvkey, date)

    processed_data <- processed_data |>
      left_join(valid_links, join_by(permno, date))
  }

  processed_data
}

#' Check if WRDS CRSP dataset is supported
#' @noRd
check_supported_dataset_wrds_crsp <- function(dataset) {
  supported_datasets <- c("crsp_monthly", "crsp_daily")

  if (!dataset %in% supported_datasets) {
    cli::cli_abort(c(
      "Unsupported CRSP dataset: {.val {dataset}}",
      "i" = "Supported datasets: {.val {supported_datasets}}"
    ))
  }
}

#' Establish a Connection to the WRDS Database
#'
#' Establishes a connection to the Wharton Research Data Services (WRDS)
#' database using the `RPostgres` package. It requires that the `RPostgres`
#' package be installed and that valid WRDS credentials be set as environment
#' variables.
#'
#' @details The function checks if the `RPostgres` package is installed
#'   before attempting to establish a connection. It uses the host, dbname,
#'   port, and sslmode as fixed parameters for the connection. Users must
#'   set their WRDS username and password as environment variables
#'   `WRDS_USER` and `WRDS_PASSWORD`, respectively, before using this
#'   function.
#'
#' @returns An object of class `DBIConnection` representing the connection
#'   to the WRDS database. This object can be used with other DBI-compliant
#'   functions to interact with the database.
#'
#' @seealso [RPostgres::Postgres()], [DBI::dbDisconnect()] for more
#'   information on managing database connections.
#'
#' @family WRDS functions
#' @export
#'
#' @examples
#' \dontrun{
#'   # Before using this function, set your WRDS credentials:
#'   # Sys.setenv(WRDS_USER = "your_username", WRDS_PASSWORD = "your_password")
#'
#'   # con <- get_wrds_connection()
#'   # Use `con` with DBI-compliant functions to interact with the WRDS database
#'   # Remember to disconnect after use:
#'   # disconnect_connection(con)
#' }
get_wrds_connection <- function() {
  if (
    !nzchar(Sys.getenv("WRDS_USER")) || !nzchar(Sys.getenv("WRDS_PASSWORD"))
  ) {
    cli::cli_inform(
      paste(
        "WRDS credentials not found. Please set them using",
        "{.fn set_wrds_credentials}."
      )
    )
  }

  DBI::dbConnect(
    RPostgres::Postgres(),
    host = "wrds-pgdata.wharton.upenn.edu",
    dbname = "wrds",
    port = 9737,
    sslmode = "require",
    user = Sys.getenv("WRDS_USER"),
    password = Sys.getenv("WRDS_PASSWORD")
  )
}

#' Compute Long-Short Returns
#'
#' Calculates long-short returns based on the returns of portfolios. The
#' long-short return is computed as the difference between the returns of the
#' "top" and "bottom" portfolios. The direction of the calculation can be
#' adjusted based on whether the return from the "bottom" portfolio is
#' subtracted from or added to the return from the "top" portfolio.
#'
#' @param data A data frame containing portfolio returns. The data frame must
#'   include columns for the portfolio identifier, date, and return
#'   measurements (as specified in `data_options`).
#' @param direction A character string specifying the direction of the
#'   long-short return calculation. It can be either `"top_minus_bottom"` or
#'   `"bottom_minus_top"`. Default is `"top_minus_bottom"`. If set to
#'   `"bottom_minus_top"`, the return will be computed as (bottom - top).
#' @param data_options A list of class `tidyfinance_data_options` (created via
#'   [data_options()]) specifying column name mappings. The `date` element is
#'   used to specify the date column, the `ret_excess` element is used to
#'   specify the excess return column, and `portfolio` is used to specify the
#'   assigned portfolio. Uses [data_options()] default if `NULL`:
#'   `"date" = "date"`, `"ret_excess" = "ret_excess"`, and
#'   `"portfolio" = "portfolio"`.
#'
#' @returns A data frame with columns for date, return measurement types (from
#'   the "ret_measure" column), and the computed long-short returns. The data
#'   frame is arranged by date and pivoted to have return measurement types as
#'   columns with their corresponding long-short returns.
#'
#' @family portfolio functions
#' @export
#'
#' @examples
#' set.seed(42)
#' data <- data.frame(
#'   permno = 1:100,
#'   date = rep(
#'     seq.Date(from = as.Date("2020-01-01"), by = "month", length.out = 100),
#'     each = 10
#'   ),
#'   mktcap_lag = runif(100, 100, 1000),
#'   ret_excess = rnorm(100),
#'   size = runif(100, 50, 150)
#' )
#'
#' portfolio_returns <- compute_portfolio_returns(
#'   data, "size", "univariate",
#'   breakpoint_options_main = breakpoint_options(n_portfolios = 5)
#' )
#'
#' compute_long_short_returns(portfolio_returns)
#'
compute_long_short_returns <- function(
  data,
  direction = "top_minus_bottom",
  data_options = NULL
) {
  if (is.null(data_options)) {
    data_options <- data_options()
  }

  data <- data |>
    check_new_col("..date") |>
    dplyr::mutate(..date = .data[[data_options$date]])

  data |>
    group_by(..date) |>
    filter(
      .data[[data_options$portfolio]] %in%
        c(
          min(.data[[data_options$portfolio]]),
          max(.data[[data_options$portfolio]])
        )
    ) |>
    mutate(
      "{data_options$portfolio}" := if_else(
        .data[[data_options$portfolio]] == min(.data[[data_options$portfolio]]),
        "bottom",
        "top"
      )
    ) |>
    ungroup() |>
    tidyr::pivot_longer(
      contains(data_options$ret_excess),
      names_to = "ret_measure",
      values_to = "ret"
    ) |>
    tidyr::pivot_wider(names_from = portfolio, values_from = ret) |>
    check_new_col("long_short_return") |>
    mutate(
      long_short_return = (top - bottom) *
        if_else(direction == "bottom_minus_top", -1, 1)
    ) |>
    arrange(..date) |>
    select(-c(top, bottom, ..date)) |>
    tidyr::pivot_wider(
      names_from = ret_measure,
      values_from = long_short_return
    )
}

#' Create Data Options
#'
#' Creates a list of data options of class `tidyfinance_data_options` used for
#' TidyFinance-related functions. These options map the specific data variables
#' to the TidyFinance naming conventions, allowing functions to flexibly work
#' with different datasets by specifying the relevant column names. Additional
#' options can be passed through `...`.
#'
#' @param id A character string representing the entity variable (defaults to
#'   `"permno"`).
#' @param date A character string representing the date variable (defaults to
#'   `"date"`).
#' @param exchange A character string representing the exchange variable
#'   (defaults to `"exchange"`).
#' @param mktcap_lag A character string representing the market capitalization
#'   lag variable (defaults to `"mktcap_lag"`).
#' @param ret_excess A character string representing the excess return variable
#'   (defaults to `"ret_excess"`).
#' @param portfolio A character string representing the portfolio variable
#'   (defaults to `"portfolio"`).
#' @param siccd A character string representing the Standard Industrial
#'   Classification code variable (defaults to `"siccd"`).
#' @param price A character string representing the (adjusted) price variable
#'   (defaults to `"prc_adj"`).
#' @param listing_age A character string representing the listing age variable
#'   (defaults to `"listing_age"`).
#' @param be A character string representing the book equity variable (defaults
#'   to `"be"`).
#' @param earnings A character string representing the earnings variable
#'   (defaults to `"ib"`, the Compustat income before extraordinary items
#'   column).
#' @param ... Additional arguments to be included in the data options list.
#'
#' @returns A list of class `tidyfinance_data_options` containing the specified
#'   data options.
#'
#' @family portfolio functions
#' @export
#'
#' @examples
#' data_options(
#'   id = "permno",
#'   date = "date",
#'   exchange = "exchange"
#' )
#'
data_options <- function(
  id = "permno",
  date = "date",
  exchange = "exchange",
  mktcap_lag = "mktcap_lag",
  ret_excess = "ret_excess",
  portfolio = "portfolio",
  siccd = "siccd",
  price = "prc_adj",
  listing_age = "listing_age",
  be = "be",
  earnings = "ib",
  ...
) {
  # Error handling for id
  if (!is.character(id) || length(id) != 1) {
    cli::cli_abort(
      paste0(
        "{.arg id} must be a string indicating the column name ",
        "for the entity variable."
      )
    )
  }

  # Error handling for date
  if (!is.character(date) || length(date) != 1) {
    cli::cli_abort(
      paste0(
        "{.arg date} must be a string indicating the column name ",
        "for the date variable."
      )
    )
  }

  # Error handling for exchange
  if (!is.character(exchange) || length(exchange) != 1) {
    cli::cli_abort(
      paste0(
        "{.arg exchange} must be a string indicating the column name ",
        "for the exchange variable."
      )
    )
  }

  # Error handling for mktcap_lag
  if (!is.character(mktcap_lag) || length(mktcap_lag) != 1) {
    cli::cli_abort(
      paste0(
        "{.arg mktcap_lag} must be a string indicating the column name ",
        "for the market capitalization lag variable."
      )
    )
  }

  # Error handling for ret_excess
  if (!is.character(ret_excess) || length(ret_excess) != 1) {
    cli::cli_abort(
      paste0(
        "{.arg ret_excess} must be a string indicating the column name ",
        "for the excess return variable."
      )
    )
  }

  # Error handling for portfolio
  if (!is.character(portfolio) || length(portfolio) != 1) {
    cli::cli_abort(
      paste0(
        "{.arg portfolio} must be a string indicating the column name ",
        "for the portfolio variable."
      )
    )
  }

  # Error handling for siccd
  if (!is.character(siccd) || length(siccd) != 1) {
    cli::cli_abort(
      paste0(
        "{.arg siccd} must be a string indicating the column name for ",
        "the Standard Industrial Classification code variable."
      )
    )
  }

  # Error handling for price
  if (!is.character(price) || length(price) != 1) {
    cli::cli_abort(
      paste0(
        "{.arg price} must be a string indicating the column name ",
        "for the (adjusted) price variable."
      )
    )
  }

  # Error handling for listing_age
  if (!is.character(listing_age) || length(listing_age) != 1) {
    cli::cli_abort(
      paste0(
        "{.arg listing_age} must be a string indicating the column name ",
        "for the listing age variable."
      )
    )
  }

  # Error handling for be
  if (!is.character(be) || length(be) != 1) {
    cli::cli_abort(
      paste0(
        "{.arg be} must be a string indicating the column name ",
        "for the book equity variable."
      )
    )
  }

  # Error handling for earnings
  if (!is.character(earnings) || length(earnings) != 1) {
    cli::cli_abort(
      paste0(
        "{.arg earnings} must be a string indicating the column ",
        "name for the earnings variable."
      )
    )
  }

  # Create the list structure with class attribute
  structure(
    list(
      "id" = id,
      "date" = date,
      "exchange" = exchange,
      "mktcap_lag" = mktcap_lag,
      "ret_excess" = ret_excess,
      "portfolio" = portfolio,
      "siccd" = siccd,
      "price" = price,
      "listing_age" = listing_age,
      "be" = be,
      "earnings" = earnings,
      ...
    ),
    class = "tidyfinance_data_options"
  )
}

test_that("data_options returns object of class tidyfinance_data_options", {
  opts <- data_options()
  expect_s3_class(opts, "tidyfinance_data_options")
})

test_that("data_options returns a list", {
  opts <- data_options()
  expect_type(opts, "list")
})

test_that("data_options has correct default values", {
  opts <- data_options()
  expect_equal(opts$id, "permno")
  expect_equal(opts$date, "date")
  expect_equal(opts$exchange, "exchange")
  expect_equal(opts$mktcap_lag, "mktcap_lag")
  expect_equal(opts$ret_excess, "ret_excess")
  expect_equal(opts$portfolio, "portfolio")
  expect_equal(opts$siccd, "siccd")
  expect_equal(opts$price, "prc_adj")
  expect_equal(opts$listing_age, "listing_age")
  expect_equal(opts$be, "be")
  expect_equal(opts$earnings, "ib")
})

test_that("data_options accepts custom column names for core columns", {
  opts <- data_options(
    id = "firm_id",
    date = "month",
    exchange = "exch",
    mktcap_lag = "lag_mcap",
    ret_excess = "excess_ret",
    portfolio = "port"
  )
  expect_equal(opts$id, "firm_id")
  expect_equal(opts$date, "month")
  expect_equal(opts$exchange, "exch")
  expect_equal(opts$mktcap_lag, "lag_mcap")
  expect_equal(opts$ret_excess, "excess_ret")
  expect_equal(opts$portfolio, "port")
})

test_that(
  "data_options accepts custom column names for filter-related columns", {
  opts <- data_options(
    siccd = "sic_code",
    price = "adj_price",
    listing_age = "age_months",
    be = "book_equity",
    earnings = "ni"
  )
  expect_equal(opts$siccd, "sic_code")
  expect_equal(opts$price, "adj_price")
  expect_equal(opts$listing_age, "age_months")
  expect_equal(opts$be, "book_equity")
  expect_equal(opts$earnings, "ni")
})

test_that("data_options stores extra arguments via ...", {
  opts <- data_options(custom_col = "my_column")
  expect_equal(opts$custom_col, "my_column")
})

test_that("data_options errors for non-character id", {
  expect_error(data_options(id = 123), "id")
  expect_error(data_options(id = c("a", "b")), "id")
})

test_that("data_options errors for non-character date", {
  expect_error(data_options(date = 1), "date")
  expect_error(data_options(date = c("date1", "date2")), "date")
})

test_that("data_options errors for non-character exchange", {
  expect_error(data_options(exchange = TRUE), "exchange")
  expect_error(data_options(exchange = c("NYSE", "NASDAQ")), "exchange")
})

test_that("data_options errors for non-character mktcap_lag", {
  expect_error(data_options(mktcap_lag = 0), "mktcap_lag")
})

test_that("data_options errors for non-character ret_excess", {
  expect_error(data_options(ret_excess = NULL), "ret_excess")
})

test_that("data_options errors for non-character portfolio", {
  expect_error(data_options(portfolio = 5L), "portfolio")
})

test_that("data_options errors for non-character siccd", {
  expect_error(data_options(siccd = 6000), "siccd")
})

test_that("data_options errors for non-character price", {
  expect_error(data_options(price = TRUE), "price")
})

test_that("data_options errors for non-character listing_age", {
  expect_error(data_options(listing_age = 12), "listing_age")
})

test_that("data_options errors for non-character be", {
  expect_error(data_options(be = 1.0), "be")
})

test_that("data_options errors for non-character earnings", {
  expect_error(data_options(earnings = 1.0), "earnings")
  expect_error(data_options(earnings = c("ib", "ni")), "earnings")
})

# Panel data: n_stocks × n_months with periodic rebalancing data
make_panel <- function(n_stocks = 50, n_months = 24, seed = 42) {
  set.seed(seed)
  dates <- seq.Date(
    from = as.Date("2020-01-01"),
    by = "month",
    length.out = n_months
  )
  expand.grid(permno = seq_len(n_stocks), date = dates) |>
    mutate(
      mktcap_lag = runif(n(), 100, 10000),
      ret_excess = rnorm(n(), 0, 0.05),
      size = runif(n(), 1, 100),
      bm = runif(n(), 0.5, 3),
      exchange = sample(c("NYSE", "NASDAQ"), n(), replace = TRUE)
    ) |>
    arrange(date, permno) |>
    as_tibble()
}

test_that("error if sorting_variables is NULL or empty", {
  data <- make_panel()
  expect_error(
    compute_portfolio_returns(
      data,
      NULL,
      "univariate",
      breakpoint_options_main = breakpoint_options(n_portfolios = 5)
    ),
    "at least one"
  )
  expect_error(
    compute_portfolio_returns(
      data,
      character(0),
      "univariate",
      breakpoint_options_main = breakpoint_options(n_portfolios = 5)
    ),
    "at least one"
  )
})

test_that("error for invalid sorting_method", {
  data <- make_panel()
  expect_error(
    compute_portfolio_returns(
      data,
      "size",
      "invalid_method",
      breakpoint_options_main = breakpoint_options(n_portfolios = 5)
    ),
    "Invalid sorting method"
  )
})

test_that("error if required columns are missing", {
  data <- make_panel() |> select(-ret_excess)
  expect_error(
    compute_portfolio_returns(
      data,
      "size",
      "univariate",
      breakpoint_options_main = breakpoint_options(n_portfolios = 5)
    ),
    "Missing"
  )
})

test_that("error if sorting variable column is missing", {
  data <- make_panel()
  expect_error(
    compute_portfolio_returns(
      data,
      "nonexistent",
      "univariate",
      breakpoint_options_main = breakpoint_options(n_portfolios = 5)
    ),
    "Missing"
  )
})

test_that("error for univariate with two sorting variables", {
  data <- make_panel()
  expect_error(
    compute_portfolio_returns(
      data,
      c("size", "bm"),
      "univariate",
      breakpoint_options_main = breakpoint_options(n_portfolios = 5)
    ),
    "one sorting variable"
  )
})

test_that("error for bivariate with one sorting variable", {
  data <- make_panel()
  expect_warning(
    expect_error(
      compute_portfolio_returns(
        data,
        "size",
        "bivariate-dependent",
        breakpoint_options_main = breakpoint_options(n_portfolios = 5)
      ),
      "two sorting variables"
    )
  )
  expect_warning(
    expect_error(
      compute_portfolio_returns(
        data,
        "size",
        "bivariate-independent",
        breakpoint_options_main = breakpoint_options(n_portfolios = 5)
      ),
      "two sorting variables"
    )
  )
})

test_that("error for rebalancing_month outside 1–12", {
  data <- make_panel()
  expect_error(
    compute_portfolio_returns(
      data,
      "size",
      "univariate",
      rebalancing_month = 13,
      breakpoint_options_main = breakpoint_options(n_portfolios = 5)
    ),
    "rebalancing_month"
  )
  expect_error(
    compute_portfolio_returns(
      data,
      "size",
      "univariate",
      rebalancing_month = 0,
      breakpoint_options_main = breakpoint_options(n_portfolios = 5)
    ),
    "rebalancing_month"
  )
})

test_that("univariate periodic returns a data.frame", {
  data <- make_panel()
  result <- compute_portfolio_returns(
    data,
    "size",
    "univariate",
    breakpoint_options_main = breakpoint_options(n_portfolios = 5)
  )
  expect_s3_class(result, "data.frame")
})

test_that("univariate periodic has correct columns with mktcap_lag", {
  data <- make_panel()
  result <- compute_portfolio_returns(
    data,
    "size",
    "univariate",
    breakpoint_options_main = breakpoint_options(n_portfolios = 5)
  )
  expect_true(all(
    c(
      "portfolio",
      "date",
      "ret_excess_vw",
      "ret_excess_ew",
      "ret_excess_vw_capped"
    ) %in%
      colnames(result)
  ))
})

test_that("univariate periodic has correct columns without mktcap_lag", {
  data <- make_panel() |> select(-mktcap_lag)
  result <- compute_portfolio_returns(
    data,
    "size",
    "univariate",
    breakpoint_options_main = breakpoint_options(n_portfolios = 5)
  )
  expect_true("ret_excess_ew" %in% colnames(result))
  expect_false("ret_excess_vw" %in% colnames(result))
  expect_false("ret_excess_vw_capped" %in% colnames(result))
})

test_that("univariate periodic returns correct number of portfolios", {
  data <- make_panel()
  result <- compute_portfolio_returns(
    data,
    "size",
    "univariate",
    breakpoint_options_main = breakpoint_options(n_portfolios = 5)
  )
  expect_equal(sort(unique(result$portfolio)), 1:5)
})

test_that("univariate periodic returns all dates", {
  data <- make_panel(n_months = 12)
  result <- compute_portfolio_returns(
    data,
    "size",
    "univariate",
    breakpoint_options_main = breakpoint_options(n_portfolios = 3)
  )
  expect_equal(length(unique(result$date)), 12)
})

test_that("univariate periodic: each date × portfolio has one row", {
  data <- make_panel()
  result <- compute_portfolio_returns(
    data,
    "size",
    "univariate",
    breakpoint_options_main = breakpoint_options(n_portfolios = 5)
  )
  counts <- result |> count(date, portfolio)
  expect_true(all(counts$n == 1))
})

test_that("univariate periodic: EW returns are within reasonable range", {
  data <- make_panel(n_stocks = 100, n_months = 60)
  result <- compute_portfolio_returns(
    data,
    "size",
    "univariate",
    breakpoint_options_main = breakpoint_options(n_portfolios = 5)
  )
  # Monthly returns should be small (data has sd = 0.05)
  expect_true(all(abs(result$ret_excess_ew) < 1, na.rm = TRUE))
})

test_that("univariate periodic: VW returns differ from EW returns", {
  data <- make_panel(n_stocks = 100, n_months = 24, seed = 1)
  result <- compute_portfolio_returns(
    data,
    "size",
    "univariate",
    breakpoint_options_main = breakpoint_options(n_portfolios = 5)
  )
  # With heterogeneous market caps, VW and EW should not be identical
  expect_false(isTRUE(all.equal(result$ret_excess_vw, result$ret_excess_ew)))
})

test_that("univariate annual rebalancing returns a data.frame", {
  data <- make_panel(n_months = 36)
  result <- compute_portfolio_returns(
    data,
    "size",
    "univariate",
    rebalancing_month = 7,
    breakpoint_options_main = breakpoint_options(n_portfolios = 5)
  )
  expect_s3_class(result, "data.frame")
})

test_that("univariate annual rebalancing has portfolio column", {
  data <- make_panel(n_months = 36)
  result <- compute_portfolio_returns(
    data,
    "size",
    "univariate",
    rebalancing_month = 7,
    breakpoint_options_main = breakpoint_options(n_portfolios = 5)
  )
  expect_true("portfolio" %in% colnames(result))
})

test_that(
  paste(
    "univariate annual: portfolios are held constant between rebalancing months"
  ),
  {
    # With rebalancing in July, a stock's portfolio should not change from
    # July to June of the following year.
    data <- make_panel(n_stocks = 30, n_months = 36)
    result_periodic <- compute_portfolio_returns(
      data,
      "size",
      "univariate",
      breakpoint_options_main = breakpoint_options(n_portfolios = 3)
    )
    result_annual <- compute_portfolio_returns(
      data,
      "size",
      "univariate",
      rebalancing_month = 7,
      breakpoint_options_main = breakpoint_options(n_portfolios = 3)
    )
    # Annual rebalancing should produce different results than periodic
    # (unless data coincidentally aligns)
    expect_false(identical(result_periodic, result_annual))
  }
)

test_that("univariate annual: works with rebalancing_month = 1", {
  data <- make_panel(n_months = 36)
  expect_no_error(
    compute_portfolio_returns(
      data,
      "size",
      "univariate",
      rebalancing_month = 1,
      breakpoint_options_main = breakpoint_options(n_portfolios = 5)
    )
  )
})

test_that("annual rebalancing: output dates start at rebalancing_month", {
  data <- make_panel(n_months = 24) # starts 2020-01
  result <- compute_portfolio_returns(
    data,
    "size",
    "univariate",
    rebalancing_month = 7,
    breakpoint_options_main = breakpoint_options(n_portfolios = 5)
  )
  expect_true(min(result$date) >= as.Date("2020-07-01"))
})

test_that("annual rebalancing: error when no dates match rebalancing_month", {
  # Data spans only 6 months starting in January — month 9 never appears
  data <- make_panel(n_months = 6) # 2020-01 through 2020-06
  expect_error(
    compute_portfolio_returns(
      data,
      "size",
      "univariate",
      rebalancing_month = 9,
      breakpoint_options_main = breakpoint_options(n_portfolios = 5)
    ),
    "rebalancing_month"
  )
})

test_that("univariate annual: works with rebalancing_month = 12", {
  data <- make_panel(n_months = 36)
  expect_no_error(
    compute_portfolio_returns(
      data,
      "size",
      "univariate",
      rebalancing_month = 12,
      breakpoint_options_main = breakpoint_options(n_portfolios = 5)
    )
  )
})

test_that("bivariate-dependent periodic returns a data.frame", {
  data <- make_panel(n_stocks = 100, n_months = 24)
  result <- suppressWarnings(
    compute_portfolio_returns(
      data,
      c("size", "bm"),
      "bivariate-dependent",
      breakpoint_options_main = breakpoint_options(n_portfolios = 5),
      breakpoint_options_secondary = breakpoint_options(n_portfolios = 3)
    )
  )
  expect_s3_class(result, "data.frame")
})

test_that("bivariate-dependent returns only main portfolio ids", {
  data <- make_panel(n_stocks = 100, n_months = 12)
  result <- suppressWarnings(
    compute_portfolio_returns(
      data,
      c("size", "bm"),
      "bivariate-dependent",
      breakpoint_options_main = breakpoint_options(n_portfolios = 5),
      breakpoint_options_secondary = breakpoint_options(n_portfolios = 3)
    )
  )
  # Should have portfolios 1–5 (main), averaged over secondary
  expect_true(all(result$portfolio %in% 1:5))
  expect_false("portfolio_secondary" %in% colnames(result))
})

test_that("bivariate-dependent has correct output columns", {
  data <- make_panel(n_stocks = 100, n_months = 12)
  result <- suppressWarnings(
    compute_portfolio_returns(
      data,
      c("size", "bm"),
      "bivariate-dependent",
      breakpoint_options_main = breakpoint_options(n_portfolios = 5),
      breakpoint_options_secondary = breakpoint_options(n_portfolios = 3)
    )
  )
  expect_true(all(
    c(
      "portfolio",
      "date",
      "ret_excess_vw",
      "ret_excess_ew",
      "ret_excess_vw_capped"
    ) %in%
      colnames(result)
  ))
})

test_that("bivariate-dependent: each date × portfolio has one row", {
  data <- make_panel(n_stocks = 100, n_months = 12)
  result <- suppressWarnings(
    compute_portfolio_returns(
      data,
      c("size", "bm"),
      "bivariate-dependent",
      breakpoint_options_main = breakpoint_options(n_portfolios = 5),
      breakpoint_options_secondary = breakpoint_options(n_portfolios = 3)
    )
  )
  counts <- result |> count(date, portfolio)
  expect_true(all(counts$n == 1))
})

test_that("bivariate-independent periodic returns a data.frame", {
  data <- make_panel(n_stocks = 100, n_months = 12)
  result <- compute_portfolio_returns(
    data,
    c("size", "bm"),
    "bivariate-independent",
    breakpoint_options_main = breakpoint_options(n_portfolios = 5),
    breakpoint_options_secondary = breakpoint_options(n_portfolios = 3)
  )
  expect_s3_class(result, "data.frame")
})

test_that("bivariate-independent returns only main portfolio ids", {
  data <- make_panel(n_stocks = 100, n_months = 12)
  result <- compute_portfolio_returns(
    data,
    c("size", "bm"),
    "bivariate-independent",
    breakpoint_options_main = breakpoint_options(n_portfolios = 5),
    breakpoint_options_secondary = breakpoint_options(n_portfolios = 3)
  )
  expect_true(all(result$portfolio %in% 1:5))
})

test_that("bivariate-independent: each date x portfolio has one row", {
  data <- make_panel(n_stocks = 100, n_months = 12)
  result <- compute_portfolio_returns(
    data,
    c("size", "bm"),
    "bivariate-independent",
    breakpoint_options_main = breakpoint_options(n_portfolios = 5),
    breakpoint_options_secondary = breakpoint_options(n_portfolios = 3)
  )
  counts <- result |> count(date, portfolio)
  expect_true(all(counts$n == 1))
})

test_that("bivariate-independent differs from bivariate-dependent", {
  data <- make_panel(n_stocks = 200, n_months = 24, seed = 99)
  result_dep <- suppressWarnings(
    compute_portfolio_returns(
      data,
      c("size", "bm"),
      "bivariate-dependent",
      breakpoint_options_main = breakpoint_options(n_portfolios = 5),
      breakpoint_options_secondary = breakpoint_options(n_portfolios = 3)
    )
  )
  result_ind <- compute_portfolio_returns(
    data,
    c("size", "bm"),
    "bivariate-independent",
    breakpoint_options_main = breakpoint_options(n_portfolios = 5),
    breakpoint_options_secondary = breakpoint_options(n_portfolios = 3)
  )
  # Results should differ because the sorting mechanisms are different
  combined <- inner_join(
    result_dep,
    result_ind,
    by = c("portfolio", "date"),
    suffix = c("_dep", "_ind")
  )
  expect_false(isTRUE(all.equal(
    combined$ret_excess_ew_dep,
    combined$ret_excess_ew_ind
  )))
})

test_that("bivariate-dependent annual rebalancing works", {
  data <- make_panel(n_stocks = 100, n_months = 36)
  result <- suppressWarnings(
    compute_portfolio_returns(
      data,
      c("size", "bm"),
      "bivariate-dependent",
      rebalancing_month = 7,
      breakpoint_options_main = breakpoint_options(n_portfolios = 3),
      breakpoint_options_secondary = breakpoint_options(n_portfolios = 2)
    )
  )
  expect_s3_class(result, "data.frame")
  expect_true("portfolio" %in% colnames(result))
})

test_that("bivariate-independent annual rebalancing works", {
  data <- make_panel(n_stocks = 100, n_months = 36)
  result <- compute_portfolio_returns(
    data,
    c("size", "bm"),
    "bivariate-independent",
    rebalancing_month = 7,
    breakpoint_options_main = breakpoint_options(n_portfolios = 3),
    breakpoint_options_secondary = breakpoint_options(n_portfolios = 2)
  )
  expect_s3_class(result, "data.frame")
  expect_true("portfolio" %in% colnames(result))
})

test_that("min_portfolio_size = 0 returns no NAs from the size check", {
  data <- make_panel(n_stocks = 50, n_months = 12)
  result <- compute_portfolio_returns(
    data,
    "size",
    "univariate",
    breakpoint_options_main = breakpoint_options(n_portfolios = 5),
    min_portfolio_size = 0
  )
  expect_false(any(is.na(result$ret_excess_ew)))
})

test_that("large min_portfolio_size produces NAs", {
  data <- make_panel(n_stocks = 10, n_months = 12)
  result <- compute_portfolio_returns(
    data,
    "size",
    "univariate",
    breakpoint_options_main = breakpoint_options(n_portfolios = 5),
    min_portfolio_size = 100
  )
  # With only 10 stocks and 5 portfolios (~2 per portfolio), all should be NA
  expect_true(all(is.na(result$ret_excess_ew)))
})

test_that("min_portfolio_size applies per portfolio-date group", {
  # With 50 stocks and 5 portfolios, each portfolio has ~10 stocks.
  # Setting min_portfolio_size = 11 should make some NA, but not necessarily all
  data <- make_panel(n_stocks = 50, n_months = 6)
  result <- compute_portfolio_returns(
    data,
    "size",
    "univariate",
    breakpoint_options_main = breakpoint_options(n_portfolios = 5),
    min_portfolio_size = 11
  )
  # At least some should be NA
  expect_true(any(is.na(result$ret_excess_ew)))
})

test_that(
  paste(
    "missing mktcap_lag removes ret_excess_vw and ret_excess_vw_capped columns"
  ),
  {
    data <- make_panel() |> select(-mktcap_lag)
    result <- compute_portfolio_returns(
      data,
      "size",
      "univariate",
      breakpoint_options_main = breakpoint_options(n_portfolios = 5)
    )
    expect_false("ret_excess_vw" %in% colnames(result))
    expect_false("ret_excess_vw_capped" %in% colnames(result))
    expect_true("ret_excess_ew" %in% colnames(result))
  }
)

test_that("missing mktcap_lag works for bivariate-independent", {
  data <- make_panel(n_stocks = 100) |> select(-mktcap_lag)
  result <- compute_portfolio_returns(
    data,
    c("size", "bm"),
    "bivariate-independent",
    breakpoint_options_main = breakpoint_options(n_portfolios = 3),
    breakpoint_options_secondary = breakpoint_options(n_portfolios = 2)
  )
  expect_false("ret_excess_vw" %in% colnames(result))
  expect_false("ret_excess_vw_capped" %in% colnames(result))
  expect_true("ret_excess_ew" %in% colnames(result))
})


test_that("rows with NA portfolio are excluded from output", {
  data <- make_panel(n_stocks = 50, n_months = 12)
  result <- compute_portfolio_returns(
    data,
    "size",
    "univariate",
    breakpoint_options_main = breakpoint_options(n_portfolios = 5)
  )
  expect_false(any(is.na(result$portfolio)))
})

test_that("custom data_options column names work", {
  set.seed(42)
  dates <- seq.Date(from = as.Date("2020-01-01"), by = "month", length.out = 12)
  data <- expand.grid(stock_id = 1:30, my_date = dates) |>
    mutate(
      my_mktcap = runif(n(), 100, 10000),
      my_ret = rnorm(n(), 0, 0.05),
      size = runif(n(), 1, 100)
    ) |>
    as.data.frame()

  custom_opts <- data_options(
    date = "my_date",
    id = "stock_id",
    ret_excess = "my_ret",
    mktcap_lag = "my_mktcap"
  )

  result <- compute_portfolio_returns(
    data,
    "size",
    "univariate",
    breakpoint_options_main = breakpoint_options(n_portfolios = 3),
    data_options = custom_opts
  )

  expect_s3_class(result, "data.frame")
  expect_true("my_date" %in% colnames(result))
  expect_true("portfolio" %in% colnames(result))
})

test_that("custom breakpoint_function_main is used", {
  data <- make_panel(n_stocks = 50, n_months = 12)

  # Custom function that always makes 3 portfolios regardless of options
  custom_bp <- function(data, sorting_variable, bp_options, data_options) {
    vals <- data[[sorting_variable]]
    q <- quantile(
      vals,
      probs = c(0, 1 / 3, 2 / 3, 1),
      na.rm = TRUE,
      names = FALSE
    )
    q[2:4] <- q[2:4] + 1e-20
    q
  }

  result <- compute_portfolio_returns(
    data,
    "size",
    "univariate",
    breakpoint_options_main = breakpoint_options(n_portfolios = 999),
    breakpoint_function_main = custom_bp
  )

  # Despite n_portfolios = 999, custom function forces 3 portfolios
  expect_equal(sort(unique(result$portfolio)), 1:3)
})

test_that("results are deterministic", {
  data <- make_panel(n_stocks = 50, n_months = 12)
  opts <- breakpoint_options(n_portfolios = 5)

  r1 <- compute_portfolio_returns(
    data,
    "size",
    "univariate",
    breakpoint_options_main = opts
  )
  r2 <- compute_portfolio_returns(
    data,
    "size",
    "univariate",
    breakpoint_options_main = opts
  )
  expect_identical(r1, r2)
})

test_that("single date works for univariate periodic", {
  set.seed(1)
  data <- data.frame(
    permno = 1:50,
    date = as.Date("2020-01-01"),
    mktcap_lag = runif(50, 100, 1000),
    ret_excess = rnorm(50, 0, 0.05),
    size = runif(50, 1, 100)
  )
  result <- compute_portfolio_returns(
    data,
    "size",
    "univariate",
    breakpoint_options_main = breakpoint_options(n_portfolios = 5)
  )
  expect_equal(length(unique(result$date)), 1)
  expect_equal(nrow(result), 5)
})

test_that("single stock per date works (produces 1 portfolio)", {
  dates <- seq.Date(from = as.Date("2020-01-01"), by = "month", length.out = 12)
  data <- data.frame(
    permno = 1,
    date = dates,
    mktcap_lag = 1000,
    ret_excess = rnorm(12, 0, 0.05),
    size = runif(12, 1, 100)
  )
  # Will get constant-variable warnings per date when only 1 stock
  result <- suppressWarnings(
    compute_portfolio_returns(
      data,
      "size",
      "univariate",
      breakpoint_options_main = breakpoint_options(n_portfolios = 5)
    )
  )
  expect_true(all(result$portfolio == 1))
})

test_that("EW return equals manual mean of ret_excess within portfolio", {
  set.seed(42)
  data <- data.frame(
    permno = 1:20,
    date = as.Date("2020-01-01"),
    mktcap_lag = rep(1, 20), # equal weights
    ret_excess = (1:20) / 100,
    size = 1:20
  )

  result <- compute_portfolio_returns(
    data,
    "size",
    "univariate",
    breakpoint_options_main = breakpoint_options(n_portfolios = 2)
  )

  # Manually: portfolio 1 = first 10 stocks, portfolio 2 = last 10 stocks
  # (approximately, depending on breakpoints)
  bp <- compute_breakpoints(data, "size", breakpoint_options(n_portfolios = 2))
  portfolios <- findInterval(data$size, bp, all.inside = TRUE)
  for (p in unique(portfolios)) {
    expected_ew <- mean(data$ret_excess[portfolios == p])
    actual_ew <- result$ret_excess_ew[result$portfolio == p]
    expect_equal(actual_ew, expected_ew, tolerance = 1e-10)
  }
})

test_that("VW return equals manual weighted.mean of ret_excess", {
  set.seed(42)
  data <- data.frame(
    permno = 1:20,
    date = as.Date("2020-01-01"),
    mktcap_lag = c(rep(1, 10), rep(100, 10)), # very different weights
    ret_excess = rep(0.01, 20),
    size = 1:20
  )
  data$ret_excess[11:20] <- 0.05 # high-cap stocks have higher returns

  result <- compute_portfolio_returns(
    data,
    "size",
    "univariate",
    breakpoint_options_main = breakpoint_options(n_portfolios = 2)
  )

  bp <- compute_breakpoints(data, "size", breakpoint_options(n_portfolios = 2))
  portfolios <- findInterval(data$size, bp, all.inside = TRUE)
  for (p in unique(portfolios)) {
    mask <- portfolios == p
    expected_vw <- weighted.mean(data$ret_excess[mask], data$mktcap_lag[mask])
    actual_vw <- result$ret_excess_vw[result$portfolio == p]
    expect_equal(actual_vw, expected_vw, tolerance = 1e-10)
  }
})

test_that("percentile-based breakpoints work in univariate", {
  data <- make_panel(n_stocks = 100, n_months = 12)
  result <- compute_portfolio_returns(
    data,
    "size",
    "univariate",
    breakpoint_options_main = breakpoint_options(percentiles = c(0.3, 0.7))
  )
  expect_equal(sort(unique(result$portfolio)), 1:3)
})

test_that(
  paste0(
    "bivariate-independent uses different n_portfolios for main vs secondary"
  ),
  {
    data <- make_panel(n_stocks = 200, n_months = 12)
    result <- compute_portfolio_returns(
      data,
      c("size", "bm"),
      "bivariate-independent",
      breakpoint_options_main = breakpoint_options(n_portfolios = 5),
      breakpoint_options_secondary = breakpoint_options(n_portfolios = 2)
    )
    # Output should have 5 main portfolios (averaged over 2 secondary)
    expect_equal(sort(unique(result$portfolio)), 1:5)
  }
)

test_that("no duplicate portfolio-date rows in output", {
  data <- make_panel(n_stocks = 100, n_months = 24)

  for (method in c("univariate")) {
    result <- compute_portfolio_returns(
      data,
      "size",
      method,
      breakpoint_options_main = breakpoint_options(n_portfolios = 5)
    )
    dupes <- result |> count(portfolio, date) |> filter(n > 1)
    expect_equal(nrow(dupes), 0, info = paste("method:", method))
  }

  for (method in c("bivariate-dependent", "bivariate-independent")) {
    result <- suppressWarnings(
      compute_portfolio_returns(
        data,
        c("size", "bm"),
        method,
        breakpoint_options_main = breakpoint_options(n_portfolios = 5),
        breakpoint_options_secondary = breakpoint_options(n_portfolios = 3)
      )
    )
    dupes <- result |> count(portfolio, date) |> filter(n > 1)
    expect_equal(nrow(dupes), 0, info = paste("method:", method))
  }
})

test_that("function handles moderately large data without error", {
  data <- make_panel(n_stocks = 500, n_months = 60, seed = 7)
  expect_no_error(
    compute_portfolio_returns(
      data,
      "size",
      "univariate",
      breakpoint_options_main = breakpoint_options(n_portfolios = 10)
    )
  )
})

test_that("bivariate-independent handles large data without error", {
  data <- make_panel(n_stocks = 500, n_months = 60, seed = 7)
  expect_no_error(
    compute_portfolio_returns(
      data,
      c("size", "bm"),
      "bivariate-independent",
      breakpoint_options_main = breakpoint_options(n_portfolios = 5),
      breakpoint_options_secondary = breakpoint_options(n_portfolios = 3)
    )
  )
})

test_that("output is a complete panel of portfolio x date", {
  data <- make_panel(n_stocks = 10, n_months = 12)

  result <- compute_portfolio_returns(
    data,
    "size",
    "univariate",
    breakpoint_options_main = breakpoint_options(n_portfolios = 3),
    min_portfolio_size = 0L,
    quiet = TRUE
  )

  n_portfolios <- length(unique(result$portfolio))
  n_dates_out <- length(unique(result$date))

  expect_equal(nrow(result), n_portfolios * n_dates_out)
})

test_that("changing cap_weight changes ret_excess_vw_capped", {
  data <- make_panel(n_stocks = 100, n_months = 12)

  result_80 <- compute_portfolio_returns(
    data,
    "size",
    "univariate",
    breakpoint_options_main = breakpoint_options(n_portfolios = 5),
    cap_weight = 0.8
  )
  result_50 <- compute_portfolio_returns(
    data,
    "size",
    "univariate",
    breakpoint_options_main = breakpoint_options(n_portfolios = 5),
    cap_weight = 0.5
  )

  # Capped returns should differ with different cap_weight
  expect_false(isTRUE(all.equal(
    result_80$ret_excess_vw_capped,
    result_50$ret_excess_vw_capped
  )))
  # But EW returns should be identical (unaffected by cap_weight)
  expect_equal(result_80$ret_excess_ew, result_50$ret_excess_ew)
})

test_that("cap_weight = 1 makes ret_excess_vw_capped equal to ret_excess_vw", {
  data <- make_panel(n_stocks = 100, n_months = 12)
  result <- compute_portfolio_returns(
    data,
    "size",
    "univariate",
    breakpoint_options_main = breakpoint_options(n_portfolios = 5),
    cap_weight = 1
  )
  expect_equal(result$ret_excess_vw_capped, result$ret_excess_vw)
})

test_that("error for invalid cap_weight values", {
  data <- make_panel()
  expect_error(
    compute_portfolio_returns(
      data,
      "size",
      "univariate",
      breakpoint_options_main = breakpoint_options(n_portfolios = 5),
      cap_weight = 1.5
    ),
    "cap_weight"
  )
  expect_error(
    compute_portfolio_returns(
      data,
      "size",
      "univariate",
      breakpoint_options_main = breakpoint_options(n_portfolios = 5),
      cap_weight = -0.1
    ),
    "cap_weight"
  )
  expect_error(
    compute_portfolio_returns(
      data,
      "size",
      "univariate",
      breakpoint_options_main = breakpoint_options(n_portfolios = 5),
      cap_weight = NA
    ),
    "cap_weight"
  )
  expect_error(
    compute_portfolio_returns(
      data,
      "size",
      "univariate",
      breakpoint_options_main = breakpoint_options(n_portfolios = 5),
      cap_weight = "high"
    ),
    "cap_weight"
  )
})

test_that("all-NA mktcap_lag produces NA not NaN for VW returns", {
  set.seed(42)
  data <- data.frame(
    permno = 1:20,
    date = as.Date("2020-01-01"),
    mktcap_lag = NA_real_,
    ret_excess = rnorm(20, 0, 0.05),
    size = 1:20
  )
  result <- compute_portfolio_returns(
    data,
    "size",
    "univariate",
    breakpoint_options_main = breakpoint_options(n_portfolios = 2)
  )
  # VW returns should be NA (not NaN) when all weights are zero
  expect_true(all(is.na(result$ret_excess_vw)))
  expect_false(any(is.nan(result$ret_excess_vw)))
  expect_true(all(is.na(result$ret_excess_vw_capped)))
  expect_false(any(is.nan(result$ret_excess_vw_capped)))
  # EW returns should still be computed
  expect_false(any(is.na(result$ret_excess_ew)))
})

test_that("bivariate min_portfolio_size counts per main-portfolio-date", {
  # 100 firms x 12 months, 5x5 bivariate -> ~4 firms per cell but ~20 per
  # main-portfolio-date cross-section. A per-cell threshold of 10 would
  # void every cell; a per-cross-section threshold of 10 should not.
  data <- make_panel(n_stocks = 100, n_months = 12)
  result <- compute_portfolio_returns(
    data,
    c("size", "bm"),
    "bivariate-dependent",
    breakpoint_options_main = breakpoint_options(n_portfolios = 5),
    breakpoint_options_secondary = breakpoint_options(n_portfolios = 5),
    min_portfolio_size = 10L,
    quiet = TRUE
  )
  expect_false(all(is.na(result$ret_excess_ew)))
  # With ~20 firms per main-portfolio-date, threshold = 10 should keep
  # essentially all cross-sections.
  expect_gt(mean(!is.na(result$ret_excess_ew)), 0.9)
})

test_that("bivariate min_portfolio_size voids cross-sections below threshold", {
  # 60 firms per date, 6 main x 2 secondary -> ~10 firms per main-portfolio
  # cross-section (summed across secondaries). A threshold of 50 should
  # void everything.
  data <- make_panel(n_stocks = 60, n_months = 12)
  result <- compute_portfolio_returns(
    data,
    c("size", "bm"),
    "bivariate-independent",
    breakpoint_options_main = breakpoint_options(n_portfolios = 6),
    breakpoint_options_secondary = breakpoint_options(n_portfolios = 2),
    min_portfolio_size = 50L,
    quiet = TRUE
  )
  expect_true(all(is.na(result$ret_excess_ew)))
})

test_that("univariate min_portfolio_size counts firms per portfolio-date", {
  data <- make_panel(n_stocks = 50, n_months = 12)
  # 50 firms / 5 portfolios = 10 per portfolio-date. Threshold of 11 should
  # void everything; threshold of 5 should keep everything.
  result_high <- compute_portfolio_returns(
    data,
    "size",
    "univariate",
    breakpoint_options_main = breakpoint_options(n_portfolios = 5),
    min_portfolio_size = 11L,
    quiet = TRUE
  )
  expect_true(all(is.na(result_high$ret_excess_ew)))

  result_low <- compute_portfolio_returns(
    data,
    "size",
    "univariate",
    breakpoint_options_main = breakpoint_options(n_portfolios = 5),
    min_portfolio_size = 5L,
    quiet = TRUE
  )
  expect_false(any(is.na(result_low$ret_excess_ew)))
})

test_that("error if breakpoint_options is not a list", {
  data <- data.frame(id = 1:100, value = seq_len(100))
  expect_error(
    compute_breakpoints(data, "value", "not_a_list"),
    "named list"
  )
})

test_that("error if both n_portfolios and percentiles are provided", {
  data <- data.frame(id = 1:100, value = seq_len(100))
  expect_error(
    compute_breakpoints(
      data,
      "value",
      breakpoint_options(n_portfolios = 5, percentiles = c(0.5))
    ),
    "not both"
  )
})

test_that("error if neither n_portfolios nor percentiles are provided", {
  data <- data.frame(id = 1:100, value = seq_len(100))
  expect_error(
    compute_breakpoints(data, "value", breakpoint_options()),
    "must provide"
  )
})

test_that("error if n_portfolios is 1 or less", {
  data <- data.frame(id = 1:100, value = seq_len(100))
  expect_error(
    compute_breakpoints(data, "value", breakpoint_options(n_portfolios = 1)),
  )
  expect_error(
    compute_breakpoints(data, "value", breakpoint_options(n_portfolios = 0)),
  )
})

test_that("error if breakpoints_exchanges column is missing from data", {
  data <- data.frame(id = 1:100, value = seq_len(100))
  expect_error(
    compute_breakpoints(
      data,
      "value",
      breakpoint_options(n_portfolios = 5, breakpoints_exchanges = c("NYSE"))
    ),
    "exchange"
  )
})

test_that("returns n_portfolios + 1 breakpoints", {
  data <- data.frame(id = 1:100, value = seq_len(100))
  for (n in c(2, 5, 10)) {
    bp <- compute_breakpoints(
      data,
      "value",
      breakpoint_options(n_portfolios = n)
    )
    expect_length(bp, n + 1)
  }
})

test_that("breakpoints are in ascending order", {
  data <- data.frame(id = 1:1000, value = rnorm(1000))
  bp <- compute_breakpoints(data, "value", breakpoint_options(n_portfolios = 5))
  expect_true(all(diff(bp) >= 0))
})

test_that("breakpoints are numeric", {
  data <- data.frame(id = 1:100, value = seq_len(100))
  bp <- compute_breakpoints(data, "value", breakpoint_options(n_portfolios = 5))
  expect_type(bp, "double")
})

test_that("first breakpoint equals min, last approximately equals max", {
  data <- data.frame(id = 1:1000, value = seq_len(1000))
  bp <- compute_breakpoints(data, "value", breakpoint_options(n_portfolios = 5))
  expect_equal(bp[1], min(data$value))
  # Last breakpoint = max + tiny epsilon (1e-20)
  expect_equal(bp[6], max(data$value) + 1e-20, tolerance = 1e-15)
})

test_that("n_portfolios = 2 gives 3 breakpoints (min, median-ish, max)", {
  data <- data.frame(id = 1:100, value = seq_len(100))
  bp <- compute_breakpoints(data, "value", breakpoint_options(n_portfolios = 2))
  expect_length(bp, 3)
})

test_that("percentiles produce correct number of breakpoints", {
  data <- data.frame(id = 1:1000, value = seq_len(1000))
  pcts <- c(0.2, 0.4, 0.6, 0.8)
  bp <- compute_breakpoints(
    data,
    "value",
    breakpoint_options(percentiles = pcts)
  )
  # breakpoints = c(0, percentiles, 1) → length = length(percentiles) + 2
  expect_length(bp, length(pcts) + 2)
})

test_that("single percentile produces 3 breakpoints (median split)", {
  data <- data.frame(id = 1:100, value = seq_len(100))
  bp <- compute_breakpoints(
    data,
    "value",
    breakpoint_options(percentiles = c(0.5))
  )
  expect_length(bp, 3)
})

test_that("percentile breakpoints are ascending", {
  set.seed(42)
  data <- data.frame(id = 1:500, value = rnorm(500))
  bp <- compute_breakpoints(
    data,
    "value",
    breakpoint_options(percentiles = c(0.1, 0.3, 0.5, 0.7, 0.9))
  )
  expect_true(all(diff(bp) >= 0))
})

test_that(
  paste0(
    "n_portfolios = 5 & percentiles = c(0.2,0.4,0.6,0.8) give same breakpoints"
  ),
  {
    data <- data.frame(id = 1:1000, value = seq_len(1000))
    bp_n <- compute_breakpoints(
      data,
      "value",
      breakpoint_options(n_portfolios = 5)
    )
    bp_p <- compute_breakpoints(
      data,
      "value",
      breakpoint_options(percentiles = c(0.2, 0.4, 0.6, 0.8))
    )
    expect_equal(bp_n, bp_p)
  }
)

test_that("n_portfolios = 10 & percentiles at deciles give same breakpoints", {
  set.seed(7)
  data <- data.frame(id = 1:5000, value = rnorm(5000))
  bp_n <- compute_breakpoints(
    data,
    "value",
    breakpoint_options(n_portfolios = 10)
  )
  bp_p <- compute_breakpoints(
    data,
    "value",
    breakpoint_options(percentiles = seq(0.1, 0.9, by = 0.1))
  )
  expect_equal(bp_n, bp_p)
})

test_that("breakpoints_exchanges filters data before computing breakpoints", {
  set.seed(1)
  data <- data.frame(
    id = 1:200,
    exchange = rep(c("NYSE", "NASDAQ"), each = 100),
    value = c(rnorm(100, mean = 10), rnorm(100, mean = 50))
  )

  bp_all <- compute_breakpoints(
    data,
    "value",
    breakpoint_options(n_portfolios = 5)
  )
  bp_nyse <- compute_breakpoints(
    data,
    "value",
    breakpoint_options(n_portfolios = 5, breakpoints_exchanges = c("NYSE"))
  )

  # NYSE only has values around 10, so breakpoints should differ
  expect_false(isTRUE(all.equal(bp_all, bp_nyse)))
  # NYSE breakpoints should be clustered around 10
  expect_true(all(bp_nyse < 20))
})

test_that("multiple exchanges can be specified", {
  data <- data.frame(
    id = 1:300,
    exchange = rep(c("NYSE", "NASDAQ", "AMEX"), each = 100),
    value = c(1:100, 201:300, 401:500)
  )
  bp <- compute_breakpoints(
    data,
    "value",
    breakpoint_options(
      n_portfolios = 5,
      breakpoints_exchanges = c("NYSE", "AMEX")
    )
  )
  # Breakpoints should span NYSE (1-100) and AMEX (401-500), not NASDAQ
  expect_equal(bp[1], 1)
  expect_true(bp[length(bp)] > 400)
})

test_that("custom exchange column name via data_options", {
  data <- data.frame(
    id = 1:100,
    exch = rep(c("NYSE", "NASDAQ"), each = 50),
    value = seq_len(100)
  )
  bp <- compute_breakpoints(
    data,
    "value",
    breakpoint_options(n_portfolios = 5, breakpoints_exchanges = c("NYSE")),
    data_options = data_options(exchange = "exch")
  )
  # NYSE is rows 1-50, so breakpoints should be within [1, 50]
  expect_true(bp[1] >= 1)
  expect_true(bp[length(bp)] <= 50 + 1e-15)
})

test_that("error when custom exchange column doesn't exist", {
  data <- data.frame(id = 1:100, value = seq_len(100))
  expect_error(
    compute_breakpoints(
      data,
      "value",
      breakpoint_options(n_portfolios = 5, breakpoints_exchanges = c("NYSE")),
      data_options = data_options(exchange = "nonexistent")
    ),
    "nonexistent"
  )
})

test_that("NULL data_options uses defaults without error", {
  data <- data.frame(id = 1:100, exchange = "NYSE", value = seq_len(100))
  expect_no_error(
    compute_breakpoints(data, "value", breakpoint_options(n_portfolios = 5))
  )
})


test_that("interior breakpoints are slightly larger than raw quantiles", {
  data <- data.frame(id = 1:1000, value = seq_len(1000))
  bp <- compute_breakpoints(data, "value", breakpoint_options(n_portfolios = 5))
  raw_q <- quantile(
    data$value,
    probs = seq(0, 1, length.out = 6),
    na.rm = TRUE,
    names = FALSE
  )

  # First breakpoint is untouched
  expect_equal(bp[1], raw_q[1])
  # Interior breakpoints (indices 2 through n+1) have epsilon added

  for (i in 2:6) {
    expect_equal(bp[i], raw_q[i] + 1e-20, tolerance = 1e-25)
  }
})

test_that("smooth_bunching handles clustering on both edges", {
  # Many zeros and many max-values, some spread in between
  data <- data.frame(
    id = 1:500,
    value = c(rep(0, 200), seq(1, 99, length.out = 100), rep(100, 200))
  )
  bp <- compute_breakpoints(
    data,
    "value",
    breakpoint_options(n_portfolios = 5, smooth_bunching = TRUE)
  )
  expect_length(bp, 6)
  # First breakpoint should be 0, last should be >= 100
  expect_equal(bp[1], 0)
  expect_true(bp[6] >= 100)
  # Interior breakpoints should be strictly between 0 and 100
  expect_true(all(bp[2:5] > 0))
  expect_true(all(bp[2:5] <= 100 + 1e-10))
})

test_that("smooth_bunching with both edges and percentiles emits warning", {
  data <- data.frame(
    id = 1:500,
    value = c(rep(0, 200), seq(1, 99, length.out = 100), rep(100, 200))
  )
  expect_warning(
    compute_breakpoints(
      data,
      "value",
      breakpoint_options(
        percentiles = c(0.2, 0.4, 0.6, 0.8),
        smooth_bunching = TRUE
      )
    ),
    "smooth_bunching"
  )
})

test_that("smooth_bunching handles clustering only on lower edge", {
  data <- data.frame(
    id = 1:500,
    value = c(rep(0, 200), seq(1, 100, length.out = 300))
  )
  bp <- compute_breakpoints(
    data,
    "value",
    breakpoint_options(n_portfolios = 5, smooth_bunching = TRUE)
  )
  expect_length(bp, 6)
  expect_equal(bp[1], 0)
  # After smoothing, breakpoints 2+ should be > 0
  expect_true(all(bp[2:6] > 0))
})

test_that("lower edge bunching with percentiles emits warning", {
  data <- data.frame(
    id = 1:500,
    value = c(rep(0, 200), seq(1, 100, length.out = 300))
  )
  expect_warning(
    compute_breakpoints(
      data,
      "value",
      breakpoint_options(
        percentiles = c(0.2, 0.4, 0.6, 0.8),
        smooth_bunching = TRUE
      )
    ),
    "smooth_bunching"
  )
})

test_that("smooth_bunching handles clustering only on upper edge", {
  data <- data.frame(
    id = 1:500,
    value = c(seq(0, 99, length.out = 300), rep(100, 200))
  )
  bp <- compute_breakpoints(
    data,
    "value",
    breakpoint_options(n_portfolios = 5, smooth_bunching = TRUE)
  )
  expect_length(bp, 6)
  expect_true(bp[6] >= 100)
  # Interior breakpoints should be < 100
  expect_true(all(bp[1:5] < 100))
})

test_that("upper edge bunching with percentiles emits warning", {
  data <- data.frame(
    id = 1:500,
    value = c(seq(0, 99, length.out = 300), rep(100, 200))
  )
  expect_warning(
    compute_breakpoints(
      data,
      "value",
      breakpoint_options(
        percentiles = c(0.2, 0.4, 0.6, 0.8),
        smooth_bunching = TRUE
      )
    ),
    "smooth_bunching"
  )
})

test_that("smooth_bunching = FALSE does not alter clustered breakpoints", {
  data <- data.frame(
    id = 1:500,
    value = c(rep(0, 200), seq(1, 99, length.out = 100), rep(100, 200))
  )
  bp_no_smooth <- compute_breakpoints(
    data,
    "value",
    breakpoint_options(n_portfolios = 5, smooth_bunching = FALSE)
  )
  raw_q <- quantile(
    data$value,
    probs = seq(0, 1, length.out = 6),
    na.rm = TRUE,
    names = FALSE
  )
  # Without smoothing, breakpoints should just be raw quantiles + epsilon
  expect_equal(bp_no_smooth[1], raw_q[1])
  for (i in 2:6) {
    expect_equal(bp_no_smooth[i], raw_q[i] + 1e-20, tolerance = 1e-25)
  }
})

test_that("smooth_bunching = NULL triggers error", {
  data <- data.frame(
    id = 1:500,
    value = c(rep(0, 200), seq(1, 99, length.out = 100), rep(100, 200))
  )
  expect_error(compute_breakpoints(
    data,
    "value",
    breakpoint_options(n_portfolios = 5, smooth_bunching = NULL)
  ))
})


test_that(
  paste0(
    "smooth_bunching = TRUE with no clustering gives same result as FALSE"
  ),
  {
    data <- data.frame(id = 1:1000, value = seq_len(1000))
    bp_smooth <- compute_breakpoints(
      data,
      "value",
      breakpoint_options(n_portfolios = 5, smooth_bunching = TRUE)
    )
    bp_plain <- compute_breakpoints(
      data,
      "value",
      breakpoint_options(n_portfolios = 5, smooth_bunching = FALSE)
    )
    expect_equal(bp_smooth, bp_plain)
  }
)

test_that("NA values in sorting variable are ignored (na.rm = TRUE)", {
  data_clean <- data.frame(id = 1:100, value = seq_len(100))
  data_na <- data.frame(id = 1:120, value = c(seq_len(100), rep(NA, 20)))

  bp_clean <- compute_breakpoints(
    data_clean,
    "value",
    breakpoint_options(n_portfolios = 5)
  )
  bp_na <- compute_breakpoints(
    data_na,
    "value",
    breakpoint_options(n_portfolios = 5)
  )
  expect_equal(bp_clean, bp_na)
})


test_that("works with arbitrary column names", {
  data <- data.frame(company_id = 1:200, market_cap = runif(200, 1e6, 1e9))
  bp <- compute_breakpoints(
    data,
    "market_cap",
    breakpoint_options(n_portfolios = 5)
  )
  expect_length(bp, 6)
  expect_true(all(diff(bp) >= 0))
})

test_that("works with very small data (n = 2)", {
  data <- data.frame(id = 1:2, value = c(1, 10))
  bp <- compute_breakpoints(data, "value", breakpoint_options(n_portfolios = 2))
  expect_length(bp, 3)
  expect_equal(bp[1], 1)
})

test_that("works with n = 3 and 3 portfolios (one obs per portfolio)", {
  data <- data.frame(id = 1:3, value = c(1, 5, 10))
  bp <- compute_breakpoints(data, "value", breakpoint_options(n_portfolios = 3))
  expect_length(bp, 4)
})

test_that("non-uniform percentiles produce correct breakpoints", {
  data <- data.frame(id = 1:1000, value = seq_len(1000))
  bp <- compute_breakpoints(
    data,
    "value",
    breakpoint_options(percentiles = c(0.1, 0.5, 0.9))
  )
  expect_length(bp, 5)
  # The 50th percentile breakpoint should be near 500
  expect_true(abs(bp[3] - 500) < 5)
})

test_that("100 portfolios work with sufficient data", {
  set.seed(99)
  data <- data.frame(id = 1:100000, value = rnorm(100000))
  bp <- compute_breakpoints(
    data,
    "value",
    breakpoint_options(n_portfolios = 100)
  )
  expect_length(bp, 101)
  expect_true(all(diff(bp) >= 0))
})

test_that("breakpoints produce valid portfolio assignments via findInterval", {
  set.seed(42)
  data <- data.frame(id = 1:1000, value = rnorm(1000))
  bp <- compute_breakpoints(data, "value", breakpoint_options(n_portfolios = 5))
  portfolios <- findInterval(data$value, bp, all.inside = TRUE)

  expect_true(all(portfolios >= 1 & portfolios <= 5))
  expect_equal(sort(unique(portfolios)), 1:5)
})

test_that(
  paste0(
    "exchange-filtered breakpoints still produce ",
    "valid assignments on full data"
  ),
  {
    set.seed(1)
    data <- data.frame(
      id = 1:500,
      exchange = rep(c("NYSE", "NASDAQ"), each = 250),
      value = rnorm(500)
    )
    bp <- compute_breakpoints(
      data,
      "value",
      breakpoint_options(n_portfolios = 5, breakpoints_exchanges = c("NYSE"))
    )
    portfolios <- findInterval(data$value, bp, all.inside = TRUE)

    expect_true(all(portfolios >= 1 & portfolios <= 5))
    expect_length(portfolios, 500)
  }
)

test_that(
  paste0(
    "smooth bunching produces more distinct interior values than raw quantiles"
  ),
  {
    data <- data.frame(
      id = 1:1000,
      value = c(rep(0, 400), seq(1, 99, length.out = 200), rep(100, 400))
    )

    bp_smooth <- compute_breakpoints(
      data,
      "value",
      breakpoint_options(n_portfolios = 5, smooth_bunching = TRUE)
    )
    bp_raw <- compute_breakpoints(
      data,
      "value",
      breakpoint_options(n_portfolios = 5, smooth_bunching = FALSE)
    )

    # Smoothed version should have more distinct breakpoint values
    n_distinct_smooth <- length(unique(round(bp_smooth, 10)))
    n_distinct_raw <- length(unique(round(bp_raw, 10)))
    expect_gte(n_distinct_smooth, n_distinct_raw)
  }
)

test_that("function is deterministic (same input → same output)", {
  set.seed(1)
  data <- data.frame(id = 1:500, value = rnorm(500))
  opts <- breakpoint_options(n_portfolios = 5)

  bp1 <- compute_breakpoints(data, "value", opts)
  bp2 <- compute_breakpoints(data, "value", opts)
  expect_identical(bp1, bp2)
})

test_that(
  paste0(
    "breakpoints_min_size_threshold with ",
    "breakpoints_exchanges filters small stocks"
  ),
  {
    set.seed(42)
    data <- data.frame(
      id = 1:200,
      exchange = rep(c("NYSE", "NASDAQ"), each = 100),
      mktcap_lag = c(1:100, 1:100),
      sorting_var = rnorm(200)
    )

    bp_no_filter <- compute_breakpoints(
      data,
      "sorting_var",
      breakpoint_options(n_portfolios = 5, breakpoints_exchanges = "NYSE")
    )
    bp_with_filter <- compute_breakpoints(
      data,
      "sorting_var",
      breakpoint_options(
        n_portfolios = 5,
        breakpoints_exchanges = "NYSE",
        breakpoints_min_size_threshold = 0.2
      )
    )

    expect_length(bp_with_filter, 6)
    expect_false(identical(bp_no_filter, bp_with_filter))
  }
)

test_that(
  paste0(
    "breakpoints_min_size_threshold without ",
    "breakpoints_exchanges uses full sample"
  ),
  {
    set.seed(42)
    data <- data.frame(
      id = 1:200,
      exchange = rep(c("NYSE", "NASDAQ"), each = 100),
      mktcap_lag = 1:200,
      sorting_var = rnorm(200)
    )

    bp_no_filter <- compute_breakpoints(
      data,
      "sorting_var",
      breakpoint_options(n_portfolios = 5)
    )
    bp_with_filter <- compute_breakpoints(
      data,
      "sorting_var",
      breakpoint_options(n_portfolios = 5, breakpoints_min_size_threshold = 0.2)
    )

    expect_length(bp_with_filter, 6)
    expect_false(identical(bp_no_filter, bp_with_filter))
  }
)

test_that(
  paste0(
    "breakpoints_min_size_threshold produces correct breakpoints"
  ),
  {
    set.seed(7)
    mktcap <- 1:100
    sorting_var <- rnorm(100)
    data <- data.frame(mktcap_lag = mktcap, sorting_var = sorting_var)

    # manual: cutoff = quantile(1:100, 0.2) = 20.8; keep stocks
    # with mktcap > 20.8
    size_cutoff <- quantile(mktcap, 0.2, na.rm = TRUE)
    above <- mktcap > size_cutoff
    expected <- unname(quantile(
      sorting_var[above],
      seq(0, 1, length.out = 6),
      na.rm = TRUE
    ))

    bp <- compute_breakpoints(
      data,
      "sorting_var",
      breakpoint_options(n_portfolios = 5, breakpoints_min_size_threshold = 0.2)
    )

    expect_equal(bp, expected)
  }
)

test_that(
  paste0(
    "breakpoints_min_size_threshold excludes stocks ",
    "exactly at the cutoff (strict > not >=)"
  ),
  {
    # Construct data where one stock lands exactly at the quantile cutoff.
    # With mktcap = c(10, 20, 30, 40, 50) and threshold = 0.25:
    # quantile type 7: h = 1 + 0.25*4 = 2; cutoff = mktcap[2] = 20 (exact
    # integer index)
    mktcap <- c(10, 20, 30, 40, 50)
    sorting_var <- c(1.0, 2.0, 3.0, 4.0, 5.0)
    data <- data.frame(mktcap_lag = mktcap, sorting_var = sorting_var)

    size_cutoff <- quantile(mktcap, 0.25, na.rm = TRUE)
    # stocks with mktcap strictly > cutoff are included; the stock at exactly
    # cutoff is excluded
    above <- mktcap > size_cutoff
    expect_false(above[mktcap == size_cutoff]) # boundary stock excluded

    expected <- unname(quantile(
      sorting_var[above],
      seq(0, 1, length.out = 4),
      na.rm = TRUE
    ))
    bp <- compute_breakpoints(
      data,
      "sorting_var",
      breakpoint_options(
        n_portfolios = 3,
        breakpoints_min_size_threshold = 0.25
      )
    )
    expect_equal(bp, expected)
  }
)

test_that(
  paste0(
    "breakpoints_min_size_threshold with NA mktcap values excludes NA rows"
  ),
  {
    data <- data.frame(
      mktcap_lag = c(NA, 10, 20, 30, 40, 50, 60, 70, 80, 90),
      sorting_var = c(99, 1, 2, 3, 4, 5, 6, 7, 8, 9)
    )

    bp <- compute_breakpoints(
      data,
      "sorting_var",
      breakpoint_options(
        n_portfolios = 3,
        breakpoints_min_size_threshold = 0.2
      )
    )

    # breakpoints must be finite numerics — no NAs
    expect_true(all(!is.na(bp)))
    expect_length(bp, 4)

    # manual: cutoff uses only non-NA mktcap values; row with mktcap=NA is
    # excluded
    size_cutoff <- quantile(
      c(10, 20, 30, 40, 50, 60, 70, 80, 90),
      0.2,
      na.rm = TRUE
    )
    above <- !is.na(data$mktcap_lag) & data$mktcap_lag > size_cutoff
    expected <- unname(quantile(
      data$sorting_var[above],
      seq(0, 1, length.out = 4),
      na.rm = TRUE
    ))
    expect_equal(bp, expected)
  }
)

test_that("breakpoints_min_size_threshold = NULL (default) has no effect", {
  set.seed(42)
  data <- data.frame(
    id = 1:100,
    exchange = rep("NYSE", 100),
    mktcap_lag = 1:100,
    sorting_var = rnorm(100)
  )

  bp_default <- compute_breakpoints(
    data,
    "sorting_var",
    breakpoint_options(n_portfolios = 5)
  )
  bp_explicit_null <- compute_breakpoints(
    data,
    "sorting_var",
    breakpoint_options(n_portfolios = 5, breakpoints_min_size_threshold = NULL)
  )

  expect_identical(bp_default, bp_explicit_null)
})

test_that(
  paste0(
    "breakpoints_min_size_threshold errors when ",
    "mktcap column is missing"
  ),
  {
    data <- data.frame(
      id = 1:100,
      sorting_var = rnorm(100)
    )

    expect_error(
      compute_breakpoints(
        data,
        "sorting_var",
        breakpoint_options(
          n_portfolios = 5,
          breakpoints_min_size_threshold = 0.2
        )
      ),
      "mktcap_lag"
    )
  }
)

test_that("breakpoints_min_size_threshold works with custom data_options", {
  set.seed(42)
  data <- data.frame(
    id = 1:100,
    listing = rep("NYSE", 100),
    mcap = 1:100,
    sorting_var = rnorm(100)
  )

  bp <- compute_breakpoints(
    data,
    "sorting_var",
    breakpoint_options(
      n_portfolios = 5,
      breakpoints_exchanges = "NYSE",
      breakpoints_min_size_threshold = 0.2
    ),
    data_options = data_options(exchange = "listing", mktcap_lag = "mcap")
  )

  expect_length(bp, 6)
})

test_that("breakpoint_options validates breakpoints_min_size_threshold", {
  expect_error(
    breakpoint_options(
      n_portfolios = 5,
      breakpoints_min_size_threshold = 0
    ),
    "breakpoints_min_size_threshold"
  )
  expect_error(
    breakpoint_options(
      n_portfolios = 5,
      breakpoints_min_size_threshold = 1
    ),
    "breakpoints_min_size_threshold"
  )
  expect_error(
    breakpoint_options(
      n_portfolios = 5,
      breakpoints_min_size_threshold = -0.1
    ),
    "breakpoints_min_size_threshold"
  )
  expect_error(
    breakpoint_options(
      n_portfolios = 5,
      breakpoints_min_size_threshold = "abc"
    ),
    "breakpoints_min_size_threshold"
  )
  expect_error(
    breakpoint_options(
      n_portfolios = 5,
      breakpoints_min_size_threshold = c(0.2, 0.3)
    ),
    "breakpoints_min_size_threshold"
  )
  expect_error(
    breakpoint_options(
      n_portfolios = 5,
      breakpoints_min_size_threshold = NA
    ),
    "breakpoints_min_size_threshold"
  )
})

make_monthly_df <- function(n = 12, start = "2020-01-01", value = seq_len(n)) {
  tibble::tibble(
    date = seq.Date(as.Date(start), by = "month", length.out = n),
    value = value
  )
}

test_that("errors when data has no 'date' column", {
  df <- tibble::tibble(x = 1:5)
  expect_error(
    compute_rolling_value(df, .f = ~ mean(.x$x), period = "month", periods = 3),
    "date"
  )
})

test_that("errors when 'date' column is not Date class", {
  df <- tibble::tibble(date = as.character(Sys.Date() + 0:4), value = 1:5)
  expect_error(
    compute_rolling_value(
      df,
      .f = ~ mean(.x$value),
      period = "month",
      periods = 3
    ),
    "Date"
  )
})

test_that("errors when period is not a single string", {
  df <- make_monthly_df(5)
  expect_error(
    compute_rolling_value(df, .f = ~ mean(.x$value), period = 123, periods = 3)
  )
  expect_error(
    compute_rolling_value(
      df,
      .f = ~ mean(.x$value),
      period = c("month", "year"),
      periods = 3
    )
  )
})

test_that("returns a numeric vector with length equal to nrow(data)", {
  df <- make_monthly_df(12)
  result <- compute_rolling_value(
    df,
    .f = ~ mean(.x$value),
    period = "month",
    periods = 3
  )
  expect_type(result, "double")
  expect_length(result, nrow(df))
})

test_that("min_obs defaults to periods; early windows return NA", {
  df <- make_monthly_df(6)
  result <- compute_rolling_value(
    df,
    .f = ~ mean(.x$value),
    period = "month",
    periods = 3
  )
  # First two windows have fewer than 3 rows → NA

  expect_true(is.na(result[1]))
  expect_true(is.na(result[2]))
  # Third window onwards should be non-NA
  expect_false(any(is.na(result[3:6])))
})

test_that("custom min_obs produces fewer leading NAs", {
  df <- make_monthly_df(6)
  result <- compute_rolling_value(
    df,
    .f = ~ mean(.x$value),
    period = "month",
    periods = 3,
    min_obs = 1
  )
  # Even the first row (1 observation) should be non-NA

  expect_false(any(is.na(result)))
})

test_that("min_obs larger than periods makes more windows NA", {
  df <- make_monthly_df(6)
  # periods = 3 but min_obs = 5 → only windows with ≥5 rows produce values

  result <- compute_rolling_value(
    df,
    .f = ~ mean(.x$value),
    period = "month",
    periods = 6,
    min_obs = 5
  )
  expect_true(is.na(result[1]))
  expect_true(is.na(result[4]))
  expect_false(is.na(result[5]))
})

test_that("rows with NA values are dropped before applying .f", {
  df <- make_monthly_df(6)
  df$value[2] <- NA
  df$value[3] <- NA

  # Window for row 4 spans rows 2-4; after dropping NAs only row 4 remains
  result <- compute_rolling_value(
    df,
    .f = ~ mean(.x$value),
    period = "month",
    periods = 3,
    min_obs = 1
  )
  expect_type(result, "double")
  # Row 2 has NA in value → after complete cases the window for row 2 (rows 1-2)

  # has only row 1 → min_obs = 1 so still non-NA

  expect_false(is.na(result[1]))
})

test_that("window returns NA when complete cases < min_obs due to NAs", {
  df <- make_monthly_df(6)
  df$value[1] <- NA
  df$value[2] <- NA
  df$value[3] <- NA

  result <- compute_rolling_value(
    df,
    .f = ~ mean(.x$value),
    period = "month",
    periods = 3,
    min_obs = 2
  )
  # Windows covering rows 1-3 have 0 complete rows → NA

  expect_true(is.na(result[1]))
  expect_true(is.na(result[2]))
  expect_true(is.na(result[3]))
})

test_that("rolling mean with periods = 1 equals the original values", {
  df <- make_monthly_df(5, value = c(10, 20, 30, 40, 50))
  result <- compute_rolling_value(
    df,
    .f = ~ mean(.x$value),
    period = "month",
    periods = 1,
    min_obs = 1
  )
  expect_equal(result, c(10, 20, 30, 40, 50))
})

test_that("rolling mean with periods = 3 computes correct values", {
  df <- make_monthly_df(5, value = c(10, 20, 30, 40, 50))
  result <- compute_rolling_value(
    df,
    .f = ~ mean(.x$value),
    period = "month",
    periods = 3,
    min_obs = 1
  )
  expect_equal(result[1], 10) # window: 10

  expect_equal(result[2], mean(c(10, 20))) # window: 10, 20
  expect_equal(result[3], mean(c(10, 20, 30))) # window: 10, 20, 30
  expect_equal(result[4], mean(c(20, 30, 40))) # window: 20, 30, 40
  expect_equal(result[5], mean(c(30, 40, 50))) # window: 30, 40, 50
})

test_that("rolling sum works correctly", {
  df <- make_monthly_df(5, value = c(1, 2, 3, 4, 5))
  result <- compute_rolling_value(
    df,
    .f = ~ sum(.x$value),
    period = "month",
    periods = 3,
    min_obs = 1
  )
  expect_equal(result[3], 6)
  expect_equal(result[4], 9)
  expect_equal(result[5], 12)
})

test_that("rolling sd works correctly", {
  set.seed(42)
  df <- make_monthly_df(6, value = c(1, 3, 2, 5, 4, 6))
  result <- compute_rolling_value(
    df,
    .f = ~ sd(.x$value),
    period = "month",
    periods = 3,
    min_obs = 2
  )
  expect_equal(result[3], sd(c(1, 3, 2)))
  expect_equal(result[5], sd(c(2, 5, 4)))
})

test_that("accepts purrr-style formula (.f = ~ ...)", {
  df <- make_monthly_df(5, value = c(1, 2, 3, 4, 5))
  result <- compute_rolling_value(
    df,
    .f = ~ sum(.x$value),
    period = "month",
    periods = 2,
    min_obs = 1
  )
  expect_equal(result[4], 7)
})

test_that("accepts a regular function for .f", {
  my_fn <- function(d) max(d$value)
  df <- make_monthly_df(4, value = c(10, 5, 20, 3))
  result <- compute_rolling_value(
    df,
    .f = my_fn,
    period = "month",
    periods = 3,
    min_obs = 1
  )
  expect_equal(result[3], 20)
  expect_equal(result[4], 20)
})

test_that("works with period = 'quarter'", {
  # 8 months → 3 quarters (Q1: Jan-Mar, Q2: Apr-Jun, Q3: Jul-Aug)
  df <- tibble::tibble(
    date = seq.Date(as.Date("2020-01-01"), by = "month", length.out = 8),
    value = 1:8
  )
  result <- compute_rolling_value(
    df,
    .f = ~ mean(.x$value),
    period = "quarter",
    periods = 2,
    min_obs = 1
  )

  expect_length(result, 3)
  expect_type(result, "double")
})

test_that("works with period = 'year'", {
  df <- tibble::tibble(
    date = seq.Date(as.Date("2018-06-01"), by = "quarter", length.out = 12),
    value = 1:12
  )
  result <- compute_rolling_value(
    df,
    .f = ~ mean(.x$value),
    period = "year",
    periods = 2,
    min_obs = 1
  )
  expect_length(result, 4)
  expect_type(result, "double")
})

test_that("works with multiple columns (e.g., regression residuals)", {
  set.seed(99)
  df <- tibble::tibble(
    date = seq.Date(as.Date("2020-01-01"), by = "month", length.out = 24),
    y = rnorm(24),
    x = rnorm(24)
  )
  result <- compute_rolling_value(
    df,
    .f = ~ {
      fit <- lm(y ~ x, data = .x)
      tail(fit$residuals, 1)
    },
    period = "month",
    periods = 12,
    min_obs = 6
  )
  expect_length(result, 24)
  expect_true(all(is.na(result[1:5])))
  expect_false(any(is.na(result[12:24])))
})

test_that("single-row data frame works", {
  df <- make_monthly_df(1, value = 42)
  result <- compute_rolling_value(
    df,
    .f = ~ mean(.x$value),
    period = "month",
    periods = 1,
    min_obs = 1
  )
  expect_equal(result, 42)
})

test_that("single-row data frame returns NA when min_obs > 1", {
  df <- make_monthly_df(1, value = 42)
  result <- compute_rolling_value(
    df,
    .f = ~ mean(.x$value),
    period = "month",
    periods = 3,
    min_obs = 2
  )
  expect_true(is.na(result))
})

test_that("all-NA value column returns all NAs", {
  df <- make_monthly_df(5, value = rep(NA_real_, 5))
  result <- compute_rolling_value(
    df,
    .f = ~ mean(.x$value),
    period = "month",
    periods = 3,
    min_obs = 1
  )
  expect_true(all(is.na(result)))
})

test_that("periods = nrow(data) uses full history for last row", {
  df <- make_monthly_df(5, value = c(2, 4, 6, 8, 10))
  result <- compute_rolling_value(
    df,
    .f = ~ mean(.x$value),
    period = "month",
    periods = 5,
    min_obs = 5
  )
  # Only the last row has a full 5-period window

  expect_true(all(is.na(result[1:4])))
  expect_equal(result[5], mean(c(2, 4, 6, 8, 10)))
})

test_that("data_options with non-default date column works", {
  df <- tibble::tibble(
    month = seq.Date(as.Date("2020-01-01"), by = "month", length.out = 12),
    value = seq_len(12)
  )
  result <- compute_rolling_value(
    df,
    .f = ~ mean(.x$value),
    period = "month",
    periods = 3,
    min_obs = 1,
    data_options = data_options(date = "month")
  )
  expect_type(result, "double")
  expect_length(result, 12)
  expect_false(any(is.na(result)))
})

test_that("non-default date column produces same results as default", {
  dates <- seq.Date(as.Date("2020-01-01"), by = "month", length.out = 6)
  vals <- c(10, 20, 30, 40, 50, 60)

  df_default <- tibble::tibble(date = dates, value = vals)
  df_custom <- tibble::tibble(trade_date = dates, value = vals)

  res_default <- compute_rolling_value(
    df_default,
    .f = ~ mean(.x$value),
    period = "month",
    periods = 3,
    min_obs = 1
  )
  res_custom <- compute_rolling_value(
    df_custom,
    .f = ~ mean(.x$value),
    period = "month",
    periods = 3,
    min_obs = 1,
    data_options = data_options(date = "trade_date")
  )

  expect_equal(res_default, res_custom)
})

test_that("errors when mapped date column is absent from data", {
  df <- make_monthly_df(6)
  expect_error(
    compute_rolling_value(
      df,
      .f = ~ mean(.x$value),
      period = "month",
      periods = 3,
      data_options = data_options(date = "month")
    ),
    "month"
  )
})

test_that("errors when mapped date column is not of class Date", {
  df <- tibble::tibble(
    month = as.character(seq.Date(
      as.Date("2020-01-01"),
      by = "month",
      length.out = 6
    )),
    value = 1:6
  )
  expect_error(
    compute_rolling_value(
      df,
      .f = ~ mean(.x$value),
      period = "month",
      periods = 3,
      data_options = data_options(date = "month")
    ),
    "Date"
  )
})

test_that("errors when data_options$date is NULL", {
  df <- make_monthly_df(6)
  opts <- data_options()
  opts$date <- NULL

  expect_error(
    compute_rolling_value(
      df,
      .f = ~ mean(.x$value),
      period = "month",
      periods = 3,
      data_options = opts
    ),
    "data_options"
  )
})

test_that("errors when data_options$date is NA", {
  df <- make_monthly_df(6)
  opts <- data_options()
  opts$date <- NA_character_

  expect_error(
    compute_rolling_value(
      df,
      .f = ~ mean(.x$value),
      period = "month",
      periods = 3,
      data_options = opts
    ),
    "data_options"
  )
})

test_that("errors when data_options$date is not character", {
  df <- make_monthly_df(6)
  opts <- data_options()
  opts$date <- 42

  expect_error(
    compute_rolling_value(
      df,
      .f = ~ mean(.x$value),
      period = "month",
      periods = 3,
      data_options = opts
    ),
    "data_options"
  )
})

test_that("errors when data_options$date has length > 1", {
  df <- make_monthly_df(6)
  opts <- data_options()
  opts$date <- c("date", "month")

  expect_error(
    compute_rolling_value(
      df,
      .f = ~ mean(.x$value),
      period = "month",
      periods = 3,
      data_options = opts
    ),
    "data_options"
  )
})

set.seed(42)
data <- tibble::tibble(
  permno = rep(1:2, each = 10),
  date = rep(seq.Date(as.Date("2023-01-01"), by = "month", length.out = 10), 2),
  bm = runif(20, 0.5, 1.5),
  size = runif(20, 100, 200)
)

# --- exact lag tests ---

test_that("add_lagged_columns adds lagged columns", {
  result <- add_lagged_columns(
    data,
    cols = c("bm", "size"),
    lag = months(3),
    by = "permno"
  )
  expect_true("bm_lag" %in% colnames(result))
  expect_true("size_lag" %in% colnames(result))
  expect_equal(nrow(result), nrow(data))
})

test_that("add_lagged_columns works without grouping", {
  result <- add_lagged_columns(
    data |> dplyr::filter(permno == 1),
    cols = c("bm", "size"),
    lag = months(3)
  )
  expect_true("bm_lag" %in% colnames(result))
  expect_true("size_lag" %in% colnames(result))
})

test_that("add_lagged_columns preserves original column values", {
  result <- add_lagged_columns(
    data,
    cols = c("bm", "size"),
    lag = months(3),
    by = "permno"
  )
  expect_equal(result$bm, data$bm)
  expect_equal(result$size, data$size)
})

test_that("add_lagged_columns handles drop_na correctly", {
  data_with_na <- data
  data_with_na$bm[c(1, 11)] <- NA

  result <- add_lagged_columns(
    data_with_na,
    cols = "bm",
    lag = months(3),
    by = "permno",
    drop_na = TRUE
  )

  # Row 4 (date = 2023-04-01) looks back to 2023-01-01 which is NA and dropped
  expect_true(is.na(result$bm_lag[result$permno == 1][4]))
  # Row 5 (date = 2023-05-01) looks back to 2023-02-01 which is not NA
  expect_false(is.na(result$bm_lag[result$permno == 1][5]))
})

test_that("add_lagged_columns works with integer lag", {
  daily_data <- tibble::tibble(
    date = as.Date("2023-01-01") + 0:9,
    value = 1:10
  )
  result <- add_lagged_columns(daily_data, cols = "value", lag = 1)
  expect_equal(result$value_lag, c(NA, 1, 2, 3, 4, 5, 6, 7, 8, 9))
})

test_that("add_lagged_columns works with custom data_options", {
  custom_data <- data |> dplyr::rename(my_date = date)
  result <- add_lagged_columns(
    custom_data,
    cols = "bm",
    lag = months(3),
    by = "permno",
    data_options = data_options(date = "my_date")
  )
  expect_true("bm_lag" %in% colnames(result))
})

# --- error handling ---

test_that("add_lagged_columns errors on missing date column", {
  expect_error(
    add_lagged_columns(
      data |> dplyr::select(-date),
      cols = "bm",
      lag = months(3)
    ),
    "date"
  )
})

test_that("add_lagged_columns errors on missing cols", {
  expect_error(
    add_lagged_columns(data, cols = "nonexistent", lag = months(3)),
    "nonexistent"
  )
})

test_that("add_lagged_columns errors on missing by columns", {
  expect_error(
    add_lagged_columns(data, cols = "bm", lag = months(3), by = "nonexistent"),
    "nonexistent"
  )
})

test_that("add_lagged_columns errors on duplicate by-date combinations", {
  dup_data <- dplyr::bind_rows(data, data[1, ])
  expect_error(
    add_lagged_columns(dup_data, cols = "bm", lag = months(3), by = "permno"),
    "unique"
  )
})

test_that("add_lagged_columns errors on invalid lag/max_lag", {
  expect_error(
    add_lagged_columns(data, cols = "bm", lag = -1, by = "permno"),
    "non-negative"
  )
  expect_error(
    add_lagged_columns(
      data,
      cols = "bm",
      lag = months(6),
      max_lag = months(3),
      by = "permno"
    ),
    "max_lag"
  )
})

# --- max_lag tests (window / inequality join path) ---

test_that("add_lagged_columns with max_lag adds lagged columns", {
  result <- add_lagged_columns(
    data,
    cols = c("bm", "size"),
    lag = months(3),
    max_lag = months(6),
    by = "permno"
  )
  expect_true("bm_lag" %in% colnames(result))
  expect_true("size_lag" %in% colnames(result))
})

test_that("add_lagged_columns with max_lag preserves row count", {
  result <- add_lagged_columns(
    data,
    cols = "size",
    lag = months(2),
    max_lag = months(5),
    by = "permno"
  )
  expect_equal(nrow(result), nrow(data))
})

test_that("add_lagged_columns with max_lag and drop_na", {
  data_with_na <- data
  data_with_na$bm[c(1, 5, 11, 15)] <- NA

  result <- add_lagged_columns(
    data_with_na,
    cols = "bm",
    lag = months(2),
    max_lag = months(4),
    by = "permno",
    drop_na = TRUE
  )

  expect_equal(nrow(result), nrow(data_with_na))
  expect_true("bm_lag" %in% colnames(result))
})

test_that(
  paste("add_lagged_columns with max_lag returns NA when no match in window"),
  {
    daily_data <- tibble::tibble(
      date = as.Date("2023-01-01") + 0:9,
      value = 1:10
    )
    result <- add_lagged_columns(
      daily_data,
      cols = "value",
      lag = 10,
      max_lag = 20
    )
    expect_true(all(is.na(result$value_lag)))
  }
)

test_that("add_lagged_columns with max_lag works with integer lags", {
  daily_data <- tibble::tibble(
    date = as.Date("2023-01-01") + 0:9,
    value = 1:10
  )
  result <- add_lagged_columns(daily_data, cols = "value", lag = 1, max_lag = 3)
  expected <- c(NA, 1, 2, 3, 4, 5, 6, 7, 8, 9)
  expect_equal(result$value_lag, expected)
})

test_that("add_lagged_columns with max_lag without grouping", {
  result <- add_lagged_columns(
    data |> dplyr::filter(permno == 1),
    cols = "size",
    lag = months(2),
    max_lag = months(4)
  )
  expect_true("size_lag" %in% colnames(result))
  expect_equal(nrow(result), 10)
})

test_that(
  paste(
    "add_lagged_columns with max_lag & non-NULL by produces correct",
    "per-group values"
  ),
  {
    # Regression test: prior to fix, !!!by spliced raw strings into join_by()
    # instead of symbols, causing the group equality constraint to be silently
    # dropped and values to bleed across groups in the inequality-join path
    # (lag < max_lag).
    det_data <- tibble::tibble(
      permno = rep(1:2, each = 4),
      date = rep(
        as.Date(c("2023-01-01", "2023-02-01", "2023-03-01", "2023-04-01")),
        2
      ),
      value = c(10, 20, 30, 40, 100, 200, 300, 400)
    )

    result <- add_lagged_columns(
      det_data,
      cols = "value",
      lag = months(1),
      max_lag = months(2),
      by = "permno"
    )

    # For 2023-03-01, the window is [Jan, Feb]; most recent is Feb.
    # For 2023-04-01, the window is [Feb, Mar]; most recent is Mar.
    p1 <- result[result$permno == 1, ]
    p2 <- result[result$permno == 2, ]

    expect_equal(p1$value_lag, c(NA, 10, 20, 30))
    expect_equal(p2$value_lag, c(NA, 100, 200, 300))
  }
)

test_that("basic lag join works with single id and single variable", {
  df1 <- tibble(
    id = rep(1, 4),
    date = as.Date(c("2020-01-01", "2020-02-01", "2020-03-01", "2020-04-01"))
  )

  df2 <- tibble(
    id = rep(1, 4),
    date = as.Date(c("2020-01-01", "2020-02-01", "2020-03-01", "2020-04-01")),
    x = c(10, 20, 30, 40)
  )

  result <- join_lagged_values(
    original_data = df1,
    new_data = df2,
    id_keys = "id",
    min_lag = months(1),
    max_lag = months(3)
  )

  expect_true("x" %in% names(result))
  expect_equal(nrow(result), 4)
  # Row with date 2020-01-01: looking for new_data where .lower <= date and
  # date <= .upper new_data row date=2020-01-01 -> .lower=2020-02-01,
  # .upper=2020-04-01 -> 2020-01-01 < .lower, no
  # So first row should be NA (nothing lagged into it yet)
  expect_true(is.na(result$x[1]))
})

test_that("multiple id keys are supported", {
  df1 <- tibble(
    id1 = c(1, 1, 2, 2),
    id2 = c("a", "a", "b", "b"),
    date = as.Date(c("2020-01-01", "2020-03-01", "2020-01-01", "2020-03-01"))
  )

  df2 <- tibble(
    id1 = c(1, 1, 2, 2),
    id2 = c("a", "a", "b", "b"),
    date = as.Date(c("2020-01-01", "2020-02-01", "2020-01-01", "2020-02-01")),
    val = c(100, 200, 300, 400)
  )

  result <- join_lagged_values(
    original_data = df1,
    new_data = df2,
    id_keys = c("id1", "id2"),
    min_lag = months(1),
    max_lag = months(3)
  )

  expect_true("val" %in% names(result))
  expect_equal(nrow(result), 4)
  # IDs should not bleed across groups
  # For id1=2, id2="b", date=2020-03-01: should match new_data with
  # id1=2, id2="b"
  row_2b_march <- result |>
    filter(id1 == 2, id2 == "b", date == as.Date("2020-03-01"))
  row_1a_march <- result |>
    filter(id1 == 1, id2 == "a", date == as.Date("2020-03-01"))
  # These should come from different source rows
  expect_false(
    identical(row_2b_march$val, row_1a_march$val) && !is.na(row_2b_march$val)
  )
})

test_that("multiple new columns are joined", {
  df1 <- tibble(
    id = rep(1, 3),
    date = as.Date(c("2020-02-01", "2020-03-01", "2020-04-01"))
  )

  df2 <- tibble(
    id = rep(1, 3),
    date = as.Date(c("2020-01-01", "2020-02-01", "2020-03-01")),
    x = c(1, 2, 3),
    y = c(10, 20, 30)
  )

  result <- join_lagged_values(
    original_data = df1,
    new_data = df2,
    id_keys = "id",
    min_lag = months(1),
    max_lag = months(3)
  )

  expect_true(all(c("x", "y") %in% names(result)))
  expect_equal(ncol(result), 4) # id, date, x, y
})

test_that("no match produces NAs", {
  df1 <- tibble(
    id = 1,
    date = as.Date("2025-01-01")
  )

  df2 <- tibble(
    id = 1,
    date = as.Date("2020-01-01"),
    x = 999
  )

  result <- join_lagged_values(
    original_data = df1,
    new_data = df2,
    id_keys = "id",
    min_lag = months(1),
    max_lag = months(3)
  )

  expect_true(is.na(result$x[1]))
})

test_that("different ids do not cross-contaminate", {
  df1 <- tibble(
    id = c(1, 2),
    date = as.Date(c("2020-03-01", "2020-03-01"))
  )

  df2 <- tibble(
    id = c(1, 2),
    date = as.Date(c("2020-01-01", "2020-01-01")),
    x = c(100, 200)
  )

  result <- join_lagged_values(
    original_data = df1,
    new_data = df2,
    id_keys = "id",
    min_lag = months(1),
    max_lag = months(6)
  )

  expect_equal(result$x[result$id == 1], 100)
  expect_equal(result$x[result$id == 2], 200)
})

test_that("ff_adjustment picks latest date per year", {
  df1 <- tibble(
    id = rep(1, 2),
    date = as.Date(c("2021-06-01", "2021-09-01"))
  )

  # Two rows in 2020 for same id; ff_adjustment should pick the later one
  df2 <- tibble(
    id = rep(1, 3),
    date = as.Date(c("2020-06-01", "2020-12-01", "2021-03-01")),
    x = c(10, 20, 30)
  )

  result <- join_lagged_values(
    original_data = df1,
    new_data = df2,
    id_keys = "id",
    min_lag = months(6),
    max_lag = months(18),
    ff_adjustment = TRUE
  )

  expect_true("x" %in% names(result))
  # With ff_adjustment, only the max-date row per (id, year) is kept
  expect_equal(result$x[1], 20)
})

test_that(
  paste0(
    "input validation: duplicate columns between ",
    "new_data and original_data"
  ),
  {
    df1 <- tibble(
      id = rep(1, 3),
      date = as.Date(c("2020-02-01", "2020-03-01", "2020-04-01")),
      bm = runif(3)
    )

    df2 <- tibble(
      id = rep(1, 3),
      date = as.Date(c("2020-01-01", "2020-02-01", "2020-03-01")),
      bm = runif(3)
    )

    expect_error(
      join_lagged_values(
        df1,
        df2,
        id_keys = "id",
        min_lag = months(1),
        max_lag = months(3)
      ),
      regexp = "already exist.*original_data.*bm"
    )
  }
)

test_that("input validation: id_keys must be character", {
  df <- tibble(id = 1, date = as.Date("2020-01-01"))
  expect_error(
    join_lagged_values(
      df,
      df,
      id_keys = 1,
      min_lag = months(1),
      max_lag = months(3)
    )
  )
})

test_that("input validation: date column must exist in both datasets", {
  df1 <- tibble(id = 1, date = as.Date("2020-01-01"))
  df2 <- tibble(id = 1, dt = as.Date("2020-01-01"), x = 1)

  expect_error(
    join_lagged_values(
      df1,
      df2,
      id_keys = "id",
      min_lag = months(1),
      max_lag = months(3),
      data_options = data_options(date = "dt")
    )
  )
})

test_that("custom date column name via data_options works", {
  df1 <- tibble(id = 1, dt = as.Date(c("2020-03-01", "2020-04-01")))
  df2 <- tibble(
    id = 1,
    dt = as.Date(c("2020-01-01", "2020-02-01")),
    x = c(5, 10)
  )

  result <- join_lagged_values(
    original_data = df1,
    new_data = df2,
    id_keys = "id",
    min_lag = months(1),
    max_lag = months(4),
    data_options = data_options(date = "dt")
  )

  expect_true("x" %in% names(result))
  expect_equal(nrow(result), 2)
})

test_that("original_data columns are preserved", {
  df1 <- tibble(
    id = rep(1, 3),
    date = as.Date(c("2020-02-01", "2020-03-01", "2020-04-01")),
    existing_col = c("a", "b", "c")
  )

  df2 <- tibble(
    id = rep(1, 3),
    date = as.Date(c("2020-01-01", "2020-02-01", "2020-03-01")),
    x = c(1, 2, 3)
  )

  result <- join_lagged_values(
    original_data = df1,
    new_data = df2,
    id_keys = "id",
    min_lag = months(1),
    max_lag = months(3)
  )

  expect_true("existing_col" %in% names(result))
  expect_equal(result$existing_col, c("a", "b", "c"))
})

test_that("empty original_data returns empty result with correct columns", {
  df1 <- tibble(id = integer(), date = as.Date(character()))
  df2 <- tibble(id = 1, date = as.Date("2020-01-01"), x = 1)

  result <- join_lagged_values(
    original_data = df1,
    new_data = df2,
    id_keys = "id",
    min_lag = months(1),
    max_lag = months(3)
  )

  expect_equal(nrow(result), 0)
  expect_true("x" %in% names(result))
})

test_that("closest match is used within the lag window", {
  # With multiple new_data dates falling in the window, closest() should pick

  df1 <- tibble(
    id = 1,
    date = as.Date("2020-06-01")
  )

  df2 <- tibble(
    id = rep(1, 4),
    date = as.Date(c("2020-01-01", "2020-02-01", "2020-03-01", "2020-04-01")),
    x = c(10, 20, 30, 40)
  )

  result <- join_lagged_values(
    original_data = df1,
    new_data = df2,
    id_keys = "id",
    min_lag = months(1),
    max_lag = months(6)
  )

  # .lower for each new_data row: +1 month from date
  # 2020-01-01 -> .lower=2020-02-01, .upper=2020-08-01
  # 2020-02-01 -> .lower=2020-03-01, .upper=2020-09-01
  # 2020-03-01 -> .lower=2020-04-01, .upper=2020-10-01
  # 2020-04-01 -> .lower=2020-05-01, .upper=2020-11-01
  # original date=2020-06-01: closest(.date >= .lower)
  # picks the largest .lower <= 2020-06-01
  # That's .lower=2020-05-01 from date=2020-04-01, so x=40
  expect_equal(result$x[1], 40)
})

#' List Supported Fama-French Dataset Types
#'
#' Returns a tibble with the supported Fama-French dataset types, including
#' their names and frequencies (daily, weekly, monthly). Each dataset type
#' is associated with a specific Fama-French model (e.g., 3 factors, 5
#' factors). Additionally, it annotates each dataset with the domain
#' "Fama-French".
#'
#' @returns A tibble with columns: `type` (the type of dataset),
#'   `dataset_name` (a descriptive name of the dataset), and `domain` (the
#'   domain to which the dataset belongs, always "Fama-French").
#'
#' @family utility functions
list_supported_types_ff <- function() {
  # nolint start
  # To update this list, run data-raw/list_supported_types_ff.R and replace
  # the tribble() call below with the output of datapasta::tribble_paste().
  tribble(
    ~type                                                                                                         ,
    ~dataset_name                                                                                                 ,
    "factors_ff_3_monthly"                                                                                        ,
    "Fama/French 3 Factors"                                                                                       ,
    "factors_ff_3_weekly"                                                                                         ,
    "Fama/French 3 Factors [Weekly]"                                                                              ,
    "factors_ff_3_daily"                                                                                          ,
    "Fama/French 3 Factors [Daily]"                                                                               ,
    "factors_ff_5_2x3_monthly"                                                                                    ,
    "Fama/French 5 Factors (2x3)"                                                                                 ,
    "factors_ff_5_2x3_daily"                                                                                      ,
    "Fama/French 5 Factors (2x3) [Daily]"                                                                         ,
    "factors_ff_size_monthly"                                                                                     ,
    "Portfolios Formed on Size"                                                                                   ,
    "factors_ff_size_exdividends_monthly"                                                                         ,
    "Portfolios Formed on Size [ex.Dividends]"                                                                    ,
    "factors_ff_size_daily"                                                                                       ,
    "Portfolios Formed on Size [Daily]"                                                                           ,
    "factors_ff_bm_monthly"                                                                                       ,
    "Portfolios Formed on Book-to-Market"                                                                         ,
    "factors_ff_bm_exdividends_monthly"                                                                           ,
    "Portfolios Formed on Book-to-Market [ex. Dividends]"                                                         ,
    "factors_ff_bm_daily"                                                                                         ,
    "Portfolios Formed on Book-to-Market [Daily]"                                                                 ,
    "factors_ff_op_monthly"                                                                                       ,
    "Portfolios Formed on Operating Profitability"                                                                ,
    "factors_ff_op_exdividends_monthly"                                                                           ,
    "Portfolios Formed on Operating Profitability [ex. Dividends]"                                                ,
    "factors_ff_op_daily"                                                                                         ,
    "Portfolios Formed on Operating Profitability [Daily]"                                                        ,
    "factors_ff_inv_monthly"                                                                                      ,
    "Portfolios Formed on Investment"                                                                             ,
    "factors_ff_inv_exdividends_monthly"                                                                          ,
    "Portfolios Formed on Investment [ex. Dividends]"                                                             ,
    "factors_ff_inv_daily"                                                                                        ,
    "Portfolios Formed on Investment [Daily]"                                                                     ,
    "factors_ff_6_size_and_bm_2_x_3_monthly"                                                                      ,
    "6 Portfolios Formed on Size and Book-to-Market (2 x 3)"                                                      ,
    "factors_ff_6_size_and_bm_2_x_3_exdividends_monthly"                                                          ,
    "6 Portfolios Formed on Size and Book-to-Market (2 x 3) [ex. Dividends]"                                      ,
    "factors_ff_6_size_and_bm_2_x_3_weekly"                                                                       ,
    "6 Portfolios Formed on Size and Book-to-Market (2 x 3) [Weekly]"                                             ,
    "factors_ff_6_size_and_bm_2_x_3_daily"                                                                        ,
    "6 Portfolios Formed on Size and Book-to-Market (2 x 3) [Daily]"                                              ,
    "factors_ff_25_size_and_bm_5_x_5_monthly"                                                                     ,
    "25 Portfolios Formed on Size and Book-to-Market (5 x 5)"                                                     ,
    "factors_ff_25_size_and_bm_5_x_5_exdividends_monthly"                                                         ,
    "25 Portfolios Formed on Size and Book-to-Market (5 x 5) [ex. Dividends]"                                     ,
    "factors_ff_25_size_and_bm_5_x_5_daily"                                                                       ,
    "25 Portfolios Formed on Size and Book-to-Market (5 x 5) [Daily]"                                             ,
    "factors_ff_100_size_and_bm_10_x_10_monthly"                                                                  ,
    "100 Portfolios Formed on Size and Book-to-Market (10 x 10)"                                                  ,
    "factors_ff_100_size_and_bm_10_x_10_exdividends_monthly"                                                      ,
    "100 Portfolios Formed on Size and Book-to-Market (10 x 10) [ex. Dividends]"                                  ,
    "factors_ff_100_size_and_bm_10_x_10_daily"                                                                    ,
    "100 Portfolios Formed on Size and Book-to-Market (10 x 10) [Daily]"                                          ,
    "factors_ff_6_size_and_op_2_x_3_monthly"                                                                      ,
    "6 Portfolios Formed on Size and Operating Profitability (2 x 3)"                                             ,
    "factors_ff_6_size_and_op_2_x_3_exdividends_monthly"                                                          ,
    "6 Portfolios Formed on Size and Operating Profitability (2 x 3) [ex. Dividends]"                             ,
    "factors_ff_6_size_and_op_2_x_3_daily"                                                                        ,
    "6 Portfolios Formed on Size and Operating Profitability (2 x 3) [Daily]"                                     ,
    "factors_ff_25_size_and_op_5_x_5_monthly"                                                                     ,
    "25 Portfolios Formed on Size and Operating Profitability (5 x 5)"                                            ,
    "factors_ff_25_size_and_op_5_x_5_exdividends_monthly"                                                         ,
    "25 Portfolios Formed on Size and Operating Profitability (5 x 5) [ex. Dividends]"                            ,
    "factors_ff_25_size_and_op_5_x_5_daily"                                                                       ,
    "25 Portfolios Formed on Size and Operating Profitability_(5 x 5) [Daily]"                                    ,
    "factors_ff_100_size_and_op_10_x_10_monthly"                                                                  ,
    "100 Portfolios Formed on Size and Operating Profitability (10 x 10)"                                         ,
    "factors_ff_100_size_and_op_10_x_10_exdividends_monthly"                                                      ,
    "100 Portfolios Formed on Size and Operating Profitability (10 x 10) [ex. Dividends]"                         ,
    "factors_ff_100_size_and_op_10_x_10_daily"                                                                    ,
    "100 Portfolios Formed on Size and Operating Profitability (10 x 10) [Daily]"                                 ,
    "factors_ff_6_size_and_inv_2_x_3_monthly"                                                                     ,
    "6 Portfolios Formed on Size and Investment (2 x 3)"                                                          ,
    "factors_ff_6_size_and_inv_2_x_3_exdividends_monthly"                                                         ,
    "6 Portfolios Formed on Size and Investment (2 x 3) [ex. Dividends]"                                          ,
    "factors_ff_6_size_and_inv_2_x_3_daily"                                                                       ,
    "6 Portfolios Formed on Size and Investment (2 x 3) [Daily]"                                                  ,
    "factors_ff_25_size_and_inv_5_x_5_monthly"                                                                    ,
    "25 Portfolios Formed on Size and Investment (5 x 5)"                                                         ,
    "factors_ff_25_size_and_inv_5_x_5_exdividends_monthly"                                                        ,
    "25 Portfolios Formed on Size and Investment (5 x 5) [ex. Dividends]"                                         ,
    "factors_ff_25_size_and_inv_5_x_5_daily"                                                                      ,
    "25 Portfolios Formed on Size and Investment (5 x 5) [Daily]"                                                 ,
    "factors_ff_100_size_and_inv_10_x_10_monthly"                                                                 ,
    "100 Portfolios Formed on Size and Investment (10 x 10)"                                                      ,
    "factors_ff_100_size_and_inv_10_x_10_exdividends_monthly"                                                     ,
    "100 Portfolios Formed on Size and Investment (10 x 10) [ex. Dividends]"                                      ,
    "factors_ff_100_size_and_inv_10_x_10_daily"                                                                   ,
    "100 Portfolios Formed on Size and Investment (10 x 10) [Daily]"                                              ,
    "factors_ff_25_bm_and_op_5_x_5_monthly"                                                                       ,
    "25 Portfolios Formed on Book-to-Market and Operating Profitability (5 x 5)"                                  ,
    "factors_ff_25_bm_and_op_5_x_5_exdividends_monthly"                                                           ,
    "25 Portfolios Formed on Book-to-Market and Operating Profitability (5 x 5) [ex. Dividends]"                  ,
    "factors_ff_25_bm_and_op_5_x_5_daily"                                                                         ,
    "25 Portfolios Formed on Book-to-Market and Operating Profitability (5 x 5) [Daily]"                          ,
    "factors_ff_25_bm_and_inv_5_x_5_monthly"                                                                      ,
    "25 Portfolios Formed on Book-to-Market and Investment (5 x 5)"                                               ,
    "factors_ff_25_bm_and_inv_5_x_5_exdividends_monthly"                                                          ,
    "25 Portfolios Formed on Book-to-Market and Investment (5 x 5) [ex. Dividends]"                               ,
    "factors_ff_25_bm_and_inv_5_x_5_daily"                                                                        ,
    "25 Portfolios Formed on Book-to-Market and Investment (5 x 5) [Daily]"                                       ,
    "factors_ff_25_op_and_inv_5_x_5_monthly"                                                                      ,
    "25 Portfolios Formed on Operating Profitability and Investment (5 x 5)"                                      ,
    "factors_ff_25_op_and_inv_5_x_5_exdividends_monthly"                                                          ,
    "25 Portfolios Formed on Operating Profitability and Investment (5 x 5) [ex. Dividends]"                      ,
    "factors_ff_25_op_and_inv_5_x_5_daily"                                                                        ,
    "25 Portfolios Formed on Operating Profitability and Investment (5 x 5) [Daily]"                              ,
    "factors_ff_32_size_bm_and_op_2_x_4_x_4_monthly"                                                              ,
    "32 Portfolios Formed on Size, Book-to-Market, and Operating Profitability (2 x 4 x 4)"                       ,
    "factors_ff_32_size_bm_and_op_2_x_4_x_4_exdividends_monthly"                                                  ,
    "32 Portfolios Formed on Size, Book-to-Market, and Operating Profitability (2 x 4 x 4) [ex. Dividends]"       ,
    "factors_ff_32_size_bm_and_inv_2_x_4_x_4_monthly"                                                             ,
    "32 Portfolios Formed on Size, Book-to-Market, and Investment (2 x 4 x 4)"                                    ,
    "factors_ff_32_size_bm_and_inv_2_x_4_x_4_exdividends_monthly"                                                 ,
    "32 Portfolios Formed on Size, Book-to-Market, and Investment (2 x 4 x 4) [ex. Dividends]"                    ,
    "factors_ff_32_size_op_and_inv_2_x_4_x_4_monthly"                                                             ,
    "32 Portfolios Formed on Size, Operating Profitability, and Investment (2 x 4 x 4)"                           ,
    "factors_ff_32_size_op_and_inv_2_x_4_x_4_exdividends_monthly"                                                 ,
    "32 Portfolios Formed on Size, Operating Profitability, and Investment (2 x 4 x 4) [ex. Dividends]"           ,
    "factors_ff_earningsprice_monthly"                                                                            ,
    "Portfolios Formed on Earnings/Price"                                                                         ,
    "factors_ff_earningsprice_exdividends_monthly"                                                                ,
    "Portfolios Formed on Earnings/Price [ex. Dividends]"                                                         ,
    "factors_ff_cashflowprice_monthly"                                                                            ,
    "Portfolios Formed on Cashflow/Price"                                                                         ,
    "factors_ff_cashflowprice_exdividends_monthly"                                                                ,
    "Portfolios Formed on Cashflow/Price [ex. Dividends]"                                                         ,
    "factors_ff_dividend_yield_monthly"                                                                           ,
    "Portfolios Formed on Dividend Yield"                                                                         ,
    "factors_ff_dividend_yield_exdividends_monthly"                                                               ,
    "Portfolios Formed on Dividend Yield [ex. Dividends]"                                                         ,
    "factors_ff_6_size_and_earningsprice_monthly"                                                                 ,
    "6 Portfolios Formed on Size and Earnings/Price"                                                              ,
    "factors_ff_6_size_and_earningsprice_exdividends_monthly"                                                     ,
    "6 Portfolios Formed on Size and Earnings/Price [ex. Dividends]"                                              ,
    "factors_ff_6_size_and_cashflowprice_monthly"                                                                 ,
    "6 Portfolios Formed on Size and Cashflow/Price"                                                              ,
    "factors_ff_6_size_and_cashflowprice_exdividends_monthly"                                                     ,
    "6 Portfolios Formed on Size and Cashflow/Price [ex. Dividends]"                                              ,
    "factors_ff_6_size_and_dividend_yield_monthly"                                                                ,
    "6 Portfolios Formed on Size and Dividend Yield"                                                              ,
    "factors_ff_6_size_and_dividend_yield_exdividends_monthly"                                                    ,
    "6 Portfolios Formed on Size and Dividend Yield [ex. Dividends]"                                              ,
    "factors_ff_momentum_factor_monthly"                                                                          ,
    "Momentum Factor (Mom)"                                                                                       ,
    "factors_ff_momentum_factor_daily"                                                                            ,
    "Momentum Factor (Mom) [Daily]"                                                                               ,
    "factors_ff_6_size_and_momentum_2_x_3_monthly"                                                                ,
    "6 Portfolios Formed on Size and Momentum (2 x 3)"                                                            ,
    "factors_ff_6_size_and_momentum_2_x_3_daily"                                                                  ,
    "6 Portfolios Formed on Size and Momentum (2 x 3) [Daily]"                                                    ,
    "factors_ff_25_size_and_momentum_5_x_5_monthly"                                                               ,
    "25 Portfolios Formed on Size and Momentum (5 x 5)"                                                           ,
    "factors_ff_25_size_and_momentum_5_x_5_daily"                                                                 ,
    "25 Portfolios Formed on Size and Momentum (5 x 5) [Daily]"                                                   ,
    "factors_ff_10_momentum_monthly"                                                                              ,
    "10 Portfolios Formed on Momentum"                                                                            ,
    "factors_ff_10_momentum_daily"                                                                                ,
    "10 Portfolios Formed on Momentum [Daily]"                                                                    ,
    "factors_ff_shortterm_reversal_factor_st_rev_monthly"                                                         ,
    "Short-Term Reversal Factor (ST Rev)"                                                                         ,
    "factors_ff_shortterm_reversal_factor_st_rev_daily"                                                           ,
    "Short-Term Reversal Factor (ST Rev) [Daily]"                                                                 ,
    "factors_ff_6_size_and_shortterm_reversal_2_x_3_monthly"                                                      ,
    "6 Portfolios Formed on Size and Short-Term Reversal (2 x 3)"                                                 ,
    "factors_ff_6_size_and_shortterm_reversal_2_x_3_daily"                                                        ,
    "6 Portfolios Formed on Size and Short-Term Reversal (2 x 3) [Daily]"                                         ,
    "factors_ff_25_size_and_shortterm_reversal_5_x_5_monthly"                                                     ,
    "25 Portfolios Formed on Size and Short-Term Reversal (5 x 5)"                                                ,
    "factors_ff_25_size_and_shortterm_reversal_5_x_5_daily"                                                       ,
    "25 Portfolios Formed on Size and Short-Term Reversal (5 x 5) [Daily]"                                        ,
    "factors_ff_10_shortterm_reversal_monthly"                                                                    ,
    "10 Portfolios Formed on Short-Term Reversal"                                                                 ,
    "factors_ff_10_shortterm_reversal_daily"                                                                      ,
    "10 Portfolios Formed on Short-Term Reversal [Daily]"                                                         ,
    "factors_ff_longterm_reversal_factor_lt_rev_monthly"                                                          ,
    "Long-Term Reversal Factor (LT Rev)"                                                                          ,
    "factors_ff_longterm_reversal_factor_lt_rev_daily"                                                            ,
    "Long-Term Reversal Factor (LT Rev) [Daily]"                                                                  ,
    "factors_ff_6_size_and_longterm_reversal_2_x_3_monthly"                                                       ,
    "6 Portfolios Formed on Size and Long-Term Reversal (2 x 3)"                                                  ,
    "factors_ff_6_size_and_longterm_reversal_2_x_3_daily"                                                         ,
    "6 Portfolios Formed on Size and Long-Term Reversal (2 x 3) [Daily]"                                          ,
    "factors_ff_25_size_and_longterm_reversal_5_x_5_monthly"                                                      ,
    "25 Portfolios Formed on Size and Long-Term Reversal (5 x 5)"                                                 ,
    "factors_ff_25_size_and_longterm_reversal_5_x_5_daily"                                                        ,
    "25 Portfolios Formed on Size and Long-Term Reversal (5 x 5) [Daily]"                                         ,
    "factors_ff_10_longterm_reversal_monthly"                                                                     ,
    "10 Portfolios Formed on Long-Term Reversal"                                                                  ,
    "factors_ff_10_longterm_reversal_daily"                                                                       ,
    "10 Portfolios Formed on Long-Term Reversal [Daily]"                                                          ,
    "factors_ff_accruals_monthly"                                                                                 ,
    "Portfolios Formed on Accruals"                                                                               ,
    "factors_ff_25_size_and_accruals_monthly"                                                                     ,
    "25 Portfolios Formed on Size and Accruals"                                                                   ,
    "factors_ff_market_beta_monthly"                                                                              ,
    "Portfolios Formed on Market Beta"                                                                            ,
    "factors_ff_25_size_and_market_beta_monthly"                                                                  ,
    "25 Portfolios Formed on Size and Market Beta"                                                                ,
    "factors_ff_net_share_issues_monthly"                                                                         ,
    "Portfolios Formed on Net Share Issues"                                                                       ,
    "factors_ff_25_size_and_net_share_issues_monthly"                                                             ,
    "25 Portfolios Formed on Size and Net Share Issues"                                                           ,
    "factors_ff_variance_monthly"                                                                                 ,
    "Portfolios Formed on Variance"                                                                               ,
    "factors_ff_25_size_and_variance_monthly"                                                                     ,
    "25 Portfolios Formed on Size and Variance"                                                                   ,
    "factors_ff_residual_variance_monthly"                                                                        ,
    "Portfolios Formed on Residual Variance"                                                                      ,
    "factors_ff_25_size_and_residual_variance_monthly"                                                            ,
    "25 Portfolios Formed on Size and Residual Variance"                                                          ,
    "factors_ff_5_industry_portfolios_monthly"                                                                    ,
    "5 Industry Portfolios"                                                                                       ,
    "factors_ff_5_industry_portfolios_exdividends_monthly"                                                        ,
    "5 Industry Portfolios [ex. Dividends]"                                                                       ,
    "factors_ff_5_industry_portfolios_daily"                                                                      ,
    "5 Industry Portfolios [Daily]"                                                                               ,
    "factors_ff_10_industry_portfolios_monthly"                                                                   ,
    "10 Industry Portfolios"                                                                                      ,
    "factors_ff_10_industry_portfolios_exdividends_monthly"                                                       ,
    "10 Industry Portfolios [ex. Dividends]"                                                                      ,
    "factors_ff_10_industry_portfolios_daily"                                                                     ,
    "10 Industry Portfolios [Daily]"                                                                              ,
    "factors_ff_12_industry_portfolios_monthly"                                                                   ,
    "12 Industry Portfolios"                                                                                      ,
    "factors_ff_12_industry_portfolios_exdividends_monthly"                                                       ,
    "12 Industry Portfolios [ex. Dividends]"                                                                      ,
    "factors_ff_12_industry_portfolios_daily"                                                                     ,
    "12 Industry Portfolios [Daily]"                                                                              ,
    "factors_ff_17_industry_portfolios_monthly"                                                                   ,
    "17 Industry Portfolios"                                                                                      ,
    "factors_ff_17_industry_portfolios_exdividends_monthly"                                                       ,
    "17 Industry Portfolios [ex. Dividends]"                                                                      ,
    "factors_ff_17_industry_portfolios_daily"                                                                     ,
    "17 Industry Portfolios [Daily]"                                                                              ,
    "factors_ff_30_industry_portfolios_monthly"                                                                   ,
    "30 Industry Portfolios"                                                                                      ,
    "factors_ff_30_industry_portfolios_exdividends_monthly"                                                       ,
    "30 Industry Portfolios [ex. Dividends]"                                                                      ,
    "factors_ff_30_industry_portfolios_daily"                                                                     ,
    "30 Industry Portfolios [Daily]"                                                                              ,
    "factors_ff_38_industry_portfolios_monthly"                                                                   ,
    "38 Industry Portfolios"                                                                                      ,
    "factors_ff_38_industry_portfolios_exdividends_monthly"                                                       ,
    "38 Industry Portfolios [ex. Dividends]"                                                                      ,
    "factors_ff_38_industry_portfolios_daily"                                                                     ,
    "38 Industry Portfolios [Daily]"                                                                              ,
    "factors_ff_48_industry_portfolios_monthly"                                                                   ,
    "48 Industry Portfolios"                                                                                      ,
    "factors_ff_48_industry_portfolios_exdividends_monthly"                                                       ,
    "48 Industry Portfolios [ex. Dividends]"                                                                      ,
    "factors_ff_48_industry_portfolios_daily"                                                                     ,
    "48 Industry Portfolios [Daily]"                                                                              ,
    "factors_ff_49_industry_portfolios_monthly"                                                                   ,
    "49 Industry Portfolios"                                                                                      ,
    "factors_ff_49_industry_portfolios_exdividends_monthly"                                                       ,
    "49 Industry Portfolios [ex. Dividends]"                                                                      ,
    "factors_ff_49_industry_portfolios_daily"                                                                     ,
    "49 Industry Portfolios [Daily]"                                                                              ,
    "factors_ff_me_breakpoints_monthly"                                                                           ,
    "ME Breakpoints"                                                                                              ,
    "factors_ff_beme_breakpoints_monthly"                                                                         ,
    "BE/ME Breakpoints"                                                                                           ,
    "factors_ff_op_breakpoints_monthly"                                                                           ,
    "Operating Profitability Breakpoints"                                                                         ,
    "factors_ff_inv_breakpoints_monthly"                                                                          ,
    "Investment Breakpoints"                                                                                      ,
    "factors_ff_ep_breakpoints_monthly"                                                                           ,
    "E/P Breakpoints"                                                                                             ,
    "factors_ff_cfp_breakpoints_monthly"                                                                          ,
    "CF/P Breakpoints"                                                                                            ,
    "factors_ff_dp_breakpoints_monthly"                                                                           ,
    "D/P Breakpoints"                                                                                             ,
    "factors_ff_prior_212_return_breakpoints_monthly"                                                             ,
    "Prior (2-12) Return Breakpoints"                                                                             ,
    "factors_ff_developed_3_monthly"                                                                              ,
    "Fama/French Developed 3 Factors"                                                                             ,
    "factors_ff_developed_3_daily"                                                                                ,
    "Fama/French Developed 3 Factors [Daily]"                                                                     ,
    "factors_ff_developed_ex_us_3_monthly"                                                                        ,
    "Fama/French Developed ex US 3 Factors"                                                                       ,
    "factors_ff_developed_ex_us_3_daily"                                                                          ,
    "Fama/French Developed ex US 3 Factors [Daily]"                                                               ,
    "factors_ff_european_3_monthly"                                                                               ,
    "Fama/French European 3 Factors"                                                                              ,
    "factors_ff_european_3_daily"                                                                                 ,
    "Fama/French European 3 Factors [Daily]"                                                                      ,
    "factors_ff_japanese_3_monthly"                                                                               ,
    "Fama/French Japanese 3 Factors"                                                                              ,
    "factors_ff_japanese_3_daily"                                                                                 ,
    "Fama/French Japanese 3 Factors [Daily]"                                                                      ,
    "factors_ff_asia_pacific_ex_japan_3_monthly"                                                                  ,
    "Fama/French Asia Pacific ex Japan 3 Factors"                                                                 ,
    "factors_ff_asia_pacific_ex_japan_3_daily"                                                                    ,
    "Fama/French Asia Pacific ex Japan 3 Factors [Daily]"                                                         ,
    "factors_ff_north_american_3_monthly"                                                                         ,
    "Fama/French North American 3 Factors"                                                                        ,
    "factors_ff_north_american_3_daily"                                                                           ,
    "Fama/French North American 3 Factors [Daily]"                                                                ,
    "factors_ff_developed_5_monthly"                                                                              ,
    "Fama/French Developed 5 Factors"                                                                             ,
    "factors_ff_developed_5_daily"                                                                                ,
    "Fama/French Developed 5 Factors [Daily]"                                                                     ,
    "factors_ff_developed_ex_us_5_monthly"                                                                        ,
    "Fama/French Developed ex US 5 Factors"                                                                       ,
    "factors_ff_developed_ex_us_5_daily"                                                                          ,
    "Fama/French Developed ex US 5 Factors [Daily]"                                                               ,
    "factors_ff_european_5_monthly"                                                                               ,
    "Fama/French European 5 Factors"                                                                              ,
    "factors_ff_european_5_daily"                                                                                 ,
    "Fama/French European 5 Factors [Daily]"                                                                      ,
    "factors_ff_japanese_5_monthly"                                                                               ,
    "Fama/French Japanese 5 Factors"                                                                              ,
    "factors_ff_japanese_5_daily"                                                                                 ,
    "Fama/French Japanese 5 Factors [Daily]"                                                                      ,
    "factors_ff_asia_pacific_ex_japan_5_monthly"                                                                  ,
    "Fama/French Asia Pacific ex Japan 5 Factors"                                                                 ,
    "factors_ff_asia_pacific_ex_japan_5_daily"                                                                    ,
    "Fama/French Asia Pacific ex Japan 5 Factors [Daily]"                                                         ,
    "factors_ff_north_american_5_monthly"                                                                         ,
    "Fama/French North American 5 Factors"                                                                        ,
    "factors_ff_north_american_5_daily"                                                                           ,
    "Fama/French North American 5 Factors [Daily]"                                                                ,
    "factors_ff_developed_momentum_factor_monthly"                                                                ,
    "Developed Momentum Factor (Mom)"                                                                             ,
    "factors_ff_developed_momentum_factor_daily"                                                                  ,
    "Developed Momentum Factor (Mom) [Daily]"                                                                     ,
    "factors_ff_developed_ex_us_momentum_factor_monthly"                                                          ,
    "Developed ex US Momentum Factor (Mom)"                                                                       ,
    "factors_ff_developed_ex_us_momentum_factor_daily"                                                            ,
    "Developed ex US Momentum Factor (Mom) [Daily]"                                                               ,
    "factors_ff_european_momentum_factor_monthly"                                                                 ,
    "European Momentum Factor (Mom)"                                                                              ,
    "factors_ff_european_momentum_factor_daily"                                                                   ,
    "European Momentum Factor (Mom) [Daily]"                                                                      ,
    "factors_ff_japanese_momentum_factor_monthly"                                                                 ,
    "Japanese Momentum Factor (Mom)"                                                                              ,
    "factors_ff_japanese_momentum_factor_daily"                                                                   ,
    "Japanese Momentum Factor (Mom) [Daily]"                                                                      ,
    "factors_ff_asia_pacific_ex_japan_momentum_factor_monthly"                                                    ,
    "Asia Pacific ex Japan Momentum Factor (Mom)"                                                                 ,
    "factors_ff_asia_pacific_ex_japan_momentum_factor_daily"                                                      ,
    "Asia Pacific ex Japan Momentum Factor (Mom) [Daily]"                                                         ,
    "factors_ff_north_american_momentum_factor_monthly"                                                           ,
    "North American Momentum Factor (Mom)"                                                                        ,
    "factors_ff_north_american_momentum_factor_daily"                                                             ,
    "North American Momentum Factor (Mom) [Daily]"                                                                ,
    "factors_ff_6_developed_size_and_bm_2_x_3_monthly"                                                            ,
    "6 Developed Portfolios Formed on Size and Book-to-Market (2 x 3)"                                            ,
    "factors_ff_6_developed_size_and_bm_2_x_3_daily"                                                              ,
    "6 Developed Portfolios Formed on Size and Book-to-Market (2 x 3) [Daily]"                                    ,
    "factors_ff_6_developed_ex_us_size_and_bm_2_x_3_monthly"                                                      ,
    "6 Developed ex US Portfolios Formed on Size and Book-to-Market (2 x 3)"                                      ,
    "factors_ff_6_developed_ex_us_size_and_bm_2_x_3_daily"                                                        ,
    "6 Developed ex US Portfolios Formed on Size and Book-to-Market (2 x 3) [Daily]"                              ,
    "factors_ff_6_european_size_and_bm_2_x_3_monthly"                                                             ,
    "6 European Portfolios Formed on Size and Book-to-Market (2 x 3)"                                             ,
    "factors_ff_6_european_size_and_bm_2_x_3_daily"                                                               ,
    "6 European Portfolios Formed on Size and Book-to-Market (2 x 3) [Daily]"                                     ,
    "factors_ff_6_japanese_size_and_bm_2_x_3_monthly"                                                             ,
    "6 Japanese Portfolios Formed on Size and Book-to-Market (2 x 3)"                                             ,
    "factors_ff_6_japanese_size_and_bm_2_x_3_daily"                                                               ,
    "6 Japanese Portfolios Formed on Size and Book-to-Market (2 x 3) [Daily]"                                     ,
    "factors_ff_6_asia_pacific_ex_japan_size_and_bm_2_x_3_monthly"                                                ,
    "6 Asia Pacific ex Japan Portfolios Formed on Size and Book-to-Market (2 x 3)"                                ,
    "factors_ff_6_asia_pacific_ex_japan_size_and_bm_2_x_3_daily"                                                  ,
    "6 Asia Pacific ex Japan Portfolios Formed on Size and Book-to-Market (2 x 3) [Daily]"                        ,
    "factors_ff_6_north_american_size_and_bm_2_x_3_monthly"                                                       ,
    "6 North American Portfolios Formed on Size and Book-to-Market (2 x 3)"                                       ,
    "factors_ff_6_north_american_size_and_bm_2_x_3_daily"                                                         ,
    "6 North American Portfolios Formed on Size and Book-to-Market (2 x 3) [Daily]"                               ,
    "factors_ff_25_developed_size_and_bm_5_x_5_monthly"                                                           ,
    "25 Developed Portfolios Formed on Size and Book-to-Market (5 x 5)"                                           ,
    "factors_ff_25_developed_size_and_bm_5_x_5_daily"                                                             ,
    "25 Developed Portfolios Formed on Size and Book-to-Market (5 x 5) [Daily]"                                   ,
    "factors_ff_25_developed_ex_us_size_and_bm_5_x_5_monthly"                                                     ,
    "25 Developed ex US Portfolios Formed on Size and Book-to-Market (5 x 5)"                                     ,
    "factors_ff_25_developed_ex_us_size_and_bm_5_x_5_daily"                                                       ,
    "25 Developed ex US Portfolios Formed on Size and Book-to-Market (5 x 5) [Daily]"                             ,
    "factors_ff_25_european_size_and_bm_5_x_5_monthly"                                                            ,
    "25 European Portfolios Formed on Size and Book-to-Market (5 x 5)"                                            ,
    "factors_ff_25_european_size_and_bm_5_x_5_daily"                                                              ,
    "25 European Portfolios Formed on Size and Book-to-Market (5 x 5) [Daily]"                                    ,
    "factors_ff_25_japanese_size_and_bm_5_x_5_monthly"                                                            ,
    "25 Japanese Portfolios Formed on Size and Book-to-Market (5 x 5)"                                            ,
    "factors_ff_25_japanese_size_and_bm_5_x_5_daily"                                                              ,
    "25 Japanese Portfolios Formed on Size and Book-to-Market (5 x 5) [Daily]"                                    ,
    "factors_ff_25_asia_pacific_ex_japan_size_and_bm_5_x_5_monthly"                                               ,
    "25 Asia Pacific ex Japan Portfolios Formed on Size and Book-to-Market (5 x 5)"                               ,
    "factors_ff_25_asia_pacific_ex_japan_size_and_bm_5_x_5_daily"                                                 ,
    "25 Asia Pacific ex Japan Portfolios Formed on Size and Book-to-Market (5 x 5) [Daily]"                       ,
    "factors_ff_25_north_american_size_and_bm_5_x_5_monthly"                                                      ,
    "25 North American Portfolios Formed on Size and Book-to-Market (5 x 5)"                                      ,
    "factors_ff_25_north_american_size_and_bm_5_x_5_daily"                                                        ,
    "25 North American Portfolios Formed on Size and Book-to-Market (5 x 5) [Daily]"                              ,
    "factors_ff_6_developed_size_and_op_2_x_3_monthly"                                                            ,
    "6 Developed Portfolios Formed on Size and Operating Profitability (2 x 3)"                                   ,
    "factors_ff_6_developed_size_and_op_2_x_3_daily"                                                              ,
    "6 Developed Portfolios Formed on Size and Operating Profitability (2 x 3) [Daily]"                           ,
    "factors_ff_6_developed_ex_us_size_and_op_2_x_3_monthly"                                                      ,
    "6 Developed ex US Portfolios Formed on Size and Operating Profitability (2 x 3)"                             ,
    "factors_ff_6_developed_ex_us_size_and_op_2_x_3_daily"                                                        ,
    "6 Developed ex US Portfolios Formed on Size and Operating Profitability (2 x 3) [Daily]"                     ,
    "factors_ff_6_european_size_and_op_2_x_3_monthly"                                                             ,
    "6 European Portfolios Formed on Size and Operating Profitability (2 x 3)"                                    ,
    "factors_ff_6_european_size_and_op_2_x_3_daily"                                                               ,
    "6 European Portfolios Formed on Size and Operating Profitability (2 x 3) [Daily]"                            ,
    "factors_ff_6_japanese_size_and_op_2_x_3_monthly"                                                             ,
    "6 Japanese Portfolios Formed on Size and Operating Profitability (2 x 3)"                                    ,
    "factors_ff_6_japanese_size_and_op_2_x_3_daily"                                                               ,
    "6 Japanese Portfolios Formed on Size and Operating Profitability (2 x 3) [Daily]"                            ,
    "factors_ff_6_asia_pacific_ex_japan_size_and_op_2_x_3_monthly"                                                ,
    "6 Asia Pacific ex Japan Portfolios Formed on Size and Operating Profitability (2 x 3)"                       ,
    "factors_ff_6_asia_pacific_ex_japan_size_and_op_2_x_3_daily"                                                  ,
    "6 Asia Pacific ex Japan Portfolios Formed on Size and Operating Profitability (2 x 3) [Daily]"               ,
    "factors_ff_6_north_american_size_and_op_2_x_3_monthly"                                                       ,
    "6 North American Portfolios Formed on Size and Operating Profitability (2 x 3)"                              ,
    "factors_ff_6_north_american_size_and_op_2_x_3_daily"                                                         ,
    "6 North American Portfolios Formed on Size and Operating Profitability (2 x 3) [Daily]"                      ,
    "factors_ff_25_developed_size_and_op_5_x_5_monthly"                                                           ,
    "25 Developed Portfolios Formed on Size and Operating Profitability (5 x 5)"                                  ,
    "factors_ff_25_developed_size_and_op_5_x_5_daily"                                                             ,
    "25 Developed Portfolios Formed on Size and Operating Profitability (5 x 5) [Daily]"                          ,
    "factors_ff_25_developed_ex_us_size_and_op_5_x_5_monthly"                                                     ,
    "25 Developed ex US Portfolios Formed on Size and Operating Profitability (5 x 5)"                            ,
    "factors_ff_25_developed_ex_us_size_and_op_5_x_5_daily"                                                       ,
    "25 Developed ex US Portfolios Formed on Size and Operating Profitability (5 x 5) [Daily]"                    ,
    "factors_ff_25_european_size_and_op_5_x_5_monthly"                                                            ,
    "25 European Portfolios Formed on Size and Operating Profitability (5 x 5)"                                   ,
    "factors_ff_25_european_size_and_op_5_x_5_daily"                                                              ,
    "25 European Portfolios Formed on Size and Operating Profitability (5 x 5) [Daily]"                           ,
    "factors_ff_25_japanese_size_and_op_5_x_5_monthly"                                                            ,
    "25 Japanese Portfolios Formed on Size and Operating Profitability (5 x 5)"                                   ,
    "factors_ff_25_japanese_size_and_op_5_x_5_daily"                                                              ,
    "25 Japanese Portfolios Formed on Size and Operating Profitability (5 x 5) [Daily]"                           ,
    "factors_ff_25_asia_pacific_ex_japan_size_and_op_5_x_5_monthly"                                               ,
    "25 Asia Pacific ex Japan Portfolios Formed on Size and Operating Profitability (5 x 5)"                      ,
    "factors_ff_25_asia_pacific_ex_japan_size_and_op_5_x_5_daily"                                                 ,
    "25 Asia Pacific ex Japan Portfolios Formed on Size and Operating Profitability (5 x 5) [Daily]"              ,
    "factors_ff_25_north_american_size_and_op_5_x_5_monthly"                                                      ,
    "25 North American Portfolios Formed on Size and Operating Profitability (5 x 5)"                             ,
    "factors_ff_25_north_american_size_and_op_5_x_5_daily"                                                        ,
    "25 North American Portfolios Formed on Size and Operating Profitability (5 x 5) [Daily]"                     ,
    "factors_ff_6_developed_size_and_inv_2_x_3_monthly"                                                           ,
    "6 Developed Portfolios Formed on Size and Investment (2 x 3)"                                                ,
    "factors_ff_6_developed_size_and_inv_2_x_3_daily"                                                             ,
    "6 Developed Portfolios Formed on Size and Investment (2 x 3) [Daily]"                                        ,
    "factors_ff_6_developed_ex_us_size_and_inv_2_x_3_monthly"                                                     ,
    "6 Developed ex US Portfolios Formed on Size and Investment (2 x 3)"                                          ,
    "factors_ff_6_developed_ex_us_size_and_inv_2_x_3_daily"                                                       ,
    "6 Developed ex US Portfolios Formed on Size and Investment (2 x 3) [Daily]"                                  ,
    "factors_ff_6_european_size_and_inv_2_x_3_monthly"                                                            ,
    "6 European Portfolios Formed on Size and Investment (2 x 3)"                                                 ,
    "factors_ff_6_european_size_and_inv_2_x_3_daily"                                                              ,
    "6 European Portfolios Formed on Size and Investment (2 x 3) [Daily]"                                         ,
    "factors_ff_6_japanese_size_and_inv_2_x_3_monthly"                                                            ,
    "6 Japanese Portfolios Formed on Size and Investment (2 x 3)"                                                 ,
    "factors_ff_6_japanese_size_and_inv_2_x_3_daily"                                                              ,
    "6 Japanese Portfolios Formed on Size and Investment (2 x 3) [Daily]"                                         ,
    "factors_ff_6_asia_pacific_ex_japan_size_and_inv_2_x_3_monthly"                                               ,
    "6 Asia Pacific ex Japan Portfolios Formed on Size and Investment (2 x 3)"                                    ,
    "factors_ff_6_asia_pacific_ex_japan_size_and_inv_2_x_3_daily"                                                 ,
    "6 Asia Pacific ex Japan Portfolios Formed on Size and Investment (2 x 3) [Daily]"                            ,
    "factors_ff_6_north_american_size_and_inv_2_x_3_monthly"                                                      ,
    "6 North American Portfolios Formed on Size and Investment (2 x 3)"                                           ,
    "factors_ff_6_north_american_size_and_inv_2_x_3_daily"                                                        ,
    "6 North American Portfolios Formed on Size and Investment (2 x 3) [Daily]"                                   ,
    "factors_ff_25_developed_size_and_inv_5_x_5_monthly"                                                          ,
    "25 Developed Portfolios Formed on Size and Investment (5 x 5)"                                               ,
    "factors_ff_25_developed_size_and_inv_5_x_5_daily"                                                            ,
    "25 Developed Portfolios Formed on Size and Investment (5 x 5) [Daily]"                                       ,
    "factors_ff_25_developed_ex_us_size_and_inv_5_x_5_monthly"                                                    ,
    "25 Developed ex US Portfolios Formed on Size and Investment (5 x 5)"                                         ,
    "factors_ff_25_developed_ex_us_size_and_inv_5_x_5_daily"                                                      ,
    "25 Developed ex US Portfolios Formed on Size and Investment (5 x 5) [Daily]"                                 ,
    "factors_ff_25_european_size_and_inv_5_x_5_monthly"                                                           ,
    "25 European Portfolios Formed on Size and Investment (5 x 5)"                                                ,
    "factors_ff_25_european_size_and_inv_5_x_5_daily"                                                             ,
    "25 European Portfolios Formed on Size and Investment (5 x 5) [Daily]"                                        ,
    "factors_ff_25_japanese_size_and_inv_5_x_5_monthly"                                                           ,
    "25 Japanese Portfolios Formed on Size and Investment (5 x 5)"                                                ,
    "factors_ff_25_japanese_size_and_inv_5_x_5_daily"                                                             ,
    "25 Japanese Portfolios Formed on Size and Investment (5 x 5) [Daily]"                                        ,
    "factors_ff_25_asia_pacific_ex_japan_size_and_inv_5_x_5_monthly"                                              ,
    "25 Asia Pacific ex Japan Portfolios Formed on Size and Investment (5 x 5)"                                   ,
    "factors_ff_25_asia_pacific_ex_japan_size_and_inv_5_x_5_daily"                                                ,
    "25 Asia Pacific ex Japan Portfolios Formed on Size and Investment (5 x 5) [Daily]"                           ,
    "factors_ff_25_north_american_size_and_inv_5_x_5_monthly"                                                     ,
    "25 North American Portfolios Formed on Size and Investment (5 x 5)"                                          ,
    "factors_ff_25_north_american_size_and_inv_5_x_5_daily"                                                       ,
    "25 North American Portfolios Formed on Size and Investment (5 x 5) [Daily]"                                  ,
    "factors_ff_6_developed_size_and_momentum_2_x_3_monthly"                                                      ,
    "6 Developed Portfolios Formed on Size and Momentum (2 x 3)"                                                  ,
    "factors_ff_6_developed_size_and_momentum_2_x_3_daily"                                                        ,
    "6 Developed Portfolios Formed on Size and Momentum (2 x 3) [Daily]"                                          ,
    "factors_ff_6_developed_ex_us_size_and_momentum_2_x_3_monthly"                                                ,
    "6 Developed ex US Portfolios Formed on Size and Momentum (2 x 3)"                                            ,
    "factors_ff_6_developed_ex_us_size_and_momentum_2_x_3_daily"                                                  ,
    "6 Developed ex US Portfolios Formed on Size and Momentum (2 x 3) [Daily]"                                    ,
    "factors_ff_6_european_size_and_momentum_2_x_3_monthly"                                                       ,
    "6 European Portfolios Formed on Size and Momentum (2 x 3)"                                                   ,
    "factors_ff_6_european_size_and_momentum_2_x_3_daily"                                                         ,
    "6 European Portfolios Formed on Size and Momentum (2 x 3) [Daily]"                                           ,
    "factors_ff_6_japanese_size_and_momentum_2_x_3_monthly"                                                       ,
    "6 Japanese Portfolios Formed on Size and Momentum (2 x 3)"                                                   ,
    "factors_ff_6_japanese_size_and_momentum_2_x_3_daily"                                                         ,
    "6 Japanese Portfolios Formed on Size and Momentum (2 x 3) [Daily]"                                           ,
    "factors_ff_6_asia_pacific_ex_japan_size_and_momentum_2_x_3_monthly"                                          ,
    "6 Asia Pacific ex Japan Portfolios Formed on Size and Momentum (2 x 3)"                                      ,
    "factors_ff_6_asia_pacific_ex_japan_size_and_momentum_2_x_3_daily"                                            ,
    "6 Asia Pacific ex Japan Portfolios Formed on Size and Momentum (2 x 3) [Daily]"                              ,
    "factors_ff_6_north_american_size_and_momentum_2_x_3_monthly"                                                 ,
    "6 North American Portfolios Formed on Size and Momentum (2 x 3)"                                             ,
    "factors_ff_6_north_american_size_and_momentum_2_x_3_daily"                                                   ,
    "6 North American Portfolios Formed on Size and Momentum (2 x 3) [Daily]"                                     ,
    "factors_ff_25_developed_size_and_momentum_5_x_5_monthly"                                                     ,
    "25 Developed Portfolios Formed on Size and Momentum (5 x 5)"                                                 ,
    "factors_ff_25_developed_size_and_momentum_5_x_5_daily"                                                       ,
    "25 Developed Portfolios Formed on Size and Momentum (5 x 5) [Daily]"                                         ,
    "factors_ff_25_developed_ex_us_size_and_momentum_5_x_5_monthly"                                               ,
    "25 Developed ex US Portfolios Formed on Size and Momentum (5 x 5)"                                           ,
    "factors_ff_25_developed_ex_us_size_and_momentum_5_x_5_daily"                                                 ,
    "25 Developed ex US Portfolios Formed on Size and Momentum (5 x 5) [Daily]"                                   ,
    "factors_ff_25_european_size_and_momentum_5_x_5_monthly"                                                      ,
    "25 European Portfolios Formed on Size and Momentum (5 x 5)"                                                  ,
    "factors_ff_25_european_size_and_momentum_5_x_5_daily"                                                        ,
    "25 European Portfolios Formed on Size and Momentum (5 x 5) [Daily]"                                          ,
    "factors_ff_25_japanese_size_and_momentum_5_x_5_monthly"                                                      ,
    "25 Japanese Portfolios Formed on Size and Momentum (5 x 5)"                                                  ,
    "factors_ff_25_japanese_size_and_momentum_5_x_5_daily"                                                        ,
    "25 Japanese Portfolios Formed on Size and Momentum (5 x 5) [Daily]"                                          ,
    "factors_ff_25_asia_pacific_ex_japan_size_and_momentum_5_x_5_monthly"                                         ,
    "25 Asia Pacific ex Japan Portfolios Formed on Size and Momentum (5 x 5)"                                     ,
    "factors_ff_25_asia_pacific_ex_japan_size_and_momentum_5_x_5_daily"                                           ,
    "25 Asia Pacific ex Japan Portfolios Formed on Size and Momentum (5 x 5) [Daily]"                             ,
    "factors_ff_25_north_american_size_and_momentum_5_x_5_monthly"                                                ,
    "25 North American Portfolios Formed on Size and Momentum (5 x 5)"                                            ,
    "factors_ff_25_north_american_size_and_momentum_5_x_5_daily"                                                  ,
    "25 North American Portfolios Formed on Size and Momentum (5 x 5) [Daily]"                                    ,
    "factors_ff_32_developed_size_bm_and_op_2_x_4_x_4_monthly"                                                    ,
    "32 Developed Portfolios Formed on Size, Book-to-Market, and Operating Profitability (2 x 4 x 4)"             ,
    "factors_ff_32_developed_ex_us_size_bm_and_op_2_x_4_x_4_monthly"                                              ,
    "32 Developed ex US Portfolios Formed on Size, Book-to-Market, and Operating Profitability (2 x 4 x 4)"       ,
    "factors_ff_32_european_size_bm_and_op_2_x_4_x_4_monthly"                                                     ,
    "32 European Portfolios Formed on Size, Book-to-Market, and Operating Profitability (2 x 4 x 4)"              ,
    "factors_ff_32_japanese_size_bm_and_op_2_x_4_x_4_monthly"                                                     ,
    "32 Japanese Portfolios Formed on Size, Book-to-Market, and Operating Profitability (2 x 4 x 4)"              ,
    "factors_ff_32_asia_pacific_ex_japan_size_bm_and_op_2_x_4_x_4_monthly"                                        ,
    "32 Asia Pacific ex Japan Portfolios Formed on Size, Book-to-Market, and Operating Profitability (2 x 4 x 4)" ,
    "factors_ff_32_north_american_size_bm_and_op_2_x_4_x_4_monthly"                                               ,
    "32 North American Portfolios Formed on Size, Book-to-Market, and Operating Profitability (2 x 4 x 4)"        ,
    "factors_ff_32_developed_size_bm_and_inv_2_x_4_x_4_monthly"                                                   ,
    "32 Developed Portfolios Formed on Size, Book-to-Market, and Investment (2 x 4 x 4)"                          ,
    "factors_ff_32_developed_ex_us_size_bm_and_inv_2_x_4_x_4_monthly"                                             ,
    "32 Developed ex US Portfolios Formed on Size, Book-to-Market, and Investment (2 x 4 x 4)"                    ,
    "factors_ff_32_european_size_bm_and_inv_2_x_4_x_4_monthly"                                                    ,
    "32 European Portfolios Formed on Size, Book-to-Market, and Investment (2 x 4 x 4)"                           ,
    "factors_ff_32_japanese_size_bm_and_inv_2_x_4_x_4_monthly"                                                    ,
    "32 Japanese Portfolios Formed on Size, Book-to-Market, and Investment (2 x 4 x 4)"                           ,
    "factors_ff_32_asia_pacific_ex_japan_size_bm_and_inv_2_x_4_x_4_monthly"                                       ,
    "32 Asia Pacific ex Japan Portfolios Formed on Size, Book-to-Market, and Investment (2 x 4 x 4)"              ,
    "factors_ff_32_north_american_size_bm_and_inv_2_x_4_x_4_monthly"                                              ,
    "32 North American Portfolios Formed on Size, Book-to-Market, and Investment (2 x 4 x 4)"                     ,
    "factors_ff_32_developed_size_op_and_inv_2_x_4_x_4_monthly"                                                   ,
    "32 Developed Portfolios Formed on Size, Operating Profitability, and Investment (2 x 4 x 4)"                 ,
    "factors_ff_32_developed_ex_us_size_op_and_inv_2_x_4_x_4_monthly"                                             ,
    "32 Developed ex US Portfolios Formed on Size, Operating Profitability, and Investment (2 x 4 x 4)"           ,
    "factors_ff_32_european_size_op_and_inv_2_x_4_x_4_monthly"                                                    ,
    "32 European Portfolios Formed on Size, Operating Profitability, and Investment (2 x 4 x 4)"                  ,
    "factors_ff_32_japanese_size_op_and_inv_2_x_4_x_4_monthly"                                                    ,
    "32 Japanese Portfolios Formed on Size, Operating Profitability, and Investment (2 x 4 x 4)"                  ,
    "factors_ff_32_asia_pacific_ex_japan_size_op_and_inv_2_x_4_x_4_monthly"                                       ,
    "32 Asia Pacific ex Japan Portfolios Formed on Size, Operating Profitability, and Investment (2 x 4 x 4)"     ,
    "factors_ff_32_north_american_size_op_and_inv_2_x_4_x_4_monthly"                                              ,
    "32 North American Portfolios Formed on Size, Operating Profitability, and Investment (2 x 4 x 4)"            ,
    "factors_ff_emerging_5_monthly"                                                                               ,
    "Fama/French Emerging 5 Factors"                                                                              ,
    "factors_ff_emerging_momentum_factor_monthly"                                                                 ,
    "Emerging Momentum Factor (Mom)"                                                                              ,
    "factors_ff_6_emerging_market_size_and_bm_2_x_3_monthly"                                                      ,
    "6 Emerging Market Portfolios Formed on Size and Book-to-Market (2 x 3)"                                      ,
    "factors_ff_6_emerging_market_size_and_op_2_x_3_monthly"                                                      ,
    "6 Emerging Market Portfolios Formed on Size and Operating Profitability (2 x 3)"                             ,
    "factors_ff_6_emerging_market_size_and_inv_2_x_3_monthly"                                                     ,
    "6 Emerging Market Portfolios Formed on Size and Investment (2 x 3)"                                          ,
    "factors_ff_6_emerging_market_size_and_momentum_2_x_3_monthly"                                                ,
    "6 Emerging Market Portfolios Formed on Size and Momentum (2 x 3)"                                            ,
    "factors_ff_4_emerging_market_bm__and_op_2_x_2___monthly"                                                     ,
    "4 Emerging Market Portfolios Formed on Book-to-Market  and Operating Profitability (2 x 2)  "                ,
    "factors_ff_4_emerging_market_op_and_inv_2_x_2_monthly"                                                       ,
    "4 Emerging Market Portfolios Formed on Operating Profitability and Investment (2 x 2)"                       ,
    "factors_ff_4_emerging_market_bm_and_inv_2_x_2_monthly"                                                       ,
    "4 Emerging Market Portfolios Formed on Book-to-Market and Investment (2 x 2)"
  ) |>
    dplyr::mutate(domain = "Fama-French")
  # nolint end
}

#' List Supported Legacy Fama-French Dataset Types
#'
#' Returns a tibble with the legacy names of initially supported
#' Fama-French dataset types, including their names and frequencies (daily,
#' weekly, monthly). Each dataset type is associated with a specific Fama-French
#' model (e.g., 3 factors, 5 factors). Additionally, it annotates each dataset
#' with the domain "Fama-French". Not included in the exported
#' `list_supported_types()` function.
#'
#' @returns A tibble with columns: `type` (the type of dataset), `dataset_name`
#'   (a descriptive name of the dataset), and `domain` (the domain to which the
#'   dataset belongs, always "Fama-French").
#'
#' @family utility functions
list_supported_types_ff_legacy <- function() {
  # nolint start
  tribble(
    ~type                                 ,
    ~dataset_name                         ,
    "factors_ff3_daily"                   ,
    "Fama/French 3 Factors [Daily]"       ,
    "factors_ff3_weekly"                  ,
    "Fama/French 3 Factors [Weekly]"      ,
    "factors_ff3_monthly"                 ,
    "Fama/French 3 Factors"               ,
    "factors_ff5_daily"                   ,
    "Fama/French 5 Factors (2x3) [Daily]" ,
    "factors_ff5_monthly"                 ,
    "Fama/French 5 Factors (2x3)"         ,
    "factors_ff_industry_5_monthly"       ,
    "5 Industry Portfolios"               ,
    "factors_ff_industry_5_daily"         ,
    "5 Industry Portfolios [Daily]"       ,
    "factors_ff_industry_10_monthly"      ,
    "10 Industry Portfolios"              ,
    "factors_ff_industry_10_daily"        ,
    "10 Industry Portfolios [Daily]"      ,
    "factors_ff_industry_30_monthly"      ,
    "30 Industry Portfolios"              ,
    "factors_ff_industry_30_daily"        ,
    "30 Industry Portfolios [Daily]"      ,
    "factors_ff_industry_38_monthly"      ,
    "38 Industry Portfolios"              ,
    "factors_ff_industry_38_daily"        ,
    "38 Industry Portfolios [Daily]"      ,
    "factors_ff_industry_48_monthly"      ,
    "48 Industry Portfolios"              ,
    "factors_ff_industry_48_daily"        ,
    "48 Industry Portfolios [Daily]"      ,
    "factors_ff_industry_49_monthly"      ,
    "49 Industry Portfolios"              ,
    "factors_ff_industry_49_daily"        ,
    "49 Industry Portfolios [Daily]"
  ) |>
    dplyr::mutate(domain = "Fama-French")
  # nolint end
}

#' List Supported Global Q Dataset Types
#'
#' Returns a tibble with the supported Global Q dataset types, including
#' their names and frequencies (daily, weekly, weekly week-to-week, monthly,
#' quarterly, annual). Each dataset type is associated with the Global Q
#' model, specifically the q5 factors model for the year 2023. Additionally,
#' it annotates each dataset with the domain "Global Q".
#'
#' @returns A tibble with columns: `type` (the type of dataset),
#'   `dataset_name` (the file name of the dataset), and `domain` (the
#'   domain to which the dataset belongs, always "Global Q").
#'
list_supported_types_q <- function() {
  # nolint start
  tribble(
    ~type                        ,
    ~dataset_name                ,
    "factors_q5_daily"           ,
    "q5_factors_daily_2024"      ,
    "factors_q5_weekly"          ,
    "q5_factors_weekly_2024"     ,
    "factors_q5_weekly_w2w"      ,
    "q5_factors_weekly_w2w_2024" ,
    "factors_q5_monthly"         ,
    "q5_factors_monthly_2024"    ,
    "factors_q5_quarterly"       ,
    "q5_factors_quarterly_2024"  ,
    "factors_q5_annual"          ,
    "q5_factors_annual_2024"
  ) |>
    dplyr::mutate(domain = "Global Q")
  # nolint end
}

#' List Supported Macro Predictor Dataset Types
#'
#' Returns a tibble with the supported macro predictor dataset types
#' provided by Goyal-Welch, including their frequencies (monthly, quarterly,
#' annual). All dataset types reference the same source file,
#' "PredictorData2022.xlsx" for the year 2022. Additionally, it annotates
#' each dataset with the domain "Goyal-Welch".
#'
#' @returns A tibble with columns: `type` (the type of dataset),
#'   `dataset_name` (the file name of the dataset, which is the same for
#'   all types), and `domain` (the domain to which the dataset belongs,
#'   always "Goyal-Welch").
#'
#' @family utility functions
list_supported_types_macro_predictors <- function() {
  # nolint start
  tribble(
    ~type                        ,
    ~dataset_name                ,
    "macro_predictors_monthly"   ,
    "PredictorData2022.xlsx"     ,
    "macro_predictors_quarterly" ,
    "PredictorData2022.xlsx"     ,
    "macro_predictors_annual"    ,
    "PredictorData2022.xlsx"     ,
  ) |>
    dplyr::mutate(domain = "Goyal-Welch")
  # nolint end
}

#' List Supported WRDS Dataset Types
#'
#' Returns a tibble with the supported dataset types provided via WRDS.
#' Additionally, it annotates each dataset with the domain "WRDS".
#'
#' @returns A tibble with columns: `type` (the type of dataset),
#'   `dataset_name` (the file name of the dataset), and `domain` (the
#'   domain to which the dataset belongs, always "WRDS").
#'
#' @family utility functions
list_supported_types_wrds <- function() {
  # nolint start
  tribble(
    ~type                                           ,
    ~dataset_name                                   ,
    "wrds_crsp_monthly"                             ,
    "crsp.msf, crsp.msenames, crsp.msedelist"       ,
    "wrds_crsp_daily"                               ,
    "crsp.dsf, crsp.msenames, crsp.msedelist"       ,
    "wrds_compustat_annual"                         ,
    "comp.funda"                                    ,
    "wrds_compustat_quarterly"                      ,
    "comp.fundq"                                    ,
    "wrds_ccm_links"                                ,
    "crsp.ccmxpf_linktable"                         ,
    "wrds_fisd"                                     ,
    "fisd.fisd_mergedissue, fisd.fisd_mergedissuer" ,
    "wrds_trace_enhanced"                           ,
    "trace.trace_enhanced"
  ) |>
    dplyr::mutate(domain = "WRDS")
  # nolint end
}

#' List Supported Other Data Types
#'
#' Returns a tibble listing the supported other data types and their
#' corresponding dataset names.
#'
#' @returns A tibble with columns: `type` (the type of dataset), `dataset_name`
#'   (the name of the data source), and `domain` (the domain to which the
#'   dataset belongs).
#'
#' @family utility functions
list_supported_types_other <- function() {
  tibble(
    "type" = c(
      "stock_prices",
      "constituents",
      "fred",
      "osap",
      "risk_free",
      "high_frequency_sp500",
      "factor_library"
    ),
    "dataset_name" = c(
      "YahooFinance",
      "various",
      "various",
      "Open Source Asset Pricing",
      "Risk-Free Rate",
      "High Frequency S&P 500",
      "Factor Library"
    ),
    "domain" = c(
      "Stock Prices",
      "Index Constituents",
      "FRED",
      "Open Source Asset Pricing",
      "tidyfinance",
      "tidyfinance",
      "tidyfinance"
    )
  )
}

#' List All Supported Dataset Types
#'
#' Aggregates and returns a comprehensive tibble of all supported dataset
#' types from different domains. It includes various datasets across
#' different frequencies (daily, weekly, monthly, quarterly, annual) and
#' models (e.g., q5 factors, Fama-French 3 and 5 factors, macro predictors).
#'
#' @param domain A character vector to filter for domain-specific types
#'   (e.g., c("WRDS", "Fama-French"))
#' @param as_vector Logical indicating whether types should be returned as
#'   a character vector instead of a data frame.
#'
#' @returns A tibble aggregating all supported dataset types with columns:
#'   `type` (the type of dataset), `dataset_name` (a descriptive name or
#'   file name of the dataset), and `domain` (the domain to which the
#'   dataset belongs, e.g., "Global Q", "Fama-French", "Goyal-Welch").
#'
#' @family utility functions
#' @export
#'
#' @examples
#' # List all supported types as a data frame
#' list_supported_types()
#'
#' # Filter by domain
#' list_supported_types(domain = "WRDS")
#'
#' # List supported types as a vector
#' list_supported_types(as_vector = TRUE)
list_supported_types <- function(domain = NULL, as_vector = FALSE) {
  supported_types <- dplyr::bind_rows(
    list_supported_types_q(),
    list_supported_types_ff(),
    list_supported_types_macro_predictors(),
    list_supported_types_wrds(),
    list_supported_types_other()
  )
  if (!is.null(domain)) {
    filter_domains <- domain
    supported_types <- supported_types |>
      dplyr::filter(.data$domain %in% filter_domains)
  }
  if (as_vector) {
    supported_types$type
  } else {
    supported_types
  }
}

test_that("download_data_risk_free returns monthly data with correct columns", {
  skip_if_offline()
  skip_on_cran()
  data <- download_data_risk_free("2020-01-01", "2020-12-31")
  expect_s3_class(data, "tbl_df")
  expect_true(all(c("date", "risk_free") %in% colnames(data)))
  expect_equal(ncol(data), 2)
  expect_true(all(!is.na(data$risk_free)))
})

test_that("download_data_risk_free returns daily data with correct columns", {
  skip_if_offline()
  skip_on_cran()
  data <- download_data_risk_free(
    "2020-01-01",
    "2020-12-31",
    frequency = "daily"
  )
  expect_s3_class(data, "tbl_df")
  expect_true(all(c("date", "risk_free") %in% colnames(data)))
  expect_equal(ncol(data), 2)
  expect_true(all(!is.na(data$risk_free)))
})

test_that("download_data_risk_free monthly returns one row per month", {
  skip_if_offline()
  skip_on_cran()
  data <- download_data_risk_free("2010-01-01", "2010-12-31")
  expect_equal(nrow(data), 12)
  expect_s3_class(data$date, "Date")
})

test_that("download_data_risk_free uses correct splice date boundary", {
  skip_if_offline()
  skip_on_cran()
  data <- download_data_risk_free("2000-11-01", "2001-02-28")
  expect_true(nrow(data) >= 4)
  expect_true(all(!is.na(data$risk_free)))
})

test_that("download_data_risk_free errors on invalid frequency", {
  expect_error(
    download_data_risk_free("2020-01-01", "2020-12-31", frequency = "weekly"),
    regexp = "frequency"
  )
})

test_that("download_data_risk_free errors when start_date after end_date", {
  skip_if_offline()
  skip_on_cran()
  expect_error(
    download_data_risk_free("2021-12-31", "2020-01-01"),
    regexp = "`start_date` cannot be after `end_date`"
  )
})

test_that("download_data_risk_free returns filtered rows for date range", {
  skip_if_offline()
  skip_on_cran()
  result <- download_data_risk_free("2020-03-01", "2020-06-30")
  expect_equal(nrow(result), 4)
  expect_true(all(result$date >= as.Date("2020-03-01")))
  expect_true(all(result$date <= as.Date("2020-06-30")))
})

test_that("download_data_risk_free aborts with clear message on failure", {
  with_mocked_bindings(
    read_parquet = function(file, ...) stop("connection refused"),
    .package = "arrow",
    {
      expect_error(
        download_data_risk_free("2020-01-01", "2020-12-31"),
        regexp = "Failed to download risk-free rate data from HuggingFace"
      )
    }
  )
})

test_that("download_data via tidyfinance risk_free routes correctly", {
  skip_if_offline()
  skip_on_cran()
  data <- download_data(
    "tidyfinance",
    "risk_free",
    "2020-01-01",
    "2020-12-31"
  )
  expect_s3_class(data, "tbl_df")
  expect_true(all(c("date", "risk_free") %in% colnames(data)))
})

test_that("risk_free appears in list_supported_types for tidyfinance", {
  types <- list_supported_types()
  tf_types <- types[types$domain == "tidyfinance", ]
  expect_true("risk_free" %in% tf_types$type)
})

#' List Supported Indexes
#'
#' Returns a tibble containing information about supported financial indexes.
#' Each index is associated with a URL that points to a CSV file containing the
#' holdings of the index. Additionally, each index has a corresponding `skip`
#' value, which indicates the number of lines to skip when reading the CSV file.
#'
#' @returns A tibble with three columns:
#' \describe{
#'   \item{index}{The name of the financial index (e.g., "DAX", "S&P 500").}
#'   \item{url}{The URL to the CSV file containing the holdings data for the
#'   index.}
#'   \item{skip}{The number of lines to skip when reading the CSV file.}
#' }
#'
#' @family utility functions
#' @export
#'
#' @examples
#' supported_indexes <- list_supported_indexes()
#' print(supported_indexes)
#'
list_supported_indexes <- function() {
  # nolint start
  tribble(
    ~index                                                                                                                                                                                    ,
    ~url                                                                                                                                                                                      ,
    ~skip                                                                                                                                                                                     ,
    "DAX"                                                                                                                                                                                     ,
    "https://www.ishares.com/de/privatanleger/de/produkte/251464/ishares-dax-ucits-etf-de-fund/1478358465952.ajax?fileType=csv&fileName=DAXEX_holdings&dataType=fund"                         ,
                                                                                                                                                                                            2 ,
    "EURO STOXX 50"                                                                                                                                                                           ,
    "https://www.ishares.com/de/privatanleger/de/produkte/251783/ishares-euro-stoxx-50-ucits-etf-de-fund/1478358465952.ajax?fileType=csv&fileName=EXW1_holdings&dataType=fund"                ,
                                                                                                                                                                                            2 ,
    "Dow Jones Industrial Average"                                                                                                                                                            ,
    "https://www.ishares.com/de/privatanleger/de/produkte/251770/ishares-dow-jones-industrial-average-ucits-etf-de-fund/1478358465952.ajax?fileType=csv&fileName=EXI3_holdings&dataType=fund" ,
                                                                                                                                                                                            2 ,
    "Russell 1000"                                                                                                                                                                            ,
    "https://www.ishares.com/ch/professionelle-anleger/de/produkte/239707/ishares-russell-1000-etf/1495092304805.ajax?fileType=csv&fileName=IWB_holdings&dataType=fund"                       ,
                                                                                                                                                                                            9 ,
    "Russell 2000"                                                                                                                                                                            ,
    "https://www.ishares.com/ch/professionelle-anleger/de/produkte/239710/ishares-russell-2000-etf/1495092304805.ajax?fileType=csv&fileName=IWM_holdings&dataType=fund"                       ,
                                                                                                                                                                                            9 ,
    "Russell 3000"                                                                                                                                                                            ,
    "https://www.ishares.com/ch/professionelle-anleger/de/produkte/239714/ishares-russell-3000-etf/1495092304805.ajax?fileType=csv&fileName=IWV_holdings&dataType=fund"                       ,
                                                                                                                                                                                            9 ,
    "S&P 100"                                                                                                                                                                                 ,
    "https://www.ishares.com/ch/professionelle-anleger/de/produkte/239723/ishares-sp-100-etf/1495092304805.ajax?fileType=csv&fileName=OEF_holdings&dataType=fund"                             ,
                                                                                                                                                                                            9 ,
    "S&P 500"                                                                                                                                                                                 ,
    "https://www.ishares.com/de/privatanleger/de/produkte/253743/ishares-sp-500-b-ucits-etf-acc-fund/1478358465952.ajax?fileType=csv&fileName=SXR8_holdings&dataType=fund"                    ,
                                                                                                                                                                                            2 ,
    "Nasdaq 100"                                                                                                                                                                              ,
    "https://www.ishares.com/de/privatanleger/de/produkte/251896/ishares-nasdaq100-ucits-etf-de-fund/1478358465952.ajax?fileType=csv&fileName=EXXT_holdings&dataType=fund"                    ,
                                                                                                                                                                                            2 ,
    "FTSE 100"                                                                                                                                                                                ,
    "https://www.ishares.com/de/privatanleger/de/produkte/251795/ishares-ftse-100-ucits-etf-inc-fund/1478358465952.ajax?fileType=csv&fileName=IUSZ_holdings&dataType=fund"                    ,
                                                                                                                                                                                            2 ,
    "MSCI World"                                                                                                                                                                              ,
    "https://www.ishares.com/de/privatanleger/de/produkte/251882/ishares-msci-world-ucits-etf-acc-fund/1478358465952.ajax?fileType=csv&fileName=EUNL_holdings&dataType=fund"                  ,
                                                                                                                                                                                            2 ,
    "Nikkei 225"                                                                                                                                                                              ,
    "https://www.ishares.com/ch/professionelle-anleger/de/produkte/253742/ishares-nikkei-225-ucits-etf/1495092304805.ajax?fileType=csv&fileName=CSNKY_holdings&dataType=fund"                 ,
                                                                                                                                                                                            2 ,
    "TOPIX"                                                                                                                                                                                   ,
    "https://www.blackrock.com/jp/individual-en/en/products/279438/fund/1480664184455.ajax?fileType=csv&fileName=1475_holdings&dataType=fund"                                                 ,
                                                                                                                                                                                            2 ,
    "STOXX Europe 600"                                                                                                                                                                        ,
    "https://www.ishares.com/de/privatanleger/de/produkte/251931/ishares-stoxx-europe-600-ucits-etf-de-fund/1478358465952.ajax?fileType=csv&fileName=EXSA_holdings&dataType=fund"             ,
                                                                                                                                                                                            2 ,
    "MDAX"                                                                                                                                                                                    ,
    "https://www.ishares.com/de/privatanleger/de/produkte/251845/ishares-mdax-ucits-etf-de-fund/1478358465952.ajax?fileType=csv&fileName=EXS3_holdings&dataType=fund"                         ,
                                                                                                                                                                                            2 ,
    "TecDAX"                                                                                                                                                                                  ,
    "https://www.ishares.com/de/privatanleger/de/produkte/251948/ishares-tecdax-ucits-etf-de-fund/1478358465952.ajax?fileType=csv&fileName=EXS2_holdings&dataType=fund"                       ,
                                                                                                                                                                                            2 ,
    "MSCI Emerging Markets"                                                                                                                                                                   ,
    "https://www.ishares.com/de/privatanleger/de/produkte/251857/ishares-msci-emerging-markets-ucits-etf-inc-fund/1478358465952.ajax?fileType=csv&fileName=IQQE_holdings&dataType=fund"       ,
                                                                                                                                                                                            2 ,
    "MSCI Europe"                                                                                                                                                                             ,
    "https://www.ishares.com/de/privatanleger/de/produkte/251860/ishares-msci-europe-ucits-etf-inc-fund/1478358465952.ajax?fileType=csv&fileName=IQQY_holdings&dataType=fund"                 ,
                                                                                                                                                                                            2 ,
    "MSCI ACWI"                                                                                                                                                                               ,
    "https://www.ishares.com/ch/professionelle-anleger/de/produkte/239600/ishares-msci-acwi-etf/1495092304805.ajax?fileType=csv&fileName=ACWI_holdings&dataType=fund"                         ,
                                                                                                                                                                                            9 ,
    "S&P SmallCap 600"                                                                                                                                                                        ,
    "https://www.ishares.com/ch/professionelle-anleger/de/produkte/239774/ishares-sp-smallcap-600-etf/1495092304805.ajax?fileType=csv&fileName=IJR_holdings&dataType=fund"                    ,
                                                                                                                                                                                            9
  )
  # nolint end
}

#' Download CCM Links from WRDS
#'
#' Downloads data from the WRDS CRSP/Compustat Merged (CCM) links
#' database. It allows users to specify the type of links (`linktype`) and the
#' primacy of the link (`linkprim`).
#'
#' @param linktype A character vector indicating the type of link to download.
#'   The default is `c("LU", "LC")`, where "LU" stands for "Link Up" and "LC"
#'   for "Link CRSP".
#' @param linkprim A character vector indicating the primacy of the link.
#'   Default is `c("P", "C")`, where "P" indicates primary and "C" indicates
#'   conditional links.
#'
#' @returns A data frame with the columns `permno`, `gvkey`, `linkdt`, and
#'   `linkenddt`, where `linkenddt` is the end date of the link, and missing end
#'   dates are replaced with today's date.
#'
#' @family WRDS functions
#' @export
#'
#' @examples
#' \dontrun{
#'   ccm_links <- download_data_wrds_ccm_links(linktype = "LU", linkprim = "P")
#' }
#'
download_data_wrds_ccm_links <- function(
  linktype = c("LU", "LC"),
  linkprim = c("P", "C")
) {
  con <- get_wrds_connection()

  ccm_linking_table_db <- tbl(con, I("crsp.ccmxpf_lnkhist"))

  par <- list(linktype = linktype, linkprim = linkprim)

  ccm_links <- ccm_linking_table_db |>
    filter(
      linktype %in% par$linktype & linkprim %in% par$linkprim
    ) |>
    select(permno = lpermno, gvkey, linkdt, linkenddt) |>
    collect() |>
    mutate(linkenddt = tidyr::replace_na(linkenddt, today()))

  disconnect_connection(con)

  ccm_links
}

#' Download and Process Data Based on Domain and Dataset
#'
#' Downloads and processes data based on the specified domain (e.g.,
#' Fama-French factors, Global Q factors, or macro predictors), dataset,
#' and date range. This function checks if the specified domain is supported
#' and then delegates to the appropriate function for downloading and
#' processing the data.
#'
#' @param domain The domain of the dataset to download (e.g.,
#'   "famafrench", "globalq", "macro_predictors", "wrds",
#'   "constituents", "fred", "stock_prices", "osap", "tidyfinance").
#' @param dataset Optional. The specific dataset to download within the
#'   domain.
#' @param start_date Optional. A character string or Date object in
#'   "YYYY-MM-DD" format specifying the start date for the data. If not
#'   provided, the full dataset or a subset is returned, depending on the
#'   dataset type.
#' @param end_date Optional. A character string or Date object in
#'   "YYYY-MM-DD" format specifying the end date for the data. If not
#'   provided, the full dataset or a subset is returned, depending on the
#'   dataset type.
#' @param type `r lifecycle::badge("deprecated")` Use `domain` and
#'   `dataset` instead.
#' @param ... Additional arguments passed to specific download functions
#'   depending on the `domain`. For instance, if `domain` is
#'   `"constituents"`, arguments are passed to
#'   `download_data_constituents()`. If `domain` is `"tidyfinance"` and
#'   `dataset` is `"factor_library"`, arguments are used to filter the
#'   portfolio grid (e.g., `sorting_variable`, `rebalancing`, `fill_all`);
#'   see `download_data_huggingface()` for details.
#'
#' @returns A tibble with processed data, including dates and the relevant
#'   financial metrics, filtered by the specified date range.
#'
#' @family download functions
#' @export
#'
#' @examples
#' \donttest{
#' download_data(
#'   "famafrench",
#'   "Fama/French 5 Factors (2x3) [Daily]",
#'   "2000-01-01",
#'   "2020-12-31"
#' )
#' download_data("macro_predictors", "monthly", "2000-01-01", "2020-12-31")
#' download_data("constituents", index = "DAX")
#' download_data("fred", series = c("GDP", "CPIAUCNS"))
#' download_data("stock_prices", symbols = c("AAPL", "MSFT"))
#' download_data(
#'   "tidyfinance",
#'   "risk_free",
#'   "2020-01-01",
#'   "2020-12-31"
#' )
#' download_data(
#'   "tidyfinance",
#'   "high_frequency_sp500",
#'   "2007-07-26",
#'   "2007-07-27"
#' )
#' download_data(
#'   "tidyfinance",
#'   "factor_library",
#'   sorting_variable = "52w",
#'   rebalancing = "annual"
#' )
#' }
download_data <- function(
  domain = NULL,
  dataset = NULL,
  start_date = NULL,
  end_date = NULL,
  type = deprecated(),
  ...
) {
  # Handle explicit type argument
  if (lifecycle::is_present(type)) {
    lifecycle::deprecate_warn(
      when = "0.5.0",
      what = "download_data(type)",
      details = "Use the `domain` and `dataset` arguments instead."
    )
    parsed <- parse_type_to_domain_dataset(type)
    domain <- parsed$domain
    dataset <- parsed$dataset
  }

  # Handle legacy type passed as domain argument
  if (!is.null(domain) && is_legacy_type(domain)) {
    lifecycle::deprecate_warn(
      when = "0.5.0",
      what = "download_data(type)",
      details = paste0(
        "Column type should be replaced with domain and dataset. ",
        "Use `list_supported_types()` to see the mapping."
      )
    )
    parsed <- parse_type_to_domain_dataset(domain)
    domain <- parsed$domain
    dataset <- parsed$dataset
  }

  if (is.null(domain)) {
    cli::cli_abort("Argument {.arg domain} is required.")
  }

  check_supported_domain(domain)

  if (domain %in% c("famafrench", "factors_ff")) {
    processed_data <- download_data_factors_ff(
      dataset = dataset,
      start_date = start_date,
      end_date = end_date,
      ...
    )
  } else if (domain %in% c("globalq", "factors_q")) {
    processed_data <- download_data_factors_q(
      dataset = dataset,
      start_date = start_date,
      end_date = end_date,
      ...
    )
  } else if (domain == "macro_predictors") {
    processed_data <- download_data_macro_predictors(
      dataset = dataset,
      start_date = start_date,
      end_date = end_date,
      ...
    )
  } else if (domain == "wrds") {
    processed_data <- download_data_wrds(
      dataset = dataset,
      start_date = start_date,
      end_date = end_date,
      ...
    )
  } else if (domain == "constituents") {
    processed_data <- download_data_constituents(...)
  } else if (domain == "fred") {
    processed_data <- download_data_fred(
      start_date = start_date,
      end_date = end_date,
      ...
    )
  } else if (domain == "stock_prices") {
    processed_data <- download_data_stock_prices(
      start_date = start_date,
      end_date = end_date,
      ...
    )
  } else if (domain == "osap") {
    processed_data <- download_data_osap(
      start_date = start_date,
      end_date = end_date,
      ...
    )
  } else if (domain == "tidyfinance") {
    if (!is.null(dataset) && dataset == "risk_free") {
      processed_data <- download_data_risk_free(
        start_date = start_date,
        end_date = end_date,
        ...
      )
    } else {
      processed_data <- download_data_huggingface(
        dataset = dataset,
        start_date = start_date,
        end_date = end_date,
        ...
      )
    }
  } else {
    cli::cli_abort("Unsupported domain: {.val {domain}}")
  }

  processed_data
}
#' Check if a string is a legacy type
#' @noRd
is_legacy_type <- function(x) {
  # These strings are valid domain names, not legacy types
  valid_domains <- c("constituents", "fred", "stock_prices", "osap")
  if (x %in% valid_domains) {
    return(FALSE)
  }

  # Check all known legacy type patterns
  ff_types <- dplyr::bind_rows(
    list_supported_types_ff(),
    list_supported_types_ff_legacy()
  )
  q_types <- list_supported_types_q()
  macro_types <- list_supported_types_macro_predictors()
  wrds_types <- list_supported_types_wrds()
  other_types <- list_supported_types_other() |>
    dplyr::filter(.data$domain != "tidyfinance", .data$type != "osap")

  all_types <- dplyr::bind_rows(
    ff_types,
    q_types,
    macro_types,
    wrds_types,
    other_types
  )

  x %in% all_types$type
}

#' Parse legacy type parameter to domain and dataset
#' @noRd
parse_type_to_domain_dataset <- function(type) {
  # Check Fama-French types (both current and legacy)
  ff_types <- dplyr::bind_rows(
    list_supported_types_ff(),
    list_supported_types_ff_legacy()
  )

  if (type %in% ff_types$type) {
    dataset_name <- ff_types$dataset_name[ff_types$type == type]
    return(list(domain = "factors_ff", dataset = dataset_name))
  }

  # Global Q factors: "factors_q5_*" -> domain = "factors_q"
  q_types <- list_supported_types_q()

  if (type %in% q_types$type) {
    dataset_name <- q_types$dataset_name[q_types$type == type]
    return(list(
      domain = "factors_q",
      dataset = gsub("\\.csv$", "", dataset_name)
    ))
  }

  # Macro predictors: "macro_predictors_*" -> domain = "macro_predictors"
  macro_types <- list_supported_types_macro_predictors()

  if (type %in% macro_types$type) {
    dataset <- sub("^macro_predictors_", "", type)
    return(list(domain = "macro_predictors", dataset = dataset))
  }

  wrds_types <- list_supported_types_wrds()

  if (type %in% wrds_types$type) {
    dataset <- sub("^wrds_", "", type)
    return(list(domain = "wrds", dataset = dataset))
  }

  # High frequency: "hf_*" -> domain = "tidyfinance"
  if (grepl("^hf_", type)) {
    dataset <- sub("^hf_", "", type)
    return(list(domain = "tidyfinance", dataset = dataset))
  }

  # Simple domain-only types (no dataset component)
  simple_domains <- c("constituents", "fred", "stock_prices", "osap")
  if (type %in% simple_domains) {
    return(list(domain = type, dataset = NULL))
  }

  cli::cli_abort(c(
    "Cannot parse legacy type: {.val {type}}",
    "i" = "Use {.fn list_supported_types} to see available types."
  ))
}

#' Check if domain is supported
#' @noRd
check_supported_domain <- function(domain) {
  supported_domains <- c(
    "famafrench",
    "factors_ff",
    "globalq",
    "factors_q",
    "macro_predictors",
    "wrds",
    "constituents",
    "fred",
    "stock_prices",
    "osap",
    "tidyfinance"
  )

  if (!domain %in% supported_domains) {
    cli::cli_abort(c(
      "Unsupported domain: {.val {domain}}",
      "i" = "Supported domains: {.val {supported_domains}}"
    ))
  }
}

#' Download Data from WRDS
#'
#' Acts as a wrapper to download data from various WRDS datasets including
#' CRSP, Compustat, and CCM links based on the specified dataset. It is
#' designed to handle different datasets by redirecting to the appropriate
#' specific data download function.
#'
#' @param dataset A string specifying the dataset to download. Supported
#'   values: "crsp_monthly", "crsp_daily" for CRSP data,
#'   "compustat_annual", "compustat_quarterly" for Compustat data,
#'   "ccm_links" for CCM links data, "fisd" for FISD data, or
#'   "trace_enhanced" for TRACE data.
#' @param start_date Optional. A character string or Date object in
#'   "YYYY-MM-DD" format specifying the start date for the data. If not
#'   provided, a subset of the dataset is returned.
#' @param end_date Optional. A character string or Date object in
#'   "YYYY-MM-DD" format specifying the end date for the data. If not
#'   provided, a subset of the dataset is returned.
#' @param type `r lifecycle::badge("deprecated")` Use `dataset` instead.
#' @param ... Additional arguments passed to specific download functions
#'   depending on the `dataset`.
#'
#' @returns A data frame containing the requested data, with the structure
#'   and contents depending on the specified `dataset`.
#'
#' @family WRDS functions
#' @export
#'
#' @examples
#' \dontrun{
#'   crsp_monthly <- download_data_wrds(
#'     "crsp_monthly", "2020-01-01", "2020-12-31"
#'   )
#'   compustat_annual <- download_data_wrds(
#'     "compustat_annual", "2020-01-01", "2020-12-31"
#'   )
#'   ccm_links <- download_data_wrds("ccm_links")
#'   fisd <- download_data_wrds("fisd")
#'   trace_enhanced <- download_data_wrds(
#'     "trace_enhanced", cusips = "00101JAH9"
#'   )
#' }
download_data_wrds <- function(
  dataset = NULL,
  start_date = NULL,
  end_date = NULL,
  type = deprecated(),
  ...
) {
  # Handle explicit type argument
  if (lifecycle::is_present(type)) {
    lifecycle::deprecate_warn(
      when = "0.5.0",
      what = "download_data_wrds(type)",
      details = "Use the `dataset` argument instead."
    )
    dataset <- sub("^wrds_", "", type)
  }

  # Handle legacy type passed as dataset argument
  if (!is.null(dataset) && is_legacy_type_wrds(dataset)) {
    lifecycle::deprecate_warn(
      when = "0.5.0",
      what = "download_data_wrds(type)",
      details = paste0(
        "The `type` argument is deprecated. ",
        "Use `dataset` instead",
        "(e.g., 'crsp_monthly' instead of 'wrds_crsp_monthly')."
      )
    )
    dataset <- sub("^wrds_", "", dataset)
  }

  if (is.null(dataset)) {
    cli::cli_abort("Argument {.arg dataset} is required.")
  }

  check_supported_dataset_wrds(dataset)

  if (grepl("^crsp", dataset)) {
    processed_data <- download_data_wrds_crsp(
      dataset = dataset,
      start_date = start_date,
      end_date = end_date,
      ...
    )
  } else if (grepl("^compustat", dataset)) {
    processed_data <- download_data_wrds_compustat(
      dataset = dataset,
      start_date = start_date,
      end_date = end_date,
      ...
    )
  } else if (dataset == "ccm_links") {
    processed_data <- download_data_wrds_ccm_links(...)
  } else if (dataset == "fisd") {
    processed_data <- download_data_wrds_fisd(...)
  } else if (dataset == "trace_enhanced") {
    processed_data <- download_data_wrds_trace_enhanced(
      start_date = start_date,
      end_date = end_date,
      ...
    )
  } else {
    cli::cli_abort("Unsupported WRDS dataset: {.val {dataset}}")
  }

  processed_data
}

#' Check if a string is a legacy WRDS type
#' @noRd
is_legacy_type_wrds <- function(x) {
  grepl("^wrds_", x)
}

#' Check if WRDS dataset is supported
#' @noRd
check_supported_dataset_wrds <- function(dataset) {
  supported_datasets <- c(
    "crsp_monthly",
    "crsp_daily",
    "compustat_annual",
    "compustat_quarterly",
    "ccm_links",
    "fisd",
    "trace_enhanced"
  )

  if (!dataset %in% supported_datasets) {
    cli::cli_abort(c(
      "Unsupported WRDS dataset: {.val {dataset}}",
      "i" = "Supported datasets: {.val {supported_datasets}}"
    ))
  }
}

#' Create Portfolio Sort Options
#'
#' Creates a list of options of class `tidyfinance_portfolio_sort_options` that
#' bundles sample construction filters and breakpoint specifications for use
#' with [implement_portfolio_sort()].
#'
#' @param filter_options A list of class `tidyfinance_filter_options` created
#'   by [filter_options()], or `NULL` (the default, which applies no filters).
#'   The arguments accepted by [filter_options()] include
#'   \itemize{
#'     \item `exclude_financials` A logical indicating whether to exclude
#'       financial firms (SIC codes 6000–6799). Defaults to `FALSE`.
#'     \item `exclude_utilities` A logical indicating whether to exclude
#'       utility firms (SIC codes 4900–4999). Defaults to `FALSE`.
#'     \item `min_stock_price` A single positive numeric specifying the
#'       minimum stock price required to include an observation. `NULL` (the
#'       default) applies no price filter.
#'     \item `min_size_quantile` A single numeric strictly between 0 and 1
#'       specifying the minimum cross-sectional size quantile (based on lagged
#'       market cap) required to include an observation. `NULL` (the default)
#'       applies no size quantile filter.
#'     \item `min_listing_age` A single non-negative integer or numeric
#'       specifying the minimum number of months a stock must have been listed
#'       in CRSP. `NULL` (the default) applies no listing age filter.
#'     \item `exclude_negative_book_equity` A logical indicating whether to
#'       exclude observations with non-positive book equity. Defaults to
#'       `FALSE`.
#'     \item `exclude_negative_earnings` A logical indicating whether to
#'       exclude observations with non-positive earnings. Defaults to `FALSE`.
#'   }
#' @param breakpoint_options_main A list of class
#'   `tidyfinance_breakpoint_options` created by [breakpoint_options()],
#'   specifying breakpoints for the primary sorting variable. The arguments
#'   accepted by [breakpoint_options()] include
#'   \itemize{
#'     \item `n_portfolios` An optional integer specifying the number of
#'       equally sized portfolios to create. This parameter is mutually
#'       exclusive with `percentiles`.
#'     \item `percentiles` An optional numeric vector specifying the
#'       percentiles for determining the breakpoints of the portfolios.
#'       This parameter is mutually exclusive with `n_portfolios`.
#'     \item `breakpoints_exchanges` An optional character vector specifying
#'       exchange names to filter the data before computing breakpoints.
#'       Exchanges must be stored in a column given by `data_options` (defaults
#'       to `exchange`). If `NULL`, no filtering is applied.
#'     \item `smooth_bunching` An optional logical parameter specifying if
#'       to attempt smoothing non-extreme portfolios if the sorting variable
#'       bunches on the extremes (`TRUE`), or not (`FALSE`, the default).
#'     \item `breakpoints_min_size_threshold` An optional numeric value between
#'       0 and 1 (exclusive). When set, stocks with market capitalization below
#'       this quantile are excluded from breakpoint computation.
#'   }
#' @param breakpoint_options_secondary A list of class
#'   `tidyfinance_breakpoint_options` created by [breakpoint_options()],
#'   specifying breakpoints for the secondary sorting variable, or `NULL`
#'   (the default) for univariate sorts. The arguments accepted by
#'   [breakpoint_options()] are the same as for `breakpoint_options_main`.
#' @param ... Additional arguments to be included in the options list.
#'
#' @returns A list of class `tidyfinance_portfolio_sort_options` containing
#'   the specified options.
#'
#' @family portfolio functions
#' @export
#'
#' @examples
#' portfolio_sort_options(
#'   filter_options = filter_options(exclude_financials = TRUE),
#'   breakpoint_options_main = breakpoint_options(n_portfolios = 10)
#' )
#'
portfolio_sort_options <- function(
  filter_options = NULL,
  breakpoint_options_main,
  breakpoint_options_secondary = NULL,
  ...
) {
  # Error handling for filter_options
  if (
    !is.null(filter_options) &&
      !inherits(filter_options, "tidyfinance_filter_options")
  ) {
    cli::cli_abort(
      paste0(
        "{.arg filter_options} must be {.val NULL} or of class ",
        "{.cls tidyfinance_filter_options}."
      )
    )
  }

  # Error handling for breakpoint_options_main
  if (missing(breakpoint_options_main)) {
    cli::cli_abort("{.arg breakpoint_options_main} must be provided.")
  }

  if (!inherits(breakpoint_options_main, "tidyfinance_breakpoint_options")) {
    cli::cli_abort(
      paste0(
        "{.arg breakpoint_options_main} must be of class ",
        "{.cls tidyfinance_breakpoint_options}."
      )
    )
  }

  # Error handling for breakpoint_options_secondary
  if (
    !is.null(breakpoint_options_secondary) &&
      !inherits(breakpoint_options_secondary, "tidyfinance_breakpoint_options")
  ) {
    cli::cli_abort(
      paste0(
        "{.arg breakpoint_options_secondary} must be {.val NULL} or of class ",
        "{.cls tidyfinance_breakpoint_options}."
      )
    )
  }

  # Create the list structure with class attribute
  structure(
    list(
      "filter_options" = filter_options,
      "breakpoint_options_main" = breakpoint_options_main,
      "breakpoint_options_secondary" = breakpoint_options_secondary,
      ...
    ),
    class = "tidyfinance_portfolio_sort_options"
  )
}

test_that(
  "portfolio_sort_options returns tidyfinance_portfolio_sort_options", {
  opts <- portfolio_sort_options(
    breakpoint_options_main = breakpoint_options(n_portfolios = 5)
  )
  expect_s3_class(opts, "tidyfinance_portfolio_sort_options")
})

test_that("portfolio_sort_options returns a list", {
  opts <- portfolio_sort_options(
    breakpoint_options_main = breakpoint_options(n_portfolios = 5)
  )
  expect_type(opts, "list")
})

test_that("portfolio_sort_options stores all provided components", {
  fo <- filter_options(exclude_financials = TRUE)
  bpm <- breakpoint_options(n_portfolios = 10)
  bps <- breakpoint_options(percentiles = c(0.3, 0.7))

  opts <- portfolio_sort_options(
    filter_options = fo,
    breakpoint_options_main = bpm,
    breakpoint_options_secondary = bps
  )

  expect_identical(opts$filter_options, fo)
  expect_identical(opts$breakpoint_options_main, bpm)
  expect_identical(opts$breakpoint_options_secondary, bps)
})

test_that("portfolio_sort_options accepts NULL filter_options", {
  expect_no_error(
    portfolio_sort_options(
      filter_options = NULL,
      breakpoint_options_main = breakpoint_options(n_portfolios = 5)
    )
  )
})

test_that("portfolio_sort_options accepts NULL breakpoint_options_secondary", {
  opts <- portfolio_sort_options(
    breakpoint_options_main = breakpoint_options(n_portfolios = 5),
    breakpoint_options_secondary = NULL
  )
  expect_null(opts$breakpoint_options_secondary)
})

test_that("portfolio_sort_options stores extra arguments via ...", {
  opts <- portfolio_sort_options(
    breakpoint_options_main = breakpoint_options(n_portfolios = 5),
    custom_option = "value"
  )
  expect_equal(opts$custom_option, "value")
})

test_that("portfolio_sort_options errors when filter_options has wrong class", {
  expect_error(
    portfolio_sort_options(
      filter_options = list(exclude_financials = TRUE),
      breakpoint_options_main = breakpoint_options(n_portfolios = 5)
    ),
    "tidyfinance_filter_options"
  )
})

test_that(
  "portfolio_sort_options errors when breakpoint_options_main wrong class", {
  expect_error(
    portfolio_sort_options(
      breakpoint_options_main = list(n_portfolios = 5)
    ),
    "tidyfinance_breakpoint_options"
  )
})

test_that(
  "portfolio_sort_options errors when breakpoint_options_main is missing", {
  expect_error(
    portfolio_sort_options(),
    "breakpoint_options_main"
  )
})

test_that(
  "portfolio_sort_options errors: breakpoint_options_secondary wrong class", {
  expect_error(
    portfolio_sort_options(
      breakpoint_options_main = breakpoint_options(n_portfolios = 5),
      breakpoint_options_secondary = list(n_portfolios = 3)
    ),
    "tidyfinance_breakpoint_options"
  )
})

#' Estimate Fama-MacBeth Regressions
#'
#' Estimates Fama-MacBeth regressions (Fama and MacBeth, 1973) by first running
#' cross-sectional regressions for each time period and then aggregating the
#' results over time to obtain average risk premia and corresponding
#' t-statistics.
#'
#' @param data A data frame containing the data for the regression. It must
#'   include a column representing the time periods (defaults to `date`) and
#'   the variables specified in the `model`.
#' @param model A character string describing the model to be estimated in
#'   each cross-section (e.g., `"ret_excess ~ beta + bm + log_mktcap"`).
#' @param vcov A character string indicating the type of standard errors to
#'   compute. Options are `"iid"` for independent and identically distributed
#'   errors or `"newey-west"` for Newey-West standard errors. Default is
#'   `"newey-west"`.
#' @param vcov_options A list of additional arguments to be passed to the
#'   `NeweyWest()` function when `vcov = "newey-west"`. These can include
#'   options such as `lag`, which specifies the number of lags to use in the
#'   Newey-West covariance matrix estimation, and `prewhite`, which indicates
#'   whether to apply a prewhitening transformation. Default is an empty list.
#' @param data_options A list of class `tidyfinance_data_options` (created via
#'   [data_options()]) specifying column name mappings. The `date` element is
#'   used to specify the date column. Uses [data_options()] default if `NULL`:
#'   `"date" = "date"`.
#' @param detail A logical value indicating whether to return additional
#'   summary statistics. If `FALSE` (default), the function returns only the
#'   coefficient estimates. If `TRUE`, it returns a list with two elements:
#'   `coefficients` (the usual estimates table) and `summary_statistics` (a
#'   one-row tibble with the average cross-sectional R-squared and the average
#'   number of observations per cross-section).
#'
#' @returns If `detail = FALSE` (default), a tibble with columns
#'   `factor`, `risk_premium`, `n` (number of time periods),
#'   `standard_error`, and `t_statistic`.
#'
#'   If `detail = TRUE`, a named list with two elements:
#'   \describe{
#'     \item{coefficients}{The same tibble described above.}
#'     \item{summary_statistics}{A one-row tibble with `r_squared` (mean
#'       cross-sectional R-squared) and `n_obs` (mean cross-sectional
#'       observation count).}
#'   }
#'
#' @references
#'   Fama, E. F., & MacBeth, J. D. (1973). Risk, return, and equilibrium:
#'   Empirical tests. *Journal of Political Economy*, 81(3), 607-636.
#'   \doi{10.1086/260061}
#'
#'   Newey, W. K., & West, K. D. (1987). A simple, positive semi-definite,
#'   heteroskedasticity and autocorrelation consistent covariance matrix.
#'   *Econometrica*, 55(3), 703-708.
#'   \doi{10.2307/1913610}
#'
#' @family estimation functions
#' @export
#'
#' @examples
#' set.seed(1234)
#'
#' data <- tibble::tibble(
#'   date = rep(seq.Date(from = as.Date("2020-01-01"),
#'                       to = as.Date("2020-12-01"), by = "month"), each = 50),
#'   permno = rep(1:50, times = 12),
#'   ret_excess = rnorm(600, 0, 0.1),
#'   beta = rnorm(600, 1, 0.2),
#'   bm = rnorm(600, 0.5, 0.1),
#'   log_mktcap = rnorm(600, 10, 1)
#' )
#'
#' estimate_fama_macbeth(data, "ret_excess ~ beta + bm + log_mktcap")
#' estimate_fama_macbeth(
#'   data,
#'   "ret_excess ~ beta + bm + log_mktcap",
#'   vcov = "iid"
#' )
#' estimate_fama_macbeth(
#'   data,
#'   "ret_excess ~ beta + bm + log_mktcap",
#'   vcov = "newey-west",
#'   vcov_options = list(lag = 6, prewhite = FALSE)
#' )
#'
#' # Return detailed output including R-squared and observation counts
#' estimate_fama_macbeth(
#'   data,
#'   "ret_excess ~ beta + bm + log_mktcap",
#'   detail = TRUE
#' )
#'
#' # Use different column name for date
#' data |>
#'   dplyr::rename(month = date) |>
#'   estimate_fama_macbeth(
#'     "ret_excess ~ beta + bm + log_mktcap",
#'     data_options = data_options(date = "month")
#'  )
#'
estimate_fama_macbeth <- function(
  data,
  model,
  vcov = "newey-west",
  vcov_options = NULL,
  data_options = NULL,
  detail = FALSE
) {
  if (is.null(data_options)) {
    data_options <- data_options()
  }

  if (!vcov %in% c("iid", "newey-west")) {
    cli::cli_abort("{.arg vcov} must be either 'iid' or 'newey-west'.")
  }

  if (!data_options$date %in% colnames(data)) {
    cli::cli_abort("The data must contain a {data_options$date} column.")
  }

  cross_sections <- data |>
    tidyr::nest(data = -all_of(data_options$date)) |>
    mutate(
      row_check = purrr::map_lgl(
        data,
        ~ nrow(.) > length(all.vars(as.formula(model)))
      )
    )

  if (any(!cross_sections$row_check)) {
    cli::cli_abort(
      paste(
        "Each date grouping must have more rows than the number of predictors",
        "in the model to estimate coefficients. Please check your data."
      )
    )
  }

  cross_sections <- cross_sections |>
    select(-row_check) |>
    mutate(
      cross_fit = purrr::map(data, ~ lm(as.formula(model), data = .)),
      estimates = purrr::map(
        .data$cross_fit,
        ~ {
          coefs <- stats::coef(.)
          if ("(Intercept)" %in% names(coefs)) {
            names(coefs)[names(coefs) == "(Intercept)"] <- "intercept"
          }
          tibble::as_tibble(t(coefs))
        }
      ),
      r_squared = purrr::map_dbl(.data$cross_fit, ~ summary(.)$r.squared),
      adj_r_squared = purrr::map_dbl(
        .data$cross_fit,
        ~ summary(.)$adj.r.squared
      ),
      n_obs = purrr::map_dbl(.data$cross_fit, ~ nrow(.$model))
    ) |>
    select(-"cross_fit")

  cross_section_stats <- cross_sections |>
    select(all_of(data_options$date), "r_squared", "adj_r_squared", "n_obs")

  cross_sections <- cross_sections |>
    select(-"r_squared", -"adj_r_squared", -"n_obs") |>
    tidyr::unnest(estimates) |>
    select(-data) |>
    tidyr::pivot_longer(-all_of(data_options$date))

  compute_standard_error <- function(model, vcov, vcov_options = NULL) {
    if (vcov == "iid") {
      sqrt(stats::vcov(model)[1, 1])
    } else if (vcov == "newey-west") {
      sqrt(do.call(sandwich::NeweyWest, c(list(model), vcov_options)))
    }
  }

  aggregations <- cross_sections |>
    tidyr::nest(data = c(all_of(data_options$date), value)) |>
    mutate(
      model = purrr::map(data, ~ lm("value ~ 1", data = .)),
      risk_premium = purrr::map_dbl(model, ~ .$coefficients),
      n = purrr::map_dbl(data, nrow),
      standard_error = purrr::map_dbl(
        model,
        \(x) compute_standard_error(x, vcov, vcov_options)
      ),
      t_statistic = risk_premium / standard_error
    )

  if (vcov == "iid") {
    aggregations <- aggregations |>
      mutate(t_statistic = t_statistic * sqrt(n))
  }

  aggregations <- aggregations |>
    select(factor = name, risk_premium, n, standard_error, t_statistic)

  if (detail) {
    avg_r_squared <- mean(cross_section_stats$r_squared)
    avg_adj_r_squared <- mean(cross_section_stats$adj_r_squared)
    avg_n_obs <- mean(cross_section_stats$n_obs)

    summary_statistics <- tibble::tibble(
      r_squared = avg_r_squared,
      adj_r_squared = avg_adj_r_squared,
      n_obs = avg_n_obs
    )

    list(
      coefficients = aggregations,
      summary_statistics = summary_statistics
    )
  } else {
    aggregations
  }
}

test_that("estimate_fama_macbeth handles missing date column", {
  data <- tibble(
    permno = rep(1:50, times = 12),
    ret_excess = rnorm(600, 0, 0.1),
    beta = rnorm(600, 1, 0.2),
    bm = rnorm(600, 0.5, 0.1),
    log_mktcap = rnorm(600, 10, 1)
  )

  expect_error(
    estimate_fama_macbeth(data, "ret_excess ~ beta + bm + log_mktcap"),
  )
})

test_that("estimate_fama_macbeth handles invalid vcov option", {
  data <- tibble(
    date = rep(
      seq.Date(
        from = as.Date("2020-01-01"),
        to = as.Date("2020-12-01"),
        by = "month"
      ),
      each = 50
    ),
    permno = rep(1:50, times = 12),
    ret_excess = rnorm(600, 0, 0.1),
    beta = rnorm(600, 1, 0.2),
    bm = rnorm(600, 0.5, 0.1),
    log_mktcap = rnorm(600, 10, 1)
  )

  expect_error(
    estimate_fama_macbeth(
      data,
      "ret_excess ~ beta + bm + log_mktcap",
      vcov = "invalid_vcov"
    ),
    regexp = "`vcov` must be either 'iid' or 'newey-west'."
  )
})

test_that("estimate_fama_macbeth works with valid iid vcov", {
  data <- tibble(
    date = rep(
      seq.Date(
        from = as.Date("2020-01-01"),
        to = as.Date("2020-12-01"),
        by = "month"
      ),
      each = 50
    ),
    permno = rep(1:50, times = 12),
    ret_excess = rnorm(600, 0, 0.1),
    beta = rnorm(600, 1, 0.2),
    bm = rnorm(600, 0.5, 0.1),
    log_mktcap = rnorm(600, 10, 1)
  )

  result <- estimate_fama_macbeth(
    data,
    "ret_excess ~ beta + bm + log_mktcap",
    vcov = "iid"
  )

  expect_s3_class(result, "tbl_df")
  expect_true(all(
    c(
      "factor",
      "risk_premium",
      "n",
      "standard_error",
      "t_statistic"
    ) %in%
      colnames(result)
  ))
})

test_that("estimate_fama_macbeth works with valid newey-west vcov", {
  data <- tibble(
    date = rep(
      seq.Date(
        from = as.Date("2020-01-01"),
        to = as.Date("2020-12-01"),
        by = "month"
      ),
      each = 50
    ),
    permno = rep(1:50, times = 12),
    ret_excess = rnorm(600, 0, 0.1),
    beta = rnorm(600, 1, 0.2),
    bm = rnorm(600, 0.5, 0.1),
    log_mktcap = rnorm(600, 10, 1)
  )

  result <- estimate_fama_macbeth(
    data,
    "ret_excess ~ beta + bm + log_mktcap",
    vcov = "newey-west",
    vcov_options = list(prewhite = FALSE)
  )

  expect_s3_class(result, "tbl_df")
  expect_true(all(
    c(
      "factor",
      "risk_premium",
      "n",
      "standard_error",
      "t_statistic"
    ) %in%
      colnames(result)
  ))
})

test_that("estimate_fama_macbeth computes correct number of rows", {
  data <- tibble(
    date = rep(
      seq.Date(
        from = as.Date("2020-01-01"),
        to = as.Date("2020-12-01"),
        by = "month"
      ),
      each = 50
    ),
    permno = rep(1:50, times = 12),
    ret_excess = rnorm(600, 0, 0.1),
    beta = rnorm(600, 1, 0.2),
    bm = rnorm(600, 0.5, 0.1),
    log_mktcap = rnorm(600, 10, 1)
  )

  result <- estimate_fama_macbeth(
    data,
    "ret_excess ~ beta + bm + log_mktcap",
    vcov_options = list(prewhite = FALSE)
  )

  expect_equal(nrow(result), 4)
})

test_that("estimate_fama_macbeth handles edge cases", {
  data <- tibble(
    date = rep(
      seq.Date(
        from = as.Date("2020-01-01"),
        to = as.Date("2020-12-01"),
        by = "month"
      ),
      each = 1
    ),
    permno = 1,
    ret_excess = rnorm(12, 0, 0.1),
    beta = rnorm(12, 1, 0.2),
    bm = rnorm(12, 0.5, 0.1),
    log_mktcap = rnorm(12, 10, 1)
  )

  expect_error(
    estimate_fama_macbeth(data, "ret_excess ~ beta + bm + log_mktcap"),
    regexp = paste(
      "Each date grouping must have more rows than the number of predictors",
      "in the model to estimate coefficients. Please check your data."
    )
  )
})

test_that(
  paste(
    "estimate_fama_macbeth detail = FALSE returns flat data frame by default"
  ),
  {
    set.seed(123)
    data <- tibble(
      date = rep(
        seq.Date(
          from = as.Date("2020-01-01"),
          to = as.Date("2020-12-01"),
          by = "month"
        ),
        each = 50
      ),
      permno = rep(1:50, times = 12),
      ret_excess = rnorm(600, 0, 0.1),
      beta = rnorm(600, 1, 0.2),
      bm = rnorm(600, 0.5, 0.1)
    )

    result <- estimate_fama_macbeth(data, "ret_excess ~ beta + bm")

    expect_s3_class(result, "tbl_df")
    expect_true(all(
      c(
        "factor",
        "risk_premium",
        "n",
        "standard_error",
        "t_statistic"
      ) %in%
        colnames(result)
    ))
  }
)

test_that(
  paste(
    "estimate_fama_macbeth detail = TRUE returns list with two data frames"
  ),
  {
    set.seed(123)
    data <- tibble(
      date = rep(
        seq.Date(
          from = as.Date("2020-01-01"),
          to = as.Date("2020-12-01"),
          by = "month"
        ),
        each = 50
      ),
      permno = rep(1:50, times = 12),
      ret_excess = rnorm(600, 0, 0.1),
      beta = rnorm(600, 1, 0.2),
      bm = rnorm(600, 0.5, 0.1)
    )

    result <- estimate_fama_macbeth(
      data,
      "ret_excess ~ beta + bm",
      detail = TRUE
    )

    expect_type(result, "list")
    expect_named(result, c("coefficients", "summary_statistics"))
    expect_s3_class(result$coefficients, "tbl_df")
    expect_s3_class(result$summary_statistics, "tbl_df")
  }
)

test_that(
  paste(
    "estimate_fama_macbeth detail = TRUE coefficients has correct structure"
  ),
  {
    set.seed(123)
    data <- tibble(
      date = rep(
        seq.Date(
          from = as.Date("2020-01-01"),
          to = as.Date("2020-12-01"),
          by = "month"
        ),
        each = 50
      ),
      permno = rep(1:50, times = 12),
      ret_excess = rnorm(600, 0, 0.1),
      beta = rnorm(600, 1, 0.2),
      bm = rnorm(600, 0.5, 0.1),
      log_mktcap = rnorm(600, 10, 1)
    )

    result <- estimate_fama_macbeth(
      data,
      "ret_excess ~ beta + bm + log_mktcap",
      detail = TRUE
    )

    coefficients <- result$coefficients
    expect_equal(
      colnames(coefficients),
      c("factor", "risk_premium", "n", "standard_error", "t_statistic")
    )
    expect_equal(nrow(coefficients), 4)
    expect_false("r_squared" %in% colnames(coefficients))
    expect_false("adj_r_squared" %in% colnames(coefficients))
    expect_false("n_obs" %in% colnames(coefficients))
  }
)

test_that(
  paste(
    "estimate_fama_macbeth detail = TRUE summary_statistics",
    "has correct structure"
  ),
  {
    set.seed(42)
    data <- tibble(
      date = rep(
        seq.Date(
          from = as.Date("2020-01-01"),
          to = as.Date("2020-12-01"),
          by = "month"
        ),
        each = 50
      ),
      permno = rep(1:50, times = 12),
      ret_excess = rnorm(600, 0, 0.1),
      beta = rnorm(600, 1, 0.2),
      bm = rnorm(600, 0.5, 0.1)
    )

    result <- estimate_fama_macbeth(
      data,
      "ret_excess ~ beta + bm",
      detail = TRUE
    )

    summary_stats <- result$summary_statistics
    expect_equal(
      colnames(summary_stats),
      c("r_squared", "adj_r_squared", "n_obs")
    )
    expect_equal(nrow(summary_stats), 1)
    expect_true(summary_stats$r_squared >= 0 & summary_stats$r_squared <= 1)
    expect_true(summary_stats$adj_r_squared <= summary_stats$r_squared)
    expect_equal(summary_stats$n_obs, 50)
  }
)

test_that("estimate_fama_macbeth detail works with different vcov options", {
  set.seed(123)
  data <- tibble(
    date = rep(
      seq.Date(
        from = as.Date("2020-01-01"),
        to = as.Date("2020-12-01"),
        by = "month"
      ),
      each = 50
    ),
    permno = rep(1:50, times = 12),
    ret_excess = rnorm(600, 0, 0.1),
    beta = rnorm(600, 1, 0.2),
    bm = rnorm(600, 0.5, 0.1)
  )

  result_iid <- estimate_fama_macbeth(
    data,
    "ret_excess ~ beta + bm",
    vcov = "iid",
    detail = TRUE
  )
  result_nw <- estimate_fama_macbeth(
    data,
    "ret_excess ~ beta + bm",
    vcov = "newey-west",
    detail = TRUE
  )

  # Both should return the same structure
  expect_type(result_iid, "list")
  expect_named(result_iid, c("coefficients", "summary_statistics"))
  expect_type(result_nw, "list")
  expect_named(result_nw, c("coefficients", "summary_statistics"))

  # Risk premia should be identical (same data, same cross-sectional fits)
  expect_equal(
    result_iid$coefficients$risk_premium,
    result_nw$coefficients$risk_premium
  )

  # Standard errors should differ between vcov methods
  expect_false(
    all(
      result_iid$coefficients$standard_error ==
        result_nw$coefficients$standard_error
    )
  )
})

#' List Parquet Files in a Hugging Face Dataset
#'
#' Query the Hugging Face Datasets API and return a tibble of files with a
#' `.parquet` suffix. The function follows pagination links returned in the
#' response `Link` header and returns path, size, and a resolved URL.
#'
#' @details
#' Uses httr2 to perform HTTP requests. Requires internet access and the
#' dataset to be publicly accessible or accessible with appropriate
#' authentication.
#'
#' @param organization Character(1). Hugging Face organization or user name.
#' @param dataset Character(1). Dataset name under the organization.
#'
#' @returns A tibble with columns: `path` (character), `size` (numeric), and
#'   `url` (character).
#'
#' @family utility functions
#' @export
#'
#' @examples
#' \dontrun{
#' get_available_huggingface_files("voigtstefan", "sp500")
#' }
get_available_huggingface_files <- function(organization, dataset) {
  api_url <- paste0(
    "https://huggingface.co/api/datasets/",
    organization,
    "/",
    dataset,
    "/tree/main?recursive=1"
  )

  out <- tibble::tibble(NULL)
  next_url <- api_url

  repeat {
    resp <- httr2::request(next_url) |>
      httr2::req_user_agent("httr2") |>
      httr2::req_perform()

    body <- resp |>
      httr2::resp_body_string() |>
      jsonlite::fromJSON(simplifyDataFrame = TRUE) |>
      tibble::tibble() |>
      dplyr::filter(
        .data$type == "file" &
          grepl("\\.parquet$", .data$path, ignore.case = TRUE)
      ) |>
      dplyr::select("path", "size")

    out <- out |>
      dplyr::bind_rows(tibble::tibble(data = list(body)))
    link <- httr2::resp_headers(resp)$link
    if (is.null(link) || !grepl('rel="next"', link)) {
      break
    }
    next_url <- sub(".*<([^>]+)>; rel=\"next\".*", "\\1", link)
  }

  out |>
    tidyr::unnest("data") |>
    dplyr::mutate(
      url = glue::glue(
        "https://huggingface.co/datasets/",
        "{organization}/{dataset}/resolve/main/{path}"
      )
    )
}

#' Download data from a Hugging Face dataset
#'
#' Downloads data from a supported Hugging Face dataset. For
#' `"high_frequency_sp500"`, parquet files are filtered by date range and
#' row-bound. For `"factor_library"`, portfolio characteristics are selected via
#' `filter_factor_library_grid()` and the matching return data is downloaded.
#'
#' @param dataset Character(1). The dataset to download. Supported values are
#'   `"high_frequency_sp500"` and `"factor_library"`.
#' @param start_date Date or character. Start date (inclusive) in
#'   `"YYYY-MM-DD"` format. Only used for `"high_frequency_sp500"`.
#' @param end_date Date or character. End date (inclusive) in `"YYYY-MM-DD"`
#'   format. Only used for `"high_frequency_sp500"`.
#' @param type `r lifecycle::badge("deprecated")` Use `dataset` instead.
#' @param ... For `dataset = "factor_library"`: named arguments used to filter
#'   the portfolio grid. Each argument takes the form `column = value`, where
#'   `value` may be a vector to match multiple levels. Optionally pass
#'   `fill_all = TRUE` to leave unspecified columns unrestricted (default:
#'   `FALSE`, i.e. unspecified columns are fixed at the defaults listed below).
#'   Passing `NULL` for any parameter removes that filter entirely, returning
#'   all values for that column (e.g., `min_size_quantile = NULL` includes all
#'   size groups). Passing an unrecognised column name raises an error listing
#'   the supported names. Ignored when `dataset != "factor_library"`. See the
#'   Details section for supported columns and their defaults.
#'
#' @details
#' **Note on `dataset = "factor_library"` defaults:** The defaults below reflect
#' one common portfolio construction choice, but may not suit every research
#' question. Always verify that the selected combination matches your intended
#' design.
#'
#' Supported columns and their defaults for `...`:
#'   \itemize{
#'     \item `sorting_variable`: **Required.** The firm characteristic used
#'       to sort stocks into portfolios (e.g., `"me"` for market equity,
#'       `"bm"` for book-to-market). No default is applied.
#'     \item `min_size_quantile` (defaults to `0.2`): Fraction of the smallest
#'       stocks (by market cap) excluded from the portfolio universe. `0.2`
#'       drops the bottom 20%.
#'     \item `exclude_financials` (defaults to `FALSE`): Whether to drop
#'       financial-sector stocks (SIC 6000-6999) from the universe.
#'     \item `exclude_utilities` (default: `FALSE`): Whether to drop
#'       utility-sector stocks (SIC 4900-4999) from the universe.
#'     \item `exclude_negative_earnings` (defaults to `FALSE`): Whether to
#'       drop firms with negative earnings before sorting.
#'     \item `sorting_variable_lag` (defaults to `"6m"`): Lag applied to the
#'       sorting variable before portfolio assignment (e.g., `"6m"` = 6-month
#'       lag).
#'     \item `rebalancing` (defaults to `"monthly"`): How frequently portfolios
#'       are reformed: `"monthly"` or `"annual"`.
#'     \item `n_portfolios_main` (defaults to `10`): Number of quantile groups
#'       (e.g., `10` for decile portfolios).
#'     \item `sorting_method` (defaults to `"univariate"`): Whether portfolios
#'       are formed on a single sort (`"univariate"`) or a sequential double
#'       sort (`"sequential"`).
#'     \item `n_portfolios_secondary` (defaults to `NULL`): Number of groups
#'       for the secondary sort variable.
#'       Required when `sorting_method` is not `"univariate"`.
#'     \item `breakpoints_exchanges` (defaults to: `"NYSE"`): Exchange(s) used
#'       to compute breakpoints. `"NYSE"` uses only NYSE-listed stocks to
#'       define quantile cutoffs (the conventional Fama-French approach).
#'     \item `breakpoints_min_size_threshold` (defaults to `NULL`): Minimum
#'       market-cap threshold (in USD) applied when computing breakpoints.
#'       `NULL` means no minimum-size screen is applied.
#'     \item `weighting_scheme` (defaults to `"VW"`): Return weighting within
#'       portfolios: `"VW"` for value-weighted or `"EW"` for equal-weighted.
#'   }
#'
#' @returns A tibble with the downloaded data. For `"high_frequency_sp500"`,
#'   contains 5-second aggregated orderbook snapshots filtered to the requested
#'   date range. For `"factor_library"`, contains portfolio return data joined
#'   with the full grid metadata for the matched portfolio IDs.
#'
#' @family download functions
#' @export
#'
#' @examples
#' \dontrun{
#'   download_data_huggingface(
#'     "high_frequency_sp500", "2007-07-26", "2007-07-27"
#'   )
#'   download_data_huggingface(
#'     "factor_library",
#'     sorting_variable = "52w",
#'     rebalancing = "annual"
#'   )
#'   download_data_huggingface(
#'     "factor_library", sorting_variable = "ag", fill_all = TRUE
#'   )
#' }
download_data_huggingface <- function(
  dataset = NULL,
  start_date = "2007-06-27",
  end_date = "2007-07-27",
  type = deprecated(),
  ...
) {
  # Handle explicit type argument
  if (lifecycle::is_present(type)) {
    lifecycle::deprecate_warn(
      when = "0.5.0",
      what = "download_data_huggingface(type)",
      details = "Use the `dataset` argument instead."
    )
    dataset <- sub("^hf_", "", type)
  }

  # Handle legacy type passed as dataset argument
  if (!is.null(dataset) && is_legacy_type_hf(dataset)) {
    lifecycle::deprecate_warn(
      when = "0.5.0",
      what = "download_data_huggingface(type)",
      details = paste0(
        "The `type` argument is deprecated. ",
        "Use `dataset` instead (e.g., 'high_frequency_sp500' instead of",
        "'hf_high_frequency_sp500')."
      )
    )
    dataset <- sub("^hf_", "", dataset)
  }

  if (is.null(dataset)) {
    cli::cli_abort("Argument {.arg dataset} is required.")
  }

  check_supported_dataset_huggingface(dataset)

  if (dataset == "high_frequency_sp500") {
    organization <- "voigtstefan"
    dataset_name <- "sp500"

    date_pattern <- "date=([0-9]{4}-[0-9]{2}-[0-9]{2})"
    available_files <- get_available_huggingface_files(
      organization,
      dataset_name
    ) |>
      dplyr::mutate(
        date = as.Date(stringr::str_match(.data$path, date_pattern)[, 2])
      )

    tibble::tibble(
      date = seq.Date(as.Date(start_date), as.Date(end_date), by = "day")
    ) |>
      dplyr::inner_join(available_files, dplyr::join_by(date)) |>
      dplyr::transmute(
        data = purrr::map(url, ~ arrow::read_parquet(.x))
      ) |>
      tidyr::unnest("data")
  } else if (dataset == "factor_library") {
    download_data_hugging_face_factor_library(...)
  } else {
    cli::cli_abort("Unsupported dataset: {.val {dataset}}")
  }
}

#' Check if a string is a legacy Hugging Face type
#' @noRd
is_legacy_type_hf <- function(x) {
  grepl("^hf_", x)
}

#' Check if Hugging Face dataset is supported
#' @noRd
check_supported_dataset_huggingface <- function(dataset) {
  supported_datasets <- c("high_frequency_sp500", "factor_library")

  if (!dataset %in% supported_datasets) {
    cli::cli_abort(c(
      "Unsupported Hugging Face dataset: {.val {dataset}}",
      "i" = "Supported datasets: {.val {supported_datasets}}"
    ))
  }
}

#' Filter the factor library grid and return matching IDs
#'
#' Downloads the `tidy-finance/factor-library-grid` from Hugging Face and
#' filters it by the provided column-value pairs. Any grid column not explicitly
#' specified is held at a sensible default when `fill_all = FALSE`. The function
#' returns the integer vector of matching portfolio IDs.
#'
#' @param ... Named arguments of the form `column = value` used to filter the
#'   grid. Supported columns and their defaults are:
#'   \describe{
#'     \item{`sorting_variable`}{No default.}
#'     \item{`min_size_quantile`}{`0.2`}
#'     \item{`exclude_financials`}{`FALSE`}
#'     \item{`exclude_utilities`}{`FALSE`}
#'     \item{`exclude_negative_earnings`}{`FALSE`}
#'     \item{`sorting_variable_lag`}{`"6m"`}
#'     \item{`rebalancing`}{`"monthly"`}
#'     \item{`n_portfolios_main`}{`10`}
#'     \item{`sorting_method`}{`"univariate"`}
#'     \item{`n_portfolios_secondary`}{`NULL` for univariate sorts;
#'       required otherwise}
#'     \item{`breakpoints_exchanges`}{`"NYSE"`}
#'     \item{`breakpoints_min_size_threshold`}{`NULL`}
#'     \item{`weighting_scheme`}{`"VW"`}
#'   }
#'   Each value can be a vector to match multiple levels.
#' @param fill_all Logical(1). If `FALSE` (default), columns not present in
#'   `...` are filled with their defaults before filtering. If `TRUE`, only the
#'   explicitly provided filters are applied and all other columns are left
#'   unrestricted.
#'
#' @return A vector of portfolio IDs matching the specified criteria. Returns
#'   an empty vector when no rows satisfy the filters.
#' @noRd
filter_factor_library_grid <- function(..., fill_all = FALSE) {
  filters <- list(...)

  defaults <- list(
    min_size_quantile = 0.2,
    exclude_financials = FALSE,
    exclude_utilities = FALSE,
    exclude_negative_earnings = FALSE,
    sorting_variable_lag = "6m",
    rebalancing = "monthly",
    n_portfolios_main = 10,
    sorting_method = "univariate",
    n_portfolios_secondary = NULL,
    breakpoints_exchanges = "NYSE",
    breakpoints_min_size_threshold = NA_real_,
    weighting_scheme = "VW"
  )

  supported_names <- c("sorting_variable", names(defaults))
  unsupported <- setdiff(names(filters), supported_names)
  if (length(unsupported) > 0) {
    cli::cli_abort(c(
      paste0(
        "{length(unsupported)} unsupported filter ",
        "name{?s}: {.val {unsupported}}"
      ),
      "i" = "Supported filters: {.val {supported_names}}"
    ))
  }

  if (!fill_all) {
    for (col in names(defaults)) {
      if (!col %in% names(filters)) {
        filters[col] <- list(defaults[[col]])
      }
    }

    if (is.null(filters[["n_portfolios_secondary"]])) {
      sorting_methods <- filters[["sorting_method"]]
      if (!all(sorting_methods == "univariate")) {
        cli::cli_abort(c(
          paste0(
            "{.arg n_portfolios_secondary} must be specified ",
            "for bivariate sorts."
          ),
          "i" = paste(
            "Provide a value for {.arg n_portfolios_secondary} or",
            "use {.code fill_all = TRUE} to skip all defaults."
          )
        ))
      }
      filters["n_portfolios_secondary"] <- list(NA_real_)
    }
  }

  result <- get_available_huggingface_files(
    "tidy-finance",
    "factor-library-grid"
  ) |>
    dplyr::pull(.data$url) |>
    arrow::read_parquet() |>
    dplyr::mutate(
      sorting_variable = stringr::str_replace(.data$sorting_variable, "sv_", "")
    )

  filters <- purrr::compact(filters)

  for (col in names(filters)) {
    result <- dplyr::filter(result, .data[[col]] %in% filters[[col]])
  }

  dplyr::pull(result, .data$id)
}

#' Download factor library returns for a vector of portfolio IDs
#'
#' Given a vector of portfolio IDs from the `tidy-finance/factor-library-grid`,
#' this function downloads the corresponding return data from the
#' `tidy-finance/factor-library` dataset on Hugging Face. It identifies the
#' unique `(sorting_variable, sorting_variable_lag, sorting_method,
#' n_portfolios_main)` combinations for the requested IDs, downloads one
#' parquet file per combination in full, and then inner-joins to retain only
#' the requested IDs. The grid metadata is joined back onto the result.
#'
#' Raises an error if `ids` is empty or contains IDs that cannot be matched to
#' a parquet file (listing the affected IDs and their key columns).
#'
#' @param ids Vector of portfolio IDs to download, as returned by
#'   `filter_factor_library_grid()`.
#'
#' @return A tibble of portfolio returns with the grid metadata columns for the
#'   requested IDs appended.
#' @noRd
download_factor_library_ids <- function(ids) {
  organization <- "tidy-finance"
  dataset_name <- "factor-library"

  available_files <- get_available_huggingface_files(
    organization,
    dataset_name
  ) |>
    tidyr::extract(
      col = "path",
      into = c(
        "sorting_variable",
        "sorting_variable_lag",
        "sorting_method",
        "n_portfolios_main"
      ),
      regex = paste0(
        "sorting_variable=([^/]+)/sorting_variable_lag=([^/]+)",
        "/sorting_method=([^/]+)/n_portfolios_main=([^/]+)/"
      ),
      remove = FALSE
    )

  id_values <- data.frame(id = ids)

  id_grid <- get_available_huggingface_files(
    organization,
    "factor-library-grid"
  ) |>
    dplyr::pull(.data$url) |>
    arrow::read_parquet() |>
    dplyr::inner_join(id_values, dplyr::join_by(id)) |>
    dplyr::mutate(
      sorting_variable = stringr::str_replace(
        .data$sorting_variable,
        "sv_",
        ""
      ),
      n_portfolios_main = as.character(.data$n_portfolios_main)
    ) |>
    dplyr::left_join(
      available_files,
      dplyr::join_by(
        sorting_variable,
        sorting_variable_lag,
        sorting_method,
        n_portfolios_main
      )
    )

  relevant_urls <- id_grid |>
    dplyr::distinct(
      .data$url,
      .data$sorting_variable,
      .data$sorting_variable_lag,
      .data$sorting_method,
      .data$n_portfolios_main
    )

  if (nrow(relevant_urls) == 0) {
    cli::cli_abort(c(
      "No parquet files found for the requested portfolio IDs.",
      "i" = paste(
        "Check that the provided {.arg ids} are valid",
        "and exist in the factor library grid."
      )
    ))
  }

  missing_urls <- relevant_urls |>
    dplyr::filter(is.na(.data$url))

  if (nrow(missing_urls) > 0) {
    missing_keys <- missing_urls |> # nolint: object_usage_linter
      dplyr::inner_join(
        id_grid |>
          dplyr::select(
            "id",
            "sorting_variable",
            "sorting_variable_lag",
            "sorting_method",
            "n_portfolios_main"
          ),
        dplyr::join_by(
          sorting_variable,
          sorting_variable_lag,
          sorting_method,
          n_portfolios_main
        )
      ) |>
      dplyr::mutate(
        key = paste0(
          "id=",
          .data$id,
          " (",
          .data$sorting_variable,
          " / ",
          .data$sorting_variable_lag,
          " / ",
          .data$sorting_method,
          " / ",
          .data$n_portfolios_main,
          ")"
        )
      ) |>
      dplyr::pull(.data$key)

    cli::cli_abort(c(
      "No parquet file found for {length(missing_keys)} portfolio ID{?s}.",
      "x" = "Affected ID{?s}: {.val {missing_keys}}",
      "i" = paste(
        "Check that the {.arg sorting_variable},",
        "{.arg sorting_variable_lag}, {.arg sorting_method},",
        "and {.arg n_portfolios_main} values exist in the factor library."
      )
    ))
  }

  relevant_files <- relevant_urls$url |>
    purrr::map(~ arrow::read_parquet(.x)) |>
    dplyr::bind_rows()

  relevant_files |>
    dplyr::inner_join(
      id_grid |> dplyr::select(-c(url, path, size)),
      by = dplyr::join_by(id)
    )
}

#' Download factor library data from Hugging Face
#'
#' A thin wrapper that combines `filter_factor_library_grid()` and
#' `download_factor_library_ids()`: it resolves matching portfolio IDs
#' from the grid and then downloads the corresponding return data.
#'
#' @param ... Named filter arguments forwarded to
#'   `filter_factor_library_grid()`. See
#'   `filter_factor_library_grid()` for the full list of supported
#'   columns and their defaults.
#' @param fill_all Logical(1). Forwarded to
#'   `filter_factor_library_grid()`. When `TRUE`, columns not
#'   specified in `...` are left unrestricted rather than set to
#'   their defaults.
#'
#' @return A tibble of portfolio returns with grid metadata columns
#'   appended, one row per portfolio-period observation for the
#'   matched IDs.
#' @noRd
download_data_hugging_face_factor_library <- function(..., fill_all = FALSE) {
  ids <- filter_factor_library_grid(..., fill_all = fill_all)
  download_factor_library_ids(ids)
}

# Helpers ---------------------------------------------------------------

make_grid_parquet_file <- function(rows) {
  tmp <- tempfile(fileext = ".parquet")
  arrow::write_parquet(rows, tmp)
  tibble::tibble(path = "grid/part.parquet", size = file.size(tmp), url = tmp)
}

make_returns_parquet_file <- function(
  rows,
  sv,
  sv_lag,
  sorting_method,
  n_portfolios_main
) {
  tmp <- tempfile(fileext = ".parquet")
  arrow::write_parquet(rows, tmp)
  tibble::tibble(
    path = paste0(
      "sorting_variable=",
      sv,
      "/sorting_variable_lag=",
      sv_lag,
      "/sorting_method=",
      sorting_method,
      "/n_portfolios_main=",
      n_portfolios_main,
      "/part.parquet"
    ),
    size = file.size(tmp),
    url = tmp
  )
}

# Input validation (no network) -----------------------------------------

test_that(
  paste(
    "filter_grid() aborts with a clear message on unsupported filter names"
  ),
  {
    expect_error(
      download_data_huggingface("factor_library", not_a_column = "x"),
      regexp = "unsupported filter name"
    )
  }
)

# Empty / missing URL behaviour (mocked) --------------------------------

test_that(
  paste(
    "download_factor_library_returns_ids() aborts when no IDs match the grid"
  ),
  {
    grid_rows <- tibble::tibble(
      id = 1L,
      sorting_variable = "sv_me",
      sorting_variable_lag = "3m",
      sorting_method = "univariate",
      n_portfolios_main = 10L
    )

    mock_files <- function(organization, dataset) {
      if (dataset == "factor-library-grid") {
        make_grid_parquet_file(grid_rows)
      } else {
        make_returns_parquet_file(
          tibble::tibble(id = 1L),
          "me",
          "3m",
          "univariate",
          10L
        )
      }
    }

    with_mocked_bindings(
      get_available_huggingface_files = mock_files,
      {
        expect_error(
          download_factor_library_ids(integer(0)),
          regexp = "No parquet files found for the requested portfolio IDs"
        )
      }
    )
  }
)

test_that(
  paste(
    "download_factor_library_returns_ids() aborts when a parquet URL",
    "is missing for a portfolio ID"
  ),
  {
    grid_rows <- tibble::tibble(
      id = 1L,
      sorting_variable = "sv_me",
      sorting_variable_lag = "3m",
      sorting_method = "univariate",
      n_portfolios_main = 10L
    )

    mock_files <- function(organization, dataset) {
      if (dataset == "factor-library-grid") {
        make_grid_parquet_file(grid_rows)
      } else {
        tibble::tibble(
          path = character(0),
          size = numeric(0),
          url = character(0)
        )
      }
    }

    with_mocked_bindings(
      get_available_huggingface_files = mock_files,
      {
        expect_error(
          download_factor_library_ids(1L),
          regexp = "No parquet file found for 1 portfolio ID"
        )
      }
    )
  }
)

# NULL default / explicit NULL behaviour (mocked) ----------------------

test_that(
  paste(
    "filter_factor_library_grid(): omitted breakpoints_min_size defaults to",
    "NA (standard portfolio); explicit NULL removes the filter"
  ),
  {
    grid_rows <- tibble::tibble(
      id = c(1L, 2L),
      sorting_variable = c("sv_bm", "sv_bm"),
      min_size_quantile = c(0.2, 0.2),
      exclude_financials = c(FALSE, FALSE),
      exclude_utilities = c(FALSE, FALSE),
      exclude_negative_earnings = c(FALSE, FALSE),
      sorting_variable_lag = c("6m", "6m"),
      rebalancing = c("monthly", "monthly"),
      n_portfolios_main = c(10L, 10L),
      sorting_method = c("univariate", "univariate"),
      n_portfolios_secondary = c(NA_real_, NA_real_),
      breakpoints_exchanges = c("NYSE", "NYSE"),
      breakpoints_min_size_threshold = c(NA_real_, 1e9),
      weighting_scheme = c("VW", "VW")
    )

    mock_files <- function(organization, dataset) {
      make_grid_parquet_file(grid_rows)
    }

    with_mocked_bindings(
      get_available_huggingface_files = mock_files,
      {
        ids_default <- filter_factor_library_grid(sorting_variable = "bm")
        ids_explicit_null <- filter_factor_library_grid(
          sorting_variable = "bm",
          breakpoints_min_size_threshold = NULL
        )

        expect_length(ids_default, 1L)
        expect_length(ids_explicit_null, 2L)
      }
    )
  }
)

# Successful download (mocked) ------------------------------------------

test_that(
  paste(
    "download_data_huggingface() returns a tibble for factor_library with",
    "matched IDs"
  ),
  {
    grid_rows <- tibble::tibble(
      id = 1L,
      sorting_variable = "sv_me",
      sorting_variable_lag = "3m",
      sorting_method = "univariate",
      n_portfolios_main = 10L
    )
    returns_rows <- tibble::tibble(id = 1L, ret = 0.01)

    mock_files <- function(organization, dataset) {
      if (dataset == "factor-library-grid") {
        make_grid_parquet_file(grid_rows)
      } else {
        make_returns_parquet_file(
          returns_rows,
          "me",
          "3m",
          "univariate",
          10L
        )
      }
    }

    with_mocked_bindings(
      get_available_huggingface_files = mock_files,
      {
        result <- download_data_huggingface(
          "factor_library",
          sorting_variable = "me",
          fill_all = TRUE
        )
        expect_s3_class(result, "tbl_df")
        expect_true(nrow(result) > 0)
        expect_true("id" %in% colnames(result))
      }
    )
  }
)

# Live smoke test -------------------------------------------------------

test_that(
  paste(
    "download_data_huggingface() downloads factor_library data successfully"
  ),
  {
    skip_if_offline()
    skip_on_cran()

    result <- download_data_huggingface(
      "factor_library",
      sorting_variable = "me"
    )

    expect_s3_class(result, "tbl_df")
    expect_true(nrow(result) > 0)
    expect_true("id" %in% colnames(result))
  }
)

#' Download Risk-Free Rate Data
#'
#' Downloads pre-processed risk-free rate data from the
#' `tidy-finance/risk-free` dataset on HuggingFace. The dataset is
#' updated monthly via a scheduled GitHub Actions workflow that splices the
#' 3-Month Treasury Bill Secondary Market Rate (pre-2001) with the 4-Week
#' Treasury Bill Secondary Market Rate (from 2001 onwards) sourced from FRED.
#' For monthly data, the monthly TB3MS series is spliced with the daily DTB4WK
#' series aggregated to month-end. For daily data, the daily DTB3 series is
#' spliced with the daily DTB4WK series, both at the business-day frequency
#' provided by FRED.
#'
#' @param start_date Optional. A character string or Date object in
#'   "YYYY-MM-DD" format specifying the start date for the data. If
#'   not provided, the full dataset is returned.
#' @param end_date Optional. A character string or Date object in
#'   "YYYY-MM-DD" format specifying the end date for the data. If not
#'   provided, the full dataset is returned.
#' @param frequency A character string, either `"monthly"` (default)
#'   or `"daily"`, specifying the frequency of the returned data. Daily
#'   data starts in 1954-01-04 because of availability of DTB3, while
#'   monthly data starts in 1934-01-01.
#'
#' @details
#' Both series are quoted as annualised bank discount rates on a
#' 360-day basis. Given an annualised discount rate `d` and a T-bill
#' with `n` days to maturity, the holding-period return is
#' `HPR = d * n/360 / (1 - d * n/360)`, which is then converted to the
#' target period length via `(1 + HPR)^(target/source) - 1`.
#'
#' The series are spliced at 2001-07-01:
#' - **Pre-2001**: TB3MS (monthly) or DTB3 (daily), 3-month T-bill
#'   with n = 90. Monthly conversion uses exponent `1/3`; daily
#'   conversion uses exponent `1/63` (approx. trading days per
#'   quarter).
#' - **From 2001**: DTB4WK, 4-week T-bill with n = 28. For monthly
#'   data, the last non-NA observation per calendar month is taken
#'   and the exponent is `365/(28*12)`. For daily data, observations
#'   are used as-is and the exponent is `1/20` (approx. trading days
#'   per 4-week period).
#'
#' Business-day gaps in the daily series (e.g. holidays) are handled
#' by forward-filling the most recent available rate.
#'
#' Monthly data starts in 1934-01-01 (TB3MS). Daily data starts in
#' 1954-01-04 due to the availability of DTB3.
#'
#' @returns A tibble with two columns:
#' \describe{
#'   \item{date}{The date of the observation.}
#'   \item{risk_free}{The risk-free rate for the period.}
#' }
#'
#' @family download functions
#' @export
#'
#' @examples
#' \donttest{
#'   download_data_risk_free("2020-01-01", "2020-12-31")
#'   download_data_risk_free(
#'     "2020-01-01", "2020-12-31", frequency = "daily"
#'   )
#' }
download_data_risk_free <- function(
  start_date = NULL,
  end_date = NULL,
  frequency = "monthly"
) {
  if (!(frequency %in% c("monthly", "daily"))) {
    cli::cli_abort(
      "{.arg frequency} must be {.str monthly} or {.str daily}."
    )
  }

  dates <- validate_dates(start_date, end_date)
  start_date <- dates$start_date
  end_date <- dates$end_date

  url <- paste0(
    "https://huggingface.co/datasets/tidy-finance/risk-free/",
    "resolve/main/risk_free_",
    frequency,
    ".parquet"
  )

  risk_free_data <- tryCatch(
    arrow::read_parquet(url),
    error = function(e) {
      cli::cli_abort(c(
        "Failed to download risk-free rate data from HuggingFace.",
        "i" = "URL attempted: {url}",
        "x" = conditionMessage(e)
      ))
    }
  )

  if (!is.null(start_date)) {
    risk_free_data <- risk_free_data |>
      filter(between(date, start_date, end_date))
  }

  risk_free_data
}

#' Join Lagged Variable Values over a Date Range
#'
#' @description `r lifecycle::badge('experimental')`
#'
#' Joins lagged values of selected variables from one dataset (`new_data`)
#' into another (`original_data`), based on date ranges defined by `min_lag`
#' and `max_lag`. Unlike [add_lagged_columns()], this function supports
#' joining across data frames with different date grids (e.g., monthly source
#' data into quarterly target data).
#'
#' @param original_data A data frame containing the target panel data.
#' @param new_data A data frame containing the source variables to lag and
#'   merge. All columns besides `id_keys` and the date column will be lagged
#'   and joined.
#' @param id_keys A character vector specifying the identifier column(s).
#' @param min_lag A `lubridate::Period` specifying the lower lag bound
#'   (inclusive).
#' @param max_lag A `lubridate::Period` specifying the upper lag bound
#'   (inclusive).
#' @param ff_adjustment Logical; if `TRUE`, keeps only the last observation
#'   per identifier and year before lagging (Fama-French convention). Defaults
#'   to `FALSE`.
#' @param data_options A list of class `tidyfinance_data_options` (created via
#'   [data_options()]) specifying column name mappings. The `date` element is
#'   used to identify the date column. Uses [data_options()] default if
#'   `NULL`: `"date" = "date"`.
#'
#' @returns A data frame with all columns from `original_data` plus the
#'   lagged columns from `new_data` (keeping their original names).
#'
#' @family rolling and lagging functions
#' @export
#'
#' @examples
#' set.seed(42)
#' library(dplyr)
#' library(lubridate)
#'
#' df1 <- tibble(
#'   id = rep(1:2, each = 6),
#'   date = rep(seq(as.Date("2020-01-01"), by = "month", length.out = 6), 2)
#' )
#'
#' df2 <- df1 |>
#'   mutate(x = rnorm(n()))
#'
#' join_lagged_values(
#'   original_data = df1,
#'   new_data = df2,
#'   id_keys = "id",
#'   min_lag = months(1),
#'   max_lag = months(3)
#' )
join_lagged_values <- function(
  original_data,
  new_data,
  id_keys,
  min_lag,
  max_lag,
  ff_adjustment = FALSE,
  data_options = NULL
) {
  if (is.null(data_options)) {
    data_options <- data_options()
  }
  id_date <- data_options$date

  if (!is.character(id_keys)) {
    cli::cli_abort(
      paste(
        "{.arg id_keys} must be a character vector,",
        "not {.obj_type_friendly {id_keys}}."
      )
    )
  }
  if (!id_date %in% names(original_data)) {
    cli::cli_abort(
      "{.arg original_data} must contain the column {.field {id_date}}."
    )
  }
  if (!id_date %in% names(new_data)) {
    cli::cli_abort(
      "{.arg new_data} must contain the column {.field {id_date}}."
    )
  }

  missing_keys_original <- setdiff(id_keys, names(original_data))
  if (length(missing_keys_original) > 0) {
    cli::cli_abort(
      paste(
        "{.arg original_data} is missing id column{?s}:",
        "{.field {missing_keys_original}}."
      )
    )
  }

  missing_keys_new <- setdiff(id_keys, names(new_data))
  if (length(missing_keys_new) > 0) {
    cli::cli_abort(
      "{.arg new_data} is missing id column{?s}: {.field {missing_keys_new}}."
    )
  }

  new_column_names <- new_data |>
    dplyr::select(-dplyr::all_of(c(id_keys, id_date))) |>
    names()

  if (length(new_column_names) == 0) {
    cli::cli_abort(
      paste(
        "{.arg new_data} must contain columns besides {.field {id_keys}} and",
        "{.field {id_date}}."
      )
    )
  }

  original_non_key_cols <- setdiff(names(original_data), c(id_keys, id_date))
  duplicate_cols <- intersect(new_column_names, original_non_key_cols)
  if (length(duplicate_cols) > 0) {
    cli::cli_abort(
      c(
        paste0(
          "{cli::qty(length(duplicate_cols))}Column{?s} in {.arg new_data} ",
          "already exist in {.arg original_data}."
        ),
        "x" = paste0(
          "{cli::qty(length(duplicate_cols))}",
          "Duplicate column{?s}: {.field {duplicate_cols}}."
        ),
        "i" = paste0(
          "{cli::qty(length(duplicate_cols))}",
          "Remove or rename {?this column/these columns} ",
          "from {.arg new_data} before joining."
        )
      )
    )
  }

  new_data <- new_data |>
    check_new_col(c(".lower", ".upper")) |>
    dplyr::mutate(
      .lower = .data[[id_date]] %m+% min_lag,
      .upper = .data[[id_date]] %m+% max_lag
    )

  if (ff_adjustment) {
    new_data <- new_data |>
      check_new_col(".year") |>
      dplyr::mutate(.year = lubridate::year(.data[[id_date]]))
  }

  for (col_name in new_column_names) {
    tmp_data <- new_data |>
      dplyr::select(
        dplyr::all_of(c(id_keys, id_date, col_name, ".lower", ".upper")),
        dplyr::any_of(".year")
      )

    if (ff_adjustment) {
      tmp_data <- tmp_data |>
        dplyr::group_by(dplyr::across(dplyr::all_of(c(id_keys, ".year")))) |>
        dplyr::slice_max(order_by = .data[[id_date]], n = 1) |>
        dplyr::ungroup() |>
        dplyr::select(-dplyr::all_of(id_date))
    }

    col_values <- original_data |>
      dplyr::rename(.date = dplyr::all_of(id_date)) |>
      dplyr::left_join(
        tmp_data,
        dplyr::join_by(
          !!!rlang::syms(id_keys),
          closest(.date >= .lower),
          .date <= .upper
        )
      ) |>
      dplyr::pull(dplyr::all_of(col_name))

    original_data <- original_data |>
      check_new_col(col_name) |>
      dplyr::mutate("{col_name}" := col_values)
  }

  original_data
}

#' Download Data from WRDS Compustat
#'
#' Downloads financial data from the WRDS Compustat database for a given
#' dataset, start date, and end date. It filters the data according to industry
#' format, data format, and consolidation level, and returns the most current
#' data for each reporting period. Additionally, the annual data also includes
#' the calculated book equity (be), operating profitability (op), and
#' investment (inv) for each company following Fama & French (1993, 2015), as
#' well as income before extraordinary items (ib).
#'
#' @param dataset The dataset to download ("compustat_annual" or
#'   "compustat_quarterly").
#' @param start_date Optional. A character string or Date object in
#'   "YYYY-MM-DD" format specifying the start date for the data. If not
#'   provided, a subset of the dataset is returned.
#' @param end_date Optional. A character string or Date object in "YYYY-MM-DD"
#'   format specifying the end date for the data. If not provided, a subset of
#'   the dataset is returned.
#' @param type `r lifecycle::badge("deprecated")` Use `dataset` instead.
#' @param additional_columns Additional columns from the Compustat table as a
#'   character vector.
#' @param only_us A logical indicating whether only US firms should be returned.
#' @returns A data frame with financial data for the specified period,
#'   including variables for book equity (be), operating profitability (op),
#'   investment (inv), and others.
#'
#' @references
#'   Fama, E. F., & French, K. R. (1993). Common risk factors in the returns on
#'   stocks and bonds. *Journal of Financial Economics*, 33(1), 3-56.
#'   \doi{10.1016/0304-405X(93)90023-5}
#'
#'   Fama, E. F., & French, K. R. (2015). A five-factor asset pricing model.
#'   *Journal of Financial Economics*, 116(1), 1-22.
#'   \doi{10.1016/j.jfineco.2014.10.010}
#'
#' @family WRDS functions
#' @export
#'
#' @examples
#' \dontrun{
#' download_data_wrds_compustat("compustat_annual", "2020-01-01", "2020-12-31")
#' download_data_wrds_compustat(
#'   "compustat_quarterly",
#'   "2020-01-01",
#'   "2020-12-31"
#' )
#'
#'   # Add additional columns
#' download_data_wrds_compustat(
#'   "compustat_annual",
#'   additional_columns = c("aodo", "aldo")
#' )
#' }
download_data_wrds_compustat <- function(
  dataset = NULL,
  start_date = NULL,
  end_date = NULL,
  type = deprecated(),
  additional_columns = NULL,
  only_us = FALSE
) {
  # Handle explicit type argument
  if (lifecycle::is_present(type)) {
    lifecycle::deprecate_warn(
      when = "0.5.0",
      what = "download_data_wrds_compustat(type)",
      details = "Use the `dataset` argument instead."
    )
    dataset <- sub("^wrds_", "", type)
  }

  # Handle legacy type passed as dataset argument
  if (!is.null(dataset) && is_legacy_type_wrds(dataset)) {
    lifecycle::deprecate_warn(
      when = "0.5.0",
      what = "download_data_wrds_compustat(type)",
      details = paste0(
        "The `type` argument is deprecated. ",
        "Use `dataset` instead (e.g., 'compustat_annual' instead of",
        "'wrds_compustat_annual')."
      )
    )
    dataset <- sub("^wrds_", "", dataset)
  }

  if (is.null(dataset)) {
    cli::cli_abort("Argument {.arg dataset} is required.")
  }

  check_supported_dataset_wrds_compustat(dataset)

  dates <- validate_dates(start_date, end_date, use_default_range = TRUE)
  start_date <- dates$start_date
  end_date <- dates$end_date

  con <- get_wrds_connection()

  if (dataset == "compustat_annual") {
    funda_db <- tbl(con, I("comp.funda"))

    has_pi <- "pi" %in% additional_columns
    additional_columns_safe <- setdiff(additional_columns, "pi")

    compustat <- funda_db |>
      filter(
        indfmt == "INDL" &
          datafmt == "STD" &
          consol == "C" &
          between(datadate, start_date, end_date)
      ) |>
      select(
        gvkey,
        datadate,
        seq,
        ceq,
        at,
        lt,
        txditc,
        txdb,
        itcb,
        pstkrv,
        pstkl,
        pstk,
        capx,
        oancf,
        sale,
        cogs,
        xint,
        xsga,
        ib,
        curcd,
        all_of(additional_columns_safe)
      ) |>
      # nolint start
      {
        \(x) if (has_pi) mutate(x, pi = sql('"pi"')) else x
      }() |>
      # nolint end
      collect()

    disconnect_connection(con)

    compustat <- compustat |>
      mutate(
        be = coalesce(seq, ceq + pstk, at - lt) +
          coalesce(txditc, txdb + itcb, 0) -
          coalesce(pstkrv, pstkl, pstk, 0),
        op = (sale -
          coalesce(cogs, 0) -
          coalesce(xsga, 0) -
          coalesce(xint, 0)) /
          be
      )

    compustat <- compustat |>
      mutate(year = year(datadate)) |>
      group_by(gvkey, year) |>
      filter(datadate == max(datadate)) |>
      ungroup() |>
      mutate(date = floor_date(datadate, "month"))

    if (isTRUE(only_us)) {
      compustat <- compustat |>
        filter(.data$curcd == "USD")
    }

    processed_data <- compustat |>
      left_join(
        compustat |>
          select(gvkey, year, at_lag = at) |>
          mutate(year = year + 1),
        join_by(gvkey, year)
      ) |>
      mutate(
        inv = at / at_lag - 1,
        inv = if_else(at_lag <= 0, NA, inv)
      ) |>
      select(gvkey, date, datadate, everything(), -year)
  } else if (dataset == "compustat_quarterly") {
    fundq_db <- tbl(con, I("comp.fundq"))

    compustat <- fundq_db |>
      filter(
        indfmt == "INDL" &
          datafmt == "STD" &
          consol == "C" &
          between(datadate, start_date, end_date)
      ) |>
      select(
        gvkey,
        datadate,
        rdq,
        fqtr,
        fyearq,
        atq,
        ceqq,
        curcdq,
        all_of(additional_columns)
      ) |>
      collect()

    disconnect_connection(con)

    compustat <- compustat |>
      tidyr::drop_na(gvkey, datadate, fyearq, fqtr) |>
      mutate(date = floor_date(datadate, "month")) |>
      group_by(gvkey, fyearq, fqtr) |>
      filter(datadate == max(datadate)) |>
      slice_head(n = 1) |>
      ungroup() |>
      group_by(gvkey, date) |>
      arrange(gvkey, date, rdq) |>
      slice_head(n = 1) |>
      ungroup() |>
      filter(if_else(is.na(rdq), TRUE, date < rdq))

    if (isTRUE(only_us)) {
      compustat <- compustat |>
        filter(.data$curcdq == "USD")
    }

    processed_data <- compustat |>
      select(gvkey, date, datadate, atq, ceqq, all_of(additional_columns))
  } else {
    cli::cli_abort("Unsupported Compustat dataset: {.val {dataset}}")
  }

  processed_data
}

#' Check if WRDS Compustat dataset is supported
#' @noRd
check_supported_dataset_wrds_compustat <- function(dataset) {
  supported_datasets <- c("compustat_annual", "compustat_quarterly")

  if (!dataset %in% supported_datasets) {
    cli::cli_abort(c(
      "Unsupported Compustat dataset: {.val {dataset}}",
      "i" = "Supported datasets: {.val {supported_datasets}}"
    ))
  }
}

#' Download Filtered FISD Data from WRDS
#'
#' Establishes a connection to the WRDS database to download a filtered subset
#' of the FISD (Fixed Income Securities Database). The function filters the
#' `fisd_mergedissue` and `fisd_mergedissuer` tables based on several criteria
#' related to the securities, such as security level, bond type, coupon type,
#' and others, focusing on specific attributes that denote the nature of the
#' securities. It finally returns a data frame with selected fields from the
#' `fisd_mergedissue` table after joining it with issuer information from the
#' `fisd_mergedissuer` table for issuers domiciled in the USA.
#'
#' @param additional_columns Additional columns from the FISD table
#'   as a character vector.
#'
#' @returns A data frame containing a subset of FISD data with fields related
#'   to the bond's characteristics and issuer information. This includes
#'   complete CUSIP, maturity date, offering amount, offering date, dated date,
#'   interest frequency, coupon, last interest date, issue ID, issuer ID, and
#'   SIC code of the issuer.
#'
#' @family WRDS functions
#' @export
#'
#' @examples
#' \dontrun{
#' fisd <- download_data_wrds_fisd()
#' fisd_extended <- download_data_wrds_fisd(
#'   additional_columns = c("asset_backed", "defeased")
#' )
#' }
download_data_wrds_fisd <- function(additional_columns = NULL) {
  con <- get_wrds_connection()

  fisd_mergedissue_db <- tbl(con, I("fisd.fisd_mergedissue"))

  fisd <- fisd_mergedissue_db |>
    filter(
      security_level == "SEN",
      slob == "N" | is.na(slob),
      is.na(security_pledge),
      asset_backed == "N" | is.na(asset_backed),
      defeased == "N" | is.na(defeased),
      is.na(defeased_date),
      bond_type %in% c("CDEB", "CMTN", "CMTZ", "CZ", "USBN"),
      pay_in_kind != "Y" | is.na(pay_in_kind),
      is.na(pay_in_kind_exp_date),
      yankee == "N" | is.na(yankee),
      canadian == "N" | is.na(canadian),
      foreign_currency == "N",
      coupon_type %in% c("F", "Z"),
      is.na(fix_frequency),
      coupon_change_indicator == "N",
      interest_frequency %in% c("0", "1", "2", "4", "12"),
      rule_144a == "N",
      private_placement == "N" | is.na(private_placement),
      defaulted == "N",
      is.na(filing_date),
      is.na(settlement),
      convertible == "N",
      is.na(exchange),
      putable == "N" | is.na(putable),
      unit_deal == "N" | is.na(unit_deal),
      exchangeable == "N" | is.na(exchangeable),
      perpetual == "N",
      preferred_security == "N" | is.na(preferred_security)
    ) |>
    select(
      complete_cusip,
      maturity,
      offering_amt,
      offering_date,
      dated_date,
      interest_frequency,
      coupon,
      last_interest_date,
      issue_id,
      issuer_id,
      all_of(additional_columns)
    ) |>
    collect()

  fisd_mergedissuer_db <- tbl(con, I("fisd.fisd_mergedissuer"))

  fisd_issuer <- fisd_mergedissuer_db |>
    filter(country_domicile == "USA") |>
    select(issuer_id, sic_code) |>
    collect()

  disconnect_connection(con)

  fisd <- fisd |>
    inner_join(fisd_issuer, join_by(issuer_id))

  fisd
}

#' Trim a Numeric Vector
#'
#' Removes the values in a numeric vector that are beyond the specified
#' quantiles, effectively trimming the distribution based on the `cut`
#' parameter. This process reduces the vector's length by excluding extreme
#' values from both tails of the distribution.
#'
#' @param x A numeric vector to be trimmed.
#' @param cut The proportion of data to be trimmed from both ends of the
#'   distribution. For example, a `cut` of 0.05 will remove the lowest and
#'   highest 5% of the data. Must be in \[0, 0.5\].
#'
#' @returns A numeric vector with the extreme values removed.
#'
#' @family utility functions
#' @export
#'
#' @examples
#' set.seed(123)
#' data <- rnorm(100)
#' trimmed_data <- trim(x = data, cut = 0.05)
#'
trim <- function(x, cut) {
  if (cut < 0 || cut > 0.5) {
    cli::cli_abort("{.arg cut} must be inside [0, 0.5].")
  }

  lb <- quantile(x, cut, na.rm = TRUE)
  up <- quantile(x, 1 - cut, na.rm = TRUE)
  x <- replace(x, x > up, NA_real_)
  x <- replace(x, x < lb, NA_real_)
  x
}

test_that("trim correctly replaces values outside quantiles with NA", {
  set.seed(123)
  x <- rnorm(100)
  cut <- 0.05

  # Trim the data
  trimmed_x <- trim(x, cut)

  # Calculate lower and upper bounds without names for comparison
  lb <- unname(quantile(x, cut, na.rm = TRUE))
  up <- unname(quantile(x, 1 - cut, na.rm = TRUE))

  # Check if values outside the quantiles are replaced with NA
  expect_true(all(is.na(trimmed_x[x < lb]) & is.na(trimmed_x[x > up])))

  # Check if values within the quantiles are unchanged
  expect_true(all(trimmed_x[x >= lb & x <= up] == x[x >= lb & x <= up]))
})

test_that("trim handles NA values correctly", {
  set.seed(123)
  x <- c(NA, rnorm(100), NA)
  cut <- 0.05

  # Trim the data, expecting NA values to remain unchanged
  trimmed_x <- trim(x, cut)

  # Check that NA values are still present
  original_na_positions <- is.na(x)
  trimmed_na_positions <- is.na(trimmed_x)
  expect_true(all(trimmed_na_positions[original_na_positions] == TRUE))

  # Check that non-NA values outside the quantiles are replaced with NA
  lb <- unname(quantile(x, cut, na.rm = TRUE, names = FALSE))
  up <- unname(quantile(x, 1 - cut, na.rm = TRUE, names = FALSE))
  expect_true(all(is.na(trimmed_x[x < lb]) & is.na(trimmed_x[x > up])))
})

test_that("trim handles edge cases", {
  # Test with an empty vector
  expect_equal(trim(numeric(0), 0.1), numeric(0))

  # Test with a vector of identical values, none should be replaced since they
  # all are at the same quantile
  x <- rep(1, 10)
  expect_equal(trim(x, 0.1), x)
})

#' Add Lagged Columns via Join
#'
#' Appends lagged versions of specified columns to a data frame using a
#' join-based approach.
#'
#' When `lag == max_lag` (the default), an equi-join is used: source dates
#' are shifted forward by `lag` and matched exactly. When `lag < max_lag`,
#' an inequality join is used: for each row, the most recent source value
#' within the window `[date - max_lag, date - lag]` is selected.
#'
#' The combination of `by` and date columns must be unique in `data`. If `by`
#' is `NULL`, dates alone must be unique.
#'
#' @param data A data frame containing the variables to lag.
#' @param cols A character vector specifying the names of the columns to be
#'   lagged. Each column produces a new column suffixed with `_lag`.
#' @param lag An integer or a `lubridate::periods()` object, e.g.,
#'   `months(1)`, specifying the minimum lag (inclusive) to apply.
#' @param max_lag An integer or a `lubridate::periods()` object specifying
#'   the maximum lag (inclusive) to apply. Defaults to `lag` (exact lag).
#' @param by An optional character vector specifying grouping columns
#'   (e.g., a stock identifier). Lagged values are matched within groups.
#'   Defaults to `NULL`.
#' @param drop_na A logical value. If `TRUE`, `NA` values in the source
#'   columns are excluded before matching, so the lookup skips over missing
#'   observations. Applied independently per column. Defaults to `FALSE`.
#' @param ff_adjustment A logical value. If `TRUE`, only the last observation
#'   per year (within each group defined by `by`) is retained as a source for
#'   lagged values, following Fama-French conventions for annual accounting
#'   data. Defaults to `FALSE`.
#' @param data_options A list of class `tidyfinance_data_options` (created via
#'   [data_options()]) specifying column name mappings. The `date` element is
#'   used to specify the date column. Uses [data_options()] default if `NULL`:
#'   `"date" = "date"`.
#'
#' @returns A data frame with the same rows as `data` and new columns appended,
#'   each suffixed with `_lag`. Unmatched rows receive `NA` in the lagged
#'   columns.
#'
#' @family rolling and lagging functions
#' @export
#'
#' @examples
#' set.seed(42)
#' data <- tibble::tibble(
#'   permno = rep(1:2, each = 10),
#'   date = rep(
#'     seq.Date(as.Date("2023-01-01"), by = "month", length.out = 10),
#'     2
#'   ),
#'   size = runif(20, 100, 200),
#'   bm = runif(20, 0.5, 1.5)
#' )
#'
#' # Exact lag: each row gets the value from exactly 2 months earlier
#' add_lagged_columns(
#'   data,
#'   cols = c("size", "bm"),
#'   lag = months(2),
#'   by = "permno"
#' )
#'
#' # Window lag: each row gets the most recent value from 2 to 4 months earlier
#' add_lagged_columns(
#'   data,
#'   cols = "size",
#'   lag = months(2),
#'   max_lag = months(4),
#'   by = "permno"
#' )
#'
add_lagged_columns <- function(
  data,
  cols,
  lag,
  max_lag = lag,
  by = NULL,
  drop_na = FALSE,
  ff_adjustment = FALSE,
  data_options = NULL
) {
  if (is.null(data_options)) {
    data_options <- data_options()
  }

  date_col <- data_options$date

  if (!date_col %in% names(data)) {
    cli::cli_abort(
      "{.arg data} must contain the date column {.field {date_col}}."
    )
  }

  if (lag < 0 || max_lag < lag) {
    cli::cli_abort(
      paste(
        "{.arg lag} and {.arg max_lag} must be non-negative",
        "and {.arg max_lag} must be >= {.arg lag}."
      )
    )
  }

  missing_cols <- setdiff(cols, names(data))
  if (length(missing_cols) > 0) {
    cli::cli_abort(
      "{.arg data} is missing column{?s}: {.field {missing_cols}}."
    )
  }

  if (!is.null(by)) {
    missing_by <- setdiff(by, names(data))
    if (length(missing_by) > 0) {
      cli::cli_abort(
        "{.arg data} is missing grouping column{?s}: {.field {missing_by}}."
      )
    }
  }

  join_cols <- c(by, date_col)

  if (anyDuplicated(data[join_cols])) {
    cli::cli_abort(
      paste(
        "The combination of {.arg by}",
        "and date columns must be unique in {.arg data}."
      )
    )
  }

  exact_lag <- (lag == max_lag)

  if (!exact_lag) {
    data <- data |>
      check_new_col(c(".upper", ".lower")) |>
      dplyr::mutate(
        .upper = .data[[date_col]] - lag,
        .lower = .data[[date_col]] - max_lag
      )
  }

  for (col in cols) {
    lagged <- data[c(join_cols, col)]

    if (drop_na) {
      lagged <- lagged[!is.na(lagged[[col]]), ]
    }

    if (ff_adjustment) {
      yr <- lubridate::year(lagged[[date_col]])
      grp <- if (!is.null(by)) interaction(lagged[by], yr) else yr
      max_dates <- stats::ave(as.numeric(lagged[[date_col]]), grp, FUN = max)
      lagged <- lagged[as.numeric(lagged[[date_col]]) == max_dates, ]
    }

    lag_col_name <- paste0(col, "_lag")

    if (exact_lag) {
      lagged[[date_col]] <- lagged[[date_col]] + lag
      names(lagged)[names(lagged) == col] <- lag_col_name

      data <- data |>
        check_new_col(lag_col_name) |>
        dplyr::left_join(lagged, by = join_cols)
    } else {
      names(lagged)[names(lagged) == date_col] <- ".src_date"
      names(lagged)[names(lagged) == col] <- lag_col_name

      data <- data |>
        check_new_col(lag_col_name) |>
        dplyr::left_join(
          lagged,
          by = dplyr::join_by(!!!rlang::syms(by), closest(.upper >= .src_date))
        ) |>
        dplyr::mutate(
          !!lag_col_name := dplyr::if_else(
            !is.na(.data[[".src_date"]]) &
              .data[[".src_date"]] >= .data[[".lower"]],
            .data[[lag_col_name]],
            .data[[lag_col_name]][NA_integer_]
          )
        ) |>
        dplyr::select(-".src_date")
    }
  }

  if (!exact_lag) {
    data[[".upper"]] <- NULL
    data[[".lower"]] <- NULL
  }

  data
}

#' Download and Process Open Source Asset Pricing Data
#'
#' Downloads the data from
#' [Open Source Asset Pricing](https://www.openassetpricing.com/data/)
#' from Google Sheets using a specified sheet ID, processes the data
#' by converting column names to snake_case, and optionally filters
#' the data based on a provided date range.
#'
#' @param start_date Optional. A character string or Date object in
#'   "YYYY-MM-DD" format specifying the start date for the data. If
#'   not provided, the full dataset is returned.
#' @param end_date Optional. A character string or Date object in
#'   "YYYY-MM-DD" format specifying the end date for the data. If not
#'   provided, the full dataset is returned.
#' @param sheet_id A character string representing the Google Sheet ID
#'   from which to download the data. Default is
#'   `"1JyhcF5PRKHcputlioxlu5j5GyLo4JYyY"`.
#'
#' @returns A tibble containing the processed data. The column names
#'   are converted to snake_case, and the data is filtered by the
#'   specified date range if `start_date` and `end_date` are provided.
#'
#' @family download functions
#' @export
#'
#' @examples
#' \donttest{
#'   osap <- download_data_osap(
#'     start_date = "2020-01-01", end_date = "2020-06-30"
#'   )
#' }
download_data_osap <- function(
  start_date = NULL,
  end_date = NULL,
  sheet_id = "1JyhcF5PRKHcputlioxlu5j5GyLo4JYyY"
) {
  dates <- validate_dates(start_date, end_date)
  start_date <- dates$start_date
  end_date <- dates$end_date

  url <- paste0("https://drive.google.com/uc?export=download&id=", sheet_id)

  raw_data <- handle_download_error(
    function() {
      suppressWarnings(
        as_tibble(read.csv(url))
      )
    },
    fallback = tibble()
  )

  if (nrow(raw_data) == 0) {
    cli::cli_inform("Returning an empty data set due to download failure.")
    return(raw_data)
  }

  processed_data <- raw_data |>
    mutate(date = ymd(date))

  colnames(processed_data) <- to_snake_case(colnames(processed_data))

  if (!is.null(start_date) && !is.null(end_date)) {
    processed_data <- processed_data |>
      filter(between(date, start_date, end_date))
  }

  processed_data
}

to_snake_case <- function(x) {
  x <- gsub("([a-z])([A-Z])", "\\1_\\2", x)
  x <- gsub("[^[:alnum:]_]+", "_", x)
  x <- tolower(x)
  x <- gsub("_+", "_", x)
  x <- gsub("^_|_$", "", x)
  x
}

# Set global variables ------------------------------------------------
utils::globalVariables(
  c(
    "date",
    "rp_div",
    "dp",
    "dy",
    "ep",
    "de",
    "svar",
    "bm",
    "b.m",
    "ntis",
    "tbl",
    "lty",
    "ltr",
    "tms",
    "dfy",
    "infl",
    "AAA",
    "BAA",
    "D12",
    "DATE",
    "E12",
    "Index",
    "IndexDiv",
    "Rfree",
    "mkt-rf",
    "mkt_excess",
    "rf",
    "year",
    "month",
    "f",
    "mkt",
    "yyyymm",
    "yyyyq",
    "quarter",
    "yyyy",
    "logret",
    "altprc",
    "at",
    "at_lag",
    "be",
    "capx",
    "ceq",
    "cfacpr",
    "cogs",
    "consol",
    "datadate",
    "datafmt",
    "dlret",
    "dlstcd",
    "dlstdt",
    "exchcd",
    "gvkey",
    "indfmt",
    "inv",
    "itcb",
    "linkdt",
    "linkenddt",
    "lpermno",
    "lt",
    "mktcap",
    "namedt",
    "nameendt",
    "oancf",
    "permno",
    "pstk",
    "pstkl",
    "pstkrv",
    "ret",
    "ret_adj",
    "ret_excess",
    "risk_free",
    "sale",
    "shrcd",
    "shrout",
    "siccd",
    "txdb",
    "txditc",
    "xint",
    "xsga",
    "ret_excess",
    "dlycaldt",
    "dlyret",
    "issuertype",
    "mthcaldt",
    "calculation_date",
    "mthprc",
    "mthret",
    "prc",
    "prc_adj",
    "primaryexch",
    "secinfoenddt",
    "secinfostartdt",
    "securitysubtype",
    "securitytype",
    "sharetype",
    "usincflg",
    "hml",
    "smb",
    "issuer_id",
    "sic_code",
    "country_domicile",
    "complete_cusip",
    "maturity",
    "offering_amt",
    "offering_date",
    "dated_date",
    "interest_frequency",
    "coupon",
    "last_interest_date",
    "issue_id",
    "security_level",
    "slob",
    "security_pledge",
    "asset_backed",
    "defeased",
    "defeased_date",
    "bond_type",
    "pay_in_kind",
    "pay_in_kind_exp_date",
    "yankee",
    "canadian",
    "foreign_currency",
    "coupon_type",
    "fix_frequency",
    "coupon_change_indicator",
    "interest_frequency",
    "rule_144a",
    "private_placement",
    "defaulted",
    "filing_date",
    "settlement",
    "convertible",
    "exchange",
    "putable",
    "unit_deal",
    "exchangeable",
    "perpetual",
    "preferred_security",
    "cusip_id",
    "trd_exctn_dt",
    "trd_exctn_tm",
    "rptd_pr",
    "entrd_vol_qt",
    "yld_pt",
    "rpt_side_cd",
    "cntra_mp_id",
    "asof_cd",
    "days_to_sttl_ct",
    "days_to_sttl_ct2",
    "msg_seq_nb",
    "orig_msg_seq_nb",
    "pr_trd_dt",
    "spcl_trd_fl",
    "stlmnt_dt",
    "trc_st",
    "trd_rpt_dt",
    "trd_rpt_tm",
    "wis_fl",
    "conditionaltype",
    "tradingstatusflg",
    "log_d12",
    "log_e12",
    "rdq",
    "fqtr",
    "fyearq",
    "atq",
    "ceqq",
    "beta_date",
    "estimates",
    "name",
    "risk_premium",
    "row_check",
    "standard_error",
    "t_statistic",
    "value",
    "VALUE",
    "vol",
    "vol_adj",
    "bottom",
    "closest",
    "long_short_return",
    "lower_bound",
    "mktcap_lag",
    "portfolio",
    "portfolio_main",
    "portfolio_secondary",
    "ret_excess_ew",
    "ret_excess_vw",
    "ret_excess_vw_capped",
    "ret_measure",
    "top",
    "upper_bound",
    "Anlageklasse",
    "Asset.Class",
    "symbol",
    "country",
    "industry",
    "ipoyear",
    "sector",
    "..date",
    "..id",
    "..lower_bound",
    "..upper_bound",
    ".date",
    ".lower",
    ".src_date",
    ".upper",
    "observation_date",
    "curcd",
    "curcdq",
    "dlyfacprc",
    "dlyprc",
    "dlyvol",
    "sorting_variable",
    "sorting_variable_lag",
    "sorting_method",
    "n_portfolios_main",
    "url",
    "path",
    "size",
    "id",
    "ret_3m",
    "ret_4wk",
    "listing_age",
    "first_crsp_date",
    ".size_cutoff",
    "ib",
    "n_firms"
  )
)

#' Create Filter Options
#'
#' Creates a list of filter options of class `tidyfinance_filter_options` used
#' for sample construction in TidyFinance-related functions. These options
#' control which observations are retained before portfolio sorting.
#'
#' @param exclude_financials A logical indicating whether to exclude financial
#'   firms (SIC codes 6000–6799). Defaults to `FALSE`.
#' @param exclude_utilities A logical indicating whether to exclude utility
#'   firms (SIC codes 4900–4999). Defaults to `FALSE`.
#' @param min_stock_price A single positive numeric specifying the minimum
#'   stock price required to include an observation. `NULL` (the default)
#'   applies no price filter.
#' @param min_size_quantile A single numeric strictly between 0 and 1
#'   specifying the minimum cross-sectional size quantile (based on lagged
#'   market cap) required to include an observation. `NULL` (the default)
#'   applies no size quantile filter. The cutoff is computed from NYSE
#'   stocks only. This requires an `exchange` column in the data (as
#'   mapped via [data_options()]); an error is raised if it is missing.
#' @param min_listing_age A single non-negative integer or numeric specifying
#'   the minimum number of months a stock must have been listed in CRSP.
#'   `NULL` (the default) applies no listing age filter.
#' @param exclude_negative_book_equity A logical indicating whether to exclude
#'   observations with non-positive book equity. Defaults to `FALSE`.
#' @param exclude_negative_earnings A logical indicating whether to exclude
#'   observations with non-positive earnings. Defaults to `FALSE`.
#' @param ... Additional arguments to be included in the filter options list.
#'
#' @returns A list of class `tidyfinance_filter_options` containing the
#'   specified filter options.
#'
#' @family portfolio functions
#' @export
#'
#' @examples
#' filter_options(
#'   exclude_financials = TRUE,
#'   exclude_utilities = TRUE,
#'   min_stock_price = 1,
#'   min_listing_age = 12
#' )
#'
filter_options <- function(
  exclude_financials = FALSE,
  exclude_utilities = FALSE,
  min_stock_price = NULL,
  min_size_quantile = NULL,
  min_listing_age = NULL,
  exclude_negative_book_equity = FALSE,
  exclude_negative_earnings = FALSE,
  ...
) {
  # Error handling for exclude_financials
  if (
    !is.logical(exclude_financials) ||
      length(exclude_financials) != 1 ||
      is.na(exclude_financials)
  ) {
    cli::cli_abort("{.arg exclude_financials} must be a single logical.")
  }

  # Error handling for exclude_utilities
  if (
    !is.logical(exclude_utilities) ||
      length(exclude_utilities) != 1 ||
      is.na(exclude_utilities)
  ) {
    cli::cli_abort("{.arg exclude_utilities} must be a single logical.")
  }

  # Error handling for min_stock_price
  if (!is.null(min_stock_price)) {
    if (
      !is.numeric(min_stock_price) ||
        length(min_stock_price) != 1 ||
        is.na(min_stock_price) ||
        min_stock_price <= 0
    ) {
      cli::cli_abort(
        "{.arg min_stock_price} must be a single positive numeric."
      )
    }
  }

  # Error handling for min_size_quantile
  if (!is.null(min_size_quantile)) {
    if (
      !is.numeric(min_size_quantile) ||
        length(min_size_quantile) != 1 ||
        is.na(min_size_quantile) ||
        min_size_quantile <= 0 ||
        min_size_quantile >= 1
    ) {
      cli::cli_abort(
        paste0(
          "{.arg min_size_quantile} must be a single numeric ",
          "strictly between 0 and 1."
        )
      )
    }
  }

  # Error handling for min_listing_age
  if (!is.null(min_listing_age)) {
    if (
      !is.numeric(min_listing_age) ||
        length(min_listing_age) != 1 ||
        is.na(min_listing_age) ||
        min_listing_age < 0
    ) {
      cli::cli_abort(
        paste0(
          "{.arg min_listing_age} must be a single ",
          "non-negative integer or numeric."
        )
      )
    }
  }

  # Error handling for exclude_negative_book_equity
  if (
    !is.logical(exclude_negative_book_equity) ||
      length(exclude_negative_book_equity) != 1 ||
      is.na(exclude_negative_book_equity)
  ) {
    cli::cli_abort(
      "{.arg exclude_negative_book_equity} must be a single logical."
    )
  }

  # Error handling for exclude_negative_earnings
  if (
    !is.logical(exclude_negative_earnings) ||
      length(exclude_negative_earnings) != 1 ||
      is.na(exclude_negative_earnings)
  ) {
    cli::cli_abort("{.arg exclude_negative_earnings} must be a single logical.")
  }

  # Create the list structure with class attribute
  structure(
    list(
      "exclude_financials" = exclude_financials,
      "exclude_utilities" = exclude_utilities,
      "min_stock_price" = min_stock_price,
      "min_size_quantile" = min_size_quantile,
      "min_listing_age" = min_listing_age,
      "exclude_negative_book_equity" = exclude_negative_book_equity,
      "exclude_negative_earnings" = exclude_negative_earnings,
      ...
    ),
    class = "tidyfinance_filter_options"
  )
}

test_that("filter_options returns object of class tidyfinance_filter_options", {
  opts <- filter_options()
  expect_s3_class(opts, "tidyfinance_filter_options")
})

test_that("filter_options returns a list", {
  opts <- filter_options()
  expect_type(opts, "list")
})

test_that("filter_options has correct defaults", {
  opts <- filter_options()
  expect_false(opts$exclude_financials)
  expect_false(opts$exclude_utilities)
  expect_null(opts$min_stock_price)
  expect_null(opts$min_size_quantile)
  expect_null(opts$min_listing_age)
  expect_false(opts$exclude_negative_book_equity)
  expect_false(opts$exclude_negative_earnings)
})

test_that("filter_options accepts non-default logical flags", {
  opts <- filter_options(
    exclude_financials = TRUE,
    exclude_utilities = TRUE,
    exclude_negative_book_equity = TRUE,
    exclude_negative_earnings = TRUE
  )
  expect_true(opts$exclude_financials)
  expect_true(opts$exclude_utilities)
  expect_true(opts$exclude_negative_book_equity)
  expect_true(opts$exclude_negative_earnings)
})

test_that("filter_options accepts valid numeric thresholds", {
  opts <- filter_options(
    min_stock_price = 5,
    min_size_quantile = 0.2,
    min_listing_age = 12
  )
  expect_equal(opts$min_stock_price, 5)
  expect_equal(opts$min_size_quantile, 0.2)
  expect_equal(opts$min_listing_age, 12)
})

test_that("filter_options stores extra arguments via ...", {
  opts <- filter_options(custom_filter = "my_value")
  expect_equal(opts$custom_filter, "my_value")
})

test_that("filter_options errors for non-logical exclude_financials", {
  expect_error(filter_options(exclude_financials = "yes"), "exclude_financials")
  expect_error(filter_options(exclude_financials = 1), "exclude_financials")
  expect_error(filter_options(exclude_financials = NA), "exclude_financials")
})

test_that("filter_options errors for non-logical exclude_utilities", {
  expect_error(filter_options(exclude_utilities = "yes"), "exclude_utilities")
  expect_error(filter_options(exclude_utilities = NA), "exclude_utilities")
})

test_that("filter_options errors for non-positive min_stock_price", {
  expect_error(filter_options(min_stock_price = 0), "min_stock_price")
  expect_error(filter_options(min_stock_price = -1), "min_stock_price")
  expect_error(filter_options(min_stock_price = NA_real_), "min_stock_price")
  expect_error(filter_options(min_stock_price = "5"), "min_stock_price")
  expect_error(filter_options(min_stock_price = c(1, 2)), "min_stock_price")
})

test_that("filter_options errors for out-of-range min_size_quantile", {
  expect_error(filter_options(min_size_quantile = 0), "min_size_quantile")
  expect_error(filter_options(min_size_quantile = 1), "min_size_quantile")
  expect_error(filter_options(min_size_quantile = -0.1), "min_size_quantile")
  expect_error(filter_options(min_size_quantile = 1.1), "min_size_quantile")
  expect_error(
    filter_options(min_size_quantile = NA_real_),
    "min_size_quantile"
  )
})

test_that("filter_options errors for invalid min_listing_age", {
  expect_error(filter_options(min_listing_age = -1), "min_listing_age")
  expect_error(filter_options(min_listing_age = NA_real_), "min_listing_age")
  expect_error(filter_options(min_listing_age = "12"), "min_listing_age")
  expect_error(filter_options(min_listing_age = c(6, 12)), "min_listing_age")
})

test_that("filter_options accepts zero as valid min_listing_age", {
  expect_no_error(filter_options(min_listing_age = 0))
})

test_that(
  paste0(
    "filter_options errors for non-logical ",
    "exclude_negative_book_equity"
  ),
  {
    expect_error(
      filter_options(exclude_negative_book_equity = "TRUE"),
      "exclude_negative_book_equity"
    )
    expect_error(
      filter_options(exclude_negative_book_equity = NA),
      "exclude_negative_book_equity"
    )
    expect_error(
      filter_options(exclude_negative_book_equity = 1),
      "exclude_negative_book_equity"
    )
  }
)

test_that("filter_options errors for non-logical exclude_negative_earnings", {
  expect_error(
    filter_options(exclude_negative_earnings = "TRUE"),
    "exclude_negative_earnings"
  )
  expect_error(
    filter_options(exclude_negative_earnings = NA),
    "exclude_negative_earnings"
  )
  expect_error(
    filter_options(exclude_negative_earnings = 1),
    "exclude_negative_earnings"
  )
})

#' Download and Process Data from FRED
#'
#' Downloads a specified data series from the Federal Reserve Economic Data
#' (FRED) website, processes the data, and returns it as a tibble.
#'
#' @details
#' Constructs the URL based on the provided FRED series ID, performs an HTTP
#' GET request to download the data in CSV format, and processes it to a tidy
#' tibble format. The resulting tibble includes the date, value, and the series
#' ID.
#'
#' This approach is inspired by `quantmod::getSymbolsFRED()` which uses a
#' different wrapper around the same FRED download data site. If you want to
#' systematically download FRED data via API, please consider using the `fredr`
#' package.
#'
#' @param series A character vector specifying the FRED series ID(s) to
#'   download.
#' @param start_date Optional. A character string or Date object in
#'   "YYYY-MM-DD" format specifying the start date for the data. If not
#'   provided, the full dataset is returned.
#' @param end_date Optional. A character string or Date object in
#'   "YYYY-MM-DD" format specifying the end date for the data. If not
#'   provided, the full dataset is returned.
#'
#' @returns A tibble containing the processed data with three columns:
#' \describe{
#'   \item{date}{The date corresponding to the data point.}
#'   \item{value}{The value of the data series at that date.}
#'   \item{series}{The FRED series ID corresponding to the data.}
#' }
#'
#' @family download functions
#' @export
#'
#' @examples
#' \donttest{
#'   download_data_fred("CPIAUCNS")
#'   download_data_fred(c("GDP", "CPIAUCNS"), "2010-01-01", "2010-12-31")
#' }
#'
download_data_fred <- function(series, start_date = NULL, end_date = NULL) {
  dates <- validate_dates(start_date, end_date)
  start_date <- dates$start_date
  end_date <- dates$end_date

  fred_processed <- list()

  cli::cli_progress_bar(
    "Downloading series",
    total = length(series),
    clear = TRUE
  )
  for (j in seq_along(series)) {
    url <- paste0(
      "https://fred.stlouisfed.org/graph/fredgraph.csv?id=",
      series[j]
    )

    response <- handle_download_error(
      function() {
        httr2::request(url) |>
          httr2::req_error(is_error = \(resp) FALSE) |>
          httr2::req_perform()
      },
      fallback = NULL
    )

    if (!is.null(response)) {
      if (response$status_code == 200) {
        fred_raw <- suppressWarnings(httr2::resp_body_string(response)) |>
          textConnection() |>
          read.csv() |>
          as_tibble()

        fred_processed[[j]] <- fred_raw |>
          mutate(
            date = as.Date(.data$observation_date),
            value = as.numeric(.data[[series[j]]]),
            series = series[j],
            .keep = "none"
          )
      } else {
        cli::cli_warn(
          paste(
            "Failed to retrieve data for series {series[j]}",
            "with status code {response$status_code}."
          )
        )
        fred_processed[[j]] <- tibble(
          date = Date(),
          value = numeric(),
          series = character()
        )
      }
    } else {
      fred_processed[[j]] <- tibble(
        date = Date(),
        value = numeric(),
        series = character()
      )
    }

    cli::cli_progress_update()
  }

  fred_processed <- bind_rows(fred_processed)

  if (!is.null(start_date) && !is.null(end_date)) {
    fred_processed <- fred_processed |>
      filter(date >= start_date & date <= end_date)
  }

  fred_processed
}

test_that(
  paste(
    "download_data_fred returns full data set when no date range is provided"
  ),
  {
    skip_if_offline()
    skip_on_cran()
    expect_message(
      data <- download_data_fred("CPIAUCNS"),
      regex = "Returning the full data set."
    )
    expect_s3_class(data, "tbl_df")
    expect_true(all(c("date", "value", "series") %in% colnames(data)))
    expect_equal(unique(data$series), "CPIAUCNS")
  }
)

test_that("download_data_fred handles invalid series ID", {
  skip_if_offline()
  skip_on_cran()
  expect_warning(
    download_data_fred(
      "INVALID_SERIES",
      start_date = "2020-12-31",
      end_date = "2021-01-01"
    ),
    regexp = paste0(
      "Failed to retrieve data for series INVALID_SERIES ",
      "with status code [0-9]{3}\\."
    )
  )
})

test_that("download_data_fred handles start_date after end_date", {
  expect_error(
    download_data_fred(
      "CPIAUCNS",
      start_date = "2021-12-31",
      end_date = "2020-01-01"
    ),
    regexp = "`start_date` cannot be after `end_date`"
  )
})

#' Compute a Rolling Value by Period
#'
#' @description `r lifecycle::badge('experimental')`
#'
#' Applies an arbitrary summary function over rolling time-period windows.
#' Each window spans `periods` units of `period` (e.g., 12 months). Before
#' calling `.f`, rows with any missing values are dropped from the window;
#' if fewer than `min_obs` rows remain, the result is `NA_real_` instead.
#'
#' @param data A data frame with a date column of class `Date`, named according
#'   to `data_options$date` (default `"date"`).
#' @param .f A function applied to each window. Receives a data-frame slice
#'   (complete cases only) and must return a single scalar value.
#' @param period A string specifying the period for rolling windows
#'   (e.g., `"month"`, `"quarter"`, `"year"`).
#' @param periods Number of periods to include in the rolling window.
#' @param min_obs Minimum number of non-missing rows required per window.
#'   Defaults to `periods`.
#' @param data_options A list of class `tidyfinance_data_options` (created via
#'   [data_options()]) specifying column name mappings. The `date` element is
#'   used to specify the date column. Uses [data_options()] default if `NULL`:
#'   `"date" = "date"`.
#'
#' @returns A numeric vector aligned with the rows of `data`.
#'
#' @family rolling and lagging functions
#' @export
#'
#' @examples
#' library(dplyr)
#'
#' # Rolling standard deviation
#' set.seed(42)
#' df <- tibble(
#'   date = seq.Date(
#'     from = as.Date("2020-01-01"),
#'     by = "month",
#'     length.out = 24
#'   ),
#'   value = rnorm(24)
#' )
#'
#' df |>
#'   mutate(
#'     rolling_sd = compute_rolling_value(
#'       pick(everything()),
#'       .f = ~ sd(.x$value, na.rm = TRUE),
#'       period = "month",
#'       periods = 4,
#'       min_obs = 2
#'     )
#'   )
#'
#' # Rolling last residual from a regression
#' set.seed(42)
#' df_reg <- tibble(
#'   date = seq.Date(
#'     from = as.Date("2020-01-01"),
#'     by = "month",
#'     length.out = 60
#'   ),
#'   ret_excess = rnorm(60, 0, 0.05),
#'   mkt_excess = rnorm(60, 0, 0.04),
#'   smb = rnorm(60, 0, 0.03),
#'   hml = rnorm(60, 0, 0.03)
#' )
#'
#' df_reg |>
#'   mutate(
#'     residual = compute_rolling_value(
#'       pick(everything()),
#'       .f = \(x) {
#'         last(lm(ret_excess ~ mkt_excess + smb + hml, data = x)$residuals)
#'       },
#'       period = "month",
#'       periods = 24,
#'       min_obs = 12
#'     )
#'   )
#'
#' # Rolling cumulative-return-to-SD ratio
#' set.seed(42)
#' df_resid <- tibble(
#'   date = seq.Date(
#'     from = as.Date("2020-01-01"),
#'     by = "month",
#'     length.out = 24
#'   ),
#'   int_roll_residual = rnorm(24, 0, 0.02)
#' )
#'
#' df_resid |>
#'   mutate(
#'     return_to_sd = compute_rolling_value(
#'       pick(everything()),
#'       .f = ~ (prod(1 + .x$int_roll_residual) - 1) / sd(.x$int_roll_residual),
#'       period = "month",
#'       periods = 12,
#'       min_obs = 12
#'     )
#'   )
#'
compute_rolling_value <- function(
  data,
  .f,
  period = "month",
  periods = 12,
  min_obs = periods,
  data_options = NULL
) {
  if (is.null(data_options)) {
    data_options <- data_options()
  }
  date_col <- data_options$date

  if (!is.character(date_col) || length(date_col) != 1 || is.na(date_col)) {
    cli::cli_abort(
      paste(
        "{.field date} in {.arg data_options} must be a single non-missing",
        "string, not {.obj_type_friendly {date_col}}."
      )
    )
  }

  if (!date_col %in% names(data)) {
    cli::cli_abort("{.arg data} must contain a {.field {date_col}} column.")
  }
  if (!inherits(data[[date_col]], "Date")) {
    cli::cli_abort(
      paste(
        "The {.field {date_col}} column must be of class {.cls Date},",
        "not {.cls {class(data[[date_col]])}}."
      )
    )
  }
  if (!is.character(period) || length(period) != 1) {
    cli::cli_abort(
      paste(
        "{.arg period} must be a single string,",
        "not {.obj_type_friendly {period}}."
      )
    )
  }

  .f <- rlang::as_function(.f)

  slider::slide_period_vec(
    .x = data,
    .i = data[[date_col]],
    .period = period,
    .f = function(.x) {
      complete <- tidyr::drop_na(.x)
      if (nrow(complete) < min_obs) {
        NA_real_
      } else {
        .f(complete)
      }
    },
    .before = periods - 1,
    .complete = FALSE
  )
}

#' Get a Random User Agent String
#'
#' Selects and returns a random user agent string from a predefined list
#' covering common browsers and operating systems (Windows, macOS, Linux,
#' Android, iOS; Chrome, Firefox, Safari, Edge).
#'
#' @returns A length-one character vector containing the randomly selected
#'   user agent string.
#'
#' @keywords internal
#' @noRd
get_random_user_agent <- function() {
  # nolint start
  user_agents <- c(
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/122.0.6261.111 Safari/537.36",
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 13_6_3) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/123.0.6312.86 Safari/537.36",
    "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/121.0.6167.185 Safari/537.36",
    "Mozilla/5.0 (iPhone; CPU iPhone OS 17_4 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/17.0 Mobile/15E148 Safari/604.1",
    "Mozilla/5.0 (iPad; CPU OS 17_4 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/17.0 Mobile/15E148 Safari/604.1",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:124.0) Gecko/20100101 Firefox/124.0",
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 14_4_1) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/17.4 Safari/605.1.15",
    "Mozilla/5.0 (Linux; Android 14; Pixel 7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/123.0.6312.105 Mobile Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Edg/123.0.2420.65 Chrome/123.0.6312.86 Safari/537.36",
    "Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:123.0) Gecko/20100101 Firefox/123.0",
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 13_5_2) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/16.6 Safari/605.1.15",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/120.0.6099.224 Safari/537.36",
    "Mozilla/5.0 (Linux; Android 13; SM-S918U) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/122.0.6261.111 Mobile Safari/537.36",
    "Mozilla/5.0 (X11; Linux x86_64; rv:122.0) Gecko/20100101 Firefox/122.0",
    "Mozilla/5.0 (iPhone; CPU iPhone OS 16_7 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/16.0 Mobile/15E148 Safari/604.1",
    "Mozilla/5.0 (Windows NT 11.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/123.0.6312.106 Safari/537.36",
    "Mozilla/5.0 (Linux; Android 12; Pixel 6) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/121.0.6167.140 Mobile Safari/537.36",
    "Mozilla/5.0 (iPad; CPU OS 16_6 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/16.6 Mobile/15E148 Safari/604.1",
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 12_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/122.0.6261.128 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:125.0) Gecko/20100101 Firefox/125.0"
  )
  # nolint end

  user_agent <- user_agents[sample(length(user_agents), 1)]

  user_agent
}

#' Download Stock Data
#'
#' Downloads historical stock data from Yahoo Finance for given symbols and date
#' range.
#'
#' @param symbols A character vector of stock symbols to download data for. At
#'   least one symbol must be provided.
#' @param start_date Optional. A character string or Date object in "YYYY-MM-DD"
#'   format specifying the start date for the data. If not provided, a one-year
#'   subset of the dataset is returned (see [validate_dates()]).
#' @param end_date Optional. A character string or Date object in "YYYY-MM-DD"
#'   format specifying the end date for the data. If not provided, a one-year
#'   subset of the dataset is returned (see [validate_dates()]).
#'
#' @returns A tibble containing the downloaded stock data with columns: symbol,
#'   date, volume, open, low, high, close, and adjusted_close.
#'
#' @family download functions
#' @export
#'
#' @examples
#' \donttest{
#'   download_data_stock_prices(c("AAPL", "MSFT"))
#'   download_data_stock_prices("GOOGL", "2021-01-01", "2022-01-01" )
#' }
download_data_stock_prices <- function(
  symbols,
  start_date = NULL,
  end_date = NULL
) {
  if (!is.character(symbols) || anyNA(symbols)) {
    cli::cli_abort(
      paste(
        "{.arg symbols} must be character vector,",
        "not {.obj_type_friendly {symbols}}."
      )
    )
  }

  dates <- validate_dates(start_date, end_date, use_default_range = TRUE)
  start_date <- dates$start_date
  end_date <- dates$end_date

  start_timestamp <- as.integer(as.POSIXct(start_date, tz = "UTC"))
  end_timestamp <- as.integer(as.POSIXct(end_date, tz = "UTC"))

  processed_data <- list()

  cli::cli_progress_bar(
    "Downloading symbols",
    total = length(symbols),
    clear = TRUE
  )

  for (j in seq_along(symbols)) {
    url <- paste0(
      "https://query2.finance.yahoo.com/v8/finance/chart/",
      symbols[j],
      "?period1=",
      start_timestamp,
      "&period2=",
      end_timestamp,
      "&interval=1d"
    )

    response <- httr2::request(url) |>
      httr2::req_error(is_error = \(resp) FALSE) |>
      httr2::req_perform()

    if (response$status_code == 200) {
      raw_data <- httr2::resp_body_json(response)$chart$result

      replace_null_with_na <- function(x) {
        if (is.list(x)) {
          lapply(x, replace_null_with_na)
        } else if (is.null(x)) {
          NA
        } else {
          x
        }
      }

      ohlcv <- unlist(raw_data[[1]]$indicators$quote, recursive = FALSE)
      ohlcv <- replace_null_with_na(ohlcv)

      indicators <- raw_data[[1]]$indicators$adjclose
      indicators <- replace_null_with_na(indicators)

      processed_data_symbol <- tibble(
        "symbol" = symbols[j],
        "date" = as.Date(as.POSIXct(
          as.numeric(raw_data[[1]]$timestamp),
          origin = "1970-01-01"
        )),
        "volume" = as.numeric(unlist(ohlcv$volume)),
        "open" = as.numeric(unlist(ohlcv$open)),
        "low" = as.numeric(unlist(ohlcv$low)),
        "high" = as.numeric(unlist(ohlcv$high)),
        "close" = as.numeric(unlist(ohlcv$close)),
        "adjusted_close" = as.numeric(unlist(indicators))
      )

      processed_data[[j]] <- processed_data_symbol
    } else {
      error_message <- # nolint: object_usage_linter.
        httr2::resp_body_json(response)$chart$error
      cli::cli_warn(
        "Failed to retrieve data for symbol {symbols[j]} with",
        "status code {response$status_code} ({error_message$code}):",
        "{error_message$description}"
      )
    }
    cli::cli_progress_update()
  }

  bind_rows(processed_data)
}

#' Create WRDS Dummy Database
#'
#' Downloads the WRDS dummy database from the respective Tidy Finance GitHub
#' repository and saves it to the specified path. If the file already exists,
#' the user is prompted before it is replaced.
#'
#' @param path The file path where the SQLite database should be saved.
#' @param url The URL where the SQLite database is stored.
#'
#' @returns Invisible `NULL`. Side effect: downloads a file to the specified
#'   path.
#'
#' @family WRDS functions
#' @export
#'
#' @examples
#' path <- paste0(tempdir(), "/tidy_finance_r.sqlite")
#' create_wrds_dummy_database(path)
create_wrds_dummy_database <- function(
  path,
  url = paste0(
    "https://github.com/tidy-finance/website/tree/main/blog/",
    "tidy-finance-dummy-data/data/tidy_finance.sqlite"
  )
) {
  if (missing(path)) {
    cli::cli_abort(c(
      "Please provide a file path for the SQLite database.",
      i = "We recommend {.path data/tidy_finance_r.sqlite}"
    ))
  }

  if (file.exists(path)) {
    response <- readline(
      prompt = paste(
        "The database file already exists at this path.",
        "Do you want to replace it? (Y/n): "
      )
    )
    if (tolower(response) != "y") {
      cli::cli_inform("Operation aborted by the user.")
      return(invisible(NULL))
    }
  }

  handle_download_error(
    function() {
      suppressWarnings(utils::download.file(
        url = url,
        destfile = path,
        quiet = TRUE
      ))
    }
  )

  if (file.exists(path)) {
    cli::cli_inform("Downloaded WRDS dummy database to {.path {path}}.")
  }

  invisible(NULL)
}

test_that(
  paste("create_wrds_dummy_database downloads database to the specified path"),
  {
    skip_if_offline()
    skip_on_cran()
    temp_path <- tempfile(fileext = ".sqlite")
    create_wrds_dummy_database(path = temp_path) |>
      expect_message("Downloaded WRDS dummy database to")
    expect_true(file.exists(temp_path))
    unlink(temp_path)
  }
)

test_that("create_wrds_dummy_database handles broken url", {
  skip_if_offline()
  skip_on_cran()
  temp_path <- tempfile(fileext = ".sqlite")
  expect_message(
    create_wrds_dummy_database(path = temp_path, url = "broken"),
    "The resource may not be available, or the URL may have changed."
  )
  unlink(temp_path)
})

#' Compute Portfolio Returns
#'
#' Computes individual portfolio returns based on specified sorting variables
#' and sorting methods. The portfolios can be rebalanced every period or on an
#' annual frequency by specifying a rebalancing month, which is only applicable
#' at a monthly return frequency. The function supports univariate and
#' bivariate sorts, with the latter supporting dependent and independent
#' sorting methods.
#'
#' @details
#' The function checks for consistency in the provided arguments. For
#' univariate sorts, a single sorting variable and a corresponding number of
#' portfolios must be provided. For bivariate sorts, two sorting variables and
#' two corresponding numbers of portfolios (or percentiles) are required. The
#' sorting method determines how portfolios are assigned and how returns are
#' computed. The function handles missing and extreme values appropriately
#' based on the specified sorting method and rebalancing frequency.
#'
#' @param data A data frame containing the dataset for portfolio assignment
#'   and return computation. The panel data must include individual stock
#'   identifiers and the time point. It must contain columns for the sorting
#'   variables and excess returns. Additionally, lagged market capitalization
#'   is required for value-weighted returns (see parameter `data_options`).
#' @param sorting_variables A character vector specifying the column names in
#'   `data` to be used for sorting and determining portfolio assignments. For
#'   univariate sorts, provide a single variable. For bivariate sorts, provide
#'   two variables, where the first string refers to the main variable and the
#'   second string refers to the secondary ("control") variable.
#' @param sorting_method A string specifying the sorting method to be used.
#'   Possible values are:
#'   \itemize{
#'     \item `"univariate"`: For a single sorting variable.
#'     \item `"bivariate-dependent"`: For two sorting variables, where the
#'       main sort depends on the secondary variable.
#'     \item `"bivariate-independent"`: For two independent sorting variables.
#'   }
#'   For bivariate sorts, the portfolio returns are averaged over the
#'   controlling sorting variable (i.e., the second sorting variable), and only
#'   portfolio returns for the main sorting variable (given as the first
#'   element of `sorting_variables`) are returned.
#' @param rebalancing_month An integer between 1 and 12 specifying the month
#'   in which to form portfolios that are held constant for one year. For
#'   example, setting it to `7` creates portfolios in July that are held
#'   constant until June of the following year. The default `NULL` corresponds
#'   to periodic rebalancing.
#' @param breakpoint_options_main A named list of [breakpoint_options()]
#'   passed to `breakpoint_function_main` for the main sorting variable.
#' @param breakpoint_options_secondary An optional named list of
#'   [breakpoint_options()] passed to `breakpoint_function_secondary`.
#' @param breakpoint_function_main A function to compute the breakpoints for
#'   the main sorting variable. The default is set to [compute_breakpoints()].
#' @param breakpoint_function_secondary A function to compute the breakpoints
#'   for the secondary sorting variable. The default is set to
#'   [compute_breakpoints()].
#' @param min_portfolio_size An integer specifying the minimum number of
#'   firms required in the reported portfolio cross-section on a given date
#'   (default `1L`, i.e. at least one observation per reported portfolio).
#'   For both univariate and bivariate sorts the threshold is applied to the
#'   firm count per `(portfolio, date)` of the reported cross-section: for
#'   univariate sorts that is firms per portfolio-date; for bivariate sorts
#'   that is firms per main-portfolio-date summed across the secondary
#'   buckets. Cross-sections below the threshold have their returns set to
#'   `NA`. A typical value is `5L` (the Fama-French convention). Set to `0L`
#'   to deactivate the check entirely.
#' @param cap_weight A numeric value between 0 and 1 specifying the percentile
#'   at which market capitalization is capped per date when computing capped
#'   value-weighted excess returns (i.e., `ret_excess_vw_capped`). Defaults to
#'   `0.8`.
#' @param data_options A list of class `tidyfinance_data_options` (created via
#'   [data_options()]) specifying column name mappings. The `id` element is
#'   used to specify the entity (i.e., firm), `date` is used to specify the
#'   date column, `ret_excess` is used to specify the excess return column, and
#'   `mktcap_lag` is used to specify the market capitalization. Uses
#'   [data_options()] default if `NULL`: `"id" = "permno"`,
#'   `"date" = "date"`, `"ret_excess" = "ret_excess"`, and
#'   `"mktcap_lag" = "mktcap_lag"`.
#' @param quiet A logical value indicating whether to suppress informational
#'   messages about missing values in the output panel (default is `FALSE`).
#'
#' @note Ensure that `data` contains all required columns: the specified
#'   sorting variables and excess returns (see options and defaults set in
#'   `data_options`). The function will stop and throw an error if any required
#'   columns are missing.
#'
#' @returns A data frame with computed portfolio returns as a complete panel
#'   (all portfolio-date combinations), containing the following columns:
#'   \itemize{
#'     \item `portfolio`: The portfolio identifier.
#'     \item `date`: The date of the portfolio return.
#'     \item `ret_excess_vw`: The value-weighted excess return of the
#'       portfolio (only computed if `data` contains a lagged market
#'       capitalization column specified by `data_options()`, which defaults
#'       to `mktcap_lag`). `NA` if insufficient observations for that
#'       portfolio-date.
#'     \item `ret_excess_ew`: The equal-weighted excess return of the
#'       portfolio. `NA` if insufficient observations for that portfolio-date.
#'     \item `ret_excess_vw_capped`: The capped value-weighted excess return
#'       of the portfolio (only computed if `data` contains a lagged market
#'       capitalization column specified by `data_options()`, which defaults
#'       to `mktcap_lag`). Weights are computed using market capitalization
#'       capped at the `cap_weight` percentile per date. `NA` if insufficient
#'       observations for that portfolio-date.
#'   }
#'
#' @family portfolio functions
#' @export
#'
#' @examples
#' set.seed(42)
#' # Univariate sorting with periodic rebalancing
#' data <- data.frame(
#'   permno = 1:500,
#'   date = rep(
#'     seq.Date(
#'       from = as.Date("2020-01-01"),
#'       by = "month",
#'       length.out = 100
#'     ),
#'     each = 10
#'   ),
#'   mktcap_lag = runif(500, 100, 1000),
#'   ret_excess = rnorm(500),
#'   size = runif(500, 50, 150)
#' )
#'
#' compute_portfolio_returns(
#'   data, "size", "univariate",
#'   breakpoint_options_main = breakpoint_options(n_portfolios = 5)
#' )
#'
#' # Bivariate dependent sorting with annual rebalancing
#' compute_portfolio_returns(
#'   data, c("size", "mktcap_lag"), "bivariate-dependent", 7,
#'   breakpoint_options_main = breakpoint_options(n_portfolios = 5),
#'   breakpoint_options_secondary = breakpoint_options(n_portfolios = 3),
#' )
#'
compute_portfolio_returns <- function(
  data,
  sorting_variables,
  sorting_method,
  rebalancing_month = NULL,
  breakpoint_options_main,
  breakpoint_options_secondary = NULL,
  breakpoint_function_main = compute_breakpoints,
  breakpoint_function_secondary = compute_breakpoints,
  min_portfolio_size = 1L,
  cap_weight = 0.8,
  data_options = NULL,
  quiet = FALSE
) {
  if (!is.logical(quiet) || length(quiet) != 1 || is.na(quiet)) {
    cli::cli_abort("{.arg quiet} must be a single logical.")
  }

  if (is.null(data_options)) {
    data_options <- data_options()
  }

  if (is.null(sorting_variables) || length(sorting_variables) == 0) {
    cli::cli_abort("You must provide at least one sorting variable.")
  }

  if (
    !sorting_method %in%
      c("univariate", "bivariate-dependent", "bivariate-independent")
  ) {
    cli::cli_abort("Invalid sorting method.")
  }

  if (
    (sorting_method %in% c("bivariate-dependent", "bivariate-independent")) &&
      is.null(breakpoint_options_secondary)
  ) {
    cli::cli_warn(
      "No 'breakpoint_options_secondary' specified in bivariate sort."
    )
  }

  if (
    !is.numeric(cap_weight) ||
      length(cap_weight) != 1L ||
      is.na(cap_weight) ||
      cap_weight < 0 ||
      cap_weight > 1
  ) {
    cli::cli_abort("'cap_weight' must be a single numeric value in [0, 1].")
  }

  id_col <- data_options$id
  date_col <- data_options$date
  ret_col <- data_options$ret_excess
  w_col <- data_options$mktcap_lag
  w_capped_col <- paste0(w_col, "_capped")

  required_columns <- c(sorting_variables, id_col, date_col, ret_col)

  missing_columns <- setdiff(required_columns, colnames(data))
  if (length(missing_columns) > 0L) {
    cli::cli_abort(
      "Missing columns: {paste(missing_columns, collapse = ', ')}."
    )
  }

  mktcap_lag_missing <- !(w_col %in% colnames(data))
  if (mktcap_lag_missing) {
    data[[w_col]] <- 1L
  }

  # Store the dates before filtering out missing values
  all_dates <- unique(data[[date_col]])

  # Filter out rows with missing values in any sorting variable, as these cannot
  # be assigned to portfolios
  data <- data |>
    tidyr::drop_na(dplyr::all_of(sorting_variables))

  # Compute capped market capitalization per date before replacing NAs,
  # so that the quantile is not distorted by zero-filled missing values
  data <- data |>
    check_new_col(w_capped_col) |>
    dplyr::group_by(.data[[date_col]]) |>
    dplyr::mutate(
      !!w_capped_col := pmin(
        .data[[w_col]],
        stats::quantile(.data[[w_col]], cap_weight, na.rm = TRUE)
      )
    ) |>
    dplyr::ungroup()

  # Replace NA market caps with 0, as these observations should not contribute
  # to the value-weighted return
  missing_mcap_data <- is.na(data[[w_col]])
  data[[w_col]][missing_mcap_data] <- 0
  data[[w_capped_col]][missing_mcap_data] <- 0

  # Handle edge case where all observations are filtered out
  if (nrow(data) == 0L) {
    if (!quiet) {
      cli::cli_inform(
        paste0(
          "Returning an empty panel: all observations were filtered out ",
          "(n() <= {min_portfolio_size} on every date)."
        )
      )
    }

    empty_result <- tibble::tibble(
      portfolio = integer(0L),
      !!date_col := as.Date(character(0L)),
      ret_excess_ew = numeric(0L)
    )
    if (!mktcap_lag_missing) {
      empty_result <- empty_result |>
        tibble::add_column(
          ret_excess_vw = numeric(0L),
          .before = "ret_excess_ew"
        ) |>
        tibble::add_column(
          ret_excess_vw_capped = numeric(0L),
          .before = "ret_excess_ew"
        )
    }
    return(empty_result)
  }

  if (
    !is.null(rebalancing_month) &&
      (rebalancing_month > 12L || rebalancing_month < 1L)
  ) {
    cli::cli_abort("Invalid rebalancing_month.")
  }

  if (sorting_method == "univariate") {
    if (length(sorting_variables) > 1L) {
      cli::cli_abort("Only provide one sorting variable for univariate sorts.")
    }

    if (is.null(rebalancing_month)) {
      portfolio_returns <- data |>
        check_new_col("portfolio") |>
        dplyr::group_by(.data[[date_col]]) |>
        dplyr::mutate(
          portfolio = assign_portfolio(
            data = dplyr::pick(dplyr::everything()),
            sorting_variable = sorting_variables,
            breakpoint_options = breakpoint_options_main,
            breakpoint_function = breakpoint_function_main,
            data_options = data_options
          )
        )
    } else {
      filtered_data <- data |>
        dplyr::filter(month(.data[[date_col]]) == rebalancing_month)

      if (nrow(filtered_data) == 0) {
        cli::cli_abort(
          paste0(
            "No observations match {.arg rebalancing_month} = ",
            "{rebalancing_month}. Check that the data contains dates in ",
            "the specified rebalancing month."
          )
        )
      }

      portfolio_data <- filtered_data |>
        check_new_col("portfolio") |>
        dplyr::group_by(.data[[date_col]]) |>
        dplyr::mutate(
          portfolio = assign_portfolio(
            data = dplyr::pick(dplyr::everything()),
            sorting_variable = sorting_variables,
            breakpoint_options = breakpoint_options_main,
            breakpoint_function = breakpoint_function_main,
            data_options = data_options
          )
        ) |>
        dplyr::ungroup() |>
        dplyr::select(all_of(c(id_col, date_col, "portfolio")))

      portfolio_returns <- join_rebalanced_portfolios(
        data,
        portfolio_data,
        date_col,
        id_col
      )
    }

    portfolio_returns <- portfolio_returns |>
      dplyr::group_by(portfolio, .data[[date_col]]) |>
      summarise_portfolio_returns(
        ret_col,
        w_col,
        w_capped_col,
        min_portfolio_size
      )
  }

  if (sorting_method == "bivariate-dependent") {
    if (length(sorting_variables) != 2L) {
      cli::cli_abort("Provide two sorting variables for bivariate sorts.")
    }

    if (is.null(rebalancing_month)) {
      portfolio_returns <- data |>
        check_new_col(c("portfolio_secondary", "portfolio_main")) |>
        dplyr::group_by(.data[[date_col]]) |>
        dplyr::mutate(
          portfolio_secondary = assign_portfolio(
            dplyr::pick(dplyr::everything()),
            sorting_variable = sorting_variables[2],
            breakpoint_options = breakpoint_options_secondary,
            breakpoint_function = breakpoint_function_secondary,
            data_options = data_options
          )
        ) |>
        dplyr::ungroup() |>
        dplyr::group_by(.data[[date_col]], portfolio_secondary) |>
        dplyr::mutate(
          portfolio_main = assign_portfolio(
            dplyr::pick(dplyr::everything()),
            sorting_variable = sorting_variables[1],
            breakpoint_options = breakpoint_options_main,
            breakpoint_function = breakpoint_function_main,
            data_options = data_options
          )
        ) |>
        dplyr::ungroup()
    } else {
      portfolio_data <- data |>
        check_new_col(c("portfolio_secondary", "portfolio_main")) |>
        dplyr::filter(month(.data[[date_col]]) == rebalancing_month) |>
        dplyr::group_by(.data[[date_col]]) |>
        dplyr::mutate(
          portfolio_secondary = assign_portfolio(
            dplyr::pick(dplyr::everything()),
            sorting_variable = sorting_variables[2L],
            breakpoint_options = breakpoint_options_secondary,
            breakpoint_function = breakpoint_function_secondary,
            data_options = data_options
          )
        ) |>
        dplyr::ungroup() |>
        dplyr::group_by(.data[[date_col]], portfolio_secondary) |>
        dplyr::mutate(
          portfolio_main = assign_portfolio(
            dplyr::pick(dplyr::everything()),
            sorting_variable = sorting_variables[1L],
            breakpoint_options = breakpoint_options_main,
            breakpoint_function = breakpoint_function_main,
            data_options = data_options
          )
        ) |>
        dplyr::ungroup() |>
        dplyr::select(all_of(c(
          id_col,
          date_col,
          "portfolio_main",
          "portfolio_secondary"
        )))

      portfolio_returns <- join_rebalanced_portfolios(
        data,
        portfolio_data,
        date_col,
        id_col
      )
    }

    portfolio_returns <- aggregate_bivariate_returns(
      portfolio_returns,
      date_col,
      ret_col,
      w_col,
      w_capped_col,
      min_portfolio_size
    )
  }

  if (sorting_method == "bivariate-independent") {
    if (length(sorting_variables) != 2L) {
      cli::cli_abort("Provide two sorting variables for bivariate sorts.")
    }

    if (is.null(rebalancing_month)) {
      portfolio_returns <- data |>
        check_new_col(c("portfolio_secondary", "portfolio_main")) |>
        dplyr::group_by(.data[[date_col]]) |>
        dplyr::mutate(
          portfolio_secondary = assign_portfolio(
            dplyr::pick(dplyr::everything()),
            sorting_variable = sorting_variables[2L],
            breakpoint_options = breakpoint_options_secondary,
            breakpoint_function = breakpoint_function_secondary,
            data_options = data_options
          ),
          portfolio_main = assign_portfolio(
            dplyr::pick(dplyr::everything()),
            sorting_variable = sorting_variables[1L],
            breakpoint_options = breakpoint_options_main,
            breakpoint_function = breakpoint_function_main,
            data_options = data_options
          )
        ) |>
        dplyr::ungroup()
    } else {
      portfolio_data <- data |>
        check_new_col(c("portfolio_secondary", "portfolio_main")) |>
        dplyr::filter(month(.data[[date_col]]) == rebalancing_month) |>
        dplyr::group_by(.data[[date_col]]) |>
        dplyr::mutate(
          portfolio_secondary = assign_portfolio(
            dplyr::pick(dplyr::everything()),
            sorting_variable = sorting_variables[2L],
            breakpoint_options = breakpoint_options_secondary,
            breakpoint_function = breakpoint_function_secondary,
            data_options = data_options
          ),
          portfolio_main = assign_portfolio(
            dplyr::pick(dplyr::everything()),
            sorting_variable = sorting_variables[1L],
            breakpoint_options = breakpoint_options_main,
            breakpoint_function = breakpoint_function_main,
            data_options = data_options
          )
        ) |>
        dplyr::ungroup() |>
        dplyr::select(all_of(c(
          id_col,
          date_col,
          "portfolio_main",
          "portfolio_secondary"
        )))

      portfolio_returns <- join_rebalanced_portfolios(
        data,
        portfolio_data,
        date_col,
        id_col
      )
    }

    portfolio_returns <- aggregate_bivariate_returns(
      portfolio_returns,
      date_col,
      ret_col,
      w_col,
      w_capped_col,
      min_portfolio_size
    )
  }

  portfolio_returns <- portfolio_returns[!is.na(portfolio_returns$portfolio), ]

  # When rebalancing annually, dates before the first rebalancing month
  # cannot receive portfolio assignments, so exclude them from the panel
  if (!is.null(rebalancing_month)) {
    matching_dates <- all_dates[month(all_dates) == rebalancing_month]
    if (length(matching_dates) == 0L) {
      cli::cli_abort("No dates in data match {rebalancing_month = }.")
    }
    first_rebalancing_date <- min(matching_dates)
    all_dates <- all_dates[all_dates >= first_rebalancing_date]
  }

  # Complete the panel: ensure all portfolio-date combinations are present
  all_portfolios <- na.omit(unique(portfolio_returns$portfolio))

  complete_panel <- tidyr::expand_grid(
    portfolio = all_portfolios,
    !!date_col := all_dates
  )

  return_cols <- if (mktcap_lag_missing) {
    "ret_excess_ew"
  } else {
    c("ret_excess_vw", "ret_excess_ew", "ret_excess_vw_capped")
  }

  portfolio_returns <- complete_panel |>
    dplyr::left_join(portfolio_returns, by = c("portfolio", date_col)) |>
    dplyr::select(dplyr::all_of(c("portfolio", date_col, return_cols)))

  # Count and report missing values
  n_missing <- sum(is.na(portfolio_returns[["ret_excess_ew"]]))

  if (!quiet && n_missing > 0L) {
    cli::cli_inform(
      paste0(
        "Returning a complete panel with {n_missing} missing value{?s} ",
        "in factor returns due to insufficient observations ",
        "(fewer than {min_portfolio_size} firm{?s} per (portfolio, date) ",
        "cross-section)."
      )
    )
  }

  portfolio_returns
}

#' Summarise portfolio returns (internal helper)
#'
#' Computes equal-weighted, value-weighted, and capped value-weighted returns
#' for an already-grouped data frame. Groups with fewer than
#' `min_portfolio_size` observations receive `NA`.
#'
#' @param data A grouped data frame.
#' @param ret_col Column name for excess returns.
#' @param w_col Column name for market capitalization weights.
#' @param w_capped_col Column name for capped market capitalization weights.
#' @param min_portfolio_size Minimum number of stocks per portfolio-date.
#'
#' @returns An ungrouped data frame with columns `ret_excess_vw`,
#'   `ret_excess_ew`, and `ret_excess_vw_capped`.
#'
#' @keywords internal
#' @noRd
summarise_portfolio_returns <- function(
  data,
  ret_col,
  w_col,
  w_capped_col,
  min_portfolio_size
) {
  data |>
    dplyr::summarise(
      ret_excess_vw = dplyr::if_else(
        dplyr::n() < min_portfolio_size | sum(.data[[w_col]]) == 0,
        NA_real_,
        stats::weighted.mean(.data[[ret_col]], .data[[w_col]])
      ),
      ret_excess_ew = dplyr::if_else(
        dplyr::n() < min_portfolio_size,
        NA_real_,
        mean(.data[[ret_col]])
      ),
      ret_excess_vw_capped = dplyr::if_else(
        dplyr::n() < min_portfolio_size | sum(.data[[w_capped_col]]) == 0,
        NA_real_,
        stats::weighted.mean(.data[[ret_col]], .data[[w_capped_col]])
      ),
      .groups = "drop"
    )
}

#' Aggregate bivariate-sort returns across the secondary dimension
#'
#' Computes cell-level returns over `(portfolio_main, portfolio_secondary,
#' date)` without an occupancy threshold, then averages across the secondary
#' buckets to obtain reported `(portfolio_main, date)` returns. The
#' `min_portfolio_size` threshold is applied to the per-`(portfolio_main,
#' date)` firm count (the reported cross-section), not per cell.
#'
#' @param portfolio_returns A panel with columns `portfolio_main`,
#'   `portfolio_secondary`, the date column, and per-stock returns/weights.
#' @param date_col,ret_col,w_col,w_capped_col Column names.
#' @param min_portfolio_size Minimum firms per reported `(portfolio_main,
#'   date)` cross-section. Cross-sections below this size receive `NA`.
#'
#' @returns A data frame with columns `portfolio`, the date column, and the
#'   three return columns.
#'
#' @keywords internal
#' @noRd
aggregate_bivariate_returns <- function(
  portfolio_returns,
  date_col,
  ret_col,
  w_col,
  w_capped_col,
  min_portfolio_size
) {
  n_per_main_date <- portfolio_returns |>
    dplyr::filter(!is.na(portfolio_main), !is.na(portfolio_secondary)) |>
    dplyr::count(portfolio_main, .data[[date_col]], name = "n_firms")

  portfolio_returns |>
    dplyr::group_by(portfolio_main, portfolio_secondary, .data[[date_col]]) |>
    summarise_portfolio_returns(
      ret_col,
      w_col,
      w_capped_col,
      min_portfolio_size = 0L
    ) |>
    dplyr::group_by(portfolio = portfolio_main, .data[[date_col]]) |>
    dplyr::summarise(
      dplyr::across(
        c(ret_excess_vw, ret_excess_ew, ret_excess_vw_capped),
        \(x) mean(x, na.rm = TRUE)
      ),
      .groups = "drop"
    ) |>
    dplyr::left_join(
      n_per_main_date,
      by = stats::setNames(
        c("portfolio_main", date_col),
        c("portfolio", date_col)
      )
    ) |>
    dplyr::mutate(
      dplyr::across(
        c(ret_excess_vw, ret_excess_ew, ret_excess_vw_capped),
        \(x) {
          dplyr::if_else(
            is.nan(x) | is.na(n_firms) | n_firms < min_portfolio_size,
            NA_real_,
            x
          )
        }
      )
    ) |>
    dplyr::select(-n_firms)
}

#' Join rebalanced portfolio assignments to sorting data (internal helper)
#'
#' Performs an inequality join to carry forward annual portfolio assignments to
#' all dates within the 12-month rebalancing window.
#'
#' @param data The full panel of stock-level data.
#' @param portfolio_data Portfolio assignments at rebalancing dates.
#' @param date_col Name of the date column.
#' @param id_col Name of the stock identifier column.
#'
#' @returns A data frame with portfolio assignments joined to all dates.
#'
#' @keywords internal
#' @noRd
join_rebalanced_portfolios <- function(
  data,
  portfolio_data,
  date_col,
  id_col
) {
  data |>
    dplyr::rename(
      "..date" = dplyr::all_of(date_col),
      "..id" = dplyr::all_of(id_col)
    ) |>
    dplyr::left_join(
      portfolio_data |>
        dplyr::rename("..id" = dplyr::all_of(id_col)) |>
        dplyr::mutate(
          lower_bound = .data[[date_col]],
          upper_bound = .data[[date_col]] + months(12)
        ) |>
        dplyr::select(-dplyr::all_of(date_col)),
      dplyr::join_by(
        ..id,
        closest(..date >= lower_bound),
        ..date <= upper_bound
      ),
      relationship = "many-to-one"
    ) |>
    dplyr::rename("{date_col}" := "..date", "{id_col}" := "..id")
}

#' Estimate a Linear Model
#'
#' @description `r lifecycle::badge('experimental')`
#'
#' Estimates a linear model specified by one or more independent
#' variables. It checks for the presence of the specified independent variables
#' in the dataset and whether the dataset has a sufficient number of
#' observations. Depending on the `output` parameter, it returns the model's
#' coefficients, t-statistics, residuals, or any combination in a named list.
#'
#' @param data A data frame containing the dependent variable and one or more
#'   independent variables.
#' @param model A character that describes the model to be estimated (e.g.,
#'   `"ret_excess ~ mkt_excess + hml + smb"`).
#' @param min_obs The minimum number of observations required to estimate the
#'   model. Defaults to 1.
#' @param output A character vector specifying what to return. Must contain one
#'   or more of `"coefficients"` (default), `"residuals"`, and `"tstats"`. If a
#'   single value is provided, the corresponding object is returned directly.
#'   If multiple values are provided, a named list is returned.
#'
#' @returns If `output` contains a single value: a data frame of coefficients
#'   or t-statistics, or a numeric vector of residuals. If `output` contains
#'   multiple values: a named list with the requested elements. Coefficients
#'   and t-statistics are returned as data frames with column names
#'   corresponding to the model terms. Residuals are returned as a numeric
#'   vector of length `nrow(data)` with `NA` for rows with missing data or
#'   insufficient observations.
#'
#' @family estimation functions
#' @export
#'
#' @examples
#' set.seed(42)
#' data <- data.frame(
#'   ret_excess = rnorm(100),
#'   mkt_excess = rnorm(100),
#'   smb = rnorm(100),
#'   hml = rnorm(100)
#' )
#'
#' # Estimate model with a single independent variable
#' estimate_model(data, "ret_excess ~ mkt_excess")
#'
#' # Estimate model with multiple independent variables
#' estimate_model(data, "ret_excess ~ mkt_excess + smb + hml")
#'
#' # Estimate model without intercept
#' estimate_model(data, "ret_excess ~ mkt_excess - 1")
#'
#' # Calculate residuals
#' estimate_model(data, "ret_excess ~ mkt_excess + smb + hml",
#'   output = "residuals"
#' )
#'
#' # Return t-statistics
#' estimate_model(data, "ret_excess ~ mkt_excess + smb + hml",
#'   output = "tstats"
#' )
#'
#' # Return coefficients, t-statistics, and residuals
#' estimate_model(data, "ret_excess ~ mkt_excess + smb + hml",
#'   output = c("coefficients", "tstats", "residuals")
#' )
#'
estimate_model <- function(data, model, min_obs = 1, output = "coefficients") {
  valid_outputs <- c("coefficients", "tstats", "residuals")
  invalid_outputs <- setdiff(output, valid_outputs)
  if (length(invalid_outputs) > 0) {
    cli::cli_abort(
      paste(
        "{.arg output} must contain one or more of {.val {valid_outputs}},",
        "not {.val {invalid_outputs}}."
      )
    )
  }

  model_parts <- strsplit(model, "~", fixed = TRUE)[[1]]
  independent_vars <- strsplit(trimws(model_parts[2]), "[ +]")[[1]]
  independent_vars <- independent_vars[nzchar(independent_vars)]
  independent_vars <- independent_vars[!independent_vars %in% c("-", "1")]

  if ("intercept" %in% independent_vars) {
    cli::cli_abort(
      paste(
        "None of the columns in {.arg model} may be called 'intercept'.",
        "Please rename the column and try again."
      )
    )
  }

  missing_vars <- independent_vars[!independent_vars %in% names(data)]
  if (length(missing_vars) > 0) {
    cli::cli_abort(
      paste(
        "The following independent variables are missing in the data:",
        "{toString(missing_vars)}."
      )
    )
  }

  formula <- stats::as.formula(model)

  return_multiple <- length(output) > 1
  needs_residuals <- "residuals" %in% output
  needs_coefficients <- "coefficients" %in% output
  needs_tstats <- "tstats" %in% output

  complete <- stats::complete.cases(stats::model.frame(
    formula,
    data = data,
    na.action = stats::na.pass
  ))

  insufficient <- sum(complete) < min_obs ||
    sum(complete) <= length(independent_vars)

  fit <- NULL
  if (!insufficient) {
    fit <- stats::lm(formula, data = data[complete, ])
  }

  to_tibble <- function(x) {
    if ("(Intercept)" %in% names(x)) {
      names(x)[names(x) == "(Intercept)"] <- "intercept"
    }
    as_tibble(t(x))
  }

  na_tibble <- function() {
    if (length(independent_vars) == 0) {
      return(NA_real_)
    }
    beta <- setNames(rep(NA_real_, length(independent_vars)), independent_vars)
    as_tibble(t(beta))
  }

  result <- list()

  if (needs_coefficients) {
    if (insufficient) {
      result$coefficients <- na_tibble()
    } else {
      result$coefficients <- to_tibble(stats::coef(fit))
    }
  }

  if (needs_tstats) {
    if (insufficient) {
      result$tstats <- na_tibble()
    } else {
      result$tstats <- to_tibble(summary(fit)$coefficients[, "t value"])
    }
  }

  if (needs_residuals) {
    if (insufficient) {
      result$residuals <- rep(NA_real_, nrow(data))
    } else {
      resid_result <- rep(NA_real_, nrow(data))
      resid_result[complete] <- fit$residuals
      result$residuals <- resid_result
    }
  }

  if (return_multiple) {
    result
  } else {
    result[[1]]
  }
}

set.seed(42)
test_data <- data.frame(
  ret_excess = rnorm(100),
  mkt_excess = rnorm(100),
  smb = rnorm(100),
  hml = rnorm(100)
)

test_data_with_na <- test_data
test_data_with_na$ret_excess[1:5] <- NA
test_data_with_na$mkt_excess[6:10] <- NA

small_data <- data.frame(
  ret_excess = rnorm(3),
  mkt_excess = rnorm(3),
  smb = rnorm(3)
)

test_that("invalid output values raise an error", {
  expect_error(
    estimate_model(test_data, "ret_excess ~ mkt_excess", output = "invalid"),
    "output"
  )
  expect_error(
    estimate_model(
      test_data,
      "ret_excess ~ mkt_excess",
      output = c("coefficients", "bogus")
    ),
    "output"
  )
})

test_that("missing independent variables raise an error", {
  expect_error(
    estimate_model(test_data, "ret_excess ~ nonexistent_var"),
    "missing"
  )
  expect_error(
    estimate_model(test_data, "ret_excess ~ mkt_excess + nonexistent_var"),
    "nonexistent_var"
  )
})

test_that("column named 'intercept' raises an error", {
  bad_data <- test_data
  bad_data$intercept <- rnorm(100)
  expect_error(
    estimate_model(bad_data, "ret_excess ~ intercept"),
    "intercept"
  )
})

test_that("default output returns a tibble of coefficients", {
  result <- estimate_model(test_data, "ret_excess ~ mkt_excess")
  expect_s3_class(result, "tbl_df")
  expect_true("intercept" %in% names(result))
  expect_true("mkt_excess" %in% names(result))
  expect_equal(ncol(result), 2)
})

test_that("coefficients match lm() output", {
  result <- estimate_model(test_data, "ret_excess ~ mkt_excess + smb + hml")
  fit <- lm(ret_excess ~ mkt_excess + smb + hml, data = test_data)
  coefs <- coef(fit)
  expect_equal(result$intercept, unname(coefs["(Intercept)"]))
  expect_equal(result$mkt_excess, unname(coefs["mkt_excess"]))
  expect_equal(result$smb, unname(coefs["smb"]))
  expect_equal(result$hml, unname(coefs["hml"]))
})

test_that("model without intercept omits intercept column", {
  result <- estimate_model(test_data, "ret_excess ~ mkt_excess - 1")
  expect_false("intercept" %in% names(result))
  expect_true("mkt_excess" %in% names(result))
})

test_that("tstats output returns t-statistics as tibble", {
  result <- estimate_model(
    test_data,
    "ret_excess ~ mkt_excess + smb",
    output = "tstats"
  )
  expect_s3_class(result, "tbl_df")
  expect_true("intercept" %in% names(result))
  expect_true("mkt_excess" %in% names(result))
  expect_true("smb" %in% names(result))

  fit <- lm(ret_excess ~ mkt_excess + smb, data = test_data)
  tvals <- summary(fit)$coefficients[, "t value"]
  expect_equal(result$mkt_excess, unname(tvals["mkt_excess"]))
})

test_that("residuals output returns numeric vector of correct length", {
  result <- estimate_model(
    test_data,
    "ret_excess ~ mkt_excess",
    output = "residuals"
  )
  expect_type(result, "double")
  expect_length(result, nrow(test_data))
})

test_that("residuals match lm() residuals", {
  result <- estimate_model(
    test_data,
    "ret_excess ~ mkt_excess + smb",
    output = "residuals"
  )
  fit <- lm(ret_excess ~ mkt_excess + smb, data = test_data)
  expect_equal(result, unname(fit$residuals))
})

test_that("residuals are NA where data has missing values", {
  result <- estimate_model(
    test_data_with_na,
    "ret_excess ~ mkt_excess",
    output = "residuals"
  )
  expect_length(result, nrow(test_data_with_na))
  expect_true(all(is.na(result[1:10])))
  expect_true(all(!is.na(result[11:100])))
})

test_that("multiple outputs return a named list", {
  result <- estimate_model(
    test_data,
    "ret_excess ~ mkt_excess",
    output = c("coefficients", "tstats", "residuals")
  )
  expect_type(result, "list")
  expect_named(result, c("coefficients", "tstats", "residuals"))
  expect_s3_class(result$coefficients, "tbl_df")
  expect_s3_class(result$tstats, "tbl_df")
  expect_type(result$residuals, "double")
})

test_that("two outputs return a named list with two elements", {
  result <- estimate_model(
    test_data,
    "ret_excess ~ mkt_excess",
    output = c("coefficients", "residuals")
  )
  expect_type(result, "list")
  expect_length(result, 2)
  expect_named(result, c("coefficients", "residuals"))
})

test_that("insufficient observations return NA coefficients", {
  result <- estimate_model(small_data, "ret_excess ~ mkt_excess", min_obs = 100)
  expect_s3_class(result, "tbl_df")
  expect_true(all(is.na(result)))
})

test_that("insufficient observations return NA residuals", {
  result <- estimate_model(
    small_data,
    "ret_excess ~ mkt_excess",
    min_obs = 100,
    output = "residuals"
  )
  expect_type(result, "double")
  expect_length(result, nrow(small_data))
  expect_true(all(is.na(result)))
})

test_that("insufficient observations return NA tstats", {
  result <- estimate_model(
    small_data,
    "ret_excess ~ mkt_excess",
    min_obs = 100,
    output = "tstats"
  )
  expect_s3_class(result, "tbl_df")
  expect_true(all(is.na(result)))
})

test_that("model with multiple independent variables works", {
  result <- estimate_model(test_data, "ret_excess ~ mkt_excess + smb + hml")
  expect_s3_class(result, "tbl_df")
  expect_equal(ncol(result), 4) # intercept + 3 vars
})

test_that("min_obs = 1 works with minimal data", {
  tiny <- data.frame(ret_excess = 1:2, mkt_excess = 1:2)
  result <- estimate_model(tiny, "ret_excess ~ mkt_excess", min_obs = 1)
  expect_s3_class(result, "tbl_df")
})

test_that("residuals are NA when too few complete cases after NA removal", {
  almost_all_na <- test_data
  almost_all_na$ret_excess[1:99] <- NA
  result <- estimate_model(
    almost_all_na,
    "ret_excess ~ mkt_excess + smb",
    min_obs = 5,
    output = "residuals"
  )
  expect_true(all(is.na(result)))
})

test_that("coefficients are NA when all rows are NA", {
  all_na <- test_data
  all_na$ret_excess[] <- NA
  expect_no_error(
    result <- estimate_model(all_na, "ret_excess ~ mkt_excess", min_obs = 1)
  )
  expect_s3_class(result, "tbl_df")
  expect_true(all(is.na(result)))
})

test_that("tstats are NA tibble when all model rows are NA", {
  all_na <- test_data
  all_na$ret_excess[] <- NA
  expect_no_error(
    result <- estimate_model(
      all_na,
      "ret_excess ~ mkt_excess",
      min_obs = 1,
      output = "tstats"
    )
  )
  expect_s3_class(result, "tbl_df")
  expect_true(all(is.na(result)))
})

#' Implement Portfolio Sort
#'
#' A convenience wrapper that combines sample construction filtering and
#' portfolio return computation into a single call. Equivalent to calling
#' [filter_sorting_data()] followed by [compute_portfolio_returns()].
#'
#' @param data A data frame containing the stock-level panel data.
#' @param sorting_variables A character vector of one or two column names to
#'   sort portfolios on.
#' @param sorting_method A string specifying the sorting method:
#'   `"univariate"`, `"bivariate-dependent"`, or `"bivariate-independent"`.
#' @param rebalancing_month An optional integer specifying the month in which
#'   portfolios are rebalanced annually. `NULL` (the default) means monthly
#'   rebalancing.
#' @param portfolio_sort_options A list of class
#'   `tidyfinance_portfolio_sort_options` created by
#'   [portfolio_sort_options()], bundling filter and breakpoint
#'   specifications. The arguments accepted by [portfolio_sort_options()]
#'   include
#'   \itemize{
#'     \item `filter_options` A list of class `tidyfinance_filter_options`
#'       created by [filter_options()], or `NULL` (the default, which applies
#'       no filters). Options include `exclude_financials`,
#'       `exclude_utilities`, `min_stock_price`, `min_size_quantile`,
#'       `min_listing_age`, `exclude_negative_book_equity`, and
#'       `exclude_negative_earnings`.
#'     \item `breakpoint_options_main` A list of class
#'       `tidyfinance_breakpoint_options` created by [breakpoint_options()],
#'       specifying breakpoints for the primary sorting variable. Options
#'       include `n_portfolios`, `percentiles`, `breakpoints_exchanges`,
#'       `smooth_bunching`, and `breakpoints_min_size_threshold`.
#'     \item `breakpoint_options_secondary` A list of class
#'       `tidyfinance_breakpoint_options` created by [breakpoint_options()],
#'       specifying breakpoints for the secondary sorting variable, or `NULL`
#'       (the default) for univariate sorts. Options are the same as for
#'       `breakpoint_options_main`.
#'   }
#' @param breakpoint_function_main The function used to compute breakpoints for
#'   the main sorting variable. Defaults to [compute_breakpoints()].
#' @param breakpoint_function_secondary The function used to compute breakpoints
#'   for the secondary sorting variable. Defaults to [compute_breakpoints()].
#' @param min_portfolio_size An integer specifying the minimum number of
#'   firms in the reported portfolio cross-section on a given date. Defaults
#'   to `1L` (at least one observation per reported portfolio). For univariate
#'   sorts that is firms per portfolio-date; for bivariate sorts that is firms
#'   per main-portfolio-date summed across the secondary buckets.
#'   Cross-sections below the threshold have their returns set to `NA`.
#'   Set to `0L` to deactivate the check. See [compute_portfolio_returns()].
#' @param cap_weight A numeric between 0 and 1 specifying the quantile at which
#'   portfolio weights are capped for the capped value-weighted return.
#'   Defaults to `0.8`.
#' @param data_options A list of class `tidyfinance_data_options` (created via
#'   [data_options()]) specifying column name mappings. All elements are
#'   forwarded to [filter_sorting_data()] and [compute_portfolio_returns()].
#'   Uses [data_options()] default if `NULL`: `"id" = "permno"`,
#'   `"date" = "date"`, `"exchange" = "exchange"`,
#'   `"mktcap_lag" = "mktcap_lag"`, `"ret_excess" = "ret_excess"`,
#'   `"siccd" = "siccd"`, `"price" = "prc_adj"`,
#'   `"listing_age" = "listing_age"`, `"be" = "be"`, and
#'   `"earnings" = "ib"`.
#' @param quiet A logical indicating whether informational messages should be
#'   suppressed. Defaults to `FALSE`.
#'
#' @returns A data frame of portfolio returns as returned by
#'   [compute_portfolio_returns()].
#'
#' @family portfolio functions
#' @export
#'
#' @examples
#'
#' set.seed(123)
#' data <- data.frame(
#'   permno = 1:500,
#'   date = rep(
#'     seq.Date(
#'       from = as.Date("2020-01-01"),
#'       by = "month",
#'       length.out = 100
#'     ),
#'     each = 10
#'   ),
#'   mktcap_lag = runif(500, 100, 1000),
#'   ret_excess = rnorm(500),
#'   prc_adj = runif(500, 0.5, 50),
#'   size = runif(500, 50, 150)
#' )
#' implement_portfolio_sort(
#'   data = data,
#'   sorting_variables = "size",
#'   sorting_method = "univariate",
#'   portfolio_sort_options = portfolio_sort_options(
#'     filter_options = filter_options(
#'       min_stock_price = 1
#'     ),
#'     breakpoint_options_main = breakpoint_options(n_portfolios = 5)
#'   )
#' )
#'
#'
implement_portfolio_sort <- function(
  data,
  sorting_variables,
  sorting_method,
  rebalancing_month = NULL,
  portfolio_sort_options,
  breakpoint_function_main = compute_breakpoints,
  breakpoint_function_secondary = compute_breakpoints,
  min_portfolio_size = 1L,
  cap_weight = 0.8,
  data_options = NULL,
  quiet = FALSE
) {
  if (!is.logical(quiet) || length(quiet) != 1 || is.na(quiet)) {
    cli::cli_abort("{.arg quiet} must be a single logical.")
  }

  if (
    !is.null(data_options) &&
      !inherits(data_options, "tidyfinance_data_options")
  ) {
    cli::cli_abort(
      paste0(
        "{.arg data_options} must be {.val NULL} or of class ",
        "{.cls tidyfinance_data_options}."
      )
    )
  }

  # Validate portfolio_sort_options
  if (!inherits(portfolio_sort_options, "tidyfinance_portfolio_sort_options")) {
    cli::cli_abort(
      paste0(
        "{.arg portfolio_sort_options} must be of class ",
        "{.cls tidyfinance_portfolio_sort_options}."
      )
    )
  }

  # Extract components from portfolio_sort_options
  filter_options <- portfolio_sort_options$filter_options
  breakpoint_options_main <- portfolio_sort_options$breakpoint_options_main
  breakpoint_options_secondary <-
    portfolio_sort_options$breakpoint_options_secondary

  # Apply sample construction filters
  data <- filter_sorting_data(
    data,
    filter_options = filter_options,
    data_options = data_options,
    quiet = quiet
  )

  # Compute and return portfolio returns
  compute_portfolio_returns(
    data,
    sorting_variables = sorting_variables,
    sorting_method = sorting_method,
    rebalancing_month = rebalancing_month,
    breakpoint_options_main = breakpoint_options_main,
    breakpoint_options_secondary = breakpoint_options_secondary,
    breakpoint_function_main = breakpoint_function_main,
    breakpoint_function_secondary = breakpoint_function_secondary,
    min_portfolio_size = min_portfolio_size,
    cap_weight = cap_weight,
    data_options = data_options,
    quiet = quiet
  )
}

#' Winsorize a Numeric Vector
#'
#' Replaces the values in a numeric vector that are beyond the specified
#' quantiles with the boundary values of those quantiles. This is done for both
#' tails of the distribution based on the `cut` parameter.
#'
#' @param x A numeric vector to be winsorized.
#' @param cut The proportion of data to be winsorized from both ends of the
#'   distribution. For example, a `cut` of 0.05 will winsorize the lowest and
#'   highest 5% of the data. Must be in \[0, 0.5\].
#'
#' @returns A numeric vector with the extreme values replaced by the
#'   corresponding quantile values.
#'
#' @family utility functions
#' @export
#'
#' @examples
#' set.seed(123)
#' data <- rnorm(100)
#' winsorized_data <- winsorize(data, 0.05)
#'
winsorize <- function(x, cut) {
  if (cut < 0 || cut > 0.5) {
    cli::cli_abort("{.arg cut} must be inside [0, 0.5].")
  }

  quantiles <- quantile(x, probs = c(cut, 1 - cut), na.rm = TRUE)
  lb <- quantiles[1]
  ub <- quantiles[2]
  x <- replace(x, x > ub, ub)
  x <- replace(x, x < lb, lb)
  x
}

# Test that winsorize function works correctly
test_that("winsorize() correctly adjusts values", {
  set.seed(123)
  x <- rnorm(100)
  cut <- 0.05

  # Winsorize the data
  winsorized_x <- winsorize(x, cut)

  # Check if the lower and upper bounds are correctly applied
  expect_equal(min(winsorized_x), unname(quantile(x, cut, na.rm = TRUE)))
  expect_equal(max(winsorized_x), unname(quantile(x, 1 - cut, na.rm = TRUE)))

  # Check if values outside the quantiles are replaced
  expect_true(all(winsorized_x >= unname(quantile(x, cut, na.rm = TRUE))))
  expect_true(all(winsorized_x <= unname(quantile(x, 1 - cut, na.rm = TRUE))))
})

# Test handling of NA values
test_that("winsorize() handles NA values", {
  x <- c(NA, 1, 2, 3, 4, 5, NA)
  cut <- 0.1

  # Winsorize the data, expecting NA values to be ignored
  winsorized_x <- winsorize(x, cut)

  # Check that the length of the output matches the input
  expect_equal(length(winsorized_x), length(x))

  # Check that NA values are still present & other values are adjusted correctly
  expect_true(all(is.na(winsorized_x) == is.na(x)))
  expect_true(all(
    winsorized_x[!is.na(winsorized_x)] >=
      unname(quantile(x, cut, na.rm = TRUE, names = FALSE))
  ))
  expect_true(all(
    winsorized_x[!is.na(winsorized_x)] <=
      unname(quantile(x, 1 - cut, na.rm = TRUE, names = FALSE))
  ))
})

# Test edge cases
test_that("winsorize() handles edge cases", {
  # Test with an empty vector
  expect_equal(winsorize(numeric(0), 0.1), numeric(0))

  # Test with a vector of identical values
  x <- rep(1, 10)
  expect_equal(winsorize(x, 0.1), x)
})

#' Assign Portfolios Based on Sorting Variable
#'
#' @description `r lifecycle::badge('experimental')`
#'
#' Assigns data points to portfolios based on a specified sorting variable and
#' the selected function to compute breakpoints. Users can specify a function
#' to compute breakpoints. The function must take `data` and
#' `sorting_variable` as the first two arguments. Additional arguments are
#' passed with a named list [breakpoint_options()]. The function needs to
#' return an ascending vector of breakpoints. By default, breakpoints are
#' computed with [compute_breakpoints()]. The default column names can be
#' modified using [data_options()].
#'
#' @param data A data frame containing the dataset for portfolio assignment.
#' @param sorting_variable A string specifying the column name in `data` to be
#'   used for sorting and determining portfolio assignments based on the
#'   breakpoints.
#' @param breakpoint_options An optional named list of arguments passed to
#'   `breakpoint_function`.
#' @param breakpoint_function A function to compute breakpoints. The default
#'   is set to [compute_breakpoints()].
#' @param data_options A list of class `tidyfinance_data_options` (created via
#'   [data_options()]) specifying column name mappings. Passed through to
#'   `breakpoint_function`. When using the default [compute_breakpoints()],
#'   the `exchange` element is used to specify the exchange column, and
#'   `mktcap_lag` is used to specify the market capitalization column. Uses
#'   [data_options()] default if `NULL`: `"exchange" = "exchange"` and
#'   `"mktcap_lag" = "mktcap_lag"`.
#'
#' @returns A vector of integer portfolio assignments for each row in the
#'   input `data`.
#'
#' @family portfolio functions
#' @export
#'
#' @examples
#' set.seed(42)
#' data <- data.frame(
#'   id = 1:100,
#'   exchange = sample(c("NYSE", "NASDAQ"), 100, replace = TRUE),
#'   market_cap = 1:100
#' )
#'
#' assign_portfolio(data, "market_cap", breakpoint_options(n_portfolios = 5))
#'
#' assign_portfolio(
#'   data,
#'   "market_cap",
#'   breakpoint_options(
#'     percentiles = c(0.2, 0.4, 0.6, 0.8),
#'     breakpoints_exchanges = c("NYSE")
#'   )
#' )
#'
assign_portfolio <- function(
  data,
  sorting_variable,
  breakpoint_options = NULL,
  breakpoint_function = compute_breakpoints,
  data_options = NULL
) {
  x <- data[[sorting_variable]]
  n <- length(x)

  ux <- unique(x[!is.na(x)])
  if (length(ux) <= 1L) {
    cli::cli_warn(
      "The sorting variable is constant and only one portfolio is returned."
    )
    return(rep.int(1L, n))
  }

  breakpoints <- breakpoint_function(
    data,
    sorting_variable,
    breakpoint_options,
    data_options
  )

  if (anyNA(breakpoints)) {
    cli::cli_warn(
      "No portfolios were assigned due to missing breakpoints."
    )
    return(NA_integer_)
  }

  portfolio_indices <- findInterval(x, breakpoints, all.inside = TRUE)

  n_expected <- length(breakpoints) - 1L
  n_actual <- length(unique(na.omit(portfolio_indices)))
  if (n_actual != n_expected) {
    cli::cli_warn(
      paste(
        "The number of portfolios differs from the",
        "specified parameter due to clusters in the sorting variable."
      )
    )
  }

  portfolio_indices
}

# A simple breakpoint function that returns quantile-based breakpoints.
# This avoids depending on the real compute_breakpoints / breakpoint_options.
mock_breakpoint_function <- function(
  data,
  sorting_variable,
  bp_options = NULL,
  data_options = NULL
) {
  n <- if (!is.null(bp_options) && !is.null(bp_options$n_portfolios)) {
    bp_options$n_portfolios
  } else {
    5
  }
  probs <- seq(0, 1, length.out = n + 1)
  quantile(data[[sorting_variable]], probs = probs, na.rm = TRUE, names = FALSE)
}

# A breakpoint function that uses explicit percentiles
mock_breakpoint_percentiles <- function(
  data,
  sorting_variable,
  bp_options = NULL,
  data_options = NULL
) {
  percentiles <- if (!is.null(bp_options) && !is.null(bp_options$percentiles)) {
    bp_options$percentiles
  } else {
    c(0.2, 0.4, 0.6, 0.8)
  }
  probs <- c(0, percentiles, 1)
  quantile(data[[sorting_variable]], probs = probs, na.rm = TRUE, names = FALSE)
}

test_that("assign_portfolio returns a vector the same length as nrow(data)", {
  data <- data.frame(id = 1:100, value = runif(100))
  result <- assign_portfolio(
    data,
    "value",
    breakpoint_options = list(n_portfolios = 5),
    breakpoint_function = mock_breakpoint_function
  )
  expect_length(result, nrow(data))
})

test_that("assign_portfolio returns integer portfolio indices", {
  data <- data.frame(id = 1:100, value = runif(100))
  result <- assign_portfolio(
    data,
    "value",
    breakpoint_options = list(n_portfolios = 5),
    breakpoint_function = mock_breakpoint_function
  )
  expect_type(result, "integer")
})

test_that("portfolio indices are within expected range for 5 portfolios", {
  set.seed(42)
  data <- data.frame(id = 1:200, value = runif(200))
  result <- assign_portfolio(
    data,
    "value",
    breakpoint_options = list(n_portfolios = 5),
    breakpoint_function = mock_breakpoint_function
  )
  expect_true(all(result >= 1 & result <= 5))
})

test_that("portfolio indices are within expected range for 10 portfolios", {
  set.seed(42)
  data <- data.frame(id = 1:1000, value = runif(1000))
  result <- assign_portfolio(
    data,
    "value",
    breakpoint_options = list(n_portfolios = 10),
    breakpoint_function = mock_breakpoint_function
  )
  expect_true(all(result >= 1 & result <= 10))
})

test_that("all portfolio buckets are populated with sufficient distinct data", {
  set.seed(123)
  data <- data.frame(id = 1:1000, value = seq_len(1000))
  result <- assign_portfolio(
    data,
    "value",
    breakpoint_options = list(n_portfolios = 5),
    breakpoint_function = mock_breakpoint_function
  )
  expect_equal(sort(unique(result)), 1:5)
})

test_that("higher sorting variable values get higher or equal portfolios", {
  data <- data.frame(id = 1:500, value = seq_len(500))
  result <- assign_portfolio(
    data,
    "value",
    breakpoint_options = list(n_portfolios = 5),
    breakpoint_function = mock_breakpoint_function
  )
  # For strictly increasing values, portfolio assignments must be non-decreasing
  expect_true(all(diff(result) >= 0))
})

test_that("rows with the same sorting variable value get the same portfolio", {
  data <- data.frame(id = 1:10, value = c(1, 1, 2, 2, 3, 3, 4, 4, 5, 5))
  result <- assign_portfolio(
    data,
    "value",
    breakpoint_options = list(n_portfolios = 3),
    breakpoint_function = mock_breakpoint_function
  )
  # Identical values must map to identical portfolios
  expect_equal(result[1], result[2])
  expect_equal(result[3], result[4])
  expect_equal(result[5], result[6])
  expect_equal(result[7], result[8])
  expect_equal(result[9], result[10])
})

test_that("constant sorting variable returns all 1s with a warning", {
  data <- data.frame(id = 1:50, value = rep(42, 50))
  expect_warning(
    result <- assign_portfolio(
      data,
      "value",
      breakpoint_options = list(n_portfolios = 5),
      breakpoint_function = mock_breakpoint_function
    ),
    "sorting variable is constant"
  )
  expect_equal(result, rep(1, 50))
})

test_that("constant sorting variable returns vector of correct length", {
  data <- data.frame(id = 1:10, value = rep(0, 10))
  expect_warning(
    result <- assign_portfolio(
      data,
      "value",
      breakpoint_function = mock_breakpoint_function
    ),
    "constant"
  )
  expect_length(result, 10)
})

test_that("two distinct values with 2 portfolios produces two groups", {
  data <- data.frame(id = 1:100, value = rep(c(1, 2), each = 50))
  result <- assign_portfolio(
    data,
    "value",
    breakpoint_options = list(n_portfolios = 2),
    breakpoint_function = mock_breakpoint_function
  )
  expect_true(all(result %in% c(1, 2)))
  # Lower values should be in portfolio 1, higher in portfolio 2
  # (or equal due to all.inside)
  expect_true(all(result[data$value == 1] <= result[data$value == 2]))
})

test_that("single-row data frame with constant variable triggers warning", {
  data <- data.frame(id = 1, value = 5)
  expect_warning(
    result <- assign_portfolio(
      data,
      "value",
      breakpoint_options = list(n_portfolios = 3),
      breakpoint_function = mock_breakpoint_function
    ),
    "constant"
  )
  expect_equal(result, 1)
})


test_that("warning is issued when clusters reduce the number of portfolios", {
  # Data heavily clustered: only 2 distinct values but requesting 5 portfolios
  data <- data.frame(id = 1:100, value = rep(c(1, 100), each = 50))

  # Mock that returns 6 breakpoints (for 5 portfolios) but clustering collapses
  # them
  mock_bp_5 <- function(data, sorting_variable, bp_options, data_options) {
    c(1, 20, 40, 60, 80, 100)
  }

  expect_warning(
    result <- assign_portfolio(data, "value", breakpoint_function = mock_bp_5),
    "number of portfolios differs"
  )
})

test_that("no cluster warning when portfolios match expected count", {
  set.seed(99)
  data <- data.frame(id = 1:1000, value = seq_len(1000))
  expect_no_warning(
    result <- assign_portfolio(
      data,
      "value",
      breakpoint_options = list(n_portfolios = 5),
      breakpoint_function = mock_breakpoint_function
    )
  )
})

test_that("custom breakpoint function is used correctly", {
  # A breakpoint function that always splits at the median (2 portfolios)
  median_bp <- function(data, sorting_variable, bp_options, data_options) {
    med <- median(data[[sorting_variable]], na.rm = TRUE)
    c(min(data[[sorting_variable]]), med, max(data[[sorting_variable]]))
  }

  data <- data.frame(id = 1:100, value = seq_len(100))
  result <- assign_portfolio(data, "value", breakpoint_function = median_bp)

  expect_true(all(result %in% c(1, 2)))
  # Values below median should be portfolio 1
  expect_true(all(result[data$value <= 50] == 1))
})

test_that("breakpoint_options are forwarded to the breakpoint function", {
  # Breakpoint function that reads a custom parameter
  custom_bp <- function(data, sorting_variable, bp_options, data_options) {
    n <- bp_options$custom_n
    probs <- seq(0, 1, length.out = n + 1)
    quantile(
      data[[sorting_variable]],
      probs = probs,
      na.rm = TRUE,
      names = FALSE
    )
  }

  data <- data.frame(id = 1:200, value = seq_len(200))
  result <- assign_portfolio(
    data,
    "value",
    breakpoint_options = list(custom_n = 4),
    breakpoint_function = custom_bp
  )

  expect_true(all(result >= 1 & result <= 4))
})

test_that("data_options are forwarded to the breakpoint function", {
  # Breakpoint function that checks data_options are received
  bp_with_opts <- function(data, sorting_variable, bp_options, data_options) {
    # Verify data_options is passed through
    stopifnot(!is.null(data_options))
    stopifnot(data_options$date == "my_date")
    probs <- seq(0, 1, length.out = 4)
    quantile(
      data[[sorting_variable]],
      probs = probs,
      na.rm = TRUE,
      names = FALSE
    )
  }

  data <- data.frame(id = 1:100, my_date = Sys.Date(), value = seq_len(100))
  # Should not error — data_options are correctly forwarded

  expect_no_error(
    assign_portfolio(
      data,
      "value",
      breakpoint_function = bp_with_opts,
      data_options = list(date = "my_date")
    )
  )
})

test_that("NULL breakpoint_options works without error", {
  bp_fn <- function(data, sorting_variable, bp_options, data_options) {
    probs <- seq(0, 1, length.out = 4)
    quantile(
      data[[sorting_variable]],
      probs = probs,
      na.rm = TRUE,
      names = FALSE
    )
  }
  data <- data.frame(id = 1:100, value = seq_len(100))
  expect_no_error(
    assign_portfolio(
      data,
      "value",
      breakpoint_options = NULL,
      breakpoint_function = bp_fn
    )
  )
})

test_that("NULL data_options works without error", {
  data <- data.frame(id = 1:100, value = seq_len(100))
  expect_no_error(
    assign_portfolio(
      data,
      "value",
      breakpoint_options = list(n_portfolios = 3),
      breakpoint_function = mock_breakpoint_function,
      data_options = NULL
    )
  )
})

test_that("negative values are handled correctly", {
  data <- data.frame(id = 1:100, value = seq(-50, 49))
  result <- assign_portfolio(
    data,
    "value",
    breakpoint_options = list(n_portfolios = 5),
    breakpoint_function = mock_breakpoint_function
  )
  expect_length(result, 100)
  expect_true(all(result >= 1 & result <= 5))
  # Monotonicity: sorted input should give non-decreasing portfolios
  expect_true(all(diff(result) >= 0))
})

test_that("decimal / floating point values are handled correctly", {
  set.seed(7)
  data <- data.frame(id = 1:200, value = rnorm(200))
  result <- assign_portfolio(
    data,
    "value",
    breakpoint_options = list(n_portfolios = 5),
    breakpoint_function = mock_breakpoint_function
  )
  expect_length(result, 200)
  expect_true(all(result >= 1 & result <= 5))
})

test_that("function works with large datasets", {
  set.seed(1)
  n <- 100000
  data <- data.frame(id = seq_len(n), value = rnorm(n))
  result <- assign_portfolio(
    data,
    "value",
    breakpoint_options = list(n_portfolios = 10),
    breakpoint_function = mock_breakpoint_function
  )
  expect_length(result, n)
  expect_true(all(result >= 1 & result <= 10))
  expect_equal(sort(unique(result)), 1:10)
})

test_that("extreme values (min and max) are assigned to boundary portfolios", {
  # With all.inside = TRUE, values at or beyond breakpoints are clamped
  data <- data.frame(id = 1:5, value = c(0, 25, 50, 75, 100))
  bp_fn <- function(data, sorting_variable, bp_options, data_options) {
    c(0, 25, 50, 75, 100)
  }
  result <- assign_portfolio(data, "value", breakpoint_function = bp_fn)
  # all.inside = TRUE means min maps to 1 and max maps to length(breakpoints)-1
  expect_equal(result[1], 1L) # value = 0 (at left boundary)
  expect_equal(result[5], 4L) # value = 100 (at right boundary, clamped inside)
})

test_that("NA values in sorting variable produce NA portfolio assignments", {
  data <- data.frame(id = 1:10, value = c(1:8, NA, NA))

  bp_fn <- function(data, sorting_variable, bp_options, data_options) {
    vals <- data[[sorting_variable]]
    probs <- seq(0, 1, length.out = 4)
    quantile(vals, probs = probs, na.rm = TRUE, names = FALSE)
  }

  # findInterval returns NA for NA inputs
  result <- assign_portfolio(data, "value", breakpoint_function = bp_fn)
  expect_true(is.na(result[9]))
  expect_true(is.na(result[10]))
  # Non-NA values should still be assigned
  expect_true(all(!is.na(result[1:8])))
})

test_that("return type is integer vector (from findInterval)", {
  data <- data.frame(id = 1:50, value = seq_len(50))
  result <- assign_portfolio(
    data,
    "value",
    breakpoint_options = list(n_portfolios = 5),
    breakpoint_function = mock_breakpoint_function
  )
  expect_true(is.integer(result))
})

test_that("constant variable early return is numeric (rep(1, n))", {
  data <- data.frame(id = 1:10, value = rep(5, 10))
  expect_warning(
    result <- assign_portfolio(
      data,
      "value",
      breakpoint_function = mock_breakpoint_function
    ),
    "constant"
  )
  # rep(1, n) returns numeric, not integer — capture this behavior
  expect_equal(result, rep(1, 10))
})

test_that("function works with different column names for sorting variable", {
  data <- data.frame(company_id = 1:100, market_cap = runif(100, 1e6, 1e9))
  result <- assign_portfolio(
    data,
    "market_cap",
    breakpoint_options = list(n_portfolios = 5),
    breakpoint_function = mock_breakpoint_function
  )
  expect_length(result, 100)
  expect_true(all(result >= 1 & result <= 5))
})

test_that("percentile-based breakpoints produce correct number of groups", {
  set.seed(42)
  data <- data.frame(id = 1:500, value = seq_len(500))
  result <- assign_portfolio(
    data,
    "value",
    breakpoint_options = list(percentiles = c(0.2, 0.4, 0.6, 0.8)),
    breakpoint_function = mock_breakpoint_percentiles
  )
  expect_true(all(result >= 1 & result <= 5))
  expect_equal(sort(unique(result)), 1:5)
})

test_that("two portfolios split data roughly in half", {
  set.seed(10)
  data <- data.frame(id = 1:1000, value = rnorm(1000))
  result <- assign_portfolio(
    data,
    "value",
    breakpoint_options = list(n_portfolios = 2),
    breakpoint_function = mock_breakpoint_function
  )
  expect_true(all(result %in% c(1, 2)))
  # Each group should have roughly 500 members (within tolerance)
  counts <- table(result)
  expect_true(all(counts > 400))
})

test_that("breakpoint function receives all four arguments", {
  received_args <- list()

  spy_bp <- function(data, sorting_variable, bp_options, data_options) {
    received_args$data <<- data
    received_args$sorting_variable <<- sorting_variable
    received_args$bp_options <<- bp_options
    received_args$data_options <<- data_options
    c(0, 0.5, 1)
  }

  data <- data.frame(id = 1:10, value = seq(0, 1, length.out = 10))
  my_bp_opts <- list(n_portfolios = 2)
  my_data_opts <- list(date = "date_col")

  assign_portfolio(
    data,
    "value",
    breakpoint_options = my_bp_opts,
    breakpoint_function = spy_bp,
    data_options = my_data_opts
  )

  expect_identical(received_args$data, data)
  expect_equal(received_args$sorting_variable, "value")
  expect_equal(received_args$bp_options, my_bp_opts)
  expect_equal(received_args$data_options, my_data_opts)
})

#' Set WRDS Credentials
#'
#' Prompts the user to input their WRDS (Wharton Research Data Services)
#' username and password, and stores these credentials in a `.Renviron` file.
#' The user can choose to store the `.Renviron` file in either the project
#' directory or the home directory. If the `.Renviron` file already contains
#' WRDS credentials, the user will be asked if they want to overwrite the
#' existing credentials. Additionally, the user has the option to add the
#' `.Renviron` file to the .gitignore file to prevent it from being tracked by
#' version control.
#'
#' @returns Invisibly returns `TRUE`. Displays messages to the user based on
#'   their input and actions taken.
#'
#' @family WRDS functions
#' @export
#'
#' @examples
#' \dontrun{
#' set_wrds_credentials()
#' }
set_wrds_credentials <- function() {
  wrds_user <- readline(prompt = "Enter your WRDS username: ")
  wrds_password <- readline(prompt = "Enter your WRDS password: ")
  location_choice <- readline(
    prompt = paste(
      "Where do you want to store the .Renviron file?",
      "Enter 'project' for project directory or 'home' for home directory: "
    )
  )

  if (tolower(location_choice) == "project") {
    renviron_path <- file.path(getwd(), ".Renviron")
    gitignore_path <- file.path(getwd(), ".gitignore")
  } else if (tolower(location_choice) == "home") {
    renviron_path <- file.path(Sys.getenv("HOME"), ".Renviron")
    gitignore_path <- file.path(Sys.getenv("HOME"), ".gitignore")
  } else {
    cli::cli_inform(
      paste(
        "Invalid choice.",
        "Please start again and enter {.str project} or {.str home}."
      )
    )
    return(invisible(TRUE))
  }

  env_lines <- if (file.exists(renviron_path)) {
    readLines(renviron_path)
  } else {
    character()
  }

  wrds_user_exists <- any(grepl("^WRDS_USER=", env_lines))
  wrds_password_exists <- any(grepl("^WRDS_PASSWORD=", env_lines))

  if (wrds_user_exists || wrds_password_exists) {
    overwrite_choice <- readline(
      prompt = paste(
        "Credentials already exist. Do you want to overwrite them?",
        "Enter 'yes' or 'no': "
      )
    )
    if (tolower(overwrite_choice) != "yes") {
      cli::cli_inform(
        "Aborted. Credentials already exist and were not overwritten."
      )
      return(invisible(TRUE))
    }
  }

  if (file.exists(gitignore_path)) {
    add_gitignore <- readline(
      prompt = paste(
        "Do you want to add .Renviron to .gitignore?",
        "It is highly recommended! Enter 'yes' or 'no': "
      )
    )
    if (tolower(add_gitignore) == "yes") {
      gitignore_lines <- readLines(gitignore_path)
      if (!any(grepl("^\\.Renviron$", gitignore_lines))) {
        gitignore_lines <- c(gitignore_lines, ".Renviron")
        writeLines(gitignore_lines, gitignore_path)
        cli::cli_inform("{.file .Renviron} added to {.file .gitignore}.")
      }
    } else if (tolower(add_gitignore) == "no") {
      cli::cli_inform("{.file .Renviron} NOT added to {.file .gitignore}.")
    } else {
      cli::cli_inform(
        "Invalid choice. Please start again and enter 'yes' or 'no'."
      )
      return(invisible(TRUE))
    }
  }

  env_lines <- env_lines[!grepl("^WRDS_USER=", env_lines)]
  env_lines <- env_lines[!grepl("^WRDS_PASSWORD=", env_lines)]

  env_lines <- c(
    env_lines,
    sprintf("WRDS_USER=%s", wrds_user),
    sprintf("WRDS_PASSWORD=%s", wrds_password)
  )

  writeLines(env_lines, renviron_path)

  readRenviron(renviron_path)

  cli::cli_inform(
    paste(
      "WRDS credentials have been set and saved in {.file .Renviron} in your",
      "{.path {location_choice}} directory."
    )
  )
}

#' Check if a Dataset Type is Supported
#'
#' Checks if a given dataset type is supported by verifying against a list of
#' all supported dataset types from different domains. If the specified type is
#' not supported, execution stops and an error message listing all supported
#' types is returned.
#'
#' @param type A character string specifying the dataset type to check.
#'
#' @returns Does not return a value; instead, it either passes silently if the
#'   type is supported or stops execution with an error message if the type is
#'   unsupported.
check_supported_type <- function(type) {
  supported_types <- list_supported_types(as_vector = TRUE)
  supported_types_legacy <- list_supported_types_ff_legacy()$type
  if (!any(type %in% c(supported_types, supported_types_legacy))) {
    cli::cli_abort(
      paste(
        "Unsupported {.arg type} specified. Call the function",
        "{.fn list_supported_types} to get all supported types."
      )
    )
  }
}

#' Estimate Rolling Betas
#'
#' Estimates rolling betas for a given model using the provided data. It
#' supports parallel processing for faster computation using the `furrr`
#' package.
#'
#' @param data A data frame containing the data with a date identifier
#'   (defaults to `date`), a stock identifier (defaults to `permno`), and other
#'   variables used in the model.
#' @param model A character string describing the model to be estimated (e.g.,
#'   `"ret_excess ~ mkt_excess + hml + smb"`).
#' @param lookback A Period object specifying the number of months, days,
#'   hours, minutes, or seconds to look back when estimating the rolling model.
#' @param min_obs An integer specifying the minimum number of observations
#'   required to estimate the model. Defaults to 80% of `lookback`.
#' @param use_furrr A logical indicating whether to use the `furrr` package
#'   and its parallelization capabilities. Defaults to `FALSE`.
#' @param data_options A list of class `tidyfinance_data_options` (created via
#'   [data_options()]) specifying column name mappings. The `id` is used to
#'   specify the entity (i.e., firm), and the `date` element is used to specify
#'   the date column. Uses [data_options()] default if `NULL`:
#'   `"id" = "permno"` and `"date" = "date"`.
#'
#' @returns A data frame with the estimated betas for each time period.
#'
#' @family estimation functions
#' @export
#'
#' @examples
#' # Estimate monthly betas using monthly return data
#' set.seed(1234)
#' data_monthly <- tibble::tibble(
#'   date = rep(seq.Date(from = as.Date("2020-01-01"),
#'                       to = as.Date("2020-12-01"), by = "month"), each = 50),
#'   permno = rep(1:50, times = 12),
#'   ret_excess = rnorm(600, 0, 0.1),
#'   mkt_excess = rnorm(600, 0, 0.1),
#'   smb = rnorm(600, 0, 0.1),
#'   hml = rnorm(600, 0, 0.1),
#' )
#'
#' estimate_betas(data_monthly,  "ret_excess ~ mkt_excess", months(3))
#' estimate_betas(
#'   data_monthly,
#'   "ret_excess ~ mkt_excess + smb + hml",
#'   months(6)
#' )
#'
#' data_monthly |>
#'   dplyr::rename(id = permno) |>
#'   estimate_betas("ret_excess ~ mkt_excess", months(3),
#'                  data_options = data_options(id = "id"))
#'
#' # Estimate monthly betas using daily return data and parallelization
#' data_daily <- tibble::tibble(
#'   date = rep(seq.Date(from = as.Date("2020-01-01"),
#'                       to = as.Date("2020-12-31"), by = "day"), each = 50),
#'   permno = rep(1:50, times = 366),
#'   ret_excess = rnorm(18300, 0, 0.02),
#'   mkt_excess = rnorm(18300, 0, 0.02),
#'   smb = rnorm(18300, 0, 0.02),
#'   hml = rnorm(18300, 0, 0.02),
#' )
#'
#' data_daily <- data_daily |>
#'   dplyr::mutate(date = lubridate::floor_date(date, "month"))
#'
#' # Change settings via future::plan(strategy = "multisession", workers = 4)
#' estimate_betas(
#'   data_daily,
#'   "ret_excess ~ mkt_excess",
#'   lubridate::days(90),
#'   use_furrr = TRUE
#' )
#'
estimate_betas <- function(
  data,
  model,
  lookback,
  min_obs = NULL,
  use_furrr = FALSE,
  data_options = NULL
) {
  if (is.null(data_options)) {
    data_options <- data_options()
  }

  if (lookback@month > 0) {
    lookback <- lookback@month
    period <- "month"
  } else if (lookback@day > 0) {
    lookback <- lookback@day
    period <- "day"
  } else if (lookback@hour > 0) {
    lookback <- lookback@hour
    period <- "hour"
  } else if (lookback@minute > 0) {
    lookback <- lookback@minute
    period <- "minute"
  } else if (lookback@second > 0) {
    lookback <- lookback@second
    period <- "second"
  } else {
    cli::cli_abort(
      paste(
        "{.arg lookback} must contain either a positive",
        "month, days, hours, minutes, or seconds"
      )
    )
  }

  if (is.null(min_obs)) {
    min_obs <- round(lookback * 0.8, 0)
  } else if (min_obs <= 0) {
    cli::cli_abort("{.arg min_obs} must be a positive integer.")
  }

  if (!is.logical(use_furrr)) {
    cli::cli_abort("{.arg use_furrr} must be a logical.")
  }

  # Warning if lookback is too low to estimate all model parameters
  num_params <- length(all.vars(as.formula(model))) - 1
  if (lookback < num_params) {
    cli::cli_warn(
      "{.arg lookback} is too low to estimate all model parameters.",
      "Consider increasing it."
    )
  }

  roll_model_estimation <- function(
    df,
    model,
    lookback,
    period,
    min_obs,
    data_options
  ) {
    df <- df |>
      arrange(data_options$date)

    betas <- slider::slide_period_dfr(
      .x = df,
      .i = df |> pull(data_options$date),
      .period = period,
      .f = ~ estimate_model(., model, min_obs),
      .before = lookback - 1,
      .complete = FALSE
    )

    bind_cols(
      tibble("{data_options$date}" := unique(df |> pull(data_options$date))),
      betas
    )
  }

  if (use_furrr) {
    rlang::check_installed(
      "furrr",
      reason = "To use furrr::future_map in estimate_betas()"
    )

    betas <- data |>
      tidyr::nest(data = -all_of(data_options$id)) |>
      mutate(
        beta = furrr::future_map(
          data,
          \(x) {
            roll_model_estimation(
              x,
              model,
              lookback,
              period = period,
              min_obs,
              data_options
            )
          }
        )
      )
  } else {
    betas <- data |>
      tidyr::nest(data = -all_of(data_options$id)) |>
      mutate(
        beta = purrr::map(
          data,
          ~ roll_model_estimation(
            .,
            model,
            lookback,
            period,
            min_obs,
            data_options
          )
        )
      )
  }

  betas <- betas |>
    tidyr::unnest(beta, names_sep = "_") |>
    select(-data) |>
    rename("{data_options$date}" := beta_date)

  if ("beta_intercept" %in% colnames(betas)) {
    colnames(betas)[colnames(betas) == "beta_intercept"] <- "intercept"
  }

  betas
}

test_that("Input validation: negative or zero lookback throws error", {
  data <- tibble(
    date = as.Date("2020-01-01"),
    permno = 1,
    ret_excess = 0,
    mkt_excess = 0
  )
  expect_error(estimate_betas(data, "ret_excess ~ mkt_excess", -1))
  expect_error(estimate_betas(data, "ret_excess ~ mkt_excess", 0))
})

test_that("Input validation: negative or zero min_obs throws error", {
  data <- tibble(
    date = as.Date("2020-01-01"),
    permno = 1,
    ret_excess = 0,
    mkt_excess = 0
  )
  expect_error(estimate_betas(data, "ret_excess ~ mkt_excess", 3, min_obs = -1))
  expect_error(estimate_betas(data, "ret_excess ~ mkt_excess", 3, min_obs = 0))
})

test_that("Input validation: invalid use_furrr throws error", {
  data <- tibble(
    date = as.Date("2020-01-01"),
    permno = 1,
    ret_excess = 0,
    mkt_excess = 0
  )
  expect_error(estimate_betas(data, "ret_excess ~ mkt_excess", 3, use_furr = 0))
})

test_that("Output structure: correct column names and number of rows", {
  data <- tibble(
    date = rep(
      seq.Date(
        from = as.Date("2020-01-01"),
        to = as.Date("2020-12-01"),
        by = "month"
      ),
      each = 2
    ),
    permno = rep(1:2, times = 12),
    ret_excess = rnorm(24, 0, 0.1),
    mkt_excess = rnorm(24, 0, 0.1)
  )

  result <- estimate_betas(data, "ret_excess ~ mkt_excess", months(3))

  expect_true(all(c("date", "beta_mkt_excess") %in% colnames(result)))
  # Should match the number of unique date-permno pairs
  expect_equal(nrow(result), 24) #
})

test_that("Correctness: Known result test", {
  data <- tibble(
    date = as.Date(c("2020-01-01", "2020-02-01", "2020-03-01")),
    permno = c(1, 1, 1),
    ret_excess = c(0.1, 0.2, 0.3),
    mkt_excess = c(0.1, 0.2, 0.3)
  )

  result <- estimate_betas(data, "ret_excess ~ mkt_excess", months(3))

  expect_equal(result$beta_mkt_excess, c(NA, 1, 1))
})

test_that("Performance: Single vs multiple workers give the same result", {
  data <- tibble(
    date = rep(
      seq.Date(
        from = as.Date("2020-01-01"),
        to = as.Date("2020-12-01"),
        by = "month"
      ),
      each = 2
    ),
    permno = rep(1:2, times = 12),
    ret_excess = rnorm(24, 0, 0.1),
    mkt_excess = rnorm(24, 0, 0.1)
  )

  suppressPackageStartupMessages(require(purrr))

  result_single <- estimate_betas(data, "ret_excess ~ mkt_excess", months(3))
  result_multi <- estimate_betas(
    data,
    "ret_excess ~ mkt_excess",
    months(3),
    use_furrr = TRUE
  )

  expect_equal(result_single, result_multi)
})

test_that("Rolling window behavior: correct handling of boundary dates", {
  data <- tibble(
    date = seq.Date(
      from = as.Date("2020-01-01"),
      to = as.Date("2020-06-01"),
      by = "month"
    ),
    permno = 1,
    ret_excess = rnorm(6, 0, 0.1),
    mkt_excess = rnorm(6, 0, 0.1)
  )

  result <- estimate_betas(data, "ret_excess ~ mkt_excess", months(3))

  expect_equal(nrow(result), 6)
  # Check if the first couple of rows have NA values where the window size
  # is not sufficient
  expect_true(all(is.na(result$beta_mkt_excess[1])))
})

test_that("Daily data test: correctly handles daily data grouped into months", {
  data <- tibble(
    date = rep(
      seq.Date(
        from = as.Date("2020-01-01"),
        to = as.Date("2020-12-31"),
        by = "day"
      ),
      each = 2
    ),
    permno = rep(1:2, times = 366),
    ret_excess = rnorm(732, 0, 0.02),
    mkt_excess = rnorm(732, 0, 0.02)
  )

  data <- data |>
    mutate(date = lubridate::floor_date(date, "month"))

  result <- estimate_betas(
    data,
    "ret_excess ~ mkt_excess",
    lookback = months(6)
  )

  # Check if rows match the expected number of date-permno combinations
  expect_equal(nrow(result), length(unique(data$date)) * 2)
  expect_true(all(c("date", "beta_mkt_excess") %in% colnames(result)))
})

test_that("Edge case: single permno", {
  data <- tibble(
    date = seq.Date(
      from = as.Date("2020-01-01"),
      to = as.Date("2020-12-01"),
      by = "month"
    ),
    permno = 1,
    ret_excess = rnorm(12, 0, 0.1),
    mkt_excess = rnorm(12, 0, 0.1)
  )

  result <- estimate_betas(data, "ret_excess ~ mkt_excess", months(3))

  expect_equal(nrow(result), 12)
  expect_true(all(c("date", "beta_mkt_excess") %in% colnames(result)))
})

#' Validate and Coerce Date Range Arguments
#'
#' Checks that `start_date` and `end_date` are a valid pair, coerces them to
#' `Date`, and handles the case where both are `NULL`. When both are `NULL` and
#' `use_default_range = TRUE`, a two-year default window ending one year ago is
#' applied and reported to the user.
#'
#' @param start_date A scalar coercible to `Date` via `as.Date()`, or `NULL`.
#' @param end_date A scalar coercible to `Date` via `as.Date()`, or `NULL`.
#' @param use_default_range A logical scalar. If `TRUE` and both date arguments
#'   are `NULL`, a default one-year range is used instead of returning
#'   `NULL`. Defaults to `FALSE`.
#'
#' @returns A named list with elements `start_date` and `end_date`, both of
#'   class `Date` (or `NULL` when no dates are provided and
#'   `use_default_range = FALSE`).
#'
#' @family utility functions
#' @export
#'
validate_dates <- function(start_date, end_date, use_default_range = FALSE) {
  if (is.null(start_date) || is.null(end_date)) {
    if (use_default_range) {
      start_date <- Sys.Date() %m-% years(2)
      end_date <- Sys.Date() %m-% years(1)
      cli::cli_inform(c(
        paste(
          "No {.arg start_date} or {.arg end_date} provided.",
          "Using the range {start_date} to {end_date} to avoid",
          "downloading large amounts of data."
        )
      ))
      list(start_date = start_date, end_date = end_date)
    } else {
      cli::cli_inform(
        paste(
          "No {.arg start_date} or {.arg end_date} provided.",
          "Returning the full data set."
        )
      )
      list(start_date = NULL, end_date = NULL)
    }
  } else {
    start_date <- as.Date(start_date)
    end_date <- as.Date(end_date)
    if (start_date > end_date) {
      cli::cli_abort("{.arg start_date} cannot be after {.arg end_date}.")
    }
    list(start_date = start_date, end_date = end_date)
  }
}

#' Create Summary Statistics for Specified Variables
#'
#' Computes a set of summary statistics for numeric and integer variables in a
#' data frame. It allows users to select specific variables for
#' summarization and can calculate statistics for the whole dataset or within
#' groups specified by the `by` argument. Additional detail levels for quantiles
#' can be included.
#'
#' @details The function first checks that all specified variables are of type
#' numeric, integer, or logical. If any variables do not meet this criterion,
#' the function stops and returns an error message indicating the non-conforming
#' variables.
#'
#' The basic set of summary statistics includes the count of non-NA values (n),
#' mean, standard deviation (sd), minimum (min), median (q50), and maximum
#' (max). If `detail` is `TRUE`, the function also computes the 1st, 5th, 10th,
#' 25th, 75th, 90th, 95th, and 99th percentiles.
#'
#' Summary statistics are computed for each variable specified in `...`. If a
#' `by` variable is provided, statistics are computed within each level of the
#' `by` variable.
#'
#' @param data A data frame containing the variables to be summarized.
#' @param ... Comma-separated list of unquoted variable names in the data frame
#'   to summarize. These variables must be either numeric, integer, or logical.
#' @param by An optional unquoted variable name to group the data before
#'   summarizing. If `NULL` (the default), summary statistics are computed
#'   across all observations.
#' @param detail A logical flag indicating whether to compute detailed summary
#'   statistics, including additional quantiles. Defaults to `FALSE`, which
#'   computes basic statistics (n, mean, sd, min, median, max). When `TRUE`,
#'   additional quantiles (1%, 5%, 10%, 25%, 75%, 90%, 95%, 99%) are computed.
#' @param drop_na A logical flag indicating whether to drop missing values for
#'   each variable (default is `FALSE`).
#'
#' @returns A tibble with summary statistics for each selected variable. If `by`
#'   is specified, the output includes the grouping variable as well. Each row
#'   represents a variable (and a group if `by` is used), and each column
#'   contains the computed statistics.
#'
#' @family utility functions
#' @export
#'
#' @examples
#' data <- data.frame(
#'   ret = c(0.01, -0.02, 0.03, NA, 0.005),
#'   size = c(100, 200, 150, 300, 250),
#'   group = c("A", "A", "B", "B", "A")
#' )
#'
#' # Basic summary across all observations
#' create_summary_statistics(data, ret, size)
#'
#' # Grouped summary
#' create_summary_statistics(data, ret, size, by = group)
#'
#' # Detailed quantiles
#' create_summary_statistics(data, ret, detail = TRUE)
create_summary_statistics <- function(
  data,
  ...,
  by = NULL,
  detail = FALSE,
  drop_na = FALSE
) {
  by <- enquo(by)
  # Check that all variables to summarize are numeric or integer
  col_types <- data |>
    select(...) |>
    sapply(class)

  if (
    sum(col_types %in% c("numeric", "integer", "logical")) < length(col_types)
  ) {
    cli::cli_abort(sprintf(
      "The following columns are neither numeric nor integer: %s",
      toString(names(col_types[
        !col_types %in% c("numeric", "integer", "logical")
      ]))
    ))
  }

  # Determine set of summary statistics to compute
  if (!detail) {
    funs <- list(
      n = function(x) sum(!is.na(x)),
      mean = mean,
      sd = stats::sd,
      min = min,
      q50 = partial(quantile_na_handler, probs = 0.50),
      max = max
    )
  } else {
    funs <- list(
      n = function(x) sum(!is.na(x)),
      mean = mean,
      sd = stats::sd,
      min = min,
      q01 = partial(quantile_na_handler, probs = 0.01),
      q05 = partial(quantile_na_handler, probs = 0.05),
      q10 = partial(quantile_na_handler, probs = 0.10),
      q25 = partial(quantile_na_handler, probs = 0.25),
      q50 = partial(quantile_na_handler, probs = 0.50),
      q75 = partial(quantile_na_handler, probs = 0.75),
      q90 = partial(quantile_na_handler, probs = 0.90),
      q95 = partial(quantile_na_handler, probs = 0.95),
      q99 = partial(quantile_na_handler, probs = 0.99),
      max = max
    )
  }

  if (rlang::quo_is_null(by)) {
    # Summarize across all observations if no "by" column is specified
    data_long <- data |>
      select(...) |>
      tidyr::pivot_longer(cols = everything(), names_to = "variable")

    if (drop_na) {
      data_long <- tidyr::drop_na(data_long)
    }

    data_long |>
      group_by(.data$variable) |>
      summarize(across(everything(), funs, .names = "{.fn}"), .groups = "drop")
  } else {
    # Summarize by group column if "by" column is specified
    data_long <- data |>
      select(!!by, ...) |>
      tidyr::pivot_longer(cols = -!!by, names_to = "variable")

    if (drop_na) {
      data_long <- tidyr::drop_na(data_long)
    }

    data_long |>
      group_by(.data$variable, !!by) |>
      summarize(across(everything(), funs, .names = "{.fn}"), .groups = "drop")
  }
}

quantile_na_handler <- function(x, probs) {
  if (anyNA(x)) {
    NA_real_
  } else {
    unname(quantile(x, probs = probs))
  }
}

test_data <- tibble(
  A = c(1, 2, 3, 4, NA),
  B = c(NA, 2, 3, 4, 5),
  C = c(1, 1, 2, 2, 2),
  D = c("a", "b", "c", "d", "e")
)

# Basic functionality with numeric columns
test_that("Basic summary statistics are computed correctly", {
  result <- create_summary_statistics(test_data, A, B, C)
  expect_true("n" %in% colnames(result))
  expect_true("mean" %in% colnames(result))
  expect_equal(nrow(result), 3)
})

# Functionality with `detail` = TRUE
test_that("Detailed summary statistics are computed correctly", {
  result <- create_summary_statistics(test_data, A, B, detail = TRUE)
  expect_true("q01" %in% colnames(result))
  expect_true("q99" %in% colnames(result))
})

# Handling of non-numeric columns
test_that("Non-numeric columns are rejected", {
  expect_error(create_summary_statistics(test_data, A, D))
})

# Grouped summaries with a `by` argument
test_that("Summary statistics by group are computed correctly", {
  result <- create_summary_statistics(test_data, A, by = C)
  expect_true("C" %in% colnames(result))
  expect_equal(
    length(unique(result$C)),
    length(unique(test_data$C[!is.na(test_data$C)]))
  )
})

# Test proper handling of NA values
test_that("NA values are handled correctly", {
  data <- data.frame(numeric_col = c(1, 2, NA, 4, 5))

  result_drop_na <- create_summary_statistics(data, numeric_col, drop_na = TRUE)
  result_keep_na <- create_summary_statistics(
    data,
    numeric_col,
    drop_na = FALSE
  )

  expect_identical(result_drop_na$n, result_keep_na$n)
  expect_identical(result_keep_na$mean, NA_real_)
  expect_identical(result_keep_na$sd, NA_real_)
  expect_identical(result_keep_na$min, NA_real_)
  expect_identical(result_keep_na$q50, NA_real_)
  expect_identical(result_keep_na$max, NA_real_)
})

#' Download Enhanced TRACE Data from WRDS
#'
#' Establishes a connection to the WRDS database to download the specified
#' CUSIPs trade messages from the Trade Reporting and Compliance Engine
#' (TRACE). The trade data is cleaned as suggested by Dick-Nielsen
#' (2009, 2014).
#'
#' @param cusips A character vector specifying the 9-digit CUSIPs to
#'   download.
#' @param start_date Optional. A character string or Date object in
#'   "YYYY-MM-DD" format specifying the start date for the data. If not
#'   provided, a subset of the dataset is returned.
#' @param end_date Optional. A character string or Date object in
#'   "YYYY-MM-DD" format specifying the end date for the data. If not
#'   provided, a subset of the dataset is returned.
#'
#' @returns A data frame containing the cleaned trade messages from TRACE
#'   for the selected CUSIPs over the time window specified. Output
#'   variables include identifying information (i.e., CUSIP, trade
#'   date/time) and trade-specific information (i.e., price/yield, volume,
#'   counterparty, and reporting side).
#'
#' @references
#'   Dick-Nielsen, J. (2009). Liquidity biases in TRACE. *Journal of Fixed
#'   Income*, 19(2), 43-55. \doi{10.3905/jfi.2009.19.2.043}
#'
#'   Dick-Nielsen, J. (2014). How to clean enhanced TRACE data. Working Paper.
#'   \doi{10.2139/ssrn.2337908}
#'
#' @family WRDS functions
#' @export
#'
#' @examples
#' \dontrun{
#' download_data_wrds_trace_enhanced("00101JAH9", "2019-01-01", "2021-12-31")
#' }
# nolint start
download_data_wrds_trace_enhanced <- function(
  cusips,
  start_date = NULL,
  end_date = NULL
) {
  if (!is.character(cusips) || anyNA(cusips) || !all(nchar(cusips) == 9)) {
    cli::cli_abort(
      paste(
        "{.arg cusips} must be a character vector of 9-digit CUSIPs,",
        "not {.obj_type_friendly {cusips}}."
      )
    )
  }

  dates <- validate_dates(start_date, end_date, use_default_range = TRUE)
  start_date <- dates$start_date
  end_date <- dates$end_date

  con <- get_wrds_connection()

  trace_enhanced_db <- tbl(con, I("trace.trace_enhanced"))

  trace_all <- trace_enhanced_db |>
    filter(
      cusip_id %in% cusips,
      between(trd_exctn_dt, start_date, end_date)
    ) |>
    select(
      cusip_id,
      msg_seq_nb,
      orig_msg_seq_nb,
      entrd_vol_qt,
      rptd_pr,
      yld_pt,
      rpt_side_cd,
      cntra_mp_id,
      trd_exctn_dt,
      trd_exctn_tm,
      trd_rpt_dt,
      trd_rpt_tm,
      pr_trd_dt,
      trc_st,
      asof_cd,
      wis_fl,
      days_to_sttl_ct,
      stlmnt_dt,
      spcl_trd_fl
    ) |>
    collect()

  disconnect_connection(con)

  # Enhanced Trace: Post 06-02-2012 -----------------------------------------
  # Trades (trc_st = T) and correction (trc_st = R)
  trace_post_TR <- trace_all |>
    filter((trc_st == "T" | trc_st == "R"), trd_rpt_dt >= as.Date("2012-02-06"))

  # Cancelations (trc_st = X) and correction cancelations (trc_st = C)
  trace_post_XC <- trace_all |>
    filter((trc_st == "X" | trc_st == "C"), trd_rpt_dt >= as.Date("2012-02-06"))

  # Cleaning corrected and cancelled trades
  trace_post_TR <- trace_post_TR |>
    anti_join(
      trace_post_XC,
      by = c(
        "cusip_id",
        "msg_seq_nb",
        "entrd_vol_qt",
        "rptd_pr",
        "rpt_side_cd",
        "cntra_mp_id",
        "trd_exctn_dt",
        "trd_exctn_tm"
      )
    )

  # Reversals (trc_st = Y)
  trace_post_Y <- trace_all |>
    filter(trc_st == "Y", trd_rpt_dt >= as.Date("2012-02-06"))

  # Clean reversals
  ## match the orig_msg_seq_nb of the Y-message to
  ## the msg_seq_nb of the main message
  trace_post <- trace_post_TR |>
    anti_join(
      trace_post_Y,
      by = join_by(
        cusip_id,
        msg_seq_nb == orig_msg_seq_nb,
        entrd_vol_qt,
        rptd_pr,
        rpt_side_cd,
        cntra_mp_id,
        trd_exctn_dt,
        trd_exctn_tm
      )
    )

  # Enhanced TRACE: Pre 06-02-2012 ------------------------------------------
  # Cancelations (trc_st = C)
  trace_pre_C <- trace_all |>
    filter(trc_st == "C", trd_rpt_dt < as.Date("2012-02-06"))

  # Trades w/o cancelations
  ## match the orig_msg_seq_nb of the C-message
  ## to the msg_seq_nb of the main message
  trace_pre_T <- trace_all |>
    filter(trc_st == "T", trd_rpt_dt < as.Date("2012-02-06")) |>
    anti_join(
      trace_pre_C,
      by = join_by(
        cusip_id,
        msg_seq_nb == orig_msg_seq_nb,
        entrd_vol_qt,
        rptd_pr,
        rpt_side_cd,
        cntra_mp_id,
        trd_exctn_dt,
        trd_exctn_tm
      )
    )

  # Corrections (trc_st = W) - W can also correct a previous W
  trace_pre_W <- trace_all |>
    filter(trc_st == "W", trd_rpt_dt < as.Date("2012-02-06"))

  # Implement corrections in a loop
  ## Correction control
  correction_control <- nrow(trace_pre_W)
  correction_control_last <- nrow(trace_pre_W)

  ## Correction loop
  while (correction_control > 0) {
    # Corrections that correct some msg
    trace_pre_W_correcting <- trace_pre_W |>
      semi_join(
        trace_pre_T,
        by = join_by(cusip_id, trd_exctn_dt, orig_msg_seq_nb == msg_seq_nb)
      )

    # Corrections that do not correct some msg
    trace_pre_W <- trace_pre_W |>
      anti_join(
        trace_pre_T,
        by = join_by(cusip_id, trd_exctn_dt, orig_msg_seq_nb == msg_seq_nb)
      )

    # Delete msgs that are corrected and add correction msgs
    trace_pre_T <- trace_pre_T |>
      anti_join(
        trace_pre_W_correcting,
        by = join_by(cusip_id, trd_exctn_dt, msg_seq_nb == orig_msg_seq_nb)
      ) |>
      union_all(trace_pre_W_correcting)

    # Escape if no corrections remain or they cannot be matched
    correction_control <- nrow(trace_pre_W)

    if (correction_control == correction_control_last) {
      correction_control <- 0
    }

    correction_control_last <- nrow(trace_pre_W)
  }

  # Clean reversals
  ## Record reversals
  trace_pre_R <- trace_pre_T |>
    filter(asof_cd == "R") |>
    group_by(
      cusip_id,
      trd_exctn_dt,
      entrd_vol_qt,
      rptd_pr,
      rpt_side_cd,
      cntra_mp_id
    ) |>
    arrange(trd_exctn_tm, trd_rpt_dt, trd_rpt_tm) |>
    mutate(seq = row_number()) |>
    ungroup()

  ## Remove reversals and the reversed trade
  trace_pre <- trace_pre_T |>
    filter(is.na(asof_cd) | !(asof_cd %in% c("R", "X", "D"))) |>
    group_by(
      cusip_id,
      trd_exctn_dt,
      entrd_vol_qt,
      rptd_pr,
      rpt_side_cd,
      cntra_mp_id
    ) |>
    arrange(trd_exctn_tm, trd_rpt_dt, trd_rpt_tm) |>
    mutate(seq = row_number()) |>
    ungroup() |>
    anti_join(
      trace_pre_R,
      by = join_by(
        cusip_id,
        trd_exctn_dt,
        entrd_vol_qt,
        rptd_pr,
        rpt_side_cd,
        cntra_mp_id,
        seq
      )
    ) |>
    select(-seq)

  # Agency trades -----------------------------------------------------------
  # Combine pre and post trades
  trace_clean <- trace_post |>
    union_all(trace_pre)

  # Keep angency sells and unmatched agency buys
  ## Agency sells
  trace_agency_sells <- trace_clean |>
    filter(cntra_mp_id == "D", rpt_side_cd == "S")

  # Agency buys that are unmatched
  trace_agency_buys_filtered <- trace_clean |>
    filter(cntra_mp_id == "D", rpt_side_cd == "B") |>
    anti_join(
      trace_agency_sells,
      by = join_by(cusip_id, trd_exctn_dt, entrd_vol_qt, rptd_pr)
    )

  # Agency clean
  trace_clean <- trace_clean |>
    filter(cntra_mp_id == "C") |>
    union_all(trace_agency_sells) |>
    union_all(trace_agency_buys_filtered)

  # Additional Filters ------------------------------------------------------
  trace_add_filters <- trace_clean |>
    mutate(days_to_sttl_ct2 = stlmnt_dt - trd_exctn_dt) |>
    filter(
      is.na(days_to_sttl_ct) | as.numeric(days_to_sttl_ct) <= 7,
      is.na(days_to_sttl_ct2) | as.numeric(days_to_sttl_ct2) <= 7,
      wis_fl == "N",
      is.na(spcl_trd_fl) | spcl_trd_fl == "",
      is.na(asof_cd) | asof_cd == ""
    )

  # Output ------------------------------------------------------------------
  # Only keep necessary columns
  trace_final <- trace_add_filters |>
    arrange(cusip_id, trd_exctn_dt, trd_exctn_tm) |>
    select(
      cusip_id,
      trd_exctn_dt,
      trd_exctn_tm,
      rptd_pr,
      entrd_vol_qt,
      yld_pt,
      rpt_side_cd,
      cntra_mp_id
    )

  trace_final
}
# nolint end

#' Create Breakpoint Options for Portfolio Sorting
#'
#' Generates a structured list of options for defining breakpoints in
#' portfolio sorting. It includes parameters for the number of portfolios,
#' percentile thresholds, exchange-specific breakpoints, and smooth bunching,
#' along with additional optional parameters.
#'
#' @param n_portfolios Integer, optional. The number of portfolios to create.
#'   Must be a positive integer. If not provided, defaults to `NULL`.
#' @param percentiles Numeric vector, optional. A vector of percentile
#'   thresholds for defining breakpoints. Each value must be between 0 and 1.
#'   If not provided, defaults to `NULL`.
#' @param breakpoints_exchanges Character vector, optional. A non-empty vector
#'   specifying the exchange from which to compute the breakpoints. If not
#'   provided, defaults to `NULL`.
#' @param smooth_bunching Logical, optional. Indicates whether smooth bunching
#'   should be applied. Defaults to `FALSE`.
#' @param breakpoints_min_size_threshold Numeric, optional. When set to a value
#'   between 0 and 1, stocks with market capitalization below this quantile are
#'   excluded from breakpoint computation. The quantile is computed among
#'   `breakpoints_exchanges` stocks if specified, otherwise among all stocks.
#'   Requires a market capitalization column in the data (see
#'   [data_options()]). Defaults to `NULL` (no size filtering).
#' @param ... Additional optional arguments. These will be captured in the
#'   resulting structure as a list.
#'
#' @returns A list of class `"tidyfinance_breakpoint_options"` containing the
#'   provided breakpoint options, including any additional arguments passed
#'   via `...`.
#'
#' @family portfolio functions
#' @export
#'
#' @examples
#' breakpoint_options(
#'   n_portfolios = 5,
#'   percentiles = c(0.2, 0.4, 0.6, 0.8),
#'   breakpoints_exchanges = "NYSE",
#'   smooth_bunching = TRUE,
#'   custom_threshold = 0.5,
#'   another_option = "example"
#' )
#'
breakpoint_options <- function(
  n_portfolios = NULL,
  percentiles = NULL,
  breakpoints_exchanges = NULL,
  smooth_bunching = FALSE,
  breakpoints_min_size_threshold = NULL,
  ...
) {
  # Error handling for n_portfolios
  if (
    !is.null(n_portfolios) &&
      (!is.numeric(n_portfolios) ||
        n_portfolios <= 0 ||
        n_portfolios %% 1 != 0)
  ) {
    cli::cli_abort("{.arg n_portfolios} must be a positive integer.")
  }

  # Error handling for percentiles
  if (
    !is.null(percentiles) &&
      (!is.numeric(percentiles) || any(percentiles < 0 | percentiles > 1))
  ) {
    cli::cli_abort(
      paste0(
        "{.arg percentiles} must be a numeric vector ",
        "with values between 0 and 1."
      )
    )
  }

  # Error handling for breakpoints_exchanges
  if (
    !is.null(breakpoints_exchanges) &&
      (!is.character(breakpoints_exchanges) ||
        length(breakpoints_exchanges) == 0)
  ) {
    cli::cli_abort(
      "{.arg breakpoints_exchanges} must be a non-empty character string."
    )
  }

  # Error handling for smooth_bunching
  if (
    !is.logical(smooth_bunching) ||
      length(smooth_bunching) != 1 ||
      is.na(smooth_bunching)
  ) {
    cli::cli_abort(
      "{.arg smooth_bunching} must be a single logical value (TRUE or FALSE)."
    )
  }

  # Error handling for breakpoints_min_size_threshold
  if (
    !is.null(breakpoints_min_size_threshold) &&
      (length(breakpoints_min_size_threshold) != 1L ||
        !is.numeric(breakpoints_min_size_threshold) ||
        breakpoints_min_size_threshold <= 0 ||
        breakpoints_min_size_threshold >= 1)
  ) {
    cli::cli_abort(
      paste0(
        "{.arg breakpoints_min_size_threshold} must be NULL or a single ",
        "numeric value between 0 and 1 (exclusive)."
      )
    )
  }

  # Create the list structure with class attribute
  structure(
    list(
      "n_portfolios" = n_portfolios,
      "percentiles" = percentiles,
      "breakpoints_exchanges" = breakpoints_exchanges,
      "smooth_bunching" = smooth_bunching,
      "breakpoints_min_size_threshold" = breakpoints_min_size_threshold,
      ...
    ),
    class = "tidyfinance_breakpoint_options"
  )
}

#' List Chapters of Tidy Finance
#'
#' Returns a character vector containing the names of the chapters available in
#' the Tidy Finance resource. It provides a quick reference to the
#' various topics covered.
#'
#' @returns A character vector where each element is the name of a chapter
#'   available in the Tidy Finance resource. These names correspond to specific
#'   chapters in Tidy Finance with R.
#'
#' @family utility functions
#' @export
#'
#' @examples
#' list_tidy_finance_chapters()
#'
list_tidy_finance_chapters <- function() {
  c(
    "setting-up-your-environment",
    "introduction-to-tidy-finance",
    "accessing-and-managing-financial-data",
    "wrds-crsp-and-compustat",
    "trace-and-fisd",
    "other-data-providers",
    "beta-estimation",
    "univariate-portfolio-sorts",
    "size-sorts-and-p-hacking",
    "value-and-bivariate-sorts",
    "replicating-fama-and-french-factors",
    "fama-macbeth-regressions",
    "fixed-effects-and-clustered-standard-errors",
    "difference-in-differences",
    "factor-selection-via-machine-learning",
    "option-pricing-via-machine-learning",
    "parametric-portfolio-policies",
    "constrained-optimization-and-backtesting",
    "wrds-dummy-data",
    "cover-and-logo-design",
    "clean-enhanced-trace-with-r",
    "proofs",
    "hex-sticker",
    "changelog"
  )
}

#' Compute Breakpoints Based on Sorting Variable
#'
#' @description `r lifecycle::badge('experimental')`
#'
#' Computes breakpoints based on a specified sorting. It can optionally filter
#' the data by exchanges or lagged size quantiles before computing the
#' breakpoints. The function requires either the number of portfolios to be
#' created or specific percentiles for the breakpoints, but not both. The
#' function also optionally handles cases where the sorting variable clusters
#' on the edges, by assigning all extreme values to the edges and attempting
#' to compute equally populated breakpoints with the remaining values.
#'
#' @param data A data frame containing the dataset for breakpoint computation.
#' @param sorting_variable A character string specifying the column name in
#'   `data` to be used for determining breakpoints.
#' @param breakpoint_options A named list of [breakpoint_options()] for the
#'   breakpoints. The arguments include
#'   \itemize{
#'     \item `n_portfolios` An optional integer specifying the number of
#'       equally sized portfolios to create. This parameter is mutually
#'       exclusive with `percentiles`.
#'     \item `percentiles` An optional numeric vector specifying the
#'       percentiles for determining the breakpoints of the portfolios.
#'       This parameter is mutually exclusive with `n_portfolios`.
#'     \item `breakpoints_exchanges` An optional character vector specifying
#'       exchange names to filter the data before computing breakpoints.
#'       Exchanges must be stored in a column given by `data_options` (defaults
#'       to `exchange`). If `NULL`, no filtering is applied.
#'     \item `smooth_bunching` An optional logical parameter specifying if
#'       to attempt smoothing non-extreme portfolios if the sorting variable
#'       bunches on the extremes (`TRUE`), or not (`FALSE`, the default).
#'       In some cases, smoothing will not result in equal-sized portfolios
#'       off the edges due to multiple clusters. If sufficiently large
#'       bunching is detected, `percentiles` is ignored and equally-spaced
#'       portfolios are returned for these cases with a warning.
#'     \item `breakpoints_min_size_threshold` An optional numeric value between
#'       0 and 1 (exclusive). When set, stocks with market capitalization below
#'       this quantile are excluded from breakpoint computation. The quantile
#'       is computed among `breakpoints_exchanges` stocks if specified,
#'       otherwise among all stocks. Requires a market capitalization column
#'       in the data (column name determined by `data_options`).
#'   }
#' @param data_options A list of class `tidyfinance_data_options` (created via
#'   [data_options()]) specifying column name mappings. The `exchange` element
#'   is used to specify the exchange column, and `mktcap_lag` is used to
#'   specify the market capitalization. Uses [data_options()] default if
#'   `NULL`: `"exchange" = "exchange"` and `"mktcap_lag" = "mktcap_lag"`.
#'
#' @note This function will stop and throw an error if both `n_portfolios` and
#'   `percentiles` are provided or missing simultaneously.
#'
#' @returns A numeric vector of breakpoints of the desired length.
#'
#' @family portfolio functions
#' @export
#'
#' @examples
#' set.seed(42)
#' data <- data.frame(
#'   id = 1:100,
#'   exchange = sample(c("NYSE", "NASDAQ"), 100, replace = TRUE),
#'   market_cap = 1:100
#' )
#'
#' compute_breakpoints(data, "market_cap", breakpoint_options(n_portfolios = 5))
#' compute_breakpoints(
#'   data,
#'   "market_cap",
#'   breakpoint_options(
#'     percentiles = c(0.2, 0.4, 0.6, 0.8),
#'     breakpoints_exchanges = c("NYSE")
#'   )
#'  )
#'
compute_breakpoints <- function(
  data,
  sorting_variable,
  breakpoint_options,
  data_options = NULL
) {
  if (!is.list(breakpoint_options)) {
    cli::cli_abort("Please provide a named list with breakpoint options.")
  }

  n_portfolios <- breakpoint_options$n_portfolios
  percentiles <- breakpoint_options$percentiles
  breakpoints_exchanges <- breakpoint_options$breakpoints_exchanges
  smooth_bunching <- breakpoint_options$smooth_bunching
  breakpoints_min_size_threshold <-
    breakpoint_options$breakpoints_min_size_threshold

  if (is.null(data_options)) {
    data_options <- data_options()
  }

  if (!is.null(n_portfolios) && !is.null(percentiles)) {
    cli::cli_abort(
      paste(
        "Please provide either {.arg n_portfolios}",
        "or {.arg percentiles}, not both."
      )
    )
  } else if (is.null(n_portfolios) && is.null(percentiles)) {
    cli::cli_abort(
      "You must provide either {.arg n_portfolios} or {.arg percentiles}"
    )
  }

  # Extract only the sorting column, filter as a vector
  sorting_values <- data[[sorting_variable]]

  if (!is.null(breakpoints_exchanges)) {
    exchange_col <- data_options$exchange
    if (!(exchange_col %in% colnames(data))) {
      cli::cli_abort(
        paste(
          "Please provide the column {exchange_col}",
          "when filtering using {.arg breakpoints_exchanges}."
        )
      )
    }
    keep <- data[[exchange_col]] %in% breakpoints_exchanges
    sorting_values <- sorting_values[keep]
  }

  if (!is.null(breakpoints_min_size_threshold)) {
    mktcap_col <- data_options$mktcap_lag
    if (!(mktcap_col %in% colnames(data))) {
      cli::cli_abort(
        paste(
          "Column {.val {mktcap_col}} is required",
          "when using {.arg breakpoints_min_size_threshold}."
        )
      )
    }
    if (!is.null(breakpoints_exchanges)) {
      mktcap_ref <- data[[mktcap_col]][keep]
    } else {
      mktcap_ref <- data[[mktcap_col]]
    }
    size_cutoff <- quantile(
      mktcap_ref,
      breakpoints_min_size_threshold,
      na.rm = TRUE
    )
    above_size <- !is.na(data[[mktcap_col]]) & data[[mktcap_col]] > size_cutoff
    if (!is.null(breakpoints_exchanges)) {
      sorting_values <- data[[sorting_variable]][keep & above_size]
    } else {
      sorting_values <- data[[sorting_variable]][above_size]
    }
  }

  if (length(sorting_values) == 0L) {
    cli::cli_warn(
      paste(
        "No breakpoints were calculated, likely due to an insufficient number",
        "of observations after filtering for breakpoint exchanges."
      )
    )
    return(NA_real_)
  }

  if (!is.null(n_portfolios)) {
    if (n_portfolios <= 1L) {
      cli::cli_abort("{.arg n_portfolios} must be larger than 1.")
    }
    probs <- seq(0, 1, length.out = n_portfolios + 1L)
  } else {
    probs <- c(0, percentiles, 1)
    n_portfolios <- length(probs) - 1L
  }

  breakpoints <- quantile(
    sorting_values,
    probs = probs,
    na.rm = TRUE,
    names = FALSE
  )

  if (isTRUE(smooth_bunching)) {
    both_edges <- breakpoints[1] == breakpoints[2] &&
      breakpoints[n_portfolios] == breakpoints[n_portfolios + 1]
    lower_edge <- breakpoints[1] == breakpoints[2]
    upper_edge <- breakpoints[n_portfolios] == breakpoints[n_portfolios + 1]

    if (both_edges) {
      if (!is.null(percentiles)) {
        cli::cli_warn(
          paste(
            "{.arg smooth_bunching} is TRUE and equally-spaced portfolios",
            "are returned for non-edge portfolios."
          )
        )
      }
      mask <- sorting_values > breakpoints[1] &
        sorting_values < breakpoints[n_portfolios + 1]
      sorting_values_new <- sorting_values[mask]
      probs_new <- seq(0, 1, length.out = n_portfolios - 1L)
      breakpoints_new <- quantile(
        sorting_values_new,
        probs = probs_new,
        na.rm = TRUE,
        names = FALSE
      )
      breakpoints_new[n_portfolios - 1L] <- breakpoints_new[n_portfolios - 1L] +
        1e-15
      breakpoints <- c(
        breakpoints[1],
        breakpoints_new,
        breakpoints[n_portfolios + 1]
      )
    } else if (lower_edge) {
      if (!is.null(percentiles)) {
        cli::cli_warn(
          paste(
            "{.arg smooth_bunching} is TRUE and equally-spaced portfolios",
            "are returned for non-edge portfolios."
          )
        )
      }

      sorting_values_new <- sorting_values[sorting_values > breakpoints[1]]
      probs_new <- seq(0, 1, length.out = n_portfolios)
      breakpoints_new <- quantile(
        sorting_values_new,
        probs = probs_new,
        na.rm = TRUE,
        names = FALSE
      )
      breakpoints <- c(breakpoints[1], breakpoints_new)
    } else if (upper_edge) {
      if (!is.null(percentiles)) {
        cli::cli_warn(
          paste(
            "{.arg smooth_bunching} is TRUE and equally-spaced portfolios",
            "are returned for non-edge portfolios."
          )
        )
      }
      sorting_values_new <- sorting_values[
        sorting_values < breakpoints[n_portfolios]
      ]
      probs_new <- seq(0, 1, length.out = n_portfolios)
      breakpoints_new <- quantile(
        sorting_values_new,
        probs = probs_new,
        na.rm = TRUE,
        names = FALSE
      )
      breakpoints_new[n_portfolios] <- breakpoints_new[n_portfolios] + 1e-15
      breakpoints <- c(breakpoints_new, breakpoints[n_portfolios + 1])
    }
  }

  breakpoints[2:(n_portfolios + 1L)] <- breakpoints[2:(n_portfolios + 1L)] +
    1e-20
  breakpoints
}

#' Handle Download Errors Gracefully
#'
#' Wraps a download function in a `tryCatch` block. On error, it emits an
#' informational message and returns `fallback` instead of stopping.
#'
#' @param download_function A function to call for downloading or processing a
#'   resource.
#' @param ... Additional arguments passed to `download_function`.
#' @param fallback The value to return when `download_function` raises an
#'   error. Defaults to `NULL`.
#'
#' @returns The result of `download_function(...)`, or `fallback` on error.
#'
#' @keywords internal
#' @noRd
handle_download_error <- function(download_function, ..., fallback = NULL) {
  tryCatch(
    {
      download_function(...)
    },
    error = function(e) {
      cli::cli_inform(
        paste0(
          "Failed to download or process the resource. ",
          "The resource may not be available, or the URL may have changed. ",
          "Error message: {.message {e$message}}"
        )
      )
      fallback
    }
  )
}

---
title: "Date Columns in `tidyfinance`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Date Columns in `tidyfinance`}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

```{r setup}
#| message: false
#| warning: false
library(tidyfinance)
library(dplyr)
library(lubridate)
```

Working with dates can be cumbersome, in particular when combining data sets from different contexts. When designing the output tables for the `download_data()` function we faced a trade-off between (i) preserving original column names that experienced users might know and for which the original sources provide documentation, and (ii) introducing consistent columns names across data sets where necessary. We decided to consistently use the `date` column with `<date>` column type across various data sets to facilitate joins and avoid unnecessary conversions. In this vignette, we walk you through some examples. 

Let us start with the popular Fama-French factors that come in daily, weekly, and monthly versions. After downloading the monthly data, you can see that each row refers to a specific month of a year and the `date` column refers to the first day of a month. 

```{r}
factors_ff_3_monthly <- download_data("factors_ff_3_monthly")
factors_ff_3_monthly
```

Why did we use the first day of a month here? This convention has the big advantage that you can easily add or subtract months or compute date differences. For instance, in the following code chunk, we subtract 3 months from the original date and then compute the difference in months:

```{r}
factors_ff_3_monthly |>
  select(date) |>
  mutate(
    date_lag3 = date %m-% months(3),
    date_difference = interval(date_lag3, date) %/% months(1)
  )
```

As another example, the `date` column is arguably straight forward for daily data, such as daily Fama-French factors:

```{r}
factors_ff_3_daily <- download_data("factors_ff_3_daily")
factors_ff_3_daily
```

Each row refers to a trading date, so `date` refers to the actual date. So computing date differences to the last row might yield an integer larger than 1. 

```{r}
factors_ff_3_daily |>
  select(date) |>
  mutate(date_difference = interval(lag(date), date) %/% days(1))
```

Similarly, working with daily CRSP data is straight-forward:

```{r}
#| output: false
#| eval: false
crsp_daily <- download_data("wrds_crsp_daily")
crsp_daily |>
  select(permno, date)
```

However, enforcing this consistency poses a challenge for monthly CRSP data. As CRSP calculates the monthly returns based on the last available price for each month, we decided to use `date` following our convention of beginning-of-month and use `calculation_date` as the date that CRSP uses to calculate the returns. In fact, the CRSP 2.0 update introduced the column `mthcaldt` instead of `date`, hence our name `calculation_date`. For most applications, you will not need `calculation_date`, so it serves primarily an informational purpose. 

```{r}
#| output: false
#| eval: false
crsp_monthly <- download_data("wrds_crsp_monthly")
crsp_monthly |>
  select(permno, date, calculation_date)
```

A similar challenge arises with Compustat data. The `datadate` column of the original Compustat tables refers to the date of the fiscal year end or quarter. For many cases, this date is the December 31st of a fiscal year, but they can be any end of month throughout a calendar year. To comply with our notion of consistency, we introduce the column `date`, which is simply the beginning of the month of each `datadate`:

```{r}
#| output: false
#| eval: false
compustat_annual <- download_data("wrds_compustat_annual")
compustat_annual |>
  select(gvkey, date, datadate)
```

If you want to use year or month information in your application, you can simply parse them from `date`:

```{r}
#| output: false
#| eval: false
compustat_annual |>
  mutate(year = year(date), month = month(date)) |>
  select(gvkey, date, datadate, year, month)
```

For completeness, you can do the same for quarterly Compustat data: 

```{r}
#| output: false
#| eval: false
compustat_quarterly <- download_data("wrds_compustat_quarterly")
compustat_quarterly |>
  select(gvkey, date, datadate) |>
  mutate(year = year(date), month = month(date), quarter = quarter(date))
```

## Joining multiple data sources

As a first example, let us combine the Fama-French factors with monthly CRSP returns. We can simply run:

```{r}
#| output: false
#| eval: false
crsp_monthly |>
  left_join(factors_ff_3_monthly, join_by(date)) |>
  select(permno, date, risk_free, mkt_excess, smb, hml)
```

Similarly, the same logic applies to daily CRSP returns:
```{r}
#| output: false
#| eval: false
crsp_daily |>
  left_join(factors_ff_3_daily, join_by(date)) |>
  select(permno, date, risk_free, mkt_excess, smb, hml)
```

Of course, be careful not to join data sets from different contexts, e.g., joining daily Fama-French data to monthly CRSP data. However, the worst that can happen is that you get meaningless or missing rows. 

Now, let us move to the involved case of joining CRSP with Compustat data. To achieve that, we have to first load the CCM links and join them to the monthly CRSP data (the same logic applies to daily CRSP data). 

```{r}
#| output: false
#| eval: false
ccm_links <- download_data("wrds_ccm_links")
crsp_monthly <- crsp_monthly |>
  left_join(ccm_links, join_by(permno), relationship = "many-to-many") |>
  filter(between(date, linkdt, linkenddt)) |>
  select(-c(linkdt, linkenddt))
```

So if you want to join information from Compustat for the **same date**, you just join by `gvkey` and `date`:

```{r}
#| output: false
#| eval: false
crsp_monthly |>
  left_join(
    compustat_annual,
    join_by(gvkey, date)
  )
```

If you want to lag the Compustat information by 6 months to incorporate the fact that there is usually a considerable lag between fiscal year ends and the release of accounting reports, you can just add months to the `date` column in Compustat: 

```{r}
#| output: false
#| eval: false
crsp_monthly |>
  left_join(
    compustat_annual |>
      mutate(date = date %m+% months(6)),
    join_by(gvkey, date)
  )
```

If you prefer to follow the Fama-French protocol for joining company fundamentals to stock returns (see our chapter on this procedure [here](https://www.tidy-finance.org/r/replicating-fama-and-french-factors.html)), you should do:

```{r}
#| output: false
#| eval: false
crsp_monthly |>
  left_join(
    compustat_annual |>
      mutate(date = ymd(paste0(year(date) + 1, "0701"))),
    join_by(gvkey, date)
  )
```

---
title: "Using the `tidyfinance` package"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Using the `tidyfinance` package}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

`tidyfinance` is an R package [on CRAN](https://CRAN.R-project.org/package=tidyfinance) that contains a set of helper functions for empirical research in financial economics, addressing a variety of topics covered in [Tidy Finance with R](https://www.tidy-finance.org/r/) (TFWR). We designed the package to provide easy shortcuts for the applications that we discuss in the book. If you want to inspect the details of the package or propose new features, feel free to visit the package repository on [Github](https://github.com/tidy-finance/r-tidyfinance).

In this vignette, we demonstrate the features of the initial release. We decided to focus on functions that allow you to download the data that we use in TFWR. 

## Install the package

You can install the released version of `tidyfinance` from CRAN via:

```{r}
#| eval: false
install.packages("tidyfinance")
```

You can install the development version of tidyfinance from GitHub using the `pak` package:

```{r}
#| eval: false
pak::pak("tidy-finance/r-tidyfinance")
```

## Download data 

Let's start by loading the package

```{r}
library(tidyfinance)
```

The main function is `download_data(type, start_date, end_date)` with supported type:
```{r}
list_supported_types()
```

So, for instance, if you want to download monthly Fama-French Three-Factor data, you can call:
```{r}
#| eval: false
download_data("factors_ff_3_monthly", "2020-01-01", "2020-12-31")
```

Under the hood, the function uses the `frenchdata` package (see its documentation [here](https://CRAN.R-project.org/package=frenchdata)) and applies some cleaning steps, as in TFWR. If you haven't installed `frenchdata` yet, you'll get prompted to install it first before you can download this specific data type. 

You can also access q-Factor data in this way, by calling:

```{r}
#| eval: false
download_data("factors_q5_daily", "2020-01-01", "2020-12-31")
```

To ensure that we can extend the functionality of the download functions for specific types, we also provide domain-specific download functions. The `download_data("factors_ff_3_monthly")` actually calls `download_data_factors("factors_ff_3_monthly")`, which in turn calls `download_data_factors_ff("factors_ff_3_monthly")`. Why did we decide to have these nested function approach?

Suppose that the q-Factor data changes its URL path and our original function does not work anymore. In this case, you can replace the default `url` value in `download_data_factors_q(type, start_date, end_date, url)` to apply the usual cleaning steps. 

This feature becomes more apparent for other data sources such as `wrds_crsp_monthly`. Note that you need to have valid WRDS credentials and need to set them correctly (check `?set_wrds_connection()` and [WRDS, CRSP, and Compustat](https://www.tidy-finance.org/r/wrds-crsp-and-compustat) in TFWR). If you want to download the standard monthly CRSP data, you can call:

```{r}
#| output: false
#| eval: false
download_data("wrds_crsp_monthly", "2020-01-01", "2020-12-31")
```

If you want to add further columns, you can add them via `additional_columns` to `download_data_wrds_crsp()`, for instance:

```{r}
#| output: false
#| eval: false
download_data_wrds_crsp(
  "wrds_crsp_monthly",
  "2020-01-01",
  "2020-12-31",
  additional_columns = "mthvol"
)
```

Note that the function downloads CRSP v2 as default, as we do in our book since February 2024. If you want to download the old version of CRSP before the update, you can use the `version = v1` parameter in `download_data_wrds_crsp()` . 

As another example, you can do the same for Compustat:

```{r}
#| output: false
#| eval: false
download_data_wrds_compustat(
  "wrds_compustat_annual",
  "2000-01-01",
  "2020-12-31",
  additional_columns = c("acoxar", "amc", "aldo")
)
```

Check out the list of supported types and the corresponding download functions for more information on the respective customization options. We decided to provide limited functionality for the initial release on purpose and rather respond to community request than overengineer the package from the start.  

## Browse content from TFWR

We include functions to check out content from TFWR in your browser. If you want to list all available R chapters, simply call the following function:

```{r}
list_tidy_finance_chapters()
```

The function returns a character vector containing the names of the chapters available in TFWR. If you want to look at a specific chapter, you can call:

```{r}
#| eval: false
open_tidy_finance_website("beta-estimation")
```

This opens either the specific chapter you requested or the main index page in your default web browser. 

## Regression helpers

We discuss winsorization in TFWR, so we figured providing this function could be useful:

```{r}
#| message: false
#| warning: false
library(dplyr)

set.seed(123)
data <- tibble(x = rnorm(100)) |>
  arrange(x)

data |>
  mutate(x_winsorized = winsorize(x, 0.01))
```

If you rather want to replace the bottom and top quantiles of your distribution with missing values, then you can use `trim()`

```{r}
data |>
  mutate(x_trimmed = trim(x, 0.01))
```

We also discuss the importance of providing summary statistics of your data, so there is also a function for that:

```{r}
create_summary_statistics(data, x, detail = TRUE)
```

## Experimental functions

We have two more experimental functions in the sense that it is unclear in which direction they might evolve. First you can assign portfolios based on a sorting variable using `assign_portfolio()`:

```{r}
data <- tibble(
  id = 1:100,
  exchange = sample(c("NYSE", "NASDAQ"), 100, replace = TRUE),
  market_cap = runif(100, 1e6, 1e9)
)

data |>
  mutate(
    portfolio = assign_portfolio(
      pick(everything()),
      "market_cap",
      list(n_portfolios = 5, breakpoint_exchanges = "NYSE")
    )
  )
```

Second, you can estimate the coefficients of a linear model specified by one or more independent variable using `estimate_model()`:

```{r}
data <- tibble(
  ret_excess = rnorm(100),
  mkt_excess = rnorm(100),
  smb = rnorm(100),
  hml = rnorm(100)
)

estimate_model(data, "ret_excess ~ mkt_excess + smb + hml")
```

## Concluding remarks

We are curious to learn in which direction we should extend the package, so please consider opening an issue in the [package repository](https://github.com/tidy-finance/r-tidyfinance/issues). For instance, we could support more data sources, add more parameters to the `download_*` family of functions, or we could put more emphasis on the generality of portfolio assignment or other modeling functions. 

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/download_data_wrds_compustat.R
\name{download_data_wrds_compustat}
\alias{download_data_wrds_compustat}
\title{Download Data from WRDS Compustat}
\usage{
download_data_wrds_compustat(
  dataset = NULL,
  start_date = NULL,
  end_date = NULL,
  type = deprecated(),
  additional_columns = NULL,
  only_us = FALSE
)
}
\arguments{
\item{dataset}{The dataset to download ("compustat_annual" or
"compustat_quarterly").}

\item{start_date}{Optional. A character string or Date object in
"YYYY-MM-DD" format specifying the start date for the data. If not
provided, a subset of the dataset is returned.}

\item{end_date}{Optional. A character string or Date object in "YYYY-MM-DD"
format specifying the end date for the data. If not provided, a subset of
the dataset is returned.}

\item{type}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use \code{dataset} instead.}

\item{additional_columns}{Additional columns from the Compustat table as a
character vector.}

\item{only_us}{A logical indicating whether only US firms should be returned.}
}
\value{
A data frame with financial data for the specified period,
including variables for book equity (be), operating profitability (op),
investment (inv), and others.
}
\description{
Downloads financial data from the WRDS Compustat database for a given
dataset, start date, and end date. It filters the data according to industry
format, data format, and consolidation level, and returns the most current
data for each reporting period. Additionally, the annual data also includes
the calculated book equity (be), operating profitability (op), and
investment (inv) for each company following Fama & French (1993, 2015), as
well as income before extraordinary items (ib).
}
\examples{
\dontrun{
download_data_wrds_compustat("compustat_annual", "2020-01-01", "2020-12-31")
download_data_wrds_compustat(
  "compustat_quarterly",
  "2020-01-01",
  "2020-12-31"
)

  # Add additional columns
download_data_wrds_compustat(
  "compustat_annual",
  additional_columns = c("aodo", "aldo")
)
}
}
\references{
Fama, E. F., & French, K. R. (1993). Common risk factors in the returns on
stocks and bonds. \emph{Journal of Financial Economics}, 33(1), 3-56.
\doi{10.1016/0304-405X(93)90023-5}

Fama, E. F., & French, K. R. (2015). A five-factor asset pricing model.
\emph{Journal of Financial Economics}, 116(1), 1-22.
\doi{10.1016/j.jfineco.2014.10.010}
}
\seealso{
Other WRDS functions: 
\code{\link{create_wrds_dummy_database}()},
\code{\link{disconnect_connection}()},
\code{\link{download_data_wrds}()},
\code{\link{download_data_wrds_ccm_links}()},
\code{\link{download_data_wrds_crsp}()},
\code{\link{download_data_wrds_fisd}()},
\code{\link{download_data_wrds_trace_enhanced}()},
\code{\link{get_wrds_connection}()},
\code{\link{set_wrds_credentials}()}
}
\concept{WRDS functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/create_wrds_dummy_database.R
\name{create_wrds_dummy_database}
\alias{create_wrds_dummy_database}
\title{Create WRDS Dummy Database}
\usage{
create_wrds_dummy_database(
  path,
  url = paste0("https://github.com/tidy-finance/website/tree/main/blog/",
    "tidy-finance-dummy-data/data/tidy_finance.sqlite")
)
}
\arguments{
\item{path}{The file path where the SQLite database should be saved.}

\item{url}{The URL where the SQLite database is stored.}
}
\value{
Invisible \code{NULL}. Side effect: downloads a file to the specified
path.
}
\description{
Downloads the WRDS dummy database from the respective Tidy Finance GitHub
repository and saves it to the specified path. If the file already exists,
the user is prompted before it is replaced.
}
\examples{
path <- paste0(tempdir(), "/tidy_finance_r.sqlite")
create_wrds_dummy_database(path)
}
\seealso{
Other WRDS functions: 
\code{\link{disconnect_connection}()},
\code{\link{download_data_wrds}()},
\code{\link{download_data_wrds_ccm_links}()},
\code{\link{download_data_wrds_compustat}()},
\code{\link{download_data_wrds_crsp}()},
\code{\link{download_data_wrds_fisd}()},
\code{\link{download_data_wrds_trace_enhanced}()},
\code{\link{get_wrds_connection}()},
\code{\link{set_wrds_credentials}()}
}
\concept{WRDS functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/breakpoint_options.R
\name{breakpoint_options}
\alias{breakpoint_options}
\title{Create Breakpoint Options for Portfolio Sorting}
\usage{
breakpoint_options(
  n_portfolios = NULL,
  percentiles = NULL,
  breakpoints_exchanges = NULL,
  smooth_bunching = FALSE,
  breakpoints_min_size_threshold = NULL,
  ...
)
}
\arguments{
\item{n_portfolios}{Integer, optional. The number of portfolios to create.
Must be a positive integer. If not provided, defaults to \code{NULL}.}

\item{percentiles}{Numeric vector, optional. A vector of percentile
thresholds for defining breakpoints. Each value must be between 0 and 1.
If not provided, defaults to \code{NULL}.}

\item{breakpoints_exchanges}{Character vector, optional. A non-empty vector
specifying the exchange from which to compute the breakpoints. If not
provided, defaults to \code{NULL}.}

\item{smooth_bunching}{Logical, optional. Indicates whether smooth bunching
should be applied. Defaults to \code{FALSE}.}

\item{breakpoints_min_size_threshold}{Numeric, optional. When set to a value
between 0 and 1, stocks with market capitalization below this quantile are
excluded from breakpoint computation. The quantile is computed among
\code{breakpoints_exchanges} stocks if specified, otherwise among all stocks.
Requires a market capitalization column in the data (see
\code{\link[=data_options]{data_options()}}). Defaults to \code{NULL} (no size filtering).}

\item{...}{Additional optional arguments. These will be captured in the
resulting structure as a list.}
}
\value{
A list of class \code{"tidyfinance_breakpoint_options"} containing the
provided breakpoint options, including any additional arguments passed
via \code{...}.
}
\description{
Generates a structured list of options for defining breakpoints in
portfolio sorting. It includes parameters for the number of portfolios,
percentile thresholds, exchange-specific breakpoints, and smooth bunching,
along with additional optional parameters.
}
\examples{
breakpoint_options(
  n_portfolios = 5,
  percentiles = c(0.2, 0.4, 0.6, 0.8),
  breakpoints_exchanges = "NYSE",
  smooth_bunching = TRUE,
  custom_threshold = 0.5,
  another_option = "example"
)

}
\seealso{
Other portfolio functions: 
\code{\link{assign_portfolio}()},
\code{\link{compute_breakpoints}()},
\code{\link{compute_long_short_returns}()},
\code{\link{compute_portfolio_returns}()},
\code{\link{data_options}()},
\code{\link{filter_options}()},
\code{\link{filter_sorting_data}()},
\code{\link{implement_portfolio_sort}()},
\code{\link{portfolio_sort_options}()}
}
\concept{portfolio functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/portfolio_sort_options.R
\name{portfolio_sort_options}
\alias{portfolio_sort_options}
\title{Create Portfolio Sort Options}
\usage{
portfolio_sort_options(
  filter_options = NULL,
  breakpoint_options_main,
  breakpoint_options_secondary = NULL,
  ...
)
}
\arguments{
\item{filter_options}{A list of class \code{tidyfinance_filter_options} created
by \code{\link[=filter_options]{filter_options()}}, or \code{NULL} (the default, which applies no filters).
The arguments accepted by \code{\link[=filter_options]{filter_options()}} include
\itemize{
\item \code{exclude_financials} A logical indicating whether to exclude
financial firms (SIC codes 6000–6799). Defaults to \code{FALSE}.
\item \code{exclude_utilities} A logical indicating whether to exclude
utility firms (SIC codes 4900–4999). Defaults to \code{FALSE}.
\item \code{min_stock_price} A single positive numeric specifying the
minimum stock price required to include an observation. \code{NULL} (the
default) applies no price filter.
\item \code{min_size_quantile} A single numeric strictly between 0 and 1
specifying the minimum cross-sectional size quantile (based on lagged
market cap) required to include an observation. \code{NULL} (the default)
applies no size quantile filter.
\item \code{min_listing_age} A single non-negative integer or numeric
specifying the minimum number of months a stock must have been listed
in CRSP. \code{NULL} (the default) applies no listing age filter.
\item \code{exclude_negative_book_equity} A logical indicating whether to
exclude observations with non-positive book equity. Defaults to
\code{FALSE}.
\item \code{exclude_negative_earnings} A logical indicating whether to
exclude observations with non-positive earnings. Defaults to \code{FALSE}.
}}

\item{breakpoint_options_main}{A list of class
\code{tidyfinance_breakpoint_options} created by \code{\link[=breakpoint_options]{breakpoint_options()}},
specifying breakpoints for the primary sorting variable. The arguments
accepted by \code{\link[=breakpoint_options]{breakpoint_options()}} include
\itemize{
\item \code{n_portfolios} An optional integer specifying the number of
equally sized portfolios to create. This parameter is mutually
exclusive with \code{percentiles}.
\item \code{percentiles} An optional numeric vector specifying the
percentiles for determining the breakpoints of the portfolios.
This parameter is mutually exclusive with \code{n_portfolios}.
\item \code{breakpoints_exchanges} An optional character vector specifying
exchange names to filter the data before computing breakpoints.
Exchanges must be stored in a column given by \code{data_options} (defaults
to \code{exchange}). If \code{NULL}, no filtering is applied.
\item \code{smooth_bunching} An optional logical parameter specifying if
to attempt smoothing non-extreme portfolios if the sorting variable
bunches on the extremes (\code{TRUE}), or not (\code{FALSE}, the default).
\item \code{breakpoints_min_size_threshold} An optional numeric value between
0 and 1 (exclusive). When set, stocks with market capitalization below
this quantile are excluded from breakpoint computation.
}}

\item{breakpoint_options_secondary}{A list of class
\code{tidyfinance_breakpoint_options} created by \code{\link[=breakpoint_options]{breakpoint_options()}},
specifying breakpoints for the secondary sorting variable, or \code{NULL}
(the default) for univariate sorts. The arguments accepted by
\code{\link[=breakpoint_options]{breakpoint_options()}} are the same as for \code{breakpoint_options_main}.}

\item{...}{Additional arguments to be included in the options list.}
}
\value{
A list of class \code{tidyfinance_portfolio_sort_options} containing
the specified options.
}
\description{
Creates a list of options of class \code{tidyfinance_portfolio_sort_options} that
bundles sample construction filters and breakpoint specifications for use
with \code{\link[=implement_portfolio_sort]{implement_portfolio_sort()}}.
}
\examples{
portfolio_sort_options(
  filter_options = filter_options(exclude_financials = TRUE),
  breakpoint_options_main = breakpoint_options(n_portfolios = 10)
)

}
\seealso{
Other portfolio functions: 
\code{\link{assign_portfolio}()},
\code{\link{breakpoint_options}()},
\code{\link{compute_breakpoints}()},
\code{\link{compute_long_short_returns}()},
\code{\link{compute_portfolio_returns}()},
\code{\link{data_options}()},
\code{\link{filter_options}()},
\code{\link{filter_sorting_data}()},
\code{\link{implement_portfolio_sort}()}
}
\concept{portfolio functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/download_data_wrds_trace_enhanced.R
\name{download_data_wrds_trace_enhanced}
\alias{download_data_wrds_trace_enhanced}
\title{Download Enhanced TRACE Data from WRDS}
\usage{
download_data_wrds_trace_enhanced(cusips, start_date = NULL, end_date = NULL)
}
\arguments{
\item{cusips}{A character vector specifying the 9-digit CUSIPs to
download.}

\item{start_date}{Optional. A character string or Date object in
"YYYY-MM-DD" format specifying the start date for the data. If not
provided, a subset of the dataset is returned.}

\item{end_date}{Optional. A character string or Date object in
"YYYY-MM-DD" format specifying the end date for the data. If not
provided, a subset of the dataset is returned.}
}
\value{
A data frame containing the cleaned trade messages from TRACE
for the selected CUSIPs over the time window specified. Output
variables include identifying information (i.e., CUSIP, trade
date/time) and trade-specific information (i.e., price/yield, volume,
counterparty, and reporting side).
}
\description{
Establishes a connection to the WRDS database to download the specified
CUSIPs trade messages from the Trade Reporting and Compliance Engine
(TRACE). The trade data is cleaned as suggested by Dick-Nielsen
(2009, 2014).
}
\examples{
\dontrun{
download_data_wrds_trace_enhanced("00101JAH9", "2019-01-01", "2021-12-31")
}
}
\references{
Dick-Nielsen, J. (2009). Liquidity biases in TRACE. \emph{Journal of Fixed
Income}, 19(2), 43-55. \doi{10.3905/jfi.2009.19.2.043}

Dick-Nielsen, J. (2014). How to clean enhanced TRACE data. Working Paper.
\doi{10.2139/ssrn.2337908}
}
\seealso{
Other WRDS functions: 
\code{\link{create_wrds_dummy_database}()},
\code{\link{disconnect_connection}()},
\code{\link{download_data_wrds}()},
\code{\link{download_data_wrds_ccm_links}()},
\code{\link{download_data_wrds_compustat}()},
\code{\link{download_data_wrds_crsp}()},
\code{\link{download_data_wrds_fisd}()},
\code{\link{get_wrds_connection}()},
\code{\link{set_wrds_credentials}()}
}
\concept{WRDS functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/download_data_wrds.R
\name{download_data_wrds}
\alias{download_data_wrds}
\title{Download Data from WRDS}
\usage{
download_data_wrds(
  dataset = NULL,
  start_date = NULL,
  end_date = NULL,
  type = deprecated(),
  ...
)
}
\arguments{
\item{dataset}{A string specifying the dataset to download. Supported
values: "crsp_monthly", "crsp_daily" for CRSP data,
"compustat_annual", "compustat_quarterly" for Compustat data,
"ccm_links" for CCM links data, "fisd" for FISD data, or
"trace_enhanced" for TRACE data.}

\item{start_date}{Optional. A character string or Date object in
"YYYY-MM-DD" format specifying the start date for the data. If not
provided, a subset of the dataset is returned.}

\item{end_date}{Optional. A character string or Date object in
"YYYY-MM-DD" format specifying the end date for the data. If not
provided, a subset of the dataset is returned.}

\item{type}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use \code{dataset} instead.}

\item{...}{Additional arguments passed to specific download functions
depending on the \code{dataset}.}
}
\value{
A data frame containing the requested data, with the structure
and contents depending on the specified \code{dataset}.
}
\description{
Acts as a wrapper to download data from various WRDS datasets including
CRSP, Compustat, and CCM links based on the specified dataset. It is
designed to handle different datasets by redirecting to the appropriate
specific data download function.
}
\examples{
\dontrun{
  crsp_monthly <- download_data_wrds(
    "crsp_monthly", "2020-01-01", "2020-12-31"
  )
  compustat_annual <- download_data_wrds(
    "compustat_annual", "2020-01-01", "2020-12-31"
  )
  ccm_links <- download_data_wrds("ccm_links")
  fisd <- download_data_wrds("fisd")
  trace_enhanced <- download_data_wrds(
    "trace_enhanced", cusips = "00101JAH9"
  )
}
}
\seealso{
Other WRDS functions: 
\code{\link{create_wrds_dummy_database}()},
\code{\link{disconnect_connection}()},
\code{\link{download_data_wrds_ccm_links}()},
\code{\link{download_data_wrds_compustat}()},
\code{\link{download_data_wrds_crsp}()},
\code{\link{download_data_wrds_fisd}()},
\code{\link{download_data_wrds_trace_enhanced}()},
\code{\link{get_wrds_connection}()},
\code{\link{set_wrds_credentials}()}
}
\concept{WRDS functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/download_data.R
\name{download_data}
\alias{download_data}
\title{Download and Process Data Based on Domain and Dataset}
\usage{
download_data(
  domain = NULL,
  dataset = NULL,
  start_date = NULL,
  end_date = NULL,
  type = deprecated(),
  ...
)
}
\arguments{
\item{domain}{The domain of the dataset to download (e.g.,
"famafrench", "globalq", "macro_predictors", "wrds",
"constituents", "fred", "stock_prices", "osap", "tidyfinance").}

\item{dataset}{Optional. The specific dataset to download within the
domain.}

\item{start_date}{Optional. A character string or Date object in
"YYYY-MM-DD" format specifying the start date for the data. If not
provided, the full dataset or a subset is returned, depending on the
dataset type.}

\item{end_date}{Optional. A character string or Date object in
"YYYY-MM-DD" format specifying the end date for the data. If not
provided, the full dataset or a subset is returned, depending on the
dataset type.}

\item{type}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use \code{domain} and
\code{dataset} instead.}

\item{...}{Additional arguments passed to specific download functions
depending on the \code{domain}. For instance, if \code{domain} is
\code{"constituents"}, arguments are passed to
\code{download_data_constituents()}. If \code{domain} is \code{"tidyfinance"} and
\code{dataset} is \code{"factor_library"}, arguments are used to filter the
portfolio grid (e.g., \code{sorting_variable}, \code{rebalancing}, \code{fill_all});
see \code{download_data_huggingface()} for details.}
}
\value{
A tibble with processed data, including dates and the relevant
financial metrics, filtered by the specified date range.
}
\description{
Downloads and processes data based on the specified domain (e.g.,
Fama-French factors, Global Q factors, or macro predictors), dataset,
and date range. This function checks if the specified domain is supported
and then delegates to the appropriate function for downloading and
processing the data.
}
\examples{
\donttest{
download_data(
  "famafrench",
  "Fama/French 5 Factors (2x3) [Daily]",
  "2000-01-01",
  "2020-12-31"
)
download_data("macro_predictors", "monthly", "2000-01-01", "2020-12-31")
download_data("constituents", index = "DAX")
download_data("fred", series = c("GDP", "CPIAUCNS"))
download_data("stock_prices", symbols = c("AAPL", "MSFT"))
download_data(
  "tidyfinance",
  "risk_free",
  "2020-01-01",
  "2020-12-31"
)
download_data(
  "tidyfinance",
  "high_frequency_sp500",
  "2007-07-26",
  "2007-07-27"
)
download_data(
  "tidyfinance",
  "factor_library",
  sorting_variable = "52w",
  rebalancing = "annual"
)
}
}
\seealso{
Other download functions: 
\code{\link{download_data_constituents}()},
\code{\link{download_data_factors_ff}()},
\code{\link{download_data_factors_q}()},
\code{\link{download_data_fred}()},
\code{\link{download_data_huggingface}()},
\code{\link{download_data_macro_predictors}()},
\code{\link{download_data_osap}()},
\code{\link{download_data_risk_free}()},
\code{\link{download_data_stock_prices}()}
}
\concept{download functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/download_data_wrds_ccm_links.R
\name{download_data_wrds_ccm_links}
\alias{download_data_wrds_ccm_links}
\title{Download CCM Links from WRDS}
\usage{
download_data_wrds_ccm_links(linktype = c("LU", "LC"), linkprim = c("P", "C"))
}
\arguments{
\item{linktype}{A character vector indicating the type of link to download.
The default is \code{c("LU", "LC")}, where "LU" stands for "Link Up" and "LC"
for "Link CRSP".}

\item{linkprim}{A character vector indicating the primacy of the link.
Default is \code{c("P", "C")}, where "P" indicates primary and "C" indicates
conditional links.}
}
\value{
A data frame with the columns \code{permno}, \code{gvkey}, \code{linkdt}, and
\code{linkenddt}, where \code{linkenddt} is the end date of the link, and missing end
dates are replaced with today's date.
}
\description{
Downloads data from the WRDS CRSP/Compustat Merged (CCM) links
database. It allows users to specify the type of links (\code{linktype}) and the
primacy of the link (\code{linkprim}).
}
\examples{
\dontrun{
  ccm_links <- download_data_wrds_ccm_links(linktype = "LU", linkprim = "P")
}

}
\seealso{
Other WRDS functions: 
\code{\link{create_wrds_dummy_database}()},
\code{\link{disconnect_connection}()},
\code{\link{download_data_wrds}()},
\code{\link{download_data_wrds_compustat}()},
\code{\link{download_data_wrds_crsp}()},
\code{\link{download_data_wrds_fisd}()},
\code{\link{download_data_wrds_trace_enhanced}()},
\code{\link{get_wrds_connection}()},
\code{\link{set_wrds_credentials}()}
}
\concept{WRDS functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/estimate_fama_macbeth.R
\name{estimate_fama_macbeth}
\alias{estimate_fama_macbeth}
\title{Estimate Fama-MacBeth Regressions}
\usage{
estimate_fama_macbeth(
  data,
  model,
  vcov = "newey-west",
  vcov_options = NULL,
  data_options = NULL,
  detail = FALSE
)
}
\arguments{
\item{data}{A data frame containing the data for the regression. It must
include a column representing the time periods (defaults to \code{date}) and
the variables specified in the \code{model}.}

\item{model}{A character string describing the model to be estimated in
each cross-section (e.g., \code{"ret_excess ~ beta + bm + log_mktcap"}).}

\item{vcov}{A character string indicating the type of standard errors to
compute. Options are \code{"iid"} for independent and identically distributed
errors or \code{"newey-west"} for Newey-West standard errors. Default is
\code{"newey-west"}.}

\item{vcov_options}{A list of additional arguments to be passed to the
\code{NeweyWest()} function when \code{vcov = "newey-west"}. These can include
options such as \code{lag}, which specifies the number of lags to use in the
Newey-West covariance matrix estimation, and \code{prewhite}, which indicates
whether to apply a prewhitening transformation. Default is an empty list.}

\item{data_options}{A list of class \code{tidyfinance_data_options} (created via
\code{\link[=data_options]{data_options()}}) specifying column name mappings. The \code{date} element is
used to specify the date column. Uses \code{\link[=data_options]{data_options()}} default if \code{NULL}:
\code{"date" = "date"}.}

\item{detail}{A logical value indicating whether to return additional
summary statistics. If \code{FALSE} (default), the function returns only the
coefficient estimates. If \code{TRUE}, it returns a list with two elements:
\code{coefficients} (the usual estimates table) and \code{summary_statistics} (a
one-row tibble with the average cross-sectional R-squared and the average
number of observations per cross-section).}
}
\value{
If \code{detail = FALSE} (default), a tibble with columns
\code{factor}, \code{risk_premium}, \code{n} (number of time periods),
\code{standard_error}, and \code{t_statistic}.

If \code{detail = TRUE}, a named list with two elements:
\describe{
\item{coefficients}{The same tibble described above.}
\item{summary_statistics}{A one-row tibble with \code{r_squared} (mean
cross-sectional R-squared) and \code{n_obs} (mean cross-sectional
observation count).}
}
}
\description{
Estimates Fama-MacBeth regressions (Fama and MacBeth, 1973) by first running
cross-sectional regressions for each time period and then aggregating the
results over time to obtain average risk premia and corresponding
t-statistics.
}
\examples{
set.seed(1234)

data <- tibble::tibble(
  date = rep(seq.Date(from = as.Date("2020-01-01"),
                      to = as.Date("2020-12-01"), by = "month"), each = 50),
  permno = rep(1:50, times = 12),
  ret_excess = rnorm(600, 0, 0.1),
  beta = rnorm(600, 1, 0.2),
  bm = rnorm(600, 0.5, 0.1),
  log_mktcap = rnorm(600, 10, 1)
)

estimate_fama_macbeth(data, "ret_excess ~ beta + bm + log_mktcap")
estimate_fama_macbeth(
  data,
  "ret_excess ~ beta + bm + log_mktcap",
  vcov = "iid"
)
estimate_fama_macbeth(
  data,
  "ret_excess ~ beta + bm + log_mktcap",
  vcov = "newey-west",
  vcov_options = list(lag = 6, prewhite = FALSE)
)

# Return detailed output including R-squared and observation counts
estimate_fama_macbeth(
  data,
  "ret_excess ~ beta + bm + log_mktcap",
  detail = TRUE
)

# Use different column name for date
data |>
  dplyr::rename(month = date) |>
  estimate_fama_macbeth(
    "ret_excess ~ beta + bm + log_mktcap",
    data_options = data_options(date = "month")
 )

}
\references{
Fama, E. F., & MacBeth, J. D. (1973). Risk, return, and equilibrium:
Empirical tests. \emph{Journal of Political Economy}, 81(3), 607-636.
\doi{10.1086/260061}

Newey, W. K., & West, K. D. (1987). A simple, positive semi-definite,
heteroskedasticity and autocorrelation consistent covariance matrix.
\emph{Econometrica}, 55(3), 703-708.
\doi{10.2307/1913610}
}
\seealso{
Other estimation functions: 
\code{\link{estimate_betas}()},
\code{\link{estimate_model}()}
}
\concept{estimation functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/download_data_stock_prices.R
\name{download_data_stock_prices}
\alias{download_data_stock_prices}
\title{Download Stock Data}
\usage{
download_data_stock_prices(symbols, start_date = NULL, end_date = NULL)
}
\arguments{
\item{symbols}{A character vector of stock symbols to download data for. At
least one symbol must be provided.}

\item{start_date}{Optional. A character string or Date object in "YYYY-MM-DD"
format specifying the start date for the data. If not provided, a one-year
subset of the dataset is returned (see \code{\link[=validate_dates]{validate_dates()}}).}

\item{end_date}{Optional. A character string or Date object in "YYYY-MM-DD"
format specifying the end date for the data. If not provided, a one-year
subset of the dataset is returned (see \code{\link[=validate_dates]{validate_dates()}}).}
}
\value{
A tibble containing the downloaded stock data with columns: symbol,
date, volume, open, low, high, close, and adjusted_close.
}
\description{
Downloads historical stock data from Yahoo Finance for given symbols and date
range.
}
\examples{
\donttest{
  download_data_stock_prices(c("AAPL", "MSFT"))
  download_data_stock_prices("GOOGL", "2021-01-01", "2022-01-01" )
}
}
\seealso{
Other download functions: 
\code{\link{download_data}()},
\code{\link{download_data_constituents}()},
\code{\link{download_data_factors_ff}()},
\code{\link{download_data_factors_q}()},
\code{\link{download_data_fred}()},
\code{\link{download_data_huggingface}()},
\code{\link{download_data_macro_predictors}()},
\code{\link{download_data_osap}()},
\code{\link{download_data_risk_free}()}
}
\concept{download functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/list_supported_types.R
\name{list_supported_types_ff}
\alias{list_supported_types_ff}
\title{List Supported Fama-French Dataset Types}
\usage{
list_supported_types_ff()
}
\value{
A tibble with columns: \code{type} (the type of dataset),
\code{dataset_name} (a descriptive name of the dataset), and \code{domain} (the
domain to which the dataset belongs, always "Fama-French").
}
\description{
Returns a tibble with the supported Fama-French dataset types, including
their names and frequencies (daily, weekly, monthly). Each dataset type
is associated with a specific Fama-French model (e.g., 3 factors, 5
factors). Additionally, it annotates each dataset with the domain
"Fama-French".
}
\seealso{
Other utility functions: 
\code{\link{create_summary_statistics}()},
\code{\link{get_available_huggingface_files}()},
\code{\link{list_supported_indexes}()},
\code{\link{list_supported_types}()},
\code{\link{list_supported_types_ff_legacy}()},
\code{\link{list_supported_types_macro_predictors}()},
\code{\link{list_supported_types_other}()},
\code{\link{list_supported_types_wrds}()},
\code{\link{list_tidy_finance_chapters}()},
\code{\link{open_tidy_finance_website}()},
\code{\link{trim}()},
\code{\link{validate_dates}()},
\code{\link{winsorize}()}
}
\concept{utility functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/check_supported_type.R
\name{check_supported_type}
\alias{check_supported_type}
\title{Check if a Dataset Type is Supported}
\usage{
check_supported_type(type)
}
\arguments{
\item{type}{A character string specifying the dataset type to check.}
}
\value{
Does not return a value; instead, it either passes silently if the
type is supported or stops execution with an error message if the type is
unsupported.
}
\description{
Checks if a given dataset type is supported by verifying against a list of
all supported dataset types from different domains. If the specified type is
not supported, execution stops and an error message listing all supported
types is returned.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/download_data_risk_free.R
\name{download_data_risk_free}
\alias{download_data_risk_free}
\title{Download Risk-Free Rate Data}
\usage{
download_data_risk_free(
  start_date = NULL,
  end_date = NULL,
  frequency = "monthly"
)
}
\arguments{
\item{start_date}{Optional. A character string or Date object in
"YYYY-MM-DD" format specifying the start date for the data. If
not provided, the full dataset is returned.}

\item{end_date}{Optional. A character string or Date object in
"YYYY-MM-DD" format specifying the end date for the data. If not
provided, the full dataset is returned.}

\item{frequency}{A character string, either \code{"monthly"} (default)
or \code{"daily"}, specifying the frequency of the returned data. Daily
data starts in 1954-01-04 because of availability of DTB3, while
monthly data starts in 1934-01-01.}
}
\value{
A tibble with two columns:
\describe{
\item{date}{The date of the observation.}
\item{risk_free}{The risk-free rate for the period.}
}
}
\description{
Downloads pre-processed risk-free rate data from the
\code{tidy-finance/risk-free} dataset on HuggingFace. The dataset is
updated monthly via a scheduled GitHub Actions workflow that splices the
3-Month Treasury Bill Secondary Market Rate (pre-2001) with the 4-Week
Treasury Bill Secondary Market Rate (from 2001 onwards) sourced from FRED.
For monthly data, the monthly TB3MS series is spliced with the daily DTB4WK
series aggregated to month-end. For daily data, the daily DTB3 series is
spliced with the daily DTB4WK series, both at the business-day frequency
provided by FRED.
}
\details{
Both series are quoted as annualised bank discount rates on a
360-day basis. Given an annualised discount rate \code{d} and a T-bill
with \code{n} days to maturity, the holding-period return is
\code{HPR = d * n/360 / (1 - d * n/360)}, which is then converted to the
target period length via \code{(1 + HPR)^(target/source) - 1}.

The series are spliced at 2001-07-01:
\itemize{
\item \strong{Pre-2001}: TB3MS (monthly) or DTB3 (daily), 3-month T-bill
with n = 90. Monthly conversion uses exponent \code{1/3}; daily
conversion uses exponent \code{1/63} (approx. trading days per
quarter).
\item \strong{From 2001}: DTB4WK, 4-week T-bill with n = 28. For monthly
data, the last non-NA observation per calendar month is taken
and the exponent is \code{365/(28*12)}. For daily data, observations
are used as-is and the exponent is \code{1/20} (approx. trading days
per 4-week period).
}

Business-day gaps in the daily series (e.g. holidays) are handled
by forward-filling the most recent available rate.

Monthly data starts in 1934-01-01 (TB3MS). Daily data starts in
1954-01-04 due to the availability of DTB3.
}
\examples{
\donttest{
  download_data_risk_free("2020-01-01", "2020-12-31")
  download_data_risk_free(
    "2020-01-01", "2020-12-31", frequency = "daily"
  )
}
}
\seealso{
Other download functions: 
\code{\link{download_data}()},
\code{\link{download_data_constituents}()},
\code{\link{download_data_factors_ff}()},
\code{\link{download_data_factors_q}()},
\code{\link{download_data_fred}()},
\code{\link{download_data_huggingface}()},
\code{\link{download_data_macro_predictors}()},
\code{\link{download_data_osap}()},
\code{\link{download_data_stock_prices}()}
}
\concept{download functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/estimate_betas.R
\name{estimate_betas}
\alias{estimate_betas}
\title{Estimate Rolling Betas}
\usage{
estimate_betas(
  data,
  model,
  lookback,
  min_obs = NULL,
  use_furrr = FALSE,
  data_options = NULL
)
}
\arguments{
\item{data}{A data frame containing the data with a date identifier
(defaults to \code{date}), a stock identifier (defaults to \code{permno}), and other
variables used in the model.}

\item{model}{A character string describing the model to be estimated (e.g.,
\code{"ret_excess ~ mkt_excess + hml + smb"}).}

\item{lookback}{A Period object specifying the number of months, days,
hours, minutes, or seconds to look back when estimating the rolling model.}

\item{min_obs}{An integer specifying the minimum number of observations
required to estimate the model. Defaults to 80\% of \code{lookback}.}

\item{use_furrr}{A logical indicating whether to use the \code{furrr} package
and its parallelization capabilities. Defaults to \code{FALSE}.}

\item{data_options}{A list of class \code{tidyfinance_data_options} (created via
\code{\link[=data_options]{data_options()}}) specifying column name mappings. The \code{id} is used to
specify the entity (i.e., firm), and the \code{date} element is used to specify
the date column. Uses \code{\link[=data_options]{data_options()}} default if \code{NULL}:
\code{"id" = "permno"} and \code{"date" = "date"}.}
}
\value{
A data frame with the estimated betas for each time period.
}
\description{
Estimates rolling betas for a given model using the provided data. It
supports parallel processing for faster computation using the \code{furrr}
package.
}
\examples{
# Estimate monthly betas using monthly return data
set.seed(1234)
data_monthly <- tibble::tibble(
  date = rep(seq.Date(from = as.Date("2020-01-01"),
                      to = as.Date("2020-12-01"), by = "month"), each = 50),
  permno = rep(1:50, times = 12),
  ret_excess = rnorm(600, 0, 0.1),
  mkt_excess = rnorm(600, 0, 0.1),
  smb = rnorm(600, 0, 0.1),
  hml = rnorm(600, 0, 0.1),
)

estimate_betas(data_monthly,  "ret_excess ~ mkt_excess", months(3))
estimate_betas(
  data_monthly,
  "ret_excess ~ mkt_excess + smb + hml",
  months(6)
)

data_monthly |>
  dplyr::rename(id = permno) |>
  estimate_betas("ret_excess ~ mkt_excess", months(3),
                 data_options = data_options(id = "id"))

# Estimate monthly betas using daily return data and parallelization
data_daily <- tibble::tibble(
  date = rep(seq.Date(from = as.Date("2020-01-01"),
                      to = as.Date("2020-12-31"), by = "day"), each = 50),
  permno = rep(1:50, times = 366),
  ret_excess = rnorm(18300, 0, 0.02),
  mkt_excess = rnorm(18300, 0, 0.02),
  smb = rnorm(18300, 0, 0.02),
  hml = rnorm(18300, 0, 0.02),
)

data_daily <- data_daily |>
  dplyr::mutate(date = lubridate::floor_date(date, "month"))

# Change settings via future::plan(strategy = "multisession", workers = 4)
estimate_betas(
  data_daily,
  "ret_excess ~ mkt_excess",
  lubridate::days(90),
  use_furrr = TRUE
)

}
\seealso{
Other estimation functions: 
\code{\link{estimate_fama_macbeth}()},
\code{\link{estimate_model}()}
}
\concept{estimation functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/assign_portfolio.R
\name{assign_portfolio}
\alias{assign_portfolio}
\title{Assign Portfolios Based on Sorting Variable}
\usage{
assign_portfolio(
  data,
  sorting_variable,
  breakpoint_options = NULL,
  breakpoint_function = compute_breakpoints,
  data_options = NULL
)
}
\arguments{
\item{data}{A data frame containing the dataset for portfolio assignment.}

\item{sorting_variable}{A string specifying the column name in \code{data} to be
used for sorting and determining portfolio assignments based on the
breakpoints.}

\item{breakpoint_options}{An optional named list of arguments passed to
\code{breakpoint_function}.}

\item{breakpoint_function}{A function to compute breakpoints. The default
is set to \code{\link[=compute_breakpoints]{compute_breakpoints()}}.}

\item{data_options}{A list of class \code{tidyfinance_data_options} (created via
\code{\link[=data_options]{data_options()}}) specifying column name mappings. Passed through to
\code{breakpoint_function}. When using the default \code{\link[=compute_breakpoints]{compute_breakpoints()}},
the \code{exchange} element is used to specify the exchange column, and
\code{mktcap_lag} is used to specify the market capitalization column. Uses
\code{\link[=data_options]{data_options()}} default if \code{NULL}: \code{"exchange" = "exchange"} and
\code{"mktcap_lag" = "mktcap_lag"}.}
}
\value{
A vector of integer portfolio assignments for each row in the
input \code{data}.
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}

Assigns data points to portfolios based on a specified sorting variable and
the selected function to compute breakpoints. Users can specify a function
to compute breakpoints. The function must take \code{data} and
\code{sorting_variable} as the first two arguments. Additional arguments are
passed with a named list \code{\link[=breakpoint_options]{breakpoint_options()}}. The function needs to
return an ascending vector of breakpoints. By default, breakpoints are
computed with \code{\link[=compute_breakpoints]{compute_breakpoints()}}. The default column names can be
modified using \code{\link[=data_options]{data_options()}}.
}
\examples{
set.seed(42)
data <- data.frame(
  id = 1:100,
  exchange = sample(c("NYSE", "NASDAQ"), 100, replace = TRUE),
  market_cap = 1:100
)

assign_portfolio(data, "market_cap", breakpoint_options(n_portfolios = 5))

assign_portfolio(
  data,
  "market_cap",
  breakpoint_options(
    percentiles = c(0.2, 0.4, 0.6, 0.8),
    breakpoints_exchanges = c("NYSE")
  )
)

}
\seealso{
Other portfolio functions: 
\code{\link{breakpoint_options}()},
\code{\link{compute_breakpoints}()},
\code{\link{compute_long_short_returns}()},
\code{\link{compute_portfolio_returns}()},
\code{\link{data_options}()},
\code{\link{filter_options}()},
\code{\link{filter_sorting_data}()},
\code{\link{implement_portfolio_sort}()},
\code{\link{portfolio_sort_options}()}
}
\concept{portfolio functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/open_tidy_finance_website.R
\name{open_tidy_finance_website}
\alias{open_tidy_finance_website}
\title{Open Tidy Finance Website or Specific Chapter in Browser}
\usage{
open_tidy_finance_website(chapter = NULL)
}
\arguments{
\item{chapter}{An optional character string specifying the chapter to open.
If \code{NULL} (the default), the function opens the main page of Tidy Finance
with R. If a chapter name is provided (e.g., "beta-estimation"), the
function opens the corresponding chapter's page (e.g.,
"beta-estimation.html"). If the chapter name does not exist, then the
function opens the main page.}
}
\value{
Invisible \code{NULL}. The function is called for its side effect of
opening a web page.
}
\description{
Opens the main Tidy Finance website or a specific chapter within the site in
the user's default web browser. If a chapter is specified, the function
constructs the URL to access the chapter directly.
}
\examples{
open_tidy_finance_website()
open_tidy_finance_website("beta-estimation")

}
\seealso{
Other utility functions: 
\code{\link{create_summary_statistics}()},
\code{\link{get_available_huggingface_files}()},
\code{\link{list_supported_indexes}()},
\code{\link{list_supported_types}()},
\code{\link{list_supported_types_ff}()},
\code{\link{list_supported_types_ff_legacy}()},
\code{\link{list_supported_types_macro_predictors}()},
\code{\link{list_supported_types_other}()},
\code{\link{list_supported_types_wrds}()},
\code{\link{list_tidy_finance_chapters}()},
\code{\link{trim}()},
\code{\link{validate_dates}()},
\code{\link{winsorize}()}
}
\concept{utility functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/download_data_osap.R
\name{download_data_osap}
\alias{download_data_osap}
\title{Download and Process Open Source Asset Pricing Data}
\usage{
download_data_osap(
  start_date = NULL,
  end_date = NULL,
  sheet_id = "1JyhcF5PRKHcputlioxlu5j5GyLo4JYyY"
)
}
\arguments{
\item{start_date}{Optional. A character string or Date object in
"YYYY-MM-DD" format specifying the start date for the data. If
not provided, the full dataset is returned.}

\item{end_date}{Optional. A character string or Date object in
"YYYY-MM-DD" format specifying the end date for the data. If not
provided, the full dataset is returned.}

\item{sheet_id}{A character string representing the Google Sheet ID
from which to download the data. Default is
\code{"1JyhcF5PRKHcputlioxlu5j5GyLo4JYyY"}.}
}
\value{
A tibble containing the processed data. The column names
are converted to snake_case, and the data is filtered by the
specified date range if \code{start_date} and \code{end_date} are provided.
}
\description{
Downloads the data from
\href{https://www.openassetpricing.com/data/}{Open Source Asset Pricing}
from Google Sheets using a specified sheet ID, processes the data
by converting column names to snake_case, and optionally filters
the data based on a provided date range.
}
\examples{
\donttest{
  osap <- download_data_osap(
    start_date = "2020-01-01", end_date = "2020-06-30"
  )
}
}
\seealso{
Other download functions: 
\code{\link{download_data}()},
\code{\link{download_data_constituents}()},
\code{\link{download_data_factors_ff}()},
\code{\link{download_data_factors_q}()},
\code{\link{download_data_fred}()},
\code{\link{download_data_huggingface}()},
\code{\link{download_data_macro_predictors}()},
\code{\link{download_data_risk_free}()},
\code{\link{download_data_stock_prices}()}
}
\concept{download functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compute_portfolio_returns.R
\name{compute_portfolio_returns}
\alias{compute_portfolio_returns}
\title{Compute Portfolio Returns}
\usage{
compute_portfolio_returns(
  data,
  sorting_variables,
  sorting_method,
  rebalancing_month = NULL,
  breakpoint_options_main,
  breakpoint_options_secondary = NULL,
  breakpoint_function_main = compute_breakpoints,
  breakpoint_function_secondary = compute_breakpoints,
  min_portfolio_size = 1L,
  cap_weight = 0.8,
  data_options = NULL,
  quiet = FALSE
)
}
\arguments{
\item{data}{A data frame containing the dataset for portfolio assignment
and return computation. The panel data must include individual stock
identifiers and the time point. It must contain columns for the sorting
variables and excess returns. Additionally, lagged market capitalization
is required for value-weighted returns (see parameter \code{data_options}).}

\item{sorting_variables}{A character vector specifying the column names in
\code{data} to be used for sorting and determining portfolio assignments. For
univariate sorts, provide a single variable. For bivariate sorts, provide
two variables, where the first string refers to the main variable and the
second string refers to the secondary ("control") variable.}

\item{sorting_method}{A string specifying the sorting method to be used.
Possible values are:
\itemize{
\item \code{"univariate"}: For a single sorting variable.
\item \code{"bivariate-dependent"}: For two sorting variables, where the
main sort depends on the secondary variable.
\item \code{"bivariate-independent"}: For two independent sorting variables.
}
For bivariate sorts, the portfolio returns are averaged over the
controlling sorting variable (i.e., the second sorting variable), and only
portfolio returns for the main sorting variable (given as the first
element of \code{sorting_variables}) are returned.}

\item{rebalancing_month}{An integer between 1 and 12 specifying the month
in which to form portfolios that are held constant for one year. For
example, setting it to \code{7} creates portfolios in July that are held
constant until June of the following year. The default \code{NULL} corresponds
to periodic rebalancing.}

\item{breakpoint_options_main}{A named list of \code{\link[=breakpoint_options]{breakpoint_options()}}
passed to \code{breakpoint_function_main} for the main sorting variable.}

\item{breakpoint_options_secondary}{An optional named list of
\code{\link[=breakpoint_options]{breakpoint_options()}} passed to \code{breakpoint_function_secondary}.}

\item{breakpoint_function_main}{A function to compute the breakpoints for
the main sorting variable. The default is set to \code{\link[=compute_breakpoints]{compute_breakpoints()}}.}

\item{breakpoint_function_secondary}{A function to compute the breakpoints
for the secondary sorting variable. The default is set to
\code{\link[=compute_breakpoints]{compute_breakpoints()}}.}

\item{min_portfolio_size}{An integer specifying the minimum number of
firms required in the reported portfolio cross-section on a given date
(default \code{1L}, i.e. at least one observation per reported portfolio).
For both univariate and bivariate sorts the threshold is applied to the
firm count per \verb{(portfolio, date)} of the reported cross-section: for
univariate sorts that is firms per portfolio-date; for bivariate sorts
that is firms per main-portfolio-date summed across the secondary
buckets. Cross-sections below the threshold have their returns set to
\code{NA}. A typical value is \code{5L} (the Fama-French convention). Set to \code{0L}
to deactivate the check entirely.}

\item{cap_weight}{A numeric value between 0 and 1 specifying the percentile
at which market capitalization is capped per date when computing capped
value-weighted excess returns (i.e., \code{ret_excess_vw_capped}). Defaults to
\code{0.8}.}

\item{data_options}{A list of class \code{tidyfinance_data_options} (created via
\code{\link[=data_options]{data_options()}}) specifying column name mappings. The \code{id} element is
used to specify the entity (i.e., firm), \code{date} is used to specify the
date column, \code{ret_excess} is used to specify the excess return column, and
\code{mktcap_lag} is used to specify the market capitalization. Uses
\code{\link[=data_options]{data_options()}} default if \code{NULL}: \code{"id" = "permno"},
\code{"date" = "date"}, \code{"ret_excess" = "ret_excess"}, and
\code{"mktcap_lag" = "mktcap_lag"}.}

\item{quiet}{A logical value indicating whether to suppress informational
messages about missing values in the output panel (default is \code{FALSE}).}
}
\value{
A data frame with computed portfolio returns as a complete panel
(all portfolio-date combinations), containing the following columns:
\itemize{
\item \code{portfolio}: The portfolio identifier.
\item \code{date}: The date of the portfolio return.
\item \code{ret_excess_vw}: The value-weighted excess return of the
portfolio (only computed if \code{data} contains a lagged market
capitalization column specified by \code{data_options()}, which defaults
to \code{mktcap_lag}). \code{NA} if insufficient observations for that
portfolio-date.
\item \code{ret_excess_ew}: The equal-weighted excess return of the
portfolio. \code{NA} if insufficient observations for that portfolio-date.
\item \code{ret_excess_vw_capped}: The capped value-weighted excess return
of the portfolio (only computed if \code{data} contains a lagged market
capitalization column specified by \code{data_options()}, which defaults
to \code{mktcap_lag}). Weights are computed using market capitalization
capped at the \code{cap_weight} percentile per date. \code{NA} if insufficient
observations for that portfolio-date.
}
}
\description{
Computes individual portfolio returns based on specified sorting variables
and sorting methods. The portfolios can be rebalanced every period or on an
annual frequency by specifying a rebalancing month, which is only applicable
at a monthly return frequency. The function supports univariate and
bivariate sorts, with the latter supporting dependent and independent
sorting methods.
}
\details{
The function checks for consistency in the provided arguments. For
univariate sorts, a single sorting variable and a corresponding number of
portfolios must be provided. For bivariate sorts, two sorting variables and
two corresponding numbers of portfolios (or percentiles) are required. The
sorting method determines how portfolios are assigned and how returns are
computed. The function handles missing and extreme values appropriately
based on the specified sorting method and rebalancing frequency.
}
\note{
Ensure that \code{data} contains all required columns: the specified
sorting variables and excess returns (see options and defaults set in
\code{data_options}). The function will stop and throw an error if any required
columns are missing.
}
\examples{
set.seed(42)
# Univariate sorting with periodic rebalancing
data <- data.frame(
  permno = 1:500,
  date = rep(
    seq.Date(
      from = as.Date("2020-01-01"),
      by = "month",
      length.out = 100
    ),
    each = 10
  ),
  mktcap_lag = runif(500, 100, 1000),
  ret_excess = rnorm(500),
  size = runif(500, 50, 150)
)

compute_portfolio_returns(
  data, "size", "univariate",
  breakpoint_options_main = breakpoint_options(n_portfolios = 5)
)

# Bivariate dependent sorting with annual rebalancing
compute_portfolio_returns(
  data, c("size", "mktcap_lag"), "bivariate-dependent", 7,
  breakpoint_options_main = breakpoint_options(n_portfolios = 5),
  breakpoint_options_secondary = breakpoint_options(n_portfolios = 3),
)

}
\seealso{
Other portfolio functions: 
\code{\link{assign_portfolio}()},
\code{\link{breakpoint_options}()},
\code{\link{compute_breakpoints}()},
\code{\link{compute_long_short_returns}()},
\code{\link{data_options}()},
\code{\link{filter_options}()},
\code{\link{filter_sorting_data}()},
\code{\link{implement_portfolio_sort}()},
\code{\link{portfolio_sort_options}()}
}
\concept{portfolio functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/create_summary_statistics.R
\name{create_summary_statistics}
\alias{create_summary_statistics}
\title{Create Summary Statistics for Specified Variables}
\usage{
create_summary_statistics(
  data,
  ...,
  by = NULL,
  detail = FALSE,
  drop_na = FALSE
)
}
\arguments{
\item{data}{A data frame containing the variables to be summarized.}

\item{...}{Comma-separated list of unquoted variable names in the data frame
to summarize. These variables must be either numeric, integer, or logical.}

\item{by}{An optional unquoted variable name to group the data before
summarizing. If \code{NULL} (the default), summary statistics are computed
across all observations.}

\item{detail}{A logical flag indicating whether to compute detailed summary
statistics, including additional quantiles. Defaults to \code{FALSE}, which
computes basic statistics (n, mean, sd, min, median, max). When \code{TRUE},
additional quantiles (1\%, 5\%, 10\%, 25\%, 75\%, 90\%, 95\%, 99\%) are computed.}

\item{drop_na}{A logical flag indicating whether to drop missing values for
each variable (default is \code{FALSE}).}
}
\value{
A tibble with summary statistics for each selected variable. If \code{by}
is specified, the output includes the grouping variable as well. Each row
represents a variable (and a group if \code{by} is used), and each column
contains the computed statistics.
}
\description{
Computes a set of summary statistics for numeric and integer variables in a
data frame. It allows users to select specific variables for
summarization and can calculate statistics for the whole dataset or within
groups specified by the \code{by} argument. Additional detail levels for quantiles
can be included.
}
\details{
The function first checks that all specified variables are of type
numeric, integer, or logical. If any variables do not meet this criterion,
the function stops and returns an error message indicating the non-conforming
variables.

The basic set of summary statistics includes the count of non-NA values (n),
mean, standard deviation (sd), minimum (min), median (q50), and maximum
(max). If \code{detail} is \code{TRUE}, the function also computes the 1st, 5th, 10th,
25th, 75th, 90th, 95th, and 99th percentiles.

Summary statistics are computed for each variable specified in \code{...}. If a
\code{by} variable is provided, statistics are computed within each level of the
\code{by} variable.
}
\examples{
data <- data.frame(
  ret = c(0.01, -0.02, 0.03, NA, 0.005),
  size = c(100, 200, 150, 300, 250),
  group = c("A", "A", "B", "B", "A")
)

# Basic summary across all observations
create_summary_statistics(data, ret, size)

# Grouped summary
create_summary_statistics(data, ret, size, by = group)

# Detailed quantiles
create_summary_statistics(data, ret, detail = TRUE)
}
\seealso{
Other utility functions: 
\code{\link{get_available_huggingface_files}()},
\code{\link{list_supported_indexes}()},
\code{\link{list_supported_types}()},
\code{\link{list_supported_types_ff}()},
\code{\link{list_supported_types_ff_legacy}()},
\code{\link{list_supported_types_macro_predictors}()},
\code{\link{list_supported_types_other}()},
\code{\link{list_supported_types_wrds}()},
\code{\link{list_tidy_finance_chapters}()},
\code{\link{open_tidy_finance_website}()},
\code{\link{trim}()},
\code{\link{validate_dates}()},
\code{\link{winsorize}()}
}
\concept{utility functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/list_supported_indexes.R
\name{list_supported_indexes}
\alias{list_supported_indexes}
\title{List Supported Indexes}
\usage{
list_supported_indexes()
}
\value{
A tibble with three columns:
\describe{
\item{index}{The name of the financial index (e.g., "DAX", "S&P 500").}
\item{url}{The URL to the CSV file containing the holdings data for the
index.}
\item{skip}{The number of lines to skip when reading the CSV file.}
}
}
\description{
Returns a tibble containing information about supported financial indexes.
Each index is associated with a URL that points to a CSV file containing the
holdings of the index. Additionally, each index has a corresponding \code{skip}
value, which indicates the number of lines to skip when reading the CSV file.
}
\examples{
supported_indexes <- list_supported_indexes()
print(supported_indexes)

}
\seealso{
Other utility functions: 
\code{\link{create_summary_statistics}()},
\code{\link{get_available_huggingface_files}()},
\code{\link{list_supported_types}()},
\code{\link{list_supported_types_ff}()},
\code{\link{list_supported_types_ff_legacy}()},
\code{\link{list_supported_types_macro_predictors}()},
\code{\link{list_supported_types_other}()},
\code{\link{list_supported_types_wrds}()},
\code{\link{list_tidy_finance_chapters}()},
\code{\link{open_tidy_finance_website}()},
\code{\link{trim}()},
\code{\link{validate_dates}()},
\code{\link{winsorize}()}
}
\concept{utility functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/download_data_huggingface.R
\name{download_data_huggingface}
\alias{download_data_huggingface}
\title{Download data from a Hugging Face dataset}
\usage{
download_data_huggingface(
  dataset = NULL,
  start_date = "2007-06-27",
  end_date = "2007-07-27",
  type = deprecated(),
  ...
)
}
\arguments{
\item{dataset}{Character(1). The dataset to download. Supported values are
\code{"high_frequency_sp500"} and \code{"factor_library"}.}

\item{start_date}{Date or character. Start date (inclusive) in
\code{"YYYY-MM-DD"} format. Only used for \code{"high_frequency_sp500"}.}

\item{end_date}{Date or character. End date (inclusive) in \code{"YYYY-MM-DD"}
format. Only used for \code{"high_frequency_sp500"}.}

\item{type}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use \code{dataset} instead.}

\item{...}{For \code{dataset = "factor_library"}: named arguments used to filter
the portfolio grid. Each argument takes the form \code{column = value}, where
\code{value} may be a vector to match multiple levels. Optionally pass
\code{fill_all = TRUE} to leave unspecified columns unrestricted (default:
\code{FALSE}, i.e. unspecified columns are fixed at the defaults listed below).
Passing \code{NULL} for any parameter removes that filter entirely, returning
all values for that column (e.g., \code{min_size_quantile = NULL} includes all
size groups). Passing an unrecognised column name raises an error listing
the supported names. Ignored when \code{dataset != "factor_library"}. See the
Details section for supported columns and their defaults.}
}
\value{
A tibble with the downloaded data. For \code{"high_frequency_sp500"},
contains 5-second aggregated orderbook snapshots filtered to the requested
date range. For \code{"factor_library"}, contains portfolio return data joined
with the full grid metadata for the matched portfolio IDs.
}
\description{
Downloads data from a supported Hugging Face dataset. For
\code{"high_frequency_sp500"}, parquet files are filtered by date range and
row-bound. For \code{"factor_library"}, portfolio characteristics are selected via
\code{filter_factor_library_grid()} and the matching return data is downloaded.
}
\details{
\strong{Note on \code{dataset = "factor_library"} defaults:} The defaults below reflect
one common portfolio construction choice, but may not suit every research
question. Always verify that the selected combination matches your intended
design.

Supported columns and their defaults for \code{...}:
\itemize{
\item \code{sorting_variable}: \strong{Required.} The firm characteristic used
to sort stocks into portfolios (e.g., \code{"me"} for market equity,
\code{"bm"} for book-to-market). No default is applied.
\item \code{min_size_quantile} (defaults to \code{0.2}): Fraction of the smallest
stocks (by market cap) excluded from the portfolio universe. \code{0.2}
drops the bottom 20\%.
\item \code{exclude_financials} (defaults to \code{FALSE}): Whether to drop
financial-sector stocks (SIC 6000-6999) from the universe.
\item \code{exclude_utilities} (default: \code{FALSE}): Whether to drop
utility-sector stocks (SIC 4900-4999) from the universe.
\item \code{exclude_negative_earnings} (defaults to \code{FALSE}): Whether to
drop firms with negative earnings before sorting.
\item \code{sorting_variable_lag} (defaults to \code{"6m"}): Lag applied to the
sorting variable before portfolio assignment (e.g., \code{"6m"} = 6-month
lag).
\item \code{rebalancing} (defaults to \code{"monthly"}): How frequently portfolios
are reformed: \code{"monthly"} or \code{"annual"}.
\item \code{n_portfolios_main} (defaults to \code{10}): Number of quantile groups
(e.g., \code{10} for decile portfolios).
\item \code{sorting_method} (defaults to \code{"univariate"}): Whether portfolios
are formed on a single sort (\code{"univariate"}) or a sequential double
sort (\code{"sequential"}).
\item \code{n_portfolios_secondary} (defaults to \code{NULL}): Number of groups
for the secondary sort variable.
Required when \code{sorting_method} is not \code{"univariate"}.
\item \code{breakpoints_exchanges} (defaults to: \code{"NYSE"}): Exchange(s) used
to compute breakpoints. \code{"NYSE"} uses only NYSE-listed stocks to
define quantile cutoffs (the conventional Fama-French approach).
\item \code{breakpoints_min_size_threshold} (defaults to \code{NULL}): Minimum
market-cap threshold (in USD) applied when computing breakpoints.
\code{NULL} means no minimum-size screen is applied.
\item \code{weighting_scheme} (defaults to \code{"VW"}): Return weighting within
portfolios: \code{"VW"} for value-weighted or \code{"EW"} for equal-weighted.
}
}
\examples{
\dontrun{
  download_data_huggingface(
    "high_frequency_sp500", "2007-07-26", "2007-07-27"
  )
  download_data_huggingface(
    "factor_library",
    sorting_variable = "52w",
    rebalancing = "annual"
  )
  download_data_huggingface(
    "factor_library", sorting_variable = "ag", fill_all = TRUE
  )
}
}
\seealso{
Other download functions: 
\code{\link{download_data}()},
\code{\link{download_data_constituents}()},
\code{\link{download_data_factors_ff}()},
\code{\link{download_data_factors_q}()},
\code{\link{download_data_fred}()},
\code{\link{download_data_macro_predictors}()},
\code{\link{download_data_osap}()},
\code{\link{download_data_risk_free}()},
\code{\link{download_data_stock_prices}()}
}
\concept{download functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/filter_sorting_data.R
\name{filter_sorting_data}
\alias{filter_sorting_data}
\title{Filter Sorting Data}
\usage{
filter_sorting_data(
  data,
  filter_options = NULL,
  data_options = NULL,
  quiet = FALSE
)
}
\arguments{
\item{data}{A data frame containing the stock-level panel data to be
filtered.}

\item{filter_options}{A list of class \code{tidyfinance_filter_options} created
by \code{\link[=filter_options]{filter_options()}}. If \code{NULL} (the default), the defaults from
\code{\link[=filter_options]{filter_options()}} are used (i.e., no filters are applied). The arguments
accepted by \code{\link[=filter_options]{filter_options()}} include
\itemize{
\item \code{exclude_financials} A logical indicating whether to exclude
financial firms (SIC codes 6000–6799). Defaults to \code{FALSE}.
\item \code{exclude_utilities} A logical indicating whether to exclude
utility firms (SIC codes 4900–4999). Defaults to \code{FALSE}.
\item \code{min_stock_price} A single positive numeric specifying the
minimum stock price required to include an observation. \code{NULL} (the
default) applies no price filter.
\item \code{min_size_quantile} A single numeric strictly between 0 and 1
specifying the minimum cross-sectional size quantile (based on lagged
market cap) required to include an observation. \code{NULL} (the default)
applies no size quantile filter. The cutoff is computed from NYSE
stocks only; the \code{exchange} column (mapped via \code{\link[=data_options]{data_options()}})
must be present in the data or an error is raised.
\item \code{min_listing_age} A single non-negative integer or numeric
specifying the minimum number of months a stock must have been listed
in CRSP. \code{NULL} (the default) applies no listing age filter.
\item \code{exclude_negative_book_equity} A logical indicating whether to
exclude observations with non-positive book equity. Defaults to
\code{FALSE}.
\item \code{exclude_negative_earnings} A logical indicating whether to
exclude observations with non-positive earnings. Defaults to \code{FALSE}.
}}

\item{data_options}{A list of class \code{tidyfinance_data_options} (created via
\code{\link[=data_options]{data_options()}}) specifying column name mappings. The \code{siccd} element is
used to specify the SIC code column, \code{price} is used to specify the
(adjusted) price column, \code{mktcap_lag} is used to specify the market
capitalization column, \code{date} is used to specify the date column,
\code{listing_age} is used to specify the listing age column, \code{be} is used to
specify the book equity column, and \code{earnings} is used to specify the
earnings column. Uses \code{\link[=data_options]{data_options()}} default if \code{NULL}:
\code{"siccd" = "siccd"}, \code{"price" = "prc_adj"}, \code{"exchange" = "exchange"}
\code{"mktcap_lag" = "mktcap_lag"}, \code{"date" = "date"},
\code{"listing_age" = "listing_age"}, \code{"be" = "be"}, and \code{"earnings" = "ib"}.}

\item{quiet}{A logical indicating whether informational messages should be
suppressed. Defaults to \code{FALSE}.}
}
\value{
The filtered data frame, preserving the class and structure of the
input.
}
\description{
Applies sample construction filters to a data frame before portfolio
sorting. Filters are applied in a fixed order: financials exclusion,
utilities exclusion, minimum stock price, minimum size quantile, minimum
listing age, positive book equity, and positive earnings. An informational
message is emitted for each filter that actually removes at least one
observation.
}
\examples{
data <- data.frame(
  permno = 1:5,
  date = as.Date("2020-01-01"),
  siccd = c(6100, 2000, 4950, 3000, 6500),
  prc_adj = c(5, 0.5, 15, 20, 10)
)

data |>
  filter_sorting_data(
    filter_options = filter_options(
      exclude_financials = TRUE,
      min_stock_price = 1
    )
  )

}
\seealso{
Other portfolio functions: 
\code{\link{assign_portfolio}()},
\code{\link{breakpoint_options}()},
\code{\link{compute_breakpoints}()},
\code{\link{compute_long_short_returns}()},
\code{\link{compute_portfolio_returns}()},
\code{\link{data_options}()},
\code{\link{filter_options}()},
\code{\link{implement_portfolio_sort}()},
\code{\link{portfolio_sort_options}()}
}
\concept{portfolio functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/download_data_wrds_crsp.R
\name{download_data_wrds_crsp}
\alias{download_data_wrds_crsp}
\title{Download Data from WRDS CRSP}
\usage{
download_data_wrds_crsp(
  dataset = NULL,
  start_date = NULL,
  end_date = NULL,
  type = deprecated(),
  batch_size = 500,
  version = "v2",
  additional_columns = NULL,
  add_ccm_links = FALSE,
  adjust_volume = FALSE
)
}
\arguments{
\item{dataset}{A string specifying the CRSP dataset to download:
"crsp_monthly" or "crsp_daily".}

\item{start_date}{Optional. A character string or Date object in
"YYYY-MM-DD" format specifying the start date for the data. If not
provided, a subset of the dataset is returned.}

\item{end_date}{Optional. A character string or Date object in
"YYYY-MM-DD" format specifying the end date for the data. If not
provided, a subset of the dataset is returned.}

\item{type}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use \code{dataset} instead.}

\item{batch_size}{An optional integer specifying the batch size for
processing daily data, with a default of 500.}

\item{version}{An optional character specifying which CRSP version to
use. "v2" (the default) uses the updated second version of CRSP, and
"v1" downloads the legacy version of CRSP.}

\item{additional_columns}{Additional columns from the CRSP monthly or
daily data as a character vector.}

\item{add_ccm_links}{A logical indicating whether CRSP-Compustat links
should be added automatically using \code{\link[=download_data_wrds_ccm_links]{download_data_wrds_ccm_links()}}.}

\item{adjust_volume}{A logical indicating whether daily CRSP trading
volume data should be adjusted according to Gao & Ritter (2010).}
}
\value{
A data frame containing CRSP stock returns, adjusted for
delistings, along with calculated market capitalization and excess returns
over the risk-free rate. The structure of the returned data frame depends
on the selected dataset.
}
\description{
Downloads and processes stock return data from the CRSP database for a
specified period. Users can choose between monthly and daily datasets.
The function also adjusts returns for delisting and calculates market
capitalization and excess returns over the risk-free rate.
}
\examples{
\dontrun{
crsp_monthly <- download_data_wrds_crsp(
  "crsp_monthly",
  "2020-11-01",
  "2020-12-31"
)
crsp_daily <- download_data_wrds_crsp(
  "crsp_daily",
  "2020-12-01",
  "2020-12-31"
)

# Add additional columns
download_data_wrds_crsp(
  "crsp_monthly",
  "2020-11-01",
  "2020-12-31",
  additional_columns = c("mthvol", "mthvolflg")
)
}
}
\references{
Gao, X., & Ritter, J. R. (2010). The marketing of seasoned equity
offerings. \emph{Journal of Financial Economics}, 97(1), 33-52.
\doi{10.1016/j.jfineco.2010.03.007}
}
\seealso{
Other WRDS functions: 
\code{\link{create_wrds_dummy_database}()},
\code{\link{disconnect_connection}()},
\code{\link{download_data_wrds}()},
\code{\link{download_data_wrds_ccm_links}()},
\code{\link{download_data_wrds_compustat}()},
\code{\link{download_data_wrds_fisd}()},
\code{\link{download_data_wrds_trace_enhanced}()},
\code{\link{get_wrds_connection}()},
\code{\link{set_wrds_credentials}()}
}
\concept{WRDS functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/winsorize.R
\name{winsorize}
\alias{winsorize}
\title{Winsorize a Numeric Vector}
\usage{
winsorize(x, cut)
}
\arguments{
\item{x}{A numeric vector to be winsorized.}

\item{cut}{The proportion of data to be winsorized from both ends of the
distribution. For example, a \code{cut} of 0.05 will winsorize the lowest and
highest 5\% of the data. Must be in [0, 0.5].}
}
\value{
A numeric vector with the extreme values replaced by the
corresponding quantile values.
}
\description{
Replaces the values in a numeric vector that are beyond the specified
quantiles with the boundary values of those quantiles. This is done for both
tails of the distribution based on the \code{cut} parameter.
}
\examples{
set.seed(123)
data <- rnorm(100)
winsorized_data <- winsorize(data, 0.05)

}
\seealso{
Other utility functions: 
\code{\link{create_summary_statistics}()},
\code{\link{get_available_huggingface_files}()},
\code{\link{list_supported_indexes}()},
\code{\link{list_supported_types}()},
\code{\link{list_supported_types_ff}()},
\code{\link{list_supported_types_ff_legacy}()},
\code{\link{list_supported_types_macro_predictors}()},
\code{\link{list_supported_types_other}()},
\code{\link{list_supported_types_wrds}()},
\code{\link{list_tidy_finance_chapters}()},
\code{\link{open_tidy_finance_website}()},
\code{\link{trim}()},
\code{\link{validate_dates}()}
}
\concept{utility functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/list_supported_types.R
\name{list_supported_types_q}
\alias{list_supported_types_q}
\title{List Supported Global Q Dataset Types}
\usage{
list_supported_types_q()
}
\value{
A tibble with columns: \code{type} (the type of dataset),
\code{dataset_name} (the file name of the dataset), and \code{domain} (the
domain to which the dataset belongs, always "Global Q").
}
\description{
Returns a tibble with the supported Global Q dataset types, including
their names and frequencies (daily, weekly, weekly week-to-week, monthly,
quarterly, annual). Each dataset type is associated with the Global Q
model, specifically the q5 factors model for the year 2023. Additionally,
it annotates each dataset with the domain "Global Q".
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/download_data_wrds_fisd.R
\name{download_data_wrds_fisd}
\alias{download_data_wrds_fisd}
\title{Download Filtered FISD Data from WRDS}
\usage{
download_data_wrds_fisd(additional_columns = NULL)
}
\arguments{
\item{additional_columns}{Additional columns from the FISD table
as a character vector.}
}
\value{
A data frame containing a subset of FISD data with fields related
to the bond's characteristics and issuer information. This includes
complete CUSIP, maturity date, offering amount, offering date, dated date,
interest frequency, coupon, last interest date, issue ID, issuer ID, and
SIC code of the issuer.
}
\description{
Establishes a connection to the WRDS database to download a filtered subset
of the FISD (Fixed Income Securities Database). The function filters the
\code{fisd_mergedissue} and \code{fisd_mergedissuer} tables based on several criteria
related to the securities, such as security level, bond type, coupon type,
and others, focusing on specific attributes that denote the nature of the
securities. It finally returns a data frame with selected fields from the
\code{fisd_mergedissue} table after joining it with issuer information from the
\code{fisd_mergedissuer} table for issuers domiciled in the USA.
}
\examples{
\dontrun{
fisd <- download_data_wrds_fisd()
fisd_extended <- download_data_wrds_fisd(
  additional_columns = c("asset_backed", "defeased")
)
}
}
\seealso{
Other WRDS functions: 
\code{\link{create_wrds_dummy_database}()},
\code{\link{disconnect_connection}()},
\code{\link{download_data_wrds}()},
\code{\link{download_data_wrds_ccm_links}()},
\code{\link{download_data_wrds_compustat}()},
\code{\link{download_data_wrds_crsp}()},
\code{\link{download_data_wrds_trace_enhanced}()},
\code{\link{get_wrds_connection}()},
\code{\link{set_wrds_credentials}()}
}
\concept{WRDS functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/download_data_huggingface.R
\name{get_available_huggingface_files}
\alias{get_available_huggingface_files}
\title{List Parquet Files in a Hugging Face Dataset}
\usage{
get_available_huggingface_files(organization, dataset)
}
\arguments{
\item{organization}{Character(1). Hugging Face organization or user name.}

\item{dataset}{Character(1). Dataset name under the organization.}
}
\value{
A tibble with columns: \code{path} (character), \code{size} (numeric), and
\code{url} (character).
}
\description{
Query the Hugging Face Datasets API and return a tibble of files with a
\code{.parquet} suffix. The function follows pagination links returned in the
response \code{Link} header and returns path, size, and a resolved URL.
}
\details{
Uses httr2 to perform HTTP requests. Requires internet access and the
dataset to be publicly accessible or accessible with appropriate
authentication.
}
\examples{
\dontrun{
get_available_huggingface_files("voigtstefan", "sp500")
}
}
\seealso{
Other utility functions: 
\code{\link{create_summary_statistics}()},
\code{\link{list_supported_indexes}()},
\code{\link{list_supported_types}()},
\code{\link{list_supported_types_ff}()},
\code{\link{list_supported_types_ff_legacy}()},
\code{\link{list_supported_types_macro_predictors}()},
\code{\link{list_supported_types_other}()},
\code{\link{list_supported_types_wrds}()},
\code{\link{list_tidy_finance_chapters}()},
\code{\link{open_tidy_finance_website}()},
\code{\link{trim}()},
\code{\link{validate_dates}()},
\code{\link{winsorize}()}
}
\concept{utility functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_wrds_connection.R
\name{get_wrds_connection}
\alias{get_wrds_connection}
\title{Establish a Connection to the WRDS Database}
\usage{
get_wrds_connection()
}
\value{
An object of class \code{DBIConnection} representing the connection
to the WRDS database. This object can be used with other DBI-compliant
functions to interact with the database.
}
\description{
Establishes a connection to the Wharton Research Data Services (WRDS)
database using the \code{RPostgres} package. It requires that the \code{RPostgres}
package be installed and that valid WRDS credentials be set as environment
variables.
}
\details{
The function checks if the \code{RPostgres} package is installed
before attempting to establish a connection. It uses the host, dbname,
port, and sslmode as fixed parameters for the connection. Users must
set their WRDS username and password as environment variables
\code{WRDS_USER} and \code{WRDS_PASSWORD}, respectively, before using this
function.
}
\examples{
\dontrun{
  # Before using this function, set your WRDS credentials:
  # Sys.setenv(WRDS_USER = "your_username", WRDS_PASSWORD = "your_password")

  # con <- get_wrds_connection()
  # Use `con` with DBI-compliant functions to interact with the WRDS database
  # Remember to disconnect after use:
  # disconnect_connection(con)
}
}
\seealso{
\code{\link[RPostgres:Postgres]{RPostgres::Postgres()}}, \code{\link[DBI:dbDisconnect]{DBI::dbDisconnect()}} for more
information on managing database connections.

Other WRDS functions: 
\code{\link{create_wrds_dummy_database}()},
\code{\link{disconnect_connection}()},
\code{\link{download_data_wrds}()},
\code{\link{download_data_wrds_ccm_links}()},
\code{\link{download_data_wrds_compustat}()},
\code{\link{download_data_wrds_crsp}()},
\code{\link{download_data_wrds_fisd}()},
\code{\link{download_data_wrds_trace_enhanced}()},
\code{\link{set_wrds_credentials}()}
}
\concept{WRDS functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compute_rolling_value.R
\name{compute_rolling_value}
\alias{compute_rolling_value}
\title{Compute a Rolling Value by Period}
\usage{
compute_rolling_value(
  data,
  .f,
  period = "month",
  periods = 12,
  min_obs = periods,
  data_options = NULL
)
}
\arguments{
\item{data}{A data frame with a date column of class \code{Date}, named according
to \code{data_options$date} (default \code{"date"}).}

\item{.f}{A function applied to each window. Receives a data-frame slice
(complete cases only) and must return a single scalar value.}

\item{period}{A string specifying the period for rolling windows
(e.g., \code{"month"}, \code{"quarter"}, \code{"year"}).}

\item{periods}{Number of periods to include in the rolling window.}

\item{min_obs}{Minimum number of non-missing rows required per window.
Defaults to \code{periods}.}

\item{data_options}{A list of class \code{tidyfinance_data_options} (created via
\code{\link[=data_options]{data_options()}}) specifying column name mappings. The \code{date} element is
used to specify the date column. Uses \code{\link[=data_options]{data_options()}} default if \code{NULL}:
\code{"date" = "date"}.}
}
\value{
A numeric vector aligned with the rows of \code{data}.
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}

Applies an arbitrary summary function over rolling time-period windows.
Each window spans \code{periods} units of \code{period} (e.g., 12 months). Before
calling \code{.f}, rows with any missing values are dropped from the window;
if fewer than \code{min_obs} rows remain, the result is \code{NA_real_} instead.
}
\examples{
library(dplyr)

# Rolling standard deviation
set.seed(42)
df <- tibble(
  date = seq.Date(
    from = as.Date("2020-01-01"),
    by = "month",
    length.out = 24
  ),
  value = rnorm(24)
)

df |>
  mutate(
    rolling_sd = compute_rolling_value(
      pick(everything()),
      .f = ~ sd(.x$value, na.rm = TRUE),
      period = "month",
      periods = 4,
      min_obs = 2
    )
  )

# Rolling last residual from a regression
set.seed(42)
df_reg <- tibble(
  date = seq.Date(
    from = as.Date("2020-01-01"),
    by = "month",
    length.out = 60
  ),
  ret_excess = rnorm(60, 0, 0.05),
  mkt_excess = rnorm(60, 0, 0.04),
  smb = rnorm(60, 0, 0.03),
  hml = rnorm(60, 0, 0.03)
)

df_reg |>
  mutate(
    residual = compute_rolling_value(
      pick(everything()),
      .f = \(x) {
        last(lm(ret_excess ~ mkt_excess + smb + hml, data = x)$residuals)
      },
      period = "month",
      periods = 24,
      min_obs = 12
    )
  )

# Rolling cumulative-return-to-SD ratio
set.seed(42)
df_resid <- tibble(
  date = seq.Date(
    from = as.Date("2020-01-01"),
    by = "month",
    length.out = 24
  ),
  int_roll_residual = rnorm(24, 0, 0.02)
)

df_resid |>
  mutate(
    return_to_sd = compute_rolling_value(
      pick(everything()),
      .f = ~ (prod(1 + .x$int_roll_residual) - 1) / sd(.x$int_roll_residual),
      period = "month",
      periods = 12,
      min_obs = 12
    )
  )

}
\seealso{
Other rolling and lagging functions: 
\code{\link{add_lagged_columns}()},
\code{\link{join_lagged_values}()}
}
\concept{rolling and lagging functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compute_long_short_returns.R
\name{compute_long_short_returns}
\alias{compute_long_short_returns}
\title{Compute Long-Short Returns}
\usage{
compute_long_short_returns(
  data,
  direction = "top_minus_bottom",
  data_options = NULL
)
}
\arguments{
\item{data}{A data frame containing portfolio returns. The data frame must
include columns for the portfolio identifier, date, and return
measurements (as specified in \code{data_options}).}

\item{direction}{A character string specifying the direction of the
long-short return calculation. It can be either \code{"top_minus_bottom"} or
\code{"bottom_minus_top"}. Default is \code{"top_minus_bottom"}. If set to
\code{"bottom_minus_top"}, the return will be computed as (bottom - top).}

\item{data_options}{A list of class \code{tidyfinance_data_options} (created via
\code{\link[=data_options]{data_options()}}) specifying column name mappings. The \code{date} element is
used to specify the date column, the \code{ret_excess} element is used to
specify the excess return column, and \code{portfolio} is used to specify the
assigned portfolio. Uses \code{\link[=data_options]{data_options()}} default if \code{NULL}:
\code{"date" = "date"}, \code{"ret_excess" = "ret_excess"}, and
\code{"portfolio" = "portfolio"}.}
}
\value{
A data frame with columns for date, return measurement types (from
the "ret_measure" column), and the computed long-short returns. The data
frame is arranged by date and pivoted to have return measurement types as
columns with their corresponding long-short returns.
}
\description{
Calculates long-short returns based on the returns of portfolios. The
long-short return is computed as the difference between the returns of the
"top" and "bottom" portfolios. The direction of the calculation can be
adjusted based on whether the return from the "bottom" portfolio is
subtracted from or added to the return from the "top" portfolio.
}
\examples{
set.seed(42)
data <- data.frame(
  permno = 1:100,
  date = rep(
    seq.Date(from = as.Date("2020-01-01"), by = "month", length.out = 100),
    each = 10
  ),
  mktcap_lag = runif(100, 100, 1000),
  ret_excess = rnorm(100),
  size = runif(100, 50, 150)
)

portfolio_returns <- compute_portfolio_returns(
  data, "size", "univariate",
  breakpoint_options_main = breakpoint_options(n_portfolios = 5)
)

compute_long_short_returns(portfolio_returns)

}
\seealso{
Other portfolio functions: 
\code{\link{assign_portfolio}()},
\code{\link{breakpoint_options}()},
\code{\link{compute_breakpoints}()},
\code{\link{compute_portfolio_returns}()},
\code{\link{data_options}()},
\code{\link{filter_options}()},
\code{\link{filter_sorting_data}()},
\code{\link{implement_portfolio_sort}()},
\code{\link{portfolio_sort_options}()}
}
\concept{portfolio functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data_options.R
\name{data_options}
\alias{data_options}
\title{Create Data Options}
\usage{
data_options(
  id = "permno",
  date = "date",
  exchange = "exchange",
  mktcap_lag = "mktcap_lag",
  ret_excess = "ret_excess",
  portfolio = "portfolio",
  siccd = "siccd",
  price = "prc_adj",
  listing_age = "listing_age",
  be = "be",
  earnings = "ib",
  ...
)
}
\arguments{
\item{id}{A character string representing the entity variable (defaults to
\code{"permno"}).}

\item{date}{A character string representing the date variable (defaults to
\code{"date"}).}

\item{exchange}{A character string representing the exchange variable
(defaults to \code{"exchange"}).}

\item{mktcap_lag}{A character string representing the market capitalization
lag variable (defaults to \code{"mktcap_lag"}).}

\item{ret_excess}{A character string representing the excess return variable
(defaults to \code{"ret_excess"}).}

\item{portfolio}{A character string representing the portfolio variable
(defaults to \code{"portfolio"}).}

\item{siccd}{A character string representing the Standard Industrial
Classification code variable (defaults to \code{"siccd"}).}

\item{price}{A character string representing the (adjusted) price variable
(defaults to \code{"prc_adj"}).}

\item{listing_age}{A character string representing the listing age variable
(defaults to \code{"listing_age"}).}

\item{be}{A character string representing the book equity variable (defaults
to \code{"be"}).}

\item{earnings}{A character string representing the earnings variable
(defaults to \code{"ib"}, the Compustat income before extraordinary items
column).}

\item{...}{Additional arguments to be included in the data options list.}
}
\value{
A list of class \code{tidyfinance_data_options} containing the specified
data options.
}
\description{
Creates a list of data options of class \code{tidyfinance_data_options} used for
TidyFinance-related functions. These options map the specific data variables
to the TidyFinance naming conventions, allowing functions to flexibly work
with different datasets by specifying the relevant column names. Additional
options can be passed through \code{...}.
}
\examples{
data_options(
  id = "permno",
  date = "date",
  exchange = "exchange"
)

}
\seealso{
Other portfolio functions: 
\code{\link{assign_portfolio}()},
\code{\link{breakpoint_options}()},
\code{\link{compute_breakpoints}()},
\code{\link{compute_long_short_returns}()},
\code{\link{compute_portfolio_returns}()},
\code{\link{filter_options}()},
\code{\link{filter_sorting_data}()},
\code{\link{implement_portfolio_sort}()},
\code{\link{portfolio_sort_options}()}
}
\concept{portfolio functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/estimate_model.R
\name{estimate_model}
\alias{estimate_model}
\title{Estimate a Linear Model}
\usage{
estimate_model(data, model, min_obs = 1, output = "coefficients")
}
\arguments{
\item{data}{A data frame containing the dependent variable and one or more
independent variables.}

\item{model}{A character that describes the model to be estimated (e.g.,
\code{"ret_excess ~ mkt_excess + hml + smb"}).}

\item{min_obs}{The minimum number of observations required to estimate the
model. Defaults to 1.}

\item{output}{A character vector specifying what to return. Must contain one
or more of \code{"coefficients"} (default), \code{"residuals"}, and \code{"tstats"}. If a
single value is provided, the corresponding object is returned directly.
If multiple values are provided, a named list is returned.}
}
\value{
If \code{output} contains a single value: a data frame of coefficients
or t-statistics, or a numeric vector of residuals. If \code{output} contains
multiple values: a named list with the requested elements. Coefficients
and t-statistics are returned as data frames with column names
corresponding to the model terms. Residuals are returned as a numeric
vector of length \code{nrow(data)} with \code{NA} for rows with missing data or
insufficient observations.
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}

Estimates a linear model specified by one or more independent
variables. It checks for the presence of the specified independent variables
in the dataset and whether the dataset has a sufficient number of
observations. Depending on the \code{output} parameter, it returns the model's
coefficients, t-statistics, residuals, or any combination in a named list.
}
\examples{
set.seed(42)
data <- data.frame(
  ret_excess = rnorm(100),
  mkt_excess = rnorm(100),
  smb = rnorm(100),
  hml = rnorm(100)
)

# Estimate model with a single independent variable
estimate_model(data, "ret_excess ~ mkt_excess")

# Estimate model with multiple independent variables
estimate_model(data, "ret_excess ~ mkt_excess + smb + hml")

# Estimate model without intercept
estimate_model(data, "ret_excess ~ mkt_excess - 1")

# Calculate residuals
estimate_model(data, "ret_excess ~ mkt_excess + smb + hml",
  output = "residuals"
)

# Return t-statistics
estimate_model(data, "ret_excess ~ mkt_excess + smb + hml",
  output = "tstats"
)

# Return coefficients, t-statistics, and residuals
estimate_model(data, "ret_excess ~ mkt_excess + smb + hml",
  output = c("coefficients", "tstats", "residuals")
)

}
\seealso{
Other estimation functions: 
\code{\link{estimate_betas}()},
\code{\link{estimate_fama_macbeth}()}
}
\concept{estimation functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/download_data_fred.R
\name{download_data_fred}
\alias{download_data_fred}
\title{Download and Process Data from FRED}
\usage{
download_data_fred(series, start_date = NULL, end_date = NULL)
}
\arguments{
\item{series}{A character vector specifying the FRED series ID(s) to
download.}

\item{start_date}{Optional. A character string or Date object in
"YYYY-MM-DD" format specifying the start date for the data. If not
provided, the full dataset is returned.}

\item{end_date}{Optional. A character string or Date object in
"YYYY-MM-DD" format specifying the end date for the data. If not
provided, the full dataset is returned.}
}
\value{
A tibble containing the processed data with three columns:
\describe{
\item{date}{The date corresponding to the data point.}
\item{value}{The value of the data series at that date.}
\item{series}{The FRED series ID corresponding to the data.}
}
}
\description{
Downloads a specified data series from the Federal Reserve Economic Data
(FRED) website, processes the data, and returns it as a tibble.
}
\details{
Constructs the URL based on the provided FRED series ID, performs an HTTP
GET request to download the data in CSV format, and processes it to a tidy
tibble format. The resulting tibble includes the date, value, and the series
ID.

This approach is inspired by \code{quantmod::getSymbolsFRED()} which uses a
different wrapper around the same FRED download data site. If you want to
systematically download FRED data via API, please consider using the \code{fredr}
package.
}
\examples{
\donttest{
  download_data_fred("CPIAUCNS")
  download_data_fred(c("GDP", "CPIAUCNS"), "2010-01-01", "2010-12-31")
}

}
\seealso{
Other download functions: 
\code{\link{download_data}()},
\code{\link{download_data_constituents}()},
\code{\link{download_data_factors_ff}()},
\code{\link{download_data_factors_q}()},
\code{\link{download_data_huggingface}()},
\code{\link{download_data_macro_predictors}()},
\code{\link{download_data_osap}()},
\code{\link{download_data_risk_free}()},
\code{\link{download_data_stock_prices}()}
}
\concept{download functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compute_breakpoints.R
\name{compute_breakpoints}
\alias{compute_breakpoints}
\title{Compute Breakpoints Based on Sorting Variable}
\usage{
compute_breakpoints(
  data,
  sorting_variable,
  breakpoint_options,
  data_options = NULL
)
}
\arguments{
\item{data}{A data frame containing the dataset for breakpoint computation.}

\item{sorting_variable}{A character string specifying the column name in
\code{data} to be used for determining breakpoints.}

\item{breakpoint_options}{A named list of \code{\link[=breakpoint_options]{breakpoint_options()}} for the
breakpoints. The arguments include
\itemize{
\item \code{n_portfolios} An optional integer specifying the number of
equally sized portfolios to create. This parameter is mutually
exclusive with \code{percentiles}.
\item \code{percentiles} An optional numeric vector specifying the
percentiles for determining the breakpoints of the portfolios.
This parameter is mutually exclusive with \code{n_portfolios}.
\item \code{breakpoints_exchanges} An optional character vector specifying
exchange names to filter the data before computing breakpoints.
Exchanges must be stored in a column given by \code{data_options} (defaults
to \code{exchange}). If \code{NULL}, no filtering is applied.
\item \code{smooth_bunching} An optional logical parameter specifying if
to attempt smoothing non-extreme portfolios if the sorting variable
bunches on the extremes (\code{TRUE}), or not (\code{FALSE}, the default).
In some cases, smoothing will not result in equal-sized portfolios
off the edges due to multiple clusters. If sufficiently large
bunching is detected, \code{percentiles} is ignored and equally-spaced
portfolios are returned for these cases with a warning.
\item \code{breakpoints_min_size_threshold} An optional numeric value between
0 and 1 (exclusive). When set, stocks with market capitalization below
this quantile are excluded from breakpoint computation. The quantile
is computed among \code{breakpoints_exchanges} stocks if specified,
otherwise among all stocks. Requires a market capitalization column
in the data (column name determined by \code{data_options}).
}}

\item{data_options}{A list of class \code{tidyfinance_data_options} (created via
\code{\link[=data_options]{data_options()}}) specifying column name mappings. The \code{exchange} element
is used to specify the exchange column, and \code{mktcap_lag} is used to
specify the market capitalization. Uses \code{\link[=data_options]{data_options()}} default if
\code{NULL}: \code{"exchange" = "exchange"} and \code{"mktcap_lag" = "mktcap_lag"}.}
}
\value{
A numeric vector of breakpoints of the desired length.
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}

Computes breakpoints based on a specified sorting. It can optionally filter
the data by exchanges or lagged size quantiles before computing the
breakpoints. The function requires either the number of portfolios to be
created or specific percentiles for the breakpoints, but not both. The
function also optionally handles cases where the sorting variable clusters
on the edges, by assigning all extreme values to the edges and attempting
to compute equally populated breakpoints with the remaining values.
}
\note{
This function will stop and throw an error if both \code{n_portfolios} and
\code{percentiles} are provided or missing simultaneously.
}
\examples{
set.seed(42)
data <- data.frame(
  id = 1:100,
  exchange = sample(c("NYSE", "NASDAQ"), 100, replace = TRUE),
  market_cap = 1:100
)

compute_breakpoints(data, "market_cap", breakpoint_options(n_portfolios = 5))
compute_breakpoints(
  data,
  "market_cap",
  breakpoint_options(
    percentiles = c(0.2, 0.4, 0.6, 0.8),
    breakpoints_exchanges = c("NYSE")
  )
 )

}
\seealso{
Other portfolio functions: 
\code{\link{assign_portfolio}()},
\code{\link{breakpoint_options}()},
\code{\link{compute_long_short_returns}()},
\code{\link{compute_portfolio_returns}()},
\code{\link{data_options}()},
\code{\link{filter_options}()},
\code{\link{filter_sorting_data}()},
\code{\link{implement_portfolio_sort}()},
\code{\link{portfolio_sort_options}()}
}
\concept{portfolio functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/filter_options.R
\name{filter_options}
\alias{filter_options}
\title{Create Filter Options}
\usage{
filter_options(
  exclude_financials = FALSE,
  exclude_utilities = FALSE,
  min_stock_price = NULL,
  min_size_quantile = NULL,
  min_listing_age = NULL,
  exclude_negative_book_equity = FALSE,
  exclude_negative_earnings = FALSE,
  ...
)
}
\arguments{
\item{exclude_financials}{A logical indicating whether to exclude financial
firms (SIC codes 6000–6799). Defaults to \code{FALSE}.}

\item{exclude_utilities}{A logical indicating whether to exclude utility
firms (SIC codes 4900–4999). Defaults to \code{FALSE}.}

\item{min_stock_price}{A single positive numeric specifying the minimum
stock price required to include an observation. \code{NULL} (the default)
applies no price filter.}

\item{min_size_quantile}{A single numeric strictly between 0 and 1
specifying the minimum cross-sectional size quantile (based on lagged
market cap) required to include an observation. \code{NULL} (the default)
applies no size quantile filter. The cutoff is computed from NYSE
stocks only. This requires an \code{exchange} column in the data (as
mapped via \code{\link[=data_options]{data_options()}}); an error is raised if it is missing.}

\item{min_listing_age}{A single non-negative integer or numeric specifying
the minimum number of months a stock must have been listed in CRSP.
\code{NULL} (the default) applies no listing age filter.}

\item{exclude_negative_book_equity}{A logical indicating whether to exclude
observations with non-positive book equity. Defaults to \code{FALSE}.}

\item{exclude_negative_earnings}{A logical indicating whether to exclude
observations with non-positive earnings. Defaults to \code{FALSE}.}

\item{...}{Additional arguments to be included in the filter options list.}
}
\value{
A list of class \code{tidyfinance_filter_options} containing the
specified filter options.
}
\description{
Creates a list of filter options of class \code{tidyfinance_filter_options} used
for sample construction in TidyFinance-related functions. These options
control which observations are retained before portfolio sorting.
}
\examples{
filter_options(
  exclude_financials = TRUE,
  exclude_utilities = TRUE,
  min_stock_price = 1,
  min_listing_age = 12
)

}
\seealso{
Other portfolio functions: 
\code{\link{assign_portfolio}()},
\code{\link{breakpoint_options}()},
\code{\link{compute_breakpoints}()},
\code{\link{compute_long_short_returns}()},
\code{\link{compute_portfolio_returns}()},
\code{\link{data_options}()},
\code{\link{filter_sorting_data}()},
\code{\link{implement_portfolio_sort}()},
\code{\link{portfolio_sort_options}()}
}
\concept{portfolio functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/implement_portfolio_sort.R
\name{implement_portfolio_sort}
\alias{implement_portfolio_sort}
\title{Implement Portfolio Sort}
\usage{
implement_portfolio_sort(
  data,
  sorting_variables,
  sorting_method,
  rebalancing_month = NULL,
  portfolio_sort_options,
  breakpoint_function_main = compute_breakpoints,
  breakpoint_function_secondary = compute_breakpoints,
  min_portfolio_size = 1L,
  cap_weight = 0.8,
  data_options = NULL,
  quiet = FALSE
)
}
\arguments{
\item{data}{A data frame containing the stock-level panel data.}

\item{sorting_variables}{A character vector of one or two column names to
sort portfolios on.}

\item{sorting_method}{A string specifying the sorting method:
\code{"univariate"}, \code{"bivariate-dependent"}, or \code{"bivariate-independent"}.}

\item{rebalancing_month}{An optional integer specifying the month in which
portfolios are rebalanced annually. \code{NULL} (the default) means monthly
rebalancing.}

\item{portfolio_sort_options}{A list of class
\code{tidyfinance_portfolio_sort_options} created by
\code{\link[=portfolio_sort_options]{portfolio_sort_options()}}, bundling filter and breakpoint
specifications. The arguments accepted by \code{\link[=portfolio_sort_options]{portfolio_sort_options()}}
include
\itemize{
\item \code{filter_options} A list of class \code{tidyfinance_filter_options}
created by \code{\link[=filter_options]{filter_options()}}, or \code{NULL} (the default, which applies
no filters). Options include \code{exclude_financials},
\code{exclude_utilities}, \code{min_stock_price}, \code{min_size_quantile},
\code{min_listing_age}, \code{exclude_negative_book_equity}, and
\code{exclude_negative_earnings}.
\item \code{breakpoint_options_main} A list of class
\code{tidyfinance_breakpoint_options} created by \code{\link[=breakpoint_options]{breakpoint_options()}},
specifying breakpoints for the primary sorting variable. Options
include \code{n_portfolios}, \code{percentiles}, \code{breakpoints_exchanges},
\code{smooth_bunching}, and \code{breakpoints_min_size_threshold}.
\item \code{breakpoint_options_secondary} A list of class
\code{tidyfinance_breakpoint_options} created by \code{\link[=breakpoint_options]{breakpoint_options()}},
specifying breakpoints for the secondary sorting variable, or \code{NULL}
(the default) for univariate sorts. Options are the same as for
\code{breakpoint_options_main}.
}}

\item{breakpoint_function_main}{The function used to compute breakpoints for
the main sorting variable. Defaults to \code{\link[=compute_breakpoints]{compute_breakpoints()}}.}

\item{breakpoint_function_secondary}{The function used to compute breakpoints
for the secondary sorting variable. Defaults to \code{\link[=compute_breakpoints]{compute_breakpoints()}}.}

\item{min_portfolio_size}{An integer specifying the minimum number of
firms in the reported portfolio cross-section on a given date. Defaults
to \code{1L} (at least one observation per reported portfolio). For univariate
sorts that is firms per portfolio-date; for bivariate sorts that is firms
per main-portfolio-date summed across the secondary buckets.
Cross-sections below the threshold have their returns set to \code{NA}.
Set to \code{0L} to deactivate the check. See \code{\link[=compute_portfolio_returns]{compute_portfolio_returns()}}.}

\item{cap_weight}{A numeric between 0 and 1 specifying the quantile at which
portfolio weights are capped for the capped value-weighted return.
Defaults to \code{0.8}.}

\item{data_options}{A list of class \code{tidyfinance_data_options} (created via
\code{\link[=data_options]{data_options()}}) specifying column name mappings. All elements are
forwarded to \code{\link[=filter_sorting_data]{filter_sorting_data()}} and \code{\link[=compute_portfolio_returns]{compute_portfolio_returns()}}.
Uses \code{\link[=data_options]{data_options()}} default if \code{NULL}: \code{"id" = "permno"},
\code{"date" = "date"}, \code{"exchange" = "exchange"},
\code{"mktcap_lag" = "mktcap_lag"}, \code{"ret_excess" = "ret_excess"},
\code{"siccd" = "siccd"}, \code{"price" = "prc_adj"},
\code{"listing_age" = "listing_age"}, \code{"be" = "be"}, and
\code{"earnings" = "ib"}.}

\item{quiet}{A logical indicating whether informational messages should be
suppressed. Defaults to \code{FALSE}.}
}
\value{
A data frame of portfolio returns as returned by
\code{\link[=compute_portfolio_returns]{compute_portfolio_returns()}}.
}
\description{
A convenience wrapper that combines sample construction filtering and
portfolio return computation into a single call. Equivalent to calling
\code{\link[=filter_sorting_data]{filter_sorting_data()}} followed by \code{\link[=compute_portfolio_returns]{compute_portfolio_returns()}}.
}
\examples{

set.seed(123)
data <- data.frame(
  permno = 1:500,
  date = rep(
    seq.Date(
      from = as.Date("2020-01-01"),
      by = "month",
      length.out = 100
    ),
    each = 10
  ),
  mktcap_lag = runif(500, 100, 1000),
  ret_excess = rnorm(500),
  prc_adj = runif(500, 0.5, 50),
  size = runif(500, 50, 150)
)
implement_portfolio_sort(
  data = data,
  sorting_variables = "size",
  sorting_method = "univariate",
  portfolio_sort_options = portfolio_sort_options(
    filter_options = filter_options(
      min_stock_price = 1
    ),
    breakpoint_options_main = breakpoint_options(n_portfolios = 5)
  )
)


}
\seealso{
Other portfolio functions: 
\code{\link{assign_portfolio}()},
\code{\link{breakpoint_options}()},
\code{\link{compute_breakpoints}()},
\code{\link{compute_long_short_returns}()},
\code{\link{compute_portfolio_returns}()},
\code{\link{data_options}()},
\code{\link{filter_options}()},
\code{\link{filter_sorting_data}()},
\code{\link{portfolio_sort_options}()}
}
\concept{portfolio functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tidyfinance-package.R
\docType{package}
\name{tidyfinance-package}
\alias{tidyfinance}
\alias{tidyfinance-package}
\title{tidyfinance: Tidy Finance Helper Functions}
\description{
Helper functions for empirical research in financial economics, addressing a variety of topics covered in Scheuch, Voigt, and Weiss (2023) \doi{10.1201/b23237}. The package is designed to provide shortcuts for issues extensively discussed in the book, facilitating easier application of its concepts. For more information and resources related to the book, visit \url{https://www.tidy-finance.org/r/index.html}.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://www.tidy-finance.org/r/}
  \item \url{https://github.com/tidy-finance/r-tidyfinance}
  \item Report bugs at \url{https://github.com/tidy-finance/r-tidyfinance/issues}
}

}
\author{
\strong{Maintainer}: Christoph Scheuch \email{christoph@tidy-intelligence.com} (\href{https://orcid.org/0009-0004-0423-6819}{ORCID}) [copyright holder]

Authors:
\itemize{
  \item Stefan Voigt \email{stefan.voigt@econ.ku.dk} (\href{https://orcid.org/0000-0001-5619-3161}{ORCID}) [copyright holder]
  \item Patrick Weiss \email{patrickw@ru.is} (\href{https://orcid.org/0000-0002-9282-5872}{ORCID}) [copyright holder]
}

Other contributors:
\itemize{
  \item Maximilian Mücke (\href{https://orcid.org/0009-0000-9432-9795}{ORCID}) [contributor]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/download_data_constituents.R
\name{download_data_constituents}
\alias{download_data_constituents}
\title{Download Constituent Data}
\usage{
download_data_constituents(index)
}
\arguments{
\item{index}{A character string specifying the name of the financial index
for which to download constituent data. The index must be one of the
supported indexes listed by \code{\link[=list_supported_indexes]{list_supported_indexes()}}.}
}
\value{
A tibble with five columns:
\describe{
\item{symbol}{The ticker symbol of the equity constituent.}
\item{name}{The name of the equity constituent.}
\item{location}{The location where the company is based.}
\item{exchange}{The exchange where the equity is traded.}
\item{currency}{The currency in which the equity is traded, derived
from the exchange.}
}
The tibble is filtered to exclude non-equity entries, blacklisted symbols,
empty names, and any entries containing the index name or "CASH".
}
\description{
Downloads and processes the constituent data for a specified financial
index. The data is fetched from a remote CSV file, filtered, and cleaned to
provide relevant information about constituents.
}
\details{
The function retrieves the URL of the CSV file for the specified index from
ETF sites, then sends an HTTP GET request to download the CSV file, and
processes the CSV file to extract equity constituents.

The approach is inspired by \code{tidyquant::tq_index()}, which uses a different
wrapper around other ETFs.
}
\examples{
\donttest{
  download_data_constituents("DAX")
}

}
\seealso{
Other download functions: 
\code{\link{download_data}()},
\code{\link{download_data_factors_ff}()},
\code{\link{download_data_factors_q}()},
\code{\link{download_data_fred}()},
\code{\link{download_data_huggingface}()},
\code{\link{download_data_macro_predictors}()},
\code{\link{download_data_osap}()},
\code{\link{download_data_risk_free}()},
\code{\link{download_data_stock_prices}()}
}
\concept{download functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/disconnect_connection.R
\name{disconnect_connection}
\alias{disconnect_connection}
\title{Disconnect Database Connection}
\usage{
disconnect_connection(con)
}
\arguments{
\item{con}{A database connection object created by DBI::dbConnect or any
similar function that establishes a connection to a database.}
}
\value{
\code{TRUE}, invisibly. Throws an error if the disconnection fails.
}
\description{
Safely disconnects an established database connection using the DBI package.
}
\examples{
\dontrun{
con <- get_wrds_connection()
disconnect_connection(con)
}
}
\seealso{
Other WRDS functions: 
\code{\link{create_wrds_dummy_database}()},
\code{\link{download_data_wrds}()},
\code{\link{download_data_wrds_ccm_links}()},
\code{\link{download_data_wrds_compustat}()},
\code{\link{download_data_wrds_crsp}()},
\code{\link{download_data_wrds_fisd}()},
\code{\link{download_data_wrds_trace_enhanced}()},
\code{\link{get_wrds_connection}()},
\code{\link{set_wrds_credentials}()}
}
\concept{WRDS functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/set_wrds_credentials.R
\name{set_wrds_credentials}
\alias{set_wrds_credentials}
\title{Set WRDS Credentials}
\usage{
set_wrds_credentials()
}
\value{
Invisibly returns \code{TRUE}. Displays messages to the user based on
their input and actions taken.
}
\description{
Prompts the user to input their WRDS (Wharton Research Data Services)
username and password, and stores these credentials in a \code{.Renviron} file.
The user can choose to store the \code{.Renviron} file in either the project
directory or the home directory. If the \code{.Renviron} file already contains
WRDS credentials, the user will be asked if they want to overwrite the
existing credentials. Additionally, the user has the option to add the
\code{.Renviron} file to the .gitignore file to prevent it from being tracked by
version control.
}
\examples{
\dontrun{
set_wrds_credentials()
}
}
\seealso{
Other WRDS functions: 
\code{\link{create_wrds_dummy_database}()},
\code{\link{disconnect_connection}()},
\code{\link{download_data_wrds}()},
\code{\link{download_data_wrds_ccm_links}()},
\code{\link{download_data_wrds_compustat}()},
\code{\link{download_data_wrds_crsp}()},
\code{\link{download_data_wrds_fisd}()},
\code{\link{download_data_wrds_trace_enhanced}()},
\code{\link{get_wrds_connection}()}
}
\concept{WRDS functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/download_data_factors.R
\name{download_data_factors_ff}
\alias{download_data_factors_ff}
\title{Download and Process Fama-French Factor Data}
\usage{
download_data_factors_ff(
  dataset = NULL,
  start_date = NULL,
  end_date = NULL,
  type = deprecated()
)
}
\arguments{
\item{dataset}{The name of the Fama-French dataset to download (e.g.,
"Fama/French 3 Factors").}

\item{start_date}{Optional. A character string or Date object in
"YYYY-MM-DD" format specifying the start date for the data. If not
provided, the full dataset is returned.}

\item{end_date}{Optional. A character string or Date object in "YYYY-MM-DD"
format specifying the end date for the data. If not provided, the full
dataset is returned.}

\item{type}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use \code{dataset} instead.}
}
\value{
A tibble with processed factor data, including the date, risk-free
rate, market excess return, and other factors, filtered by the specified
date range.
}
\description{
Downloads and processes Fama-French factor data based on the specified
dataset name and date range. The function requires the \code{frenchdata}
package to download the data. It processes the raw data into a structured
format, including date conversion, scaling factor values, and filtering by
the specified date range.
}
\details{
If there are multiple tables in the raw Fama-French data (e.g.,
value-weighted and equal-weighted returns), then the function only returns
the first table because these are the most popular. Please use the
\code{frenchdata} package directly if you need less commonly used tables.
}
\examples{
\donttest{
  download_data_factors_ff(
    "Fama/French 3 Factors", "2000-01-01", "2020-12-31"
  )
  download_data_factors_ff(
    "10 Industry Portfolios", "2000-01-01", "2020-12-31"
  )
}
}
\references{
Fama, E. F., & French, K. R. (1993). Common risk factors in the returns on
stocks and bonds. \emph{Journal of Financial Economics}, 33(1), 3-56.
\doi{10.1016/0304-405X(93)90023-5}

Fama, E. F., & French, K. R. (2015). A five-factor asset pricing model.
\emph{Journal of Financial Economics}, 116(1), 1-22.
\doi{10.1016/j.jfineco.2014.10.010}

Carhart, M. M. (1997). On persistence in mutual fund performance.
\emph{Journal of Finance}, 52(1), 57-82.
\doi{10.1111/j.1540-6261.1997.tb03808.x}
}
\seealso{
Other download functions: 
\code{\link{download_data}()},
\code{\link{download_data_constituents}()},
\code{\link{download_data_factors_q}()},
\code{\link{download_data_fred}()},
\code{\link{download_data_huggingface}()},
\code{\link{download_data_macro_predictors}()},
\code{\link{download_data_osap}()},
\code{\link{download_data_risk_free}()},
\code{\link{download_data_stock_prices}()}
}
\concept{download functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/add_lagged_columns.R
\name{add_lagged_columns}
\alias{add_lagged_columns}
\title{Add Lagged Columns via Join}
\usage{
add_lagged_columns(
  data,
  cols,
  lag,
  max_lag = lag,
  by = NULL,
  drop_na = FALSE,
  ff_adjustment = FALSE,
  data_options = NULL
)
}
\arguments{
\item{data}{A data frame containing the variables to lag.}

\item{cols}{A character vector specifying the names of the columns to be
lagged. Each column produces a new column suffixed with \verb{_lag}.}

\item{lag}{An integer or a \code{lubridate::periods()} object, e.g.,
\code{months(1)}, specifying the minimum lag (inclusive) to apply.}

\item{max_lag}{An integer or a \code{lubridate::periods()} object specifying
the maximum lag (inclusive) to apply. Defaults to \code{lag} (exact lag).}

\item{by}{An optional character vector specifying grouping columns
(e.g., a stock identifier). Lagged values are matched within groups.
Defaults to \code{NULL}.}

\item{drop_na}{A logical value. If \code{TRUE}, \code{NA} values in the source
columns are excluded before matching, so the lookup skips over missing
observations. Applied independently per column. Defaults to \code{FALSE}.}

\item{ff_adjustment}{A logical value. If \code{TRUE}, only the last observation
per year (within each group defined by \code{by}) is retained as a source for
lagged values, following Fama-French conventions for annual accounting
data. Defaults to \code{FALSE}.}

\item{data_options}{A list of class \code{tidyfinance_data_options} (created via
\code{\link[=data_options]{data_options()}}) specifying column name mappings. The \code{date} element is
used to specify the date column. Uses \code{\link[=data_options]{data_options()}} default if \code{NULL}:
\code{"date" = "date"}.}
}
\value{
A data frame with the same rows as \code{data} and new columns appended,
each suffixed with \verb{_lag}. Unmatched rows receive \code{NA} in the lagged
columns.
}
\description{
Appends lagged versions of specified columns to a data frame using a
join-based approach.
}
\details{
When \code{lag == max_lag} (the default), an equi-join is used: source dates
are shifted forward by \code{lag} and matched exactly. When \code{lag < max_lag},
an inequality join is used: for each row, the most recent source value
within the window \verb{[date - max_lag, date - lag]} is selected.

The combination of \code{by} and date columns must be unique in \code{data}. If \code{by}
is \code{NULL}, dates alone must be unique.
}
\examples{
set.seed(42)
data <- tibble::tibble(
  permno = rep(1:2, each = 10),
  date = rep(
    seq.Date(as.Date("2023-01-01"), by = "month", length.out = 10),
    2
  ),
  size = runif(20, 100, 200),
  bm = runif(20, 0.5, 1.5)
)

# Exact lag: each row gets the value from exactly 2 months earlier
add_lagged_columns(
  data,
  cols = c("size", "bm"),
  lag = months(2),
  by = "permno"
)

# Window lag: each row gets the most recent value from 2 to 4 months earlier
add_lagged_columns(
  data,
  cols = "size",
  lag = months(2),
  max_lag = months(4),
  by = "permno"
)

}
\seealso{
Other rolling and lagging functions: 
\code{\link{compute_rolling_value}()},
\code{\link{join_lagged_values}()}
}
\concept{rolling and lagging functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/download_data_macro_predictors.R
\name{download_data_macro_predictors}
\alias{download_data_macro_predictors}
\title{Download and Process Macro Predictor Data}
\usage{
download_data_macro_predictors(
  dataset = NULL,
  start_date = NULL,
  end_date = NULL,
  type = deprecated(),
  sheet_id = "1bM7vCWd3WOt95Sf9qjLPZjoiafgF_8EG"
)
}
\arguments{
\item{dataset}{The dataset to download ("monthly", "quarterly",
"annual").}

\item{start_date}{Optional. A character string or Date object in
"YYYY-MM-DD" format specifying the start date for the data. If
not provided, the full dataset is returned.}

\item{end_date}{Optional. A character string or Date object in
"YYYY-MM-DD" format specifying the end date for the data. If not
provided, the full dataset is returned.}

\item{type}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use \code{dataset}
instead.}

\item{sheet_id}{The Google Sheets ID from which to download the
dataset, with the default "1bM7vCWd3WOt95Sf9qjLPZjoiafgF_8EG".}
}
\value{
A tibble with processed data, filtered by the specified
date range and including financial metrics.
}
\description{
Downloads and processes macroeconomic predictor data based on the
specified dataset (monthly, quarterly, or annual), date range, and
source URL. The function downloads the data from a Google Sheets
export link. It processes the raw data into a structured format,
calculating additional financial metrics and filtering by the
specified date range.
}
\examples{
\donttest{
  download_data_macro_predictors("monthly")
  download_data_macro_predictors("quarterly", "2000-01-01", "2020-12-31")
}

}
\references{
Welch, I., & Goyal, A. (2008). A comprehensive look at the empirical
performance of equity premium prediction. \emph{Review of Financial Studies},
21(4), 1455-1508. \doi{10.1093/rfs/hhm014}
}
\seealso{
Other download functions: 
\code{\link{download_data}()},
\code{\link{download_data_constituents}()},
\code{\link{download_data_factors_ff}()},
\code{\link{download_data_factors_q}()},
\code{\link{download_data_fred}()},
\code{\link{download_data_huggingface}()},
\code{\link{download_data_osap}()},
\code{\link{download_data_risk_free}()},
\code{\link{download_data_stock_prices}()}
}
\concept{download functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/trim.R
\name{trim}
\alias{trim}
\title{Trim a Numeric Vector}
\usage{
trim(x, cut)
}
\arguments{
\item{x}{A numeric vector to be trimmed.}

\item{cut}{The proportion of data to be trimmed from both ends of the
distribution. For example, a \code{cut} of 0.05 will remove the lowest and
highest 5\% of the data. Must be in [0, 0.5].}
}
\value{
A numeric vector with the extreme values removed.
}
\description{
Removes the values in a numeric vector that are beyond the specified
quantiles, effectively trimming the distribution based on the \code{cut}
parameter. This process reduces the vector's length by excluding extreme
values from both tails of the distribution.
}
\examples{
set.seed(123)
data <- rnorm(100)
trimmed_data <- trim(x = data, cut = 0.05)

}
\seealso{
Other utility functions: 
\code{\link{create_summary_statistics}()},
\code{\link{get_available_huggingface_files}()},
\code{\link{list_supported_indexes}()},
\code{\link{list_supported_types}()},
\code{\link{list_supported_types_ff}()},
\code{\link{list_supported_types_ff_legacy}()},
\code{\link{list_supported_types_macro_predictors}()},
\code{\link{list_supported_types_other}()},
\code{\link{list_supported_types_wrds}()},
\code{\link{list_tidy_finance_chapters}()},
\code{\link{open_tidy_finance_website}()},
\code{\link{validate_dates}()},
\code{\link{winsorize}()}
}
\concept{utility functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/validate.R
\name{validate_dates}
\alias{validate_dates}
\title{Validate and Coerce Date Range Arguments}
\usage{
validate_dates(start_date, end_date, use_default_range = FALSE)
}
\arguments{
\item{start_date}{A scalar coercible to \code{Date} via \code{as.Date()}, or \code{NULL}.}

\item{end_date}{A scalar coercible to \code{Date} via \code{as.Date()}, or \code{NULL}.}

\item{use_default_range}{A logical scalar. If \code{TRUE} and both date arguments
are \code{NULL}, a default one-year range is used instead of returning
\code{NULL}. Defaults to \code{FALSE}.}
}
\value{
A named list with elements \code{start_date} and \code{end_date}, both of
class \code{Date} (or \code{NULL} when no dates are provided and
\code{use_default_range = FALSE}).
}
\description{
Checks that \code{start_date} and \code{end_date} are a valid pair, coerces them to
\code{Date}, and handles the case where both are \code{NULL}. When both are \code{NULL} and
\code{use_default_range = TRUE}, a two-year default window ending one year ago is
applied and reported to the user.
}
\seealso{
Other utility functions: 
\code{\link{create_summary_statistics}()},
\code{\link{get_available_huggingface_files}()},
\code{\link{list_supported_indexes}()},
\code{\link{list_supported_types}()},
\code{\link{list_supported_types_ff}()},
\code{\link{list_supported_types_ff_legacy}()},
\code{\link{list_supported_types_macro_predictors}()},
\code{\link{list_supported_types_other}()},
\code{\link{list_supported_types_wrds}()},
\code{\link{list_tidy_finance_chapters}()},
\code{\link{open_tidy_finance_website}()},
\code{\link{trim}()},
\code{\link{winsorize}()}
}
\concept{utility functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/list_supported_types.R
\name{list_supported_types}
\alias{list_supported_types}
\title{List All Supported Dataset Types}
\usage{
list_supported_types(domain = NULL, as_vector = FALSE)
}
\arguments{
\item{domain}{A character vector to filter for domain-specific types
(e.g., c("WRDS", "Fama-French"))}

\item{as_vector}{Logical indicating whether types should be returned as
a character vector instead of a data frame.}
}
\value{
A tibble aggregating all supported dataset types with columns:
\code{type} (the type of dataset), \code{dataset_name} (a descriptive name or
file name of the dataset), and \code{domain} (the domain to which the
dataset belongs, e.g., "Global Q", "Fama-French", "Goyal-Welch").
}
\description{
Aggregates and returns a comprehensive tibble of all supported dataset
types from different domains. It includes various datasets across
different frequencies (daily, weekly, monthly, quarterly, annual) and
models (e.g., q5 factors, Fama-French 3 and 5 factors, macro predictors).
}
\examples{
# List all supported types as a data frame
list_supported_types()

# Filter by domain
list_supported_types(domain = "WRDS")

# List supported types as a vector
list_supported_types(as_vector = TRUE)
}
\seealso{
Other utility functions: 
\code{\link{create_summary_statistics}()},
\code{\link{get_available_huggingface_files}()},
\code{\link{list_supported_indexes}()},
\code{\link{list_supported_types_ff}()},
\code{\link{list_supported_types_ff_legacy}()},
\code{\link{list_supported_types_macro_predictors}()},
\code{\link{list_supported_types_other}()},
\code{\link{list_supported_types_wrds}()},
\code{\link{list_tidy_finance_chapters}()},
\code{\link{open_tidy_finance_website}()},
\code{\link{trim}()},
\code{\link{validate_dates}()},
\code{\link{winsorize}()}
}
\concept{utility functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/list_tidy_finance_chapters.R
\name{list_tidy_finance_chapters}
\alias{list_tidy_finance_chapters}
\title{List Chapters of Tidy Finance}
\usage{
list_tidy_finance_chapters()
}
\value{
A character vector where each element is the name of a chapter
available in the Tidy Finance resource. These names correspond to specific
chapters in Tidy Finance with R.
}
\description{
Returns a character vector containing the names of the chapters available in
the Tidy Finance resource. It provides a quick reference to the
various topics covered.
}
\examples{
list_tidy_finance_chapters()

}
\seealso{
Other utility functions: 
\code{\link{create_summary_statistics}()},
\code{\link{get_available_huggingface_files}()},
\code{\link{list_supported_indexes}()},
\code{\link{list_supported_types}()},
\code{\link{list_supported_types_ff}()},
\code{\link{list_supported_types_ff_legacy}()},
\code{\link{list_supported_types_macro_predictors}()},
\code{\link{list_supported_types_other}()},
\code{\link{list_supported_types_wrds}()},
\code{\link{open_tidy_finance_website}()},
\code{\link{trim}()},
\code{\link{validate_dates}()},
\code{\link{winsorize}()}
}
\concept{utility functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/download_data_factors.R
\name{download_data_factors_q}
\alias{download_data_factors_q}
\title{Download and Process Global Q Factor Data}
\usage{
download_data_factors_q(
  dataset = NULL,
  start_date = NULL,
  end_date = NULL,
  type = deprecated(),
  url = "https://global-q.org/uploads/1/2/2/6/122679606/"
)
}
\arguments{
\item{dataset}{The name of the dataset to download (e.g.,
"q5_factors_daily_2023.csv", "q5_factors_monthly_2023.csv").}

\item{start_date}{Optional. A character string or Date object in
"YYYY-MM-DD" format specifying the start date for the data. If
not provided, the full dataset is returned.}

\item{end_date}{Optional. A character string or Date object in
"YYYY-MM-DD" format specifying the end date for the data. If not
provided, the full dataset is returned.}

\item{type}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use \code{dataset}
instead.}

\item{url}{The base URL from which to download the dataset files.}
}
\value{
A tibble with processed factor data, including the date,
risk-free rate, market excess return, and other factors,
filtered by the specified date range.
}
\description{
Downloads and processes Global Q factor data based on the
specified dataset, date range, and source URL. The processing
includes date conversion, renaming variables to a standardized
format, scaling factor values, and filtering by the specified
date range.
}
\examples{
\donttest{
download_data_factors_q("q5_factors_daily_2024", "2020-01-01", "2020-12-31")
download_data_factors_q("q5_factors_annual_2024")
}
}
\references{
Hou, K., Xue, C., & Zhang, L. (2015). Digesting anomalies: An investment
approach. \emph{Review of Financial Studies}, 28(3), 650-705.
\doi{10.1093/rfs/hhu068}

Hou, K., Mo, H., Xue, C., & Zhang, L. (2019). Which factors?
\emph{Review of Finance}, 23(1), 1-35.
\doi{10.1093/rof/rfy032}
}
\seealso{
Other download functions: 
\code{\link{download_data}()},
\code{\link{download_data_constituents}()},
\code{\link{download_data_factors_ff}()},
\code{\link{download_data_fred}()},
\code{\link{download_data_huggingface}()},
\code{\link{download_data_macro_predictors}()},
\code{\link{download_data_osap}()},
\code{\link{download_data_risk_free}()},
\code{\link{download_data_stock_prices}()}
}
\concept{download functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/list_supported_types.R
\name{list_supported_types_macro_predictors}
\alias{list_supported_types_macro_predictors}
\title{List Supported Macro Predictor Dataset Types}
\usage{
list_supported_types_macro_predictors()
}
\value{
A tibble with columns: \code{type} (the type of dataset),
\code{dataset_name} (the file name of the dataset, which is the same for
all types), and \code{domain} (the domain to which the dataset belongs,
always "Goyal-Welch").
}
\description{
Returns a tibble with the supported macro predictor dataset types
provided by Goyal-Welch, including their frequencies (monthly, quarterly,
annual). All dataset types reference the same source file,
"PredictorData2022.xlsx" for the year 2022. Additionally, it annotates
each dataset with the domain "Goyal-Welch".
}
\seealso{
Other utility functions: 
\code{\link{create_summary_statistics}()},
\code{\link{get_available_huggingface_files}()},
\code{\link{list_supported_indexes}()},
\code{\link{list_supported_types}()},
\code{\link{list_supported_types_ff}()},
\code{\link{list_supported_types_ff_legacy}()},
\code{\link{list_supported_types_other}()},
\code{\link{list_supported_types_wrds}()},
\code{\link{list_tidy_finance_chapters}()},
\code{\link{open_tidy_finance_website}()},
\code{\link{trim}()},
\code{\link{validate_dates}()},
\code{\link{winsorize}()}
}
\concept{utility functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/join_lagged_values.R
\name{join_lagged_values}
\alias{join_lagged_values}
\title{Join Lagged Variable Values over a Date Range}
\usage{
join_lagged_values(
  original_data,
  new_data,
  id_keys,
  min_lag,
  max_lag,
  ff_adjustment = FALSE,
  data_options = NULL
)
}
\arguments{
\item{original_data}{A data frame containing the target panel data.}

\item{new_data}{A data frame containing the source variables to lag and
merge. All columns besides \code{id_keys} and the date column will be lagged
and joined.}

\item{id_keys}{A character vector specifying the identifier column(s).}

\item{min_lag}{A \code{lubridate::Period} specifying the lower lag bound
(inclusive).}

\item{max_lag}{A \code{lubridate::Period} specifying the upper lag bound
(inclusive).}

\item{ff_adjustment}{Logical; if \code{TRUE}, keeps only the last observation
per identifier and year before lagging (Fama-French convention). Defaults
to \code{FALSE}.}

\item{data_options}{A list of class \code{tidyfinance_data_options} (created via
\code{\link[=data_options]{data_options()}}) specifying column name mappings. The \code{date} element is
used to identify the date column. Uses \code{\link[=data_options]{data_options()}} default if
\code{NULL}: \code{"date" = "date"}.}
}
\value{
A data frame with all columns from \code{original_data} plus the
lagged columns from \code{new_data} (keeping their original names).
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}

Joins lagged values of selected variables from one dataset (\code{new_data})
into another (\code{original_data}), based on date ranges defined by \code{min_lag}
and \code{max_lag}. Unlike \code{\link[=add_lagged_columns]{add_lagged_columns()}}, this function supports
joining across data frames with different date grids (e.g., monthly source
data into quarterly target data).
}
\examples{
set.seed(42)
library(dplyr)
library(lubridate)

df1 <- tibble(
  id = rep(1:2, each = 6),
  date = rep(seq(as.Date("2020-01-01"), by = "month", length.out = 6), 2)
)

df2 <- df1 |>
  mutate(x = rnorm(n()))

join_lagged_values(
  original_data = df1,
  new_data = df2,
  id_keys = "id",
  min_lag = months(1),
  max_lag = months(3)
)
}
\seealso{
Other rolling and lagging functions: 
\code{\link{add_lagged_columns}()},
\code{\link{compute_rolling_value}()}
}
\concept{rolling and lagging functions}
